/*
Program Id: myschema

Program Description: this program will generate an sql schema for a
                     specified database or a set of tables.  Adds 
                     features not found in Informix' 5.0 dbschema.

                     For 7.xx+ users adds ability to schema a busy database
                     since dbschema wants to take a lock on each table as 
                     it goes.

Author: Art S. Kagel original code from Rick Friedman's dbschema_cc for Sybase
        Collection support added by: Sergey E. Volkov (Oct. 2001)

Original Date: May of 1992  

Version: 6.32

Comments:    

Revisions:

Usage:  See usage function output.

RCS Header
----------
$Header: /home/art/GoogleDrive/Latest/myschema.d/RCS/myschema.ec,v 2.440 2023/11/15 19:07:03 art Exp $

RCS Log
-------
$Log: myschema.ec,v $
Revision 2.440  2023/11/15 19:07:03  art
Fixed the reported length of LVARCHAR type columns when the multibyte flag is set.

Revision 2.439  2023/11/15 15:12:46  art
Fixed routine parameter lists in GRANT and REVOKE statements that contain complex parameters.
Thanks to Davorin Kremenjas for reporting the issue.

Revision 2.438  2023/08/30 18:21:23  art
Added support for logged/unlogged default sbspaces.
Added support for access time logging/nologging default sbspaces.

Revision 2.437  2023/08/30 18:20:49  art
Removed trace.

Revision 2.436  2023/02/19 17:08:47  art
Fixed issue correctly reporting PUT clauses for BLOB and CLOB
type columns in tables with CRCOLS defined.

Revision 2.435  2023/01/04 18:29:46  art
Fixed issue with VIEWs under user name filtering options (-O, --set-owner,
--use-table-owner, & --use-database-owner) stripping commas from the
SELECT statement that follow closing quotes. Thanks to Øyvind Gjerstad
for reporting the issue.

Revision 2.434  2022/12/20 14:02:12  art
Fixed issued with GRANT and REVOKE EXECUTE privs for routines 
that have collection type arguments (LIST, SET, MULTISET, etc.)

Revision 2.433  2022/11/06 18:20:14  art
New option --include-sysprocs
to print system procedures.

Revision 2.432  2022/09/08 17:21:14  art
Added new options export-binary and import-binary to generate only 
the export or import script for using an external table in binary
format.

Revision 2.431  2022/07/12 22:55:43  art
Fixed a problem with running --export-scripts & --import-scripts together which
was intended to be permitted.
Changed the external filenames of the scripts for 
--import-scripts and --export-scripts to include the words 'import' and 'export'
respectively, as opposed to 'myimport' & 'myexport'.
Changed the external table name to <table>_external for both the export and the
import scripts rather than <table>_myexport or <table>_myimport.
Changed the name of the export file from the dbexport convention to 
<table>_external.
Creating the external tables for all these options to be conditional in case the
user does not want to drop the external table between exporting and importing or
wants to use a permanent external table.

Revision 2.430  2022/06/08 16:32:04  art
Removed "START AT ..." if start_val is NULL (implies that the min_val is
greater than 1 (or less than -1).

Revision 2.429  2022/02/21 14:35:12  art
Modified to return 99 if any table contains a BSON type column.
This so that myexport can exit with a warning to not export the
database using the default unload method (sqlunload) or -U.

Revision 2.428  2022/01/23 16:00:40  art
Fixed problem that only printed the first partition of indexes partitioned by expression.

Revision 2.427  2021/12/21 17:26:34  art
Updated copyright notice.

Revision 2.426  2021/12/02 18:41:47  art
Fix the fix.

Revision 2.425  2021/12/02 18:33:44  art
Fixed handling table names created with CAPs under DELIMIDENT.

Revision 2.424  2021/11/03 14:11:14  art
Added new option --myexport-binary to unload in INFORMIX binary
internal format in express mode.

Revision 2.423  2021/10/28 18:29:21  art
Changed --myexport-express output to the myimport external table script to 
set the table into RAW mode before the import and to STANDARD after.

Revision 2.422  2021/07/14 18:42:14  art
Copyright update. Removed some old noop code from print_constraints.ec.

Revision 2.421  2021/06/01 19:51:18  art
Fixed problem with not closing the myimport external table files!

Revision 2.420  2021/06/01 13:25:23  art
Fixed a problem with --dependency-order and --myexport-order writing to only one temp
dbspace. Now obeys DBSPACETEMP.

Revision 2.419  2021/05/20 22:03:24  art
Fixed printout of the traditional replcheck index when it
includes the erkey columns. Was being suppressed like the
automatic erkey index.

Revision 2.418  2021/05/14 15:44:53  art
Fixed problem with fragmented tables not displaying CDR special
columns (REPLCHECK, ERKEY, CRCOLS, & VERCOLS) or the WITH O
AUDIT attribute.

Revision 2.417  2021/05/13 23:01:26  art
Added new --myexport-dependency option to sort by dependencies
but in reverse order to prevent inconsistent data.

Revision 2.416  2021/05/06 11:27:56  art
Added new --no-export-triggers - Disable select triggers in the external table 
export scripts.

Revision 2.415  2021/05/06 00:50:06  art
Added new options --no-data-tables and --no-data-accesses to suppress 
generation of unload and load scripts when --export-scripts
--import-scripts --myexport-scripts --myexport-express is enabled.
The former suppresses a list of tables (default: tsinstancetable)
The latter suppresses tables built with any of a list of access 
methods (default:ts_rtl_vtam,ts_vtam).

Revision 2.414  2021/04/20 16:57:42  art
ER hidden columns were not being generated for partitioned tables.

Revision 2.413  2021/03/17 12:13:57  art
Improved some abort messages.

Revision 2.412  2021/03/09 16:19:40  art
New option --for-er adds WITH REPLCHECK and WITH ERKEY to each table to 
create a schema suitable for building an ER ready copy of the database.
Also generates code to create the index on the primary key column and
the ifx_replcheck column if a primary key exists.

Revision 2.411  2021/02/22 14:21:04  art
Fix for multibyte character sets.
Thanks to Hrvoje Zokovic for providing the fix.

Revision 2.410  2021/02/19 17:28:15  art
Added new options: --export-scripts & --import-scripts to create only 
the external table files for exports or imports respectively. These
obey the --myexport-delimiter and --myexport-escape option settings,
but synonyms for each that start with --export- & --import were added
for completeness. The unload filenames using these new options are 
made with the full filename instead of the dbexport compatible 
filenames produced by --myexport-scripts and --myexport-express.

Revision 2.409  2021/02/17 17:16:59  art
Fixed PUT clauses for smartblob types.

Revision 2.408  2021/02/15 17:37:30  art
Fixed condition where a quoted string immediately follows a comma which 
was causing keywords within the quotes being processed instead of ignored.
Thanks to David Grove for pointing this one out!

Revision 2.407  2021/01/12 16:42:40  art
Added new options --no-setsessionauth and --comment-setsessionauth
Needed because even the database owner cannot grant SETSESSIONAUTH to
itself so some scripts will fail if these auths are included. They 
should be processed manually. The -no suppresses and the --comment
comments out these privs.

Revision 2.406  2020/12/02 19:02:53  art
Fixed bad interaction that depended on the order of the 
-o and --dependency-order options.

Revision 2.405  2020/12/02 18:55:16  art
Fixed -o to sort systable objects other than tables in name order within type.
Fixed the interaction between -o and --dependency-order  to sort tables my name
within dependency level and to sort all non-systables objects by name (was
reverting to id order when --dependency-order is specified.

Revision 2.404  2020/11/25 12:24:06  art
Extended --no-blade-procs/--no-blade-funcs to also exclude the resources added by the 
blade management routines when any blade is registered.

Revision 2.403  2020/11/24 14:42:27  art
Fixed spacing of EXTENT and LOCK MODE clauses if --no-extent specified.
Thanks to Antonin Rozenal for pointing this one out.

Revision 2.402  2020/11/23 21:18:16  art
Added new option --alter-checks to force check constraints to be 
generated as ALTER TABLE statements rather than inline in the 
CREATE TABLE statement. Similar to --no-check but does not 
include the NOVALIDATE option.

Revision 2.401  2020/11/20 12:20:37  art
Fixed sub-version parsing. Fixed --dependency-ordering when it encounters a 
self-referencing table loop. Thanks to David Grove for reporting these issues.

Revision 2.400  2020/11/15 22:41:48  art
Fixed an initialization bug in stored procedure handling. Thanks to Jerome Outhier
for pointing it out and suggesting a fix!

Revision 2.399  2020/11/04 22:28:55  art
Fixed plogdbspace API function call in --infrastructure=sql.

Revision 2.398  2020/10/12 13:13:01  art
Fixed an issue for compressed tables. Thanks to Øyvind Gjerstad for 
reporting it problem.

Revision 2.397  2020/09/02 18:05:48  art
Fixed missing usage text.

Revision 2.396  2020/06/02 17:34:39  art
Copyright update.

Revision 2.395  2020/06/02 17:31:21  art
Fixed empty call to dbaccess in "cmd" shell style output where there are not 
storage pool entries.

Revision 2.394  2020/05/28 13:33:10  art
Added a call to se_metadatainit() when the spatial_references table is loaded in the myimport external 
table load script generated by --myexport-scripts. This fixes a problem with the spatial datablade
cache not updating which can cause the loads of tables containing spatial columns to fail. Thanks to
Daniel Tafili for identifying the problem and hunting up the solution.

Revision 2.393  2020/05/10 17:39:21  art
Fixed problem with support functions for Datablade supplied Secondary Access Methods
were being written to the procedure file. However, the Access Method definitions 
required those functions to already exist. Changed so that functions mentioned in 
sysams are written to the main file along with the AMS definition. Thanks to Kenneth
Penza for pointing out the issue.

Revision 2.392  2020/05/07 14:24:34  art
Added new option --no-private-synonyms to exclude private synonyms from the
output.
Thanks to Ognjen Orel for the suggestion.

Revision 2.391  2020/04/27 15:11:17  art
Removed references to unimplemented system call isastream() and its include file stropts.h
It was only used for information output and was no-op for several years.

Revision 2.390  2020/04/14 15:26:30  art
Added --help as a valid option. Just thought it was a good idea.

Revision 2.389  2020/04/12 15:49:31  art
Changed one of the queries related to --dependency-order because some early CSDK v3.50 releases \
did not recognize ORDER SIBLINGS BY and rejects the original code. That query is now an execute
immediate so it is passed to the engine by the compiler. Thanks to Mark Collins for exposing the
problem.

Revision 2.388  2019/12/04 16:56:33  art
Fixed extendable output for --infrastructure.

Revision 2.387  2019/12/03 15:15:39  art
Fixed an error message. Thanks to Paul Watson for pointing it out.

Revision 2.386  2019/11/26 20:59:40  art
Fixed a typo in the usage for --constraint-file

Revision 2.385  2019/11/18 22:57:47  art
Fixed REVOKE ... AS <owner> to conform to -O -A & --set-owner.
Added AS <owner> to REVOKEs for sequences.

Revision 2.384  2019/11/18 12:31:28  art
Fixed datetime printouts and improved handling of distinct types.

Revision 2.383  2019/09/17 14:43:08  art
Added support for the new TSTAMP distinct type.

Revision 2.382  2019/09/13 15:46:24  art
Added support for the v14.10.FC2 partial index feature!

Revision 2.381  2019/09/10 16:42:03  art
Moved UDT CREATE TYPE statements to the main file, GRANTS on TYPES to the grant file.
Cleaned format for DISTINCT TYPEs.

Revision 2.380  2019/08/26 14:21:32  art
Fixed support for security policy display.

Revision 2.379  2019/04/30 10:38:06  art
Added --view-only to generate only DDL for views and their associated
triggers.

Revision 2.378  2019/04/29 23:16:42  art
Added new option, --synonym-file

Revision 2.377  2019/04/23 18:24:38  art
Documented --no-proc-cleaning, improved handling of comments embedded in the proc
and function create lines.

Revision 2.376  2019/04/09 20:03:37  art
Improved stored procedure handling. Stripping out IF NOT EXISTS from the procedure
source (apparently the engine saves that. New options --no-proc-cleaning to disable
the scrubbing and rewriting of the CREATE PROCEDURE/FUNCTION statement line because
that sometimes causes issues.

Revision 2.375  2019/04/08 13:47:24  art
Added AS "owner" to REVOKE statements.
Added the owner of the referenced table in FOREIGN KEY constraints.

Revision 2.374  2019/04/03 02:47:27  art
Fixed -F references printout.

Revision 2.373  2019/03/27 17:11:26  art
Changed infrastructuer printout to not generate code to set plog dbspaces extendable
as they are extendable by default.

Revision 2.372  2019/03/26 22:01:51  art
Updated copyright statement.

Revision 2.371  2019/03/17 15:13:12  art
Added new options --replace-views and --replace-triggers to support the CREATE OR REPLACE 
feature in v14.10.

Revision 2.370  2019/03/15 14:48:19  art
Added support for CREATE OR REPLACE routines with --replace-procs option (v14.10 feature).

Revision 2.369  2019/02/06 14:16:54  art
Fixed problem with RENAME INDEX statements that include the owner.

Revision 2.368  2018/12/31 22:04:29  art
Making changes to get myschema to compile/link under GCC v7+

Revision 2.367  2018/12/31 21:18:53  art
Making minor code changes to make myschema compatible with GCC version later than 4.8.

Revision 2.366  2018/12/27 15:06:07  art
Fixed a problem with disabled triggers.

Revision 2.365  2018/11/07 15:23:44  art
Added support for the new --rename-auto-index feature to generate RENAME
INDEX statements for auto-named indexes (ie constraint indexes).

Revision 2.364  2018/07/16 13:27:54  art
Added the ability to specify a file containing the password either with
the --passwd parameter or using the $PASSWD environment variable.

Revision 2.363  2018/06/22 16:30:05  art
New options: --use-database-owner, --use-table-owner, --constraint-file
Fixed some problems with the partial schema file output.
Expanded "owner"s to all objects.

Revision 2.362  2018/02/26 21:58:29  art
Changed print_usersuths.ec to not use the same global string as myschema.ec's a1 cursor.

Revision 2.361  2018/02/20 16:47:46  art
Fixed RANGE/INTERVAL output to supress automatic partitions (not needed).
Removed extent and lock mode clauses from virtual table definitions.
Changed to generate INSTEAD of triggers before GRANT statements for 
VIEWs to eliminate errors at create time from grants to multi-table
views.

Revision 2.360  2018/01/04 17:02:07  art
Updated copyright. Fixed handling of PLOG dbspaces.

Revision 2.359  2017/10/24 10:31:28  art
Fixed BIGSERIAL and SERIAL8 current value output on 32bit compilers.

Revision 2.358  2017/07/11 22:54:12  artka
Fixed a but causing myschema to hang on exit on some AIX versions.

Revision 2.357  2017/07/09 16:41:47  artka
Fixed behavior under --myexport-express run against v11.xx databases. Was not 
generating named partitions when appropriate under these conditions.

Revision 2.356  2017/05/24 18:19:39  art
Fixed problems with --primary, --secondary and the two trailing filenames.
The long options now take precedence over the trailing names.

Revision 2.355  2017/05/24 18:04:41  art
Fixed problem if users uses the same path for multiple file options.

Revision 2.354  2017/05/24 16:33:18  art
Fixed runtime information display to include all output files.

Revision 2.353  2017/05/24 15:39:17  art
Added new options to separate objects into different files
by type: --procedure-file, --view-file, --index-file,
--trigger-file. Also documented the --grant-file alternative
to the -U file option. Thanks to Paul Watson for triggering
the idea.

Revision 2.352  2017/05/03 21:09:21  art
Fixed handling of temp smartblob spaces.

Revision 2.351  2017/04/05 14:15:43  art
Fixed timeseries printout to eliminate the 'size' clause.

Revision 2.350  2017/02/26 19:16:45  art
Updated copyright.

*/

#if defined( _AIX )  /* Needed as AIX defines a locale control structure loc_t */
#if !defined( _H_LOCALEDEF )
#define _H_LOCALEDEF
#endif
#endif

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <errno.h>
#include <fnmatch.h>
#include "getopt.h"

EXEC SQL INCLUDE sqlhdr;
EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;
EXEC SQL INCLUDE datetime;
EXEC SQL INCLUDE locator;

EXEC SQL DEFINE MAINCODE 1;
#define MAINCODE 1

EXEC SQL INCLUDE myschema.eh;

EXEC SQL INCLUDE prototypes.h;

static char rcs_header[] = "$Header: /home/art/GoogleDrive/Latest/myschema.d/RCS/myschema.ec,v 2.440 2023/11/15 19:07:03 art Exp $";
#define Version "6.32"
#define Revision "$Revision: 2.440 $"

#if ! (defined( __DGUX__ ) & defined( __ix86 ))
extern int optind;
extern char *optarg;
#endif

#if !defined( FALSE )
#define FALSE 0
#endif
#if !defined( TRUE )
#define TRUE (!FALSE)
#endif

#if defined( __GNUC__ )
#define inline_or_defs "inlines"
static inline __attribute__((always_inline)) int max( int a, int b)
{
    return (a > b ? a : b);
}
static inline __attribute__((always_inline)) void strlower( char *target, char *source, int max )
{
    while (max-- && *source) {
        *target++ = tolower(*source++);
    }
    *target = 0;
}
#else
#define inline_or_defs "defines"
#define max(a,b) ((a) > (b) ? (a) : (b) )
#define strlower( t, s, m ) {      char *target = (t), *source = (s); int max = (m);      while (max-- && *source) {          *target++ = tolower(*source++);      }      *target = 0;  }
#endif


/* MACRO */
#define limit_unit( a ) (a == 1 ? "KB" : (a == 1024 ? "MB" : "GB" ))
#define disposition( a ) (a == 8 ? "DISCARD" : "DETACH" )
#define which_disp( a ) ((((a >> 6) % 2) == 1) ? "INTERVAL ONLY" : ((((a >> 5) % 2) == 1) ? "ANY" : "INTERVAL FIRST"))

#define MINEXT (4 * system_pagesize)

/* LOCAL PROTOTYPES */
int read_pwd( char *pwd_param );

/* GLOBALS, just a few. */
FILE *outfileMI = (FILE *)NULL;
FILE *outfileME = (FILE *)NULL;
int SMICounts = FALSE, grantas = TRUE, InHerFlag, CollectionFlag;
int QuoteIdent = 0;
int wrap_at = 80;  /* Only needed in format_triggers.ec & set here and in getopt loop. */
int squish_trig = 0;  /* Only needed in format_triggers.ec for now. */
int doDebug = 0;
int NoAuditTabs = 0;
int DatabaseConnected = 0;
int AlterChecks = FALSE;
typedef struct replflags_s {
    int replcheck_flag;
    int erkey_flag;
    int vercols_flag;
} replflags_t;
replflags_t replflags;

int main(int argc, char *argv[])
{
    static int pass_no = 1, def_open = 0, type_ret = 0, hold_ret = 0;
    int c, errflag=0, dflag=0, tflag=0, hflag=0, spflag=0, nrecs = 0;
    int tabauths = TRUE, colauths = TRUE, revauths = TRUE, TypesOnly = FALSE;
    int pc = 0, version = FALSE, doall = TRUE, dorefs = FALSE, setexplain = FALSE;
    int indexlength, summary = FALSE, doreorg = 0, last_record = 0, doaggs = FALSE;
    int doViolations = TRUE, dominimize=0, NoTypes = FALSE, NoCasts = FALSE;
    int NoSequences = FALSE, NoProcs = FALSE;
    int SimpleFrags = FALSE, PrintParts = TRUE, AuthsOnly = FALSE;
    int gotfile1 = FALSE, gotfile2 = FALSE;
    int gotusername = FALSE, gotpassword = FALSE;
    int forceautostats = FALSE, do_infrastructure = FALSE, infra_out = 0;
    int do_myexport_express = FALSE, do_myexport_scripts = FALSE;
    int do_binary = FALSE;
    int do_export = FALSE, do_import = FALSE;
    int suppress_directives = 0, directives_only = 0, minext = 0, minnext = 0;
    int ViewsOnly = FALSE, No_Triggers = FALSE;
    float NoExtentClause = FALSE, AdjNext = 0.0, AdjExt = 0.0;
    int loadflag = FALSE, ix, tempint, fe, CrColFlag, AuditFlag;
    FILE *outfile = stdout, *outfile2 = stdout, *outfile3 = stdout, *outfileG = stdout;
    FILE *ProcFile = stdout, *ViewFile = stdout, *IndexFile = stdout, *TrigFile = stdout;
    FILE *ConstrFile = stdout, *SynonymFile;
    int haveProcFile = 0, haveViewFile = 0, haveIndexFile = 0, haveTrigFile = 0, haveConstrFile = 0;
    int haveSynonymFile = 0, OK;
    char ProcFileName[1025], ViewFileName[1025], IndexFileName[1025], TrigFileName[1025];
    char ConstrFileName[1025], SynonymFileName[1025], prev_owner[33];
    char revision[256], ptempl[1026];
    char filename[1025], filename2[1025], filename3[1025], filenameG[1025], filenameME[1025], filenameMI[1025];
    char haveGrantFile = 0, haveDistFile = 0;
    static char tempstr[1024], ahost[129], argcpy[257], *Flags;
    char HowName[3][20] = { "smallest", "average", "largest" };
    char myexport_isolation[101], AuditMatch[129];
    char myexport_delimiter[2] = "|", myexport_escape[1025] = "\\";
    int haveHowExtents = FALSE, RawFlag = 0x00000008;
    char *at;
    char no_data_tables[10240], no_data_accesses[10240];
    char sel_trig_list[13040] = ""; /* Room for at least 100 trigger names. */
    int tables_data_OK = TRUE, access_data_OK = TRUE;
    char am_match[150];
    char tbl_match[150];
    int export_OK;
    int allows_nulls;
    int prev_colno;
    int Have_c = 0, Have_i = -1, Have_I = FALSE;
    struct option longopts[] = {
        { "newopt", required_argument, (int *)NULL, 'N' },
        { "host", required_argument, (int *)NULL, 'h' },
        { "database", required_argument, (int *)NULL, 'd' },
        { "table", required_argument, (int *)NULL, 't' },
        { "distributions-file", optional_argument, (int *)NULL, 'U' },
        { "primary", required_argument, (int *)NULL, 0x100 },
        { "secondary", required_argument, (int *)NULL, 0x101 },
        { "user", required_argument, (int *)NULL, 0x102 },
        { "passwd", required_argument, (int *)NULL, 0x103 },
        { "filtering", no_argument, (int *)NULL, 0x104 },
        { "simple-fragments", no_argument, (int *)NULL, 0x105 },
        { "grantfile", required_argument, (int *)NULL, 0x106 },
	{ "no-extent-clause", no_argument, (int *)NULL, 0x107 },
	{ "set-owner", required_argument, (int *)NULL, 0x108 },
	{ "no-ags-procs", no_argument, (int *)NULL, 0x109 },
	{ "sequence", optional_argument, (int *)NULL, 0x110 },
	{ "wrap-at", required_argument, (int *)NULL, 0x111 },
	{ "create-conditionally", no_argument, (int *)NULL, 0x112 },
	{ "suppress-directives", no_argument, (int *)NULL, 0x113 },
	{ "directives-only", no_argument, (int *) NULL, 0x114 },
	{ "myexport-scripts", no_argument, (int *) NULL, 0x115 },
	{ "myexport-express", no_argument, (int *) NULL, 0x116 },
	{ "reorg-api", no_argument, (int *) NULL, 0x117 },
	{ "no-blade-funcs", no_argument, (int *) NULL, 0x118 },
	{ "no-blade-procs", no_argument, (int *) NULL, 0x118 },
	{ "force-autostats", no_argument, (int *) NULL, 0x119 },
	{ "force_autostats", no_argument, (int *) NULL, 0x119 },
	{ "myexport-escape", required_argument, (int *)NULL, 0x11a }, 
	{ "export-escape", required_argument, (int *)NULL, 0x11a }, 
	{ "import-escape", required_argument, (int *)NULL, 0x11a }, 
        { "grant-file", required_argument, (int *)NULL, 0x11b },
	{ "infrastructure", optional_argument, (int *)NULL, 0x120 },
	{ "drop-procs", no_argument, (int *)NULL, 0x121 },
	{ "index-online", no_argument, (int *)NULL, 0x122 },
	{ "squish-triggers", no_argument, (int *)NULL, 0x123 },
	{ "no-storage-options", no_argument, (int *)NULL, 0x124 },
	{ "dependency-order", no_argument, (int *)NULL, 0x125 },
	{ "debug", no_argument, (int *)NULL, 0x126 },
	{ "no-audit-tabs", required_argument, (int *)NULL, 0x127 },
	{ "myexport-delimiter", required_argument, (int *)NULL, 0x128 },
	{ "export-delimiter", required_argument, (int *)NULL, 0x128 },
	{ "import-delimiter", required_argument, (int *)NULL, 0x128 },
	{ "no-validate", no_argument, (int *)NULL, 0x129 },
	{ "aggregates", no_argument, (int *)NULL, 0x12a },
	{ "set-min-extent", required_argument, (int *)NULL, 0x12b },
	{ "set-min-nextent", required_argument, (int *)NULL, 0x12c },
	{ "keep-generated-names", no_argument, (int *)NULL, 0x12d },
	{ "no-check", no_argument, (int *)NULL, 0x12e },
	{ "procedure-file", required_argument, (int *)NULL, 0x12f },
	{ "view-file", required_argument, (int *)NULL, 0x130 },
	{ "index-file", required_argument, (int *)NULL, 0x131 },
	{ "trigger-file", required_argument, (int *)NULL, 0x132 },
	{ "constraint-file", required_argument, (int *)NULL, 0x133 },
	{ "use-table-owner", no_argument, (int *)NULL, 0x134 },
	{ "use-database-owner", no_argument, (int *)NULL, 0x135 },
	{ "rename-auto-index", required_argument, (int *)NULL, 0x136 },
	{ "replace-procs", no_argument, (int *)NULL, 0x137 },
	{ "replace-views", no_argument, (int *)NULL, 0x138 },
	{ "replace-triggers", no_argument, (int *)NULL, 0x139 },
	{ "no-proc-cleaning", no_argument, (int *)NULL, 0x140 },
	{ "synonym-file", required_argument, (int *)NULL, 0x141 },
	{ "views-only", no_argument, (int *)NULL, 0x142 },
	{ "help", no_argument, (int *)NULL, 0x143 },
	{ "no-private-synonyms", no_argument, (int *)NULL, 0x144 },
	{ "alter-checks", no_argument, (int *)NULL, 0x145 },
	{ "no-setsessionauth", no_argument, (int *)NULL, 0x146 },
	{ "comment-setsessionauth", no_argument, (int *)NULL, 0x147 },
	{ "export-scripts", no_argument, (int *) NULL, 0x148 },
	{ "import-scripts", no_argument, (int *) NULL, 0x149 },
	{ "for-er", no_argument, (int *) NULL, 0x150 },
	{ "no-data-tables", optional_argument, (int *)NULL, 0x151 },
	{ "no-data-access-method", optional_argument, (int *)NULL, 0x152 },
	{ "no-export-triggers", no_argument, (int *)NULL, 0x153 },
	{ "myexport-order", no_argument, (int *)NULL, 0x154 },
	{ "myexport-binary", no_argument, (int *)NULL, 0x155 },
	{ "export-binary", no_argument, (int *) NULL, 0x156 },
	{ "import-binary", no_argument, (int *) NULL, 0x157 },
	{ "include-sysprocs", no_argument, (int *) NULL, 0x158 },
        { (char *)NULL, no_argument, (int *)NULL, -1 }
    };
    int longint;

    EXEC SQL BEGIN DECLARE SECTION;
    char username[1024];
    char passwd[4097];
    int nspaces;
    int pagesize, system_pagesize;
    string serv_type[IDBYTES], level_vers[IDBYTES], taborder[2];
    int vers;
    int real_ncols;
    string dbstmp[129], tmpstr[1025], trgname[129];
    string syssecpolicy[129], prev_syssecpolicy[129];
    char seltr_txt[500];
    EXEC SQL END DECLARE SECTION;

    char itabname[129];
    int HaveIRename = 0;
    int NoSetSessionAuth = 0;
    
    prev.tabname[0] = (char)0;
    strcpy( tablename, "*" );
    host[0] = (char)0;
    filename3[0] = (char)0;
    filenameG[0] = (char)0;
    filenameMI[0] = (char)0;
    filenameME[0] = (char)0;
    sequence_str[0] = (char)0;

    if (getenv( "DELIMIDENT" ) != (char *)0)
        QuoteIdent = 1;

    /* Parse the environment for user defined default flags. */
    Flags = getenv( "MYSCHEMA" );
    while (Flags && *Flags) {
        switch (*Flags) {
        case 'A':
            grantas = !FALSE;
            break;
        case 'O':
            stripowner = TRUE;
            grantas = FALSE;
            break;
        case 'C':
            colauths = FALSE;
            break;
        case 'E':
            setexplain = TRUE;
            break;
        case 'F':
            dorefs = TRUE;
            break;
        case 'I':
            Have_I = !Have_I;
            break;
	case 'K':
	    LongKeyIndices = TRUE;
	    break;
        case 'k':
            NoKeyIndices = TRUE;
            break;
        case 'L':
            doViolations = FALSE;
            break;
        case 'R':
            revauths = FALSE;
            break;
        case 'S':
            ids_server = FALSE;
            break;
        case 'T':
            tabauths = FALSE;
            break;
        case 'v':
            verbose = TRUE;
            break;
        case 'q':
            silent = TRUE;
            break;
        case 'c':
            Have_c++;
            break;
        case 'a':
            SMICounts = TRUE;
            break;
        case 'o':
            doOrder = NameOrder;
	    if (doTabOrder == IdOrder)
	      doTabOrder = NameOrder;
            break;
        case 's':
            summary = TRUE;
            break;
        case 'l':
            loadflag = TRUE;
            break;
        case 'm':
            dominimize = TRUE;
            break;
        case '5':
            conn5 = TRUE;
            TBSFACT=0x1000000;
            break;
        default:
            fprintf( stderr,
                     "Invalid flag in $MYSCHEMA: %c.  Continuing.\n",
                     *Flags );
            break;
        }
        Flags++;
    }

    while ((c = getopt_long(argc,
                            argv,
                            "G:M:N:U::d:e:f:g:h:i:n:t:5ACDEFIKLORSTVacklmoqrsuv::?",
                            longopts,
                            &longint )) != EOF)
    {
        switch (c) {
        case 'N':
            /* catchall for 'N'ew commands/options. */
            if (strcmp( optarg, "notypes") == 0) {
                NoTypes = !NoTypes;
            } else if (strcmp( optarg, "nocasts") == 0) {
                NoCasts = !NoCasts;
            }
            break;
        case 'D':
            TypesOnly = !TypesOnly;
            doall = FALSE;
            break;
        case 'M':
            strtoupper( argcpy, optarg );
            if (strcmp( argcpy, "MAX" ) == 0)
                HowExtents = ES_Maximum;
            else if (strcmp( argcpy, "MIN" ) == 0)
                HowExtents = ES_Minimum;
            else if (strcmp( argcpy, "AVG" ) == 0)
                HowExtents = ES_Average;
            else {
                fprintf( stderr, "Invalid extent calculation method (min/avg/max)!\n" );
                rexit(1);
            }
            haveHowExtents = TRUE;
            break;
        case '5':
            conn5 = !conn5;
            TBSFACT=(conn5 ? 0x1000000 : 0x100000);
            break;
        case 'A':
            grantas = !grantas;
            break;
        case 'c':
            Have_c++;
            break;
        case 'C':
            colauths = !colauths;
            break;
        case 'E':
            setexplain = !setexplain;
            break;
        case 'F':
            dorefs = !dorefs;
            break;
        case 'i':
            Have_i = atoi( optarg );
            if (Have_i <= 0) {
                usage( argv[0] );
                rexit(1);
            }
            break;
        case 'I':
            Have_I = !Have_I;
            break;
        case 'G':
            AuthsOnly = TRUE;
            doall = FALSE;
            tflag = FALSE;
            dorefs = FALSE;
            spflag = FALSE;
        case 'g':
        case 0x106:
	case 0x11b:
            if (optarg != (char *)NULL) {
                strcpy( filenameG, optarg );
		haveGrantFile = 1;
            }
            break;
        case 'U':
            if (optarg != (char *)NULL) {
                strcpy( filename3, optarg );
		haveDistFile = 1;
            }
        case 'u':
            doDistrib = TRUE;
            break;
	case 'K':
	    LongKeyIndices = !LongKeyIndices;
	    break;
        case 'k':
            NoKeyIndices = !NoKeyIndices;
            break;
        case 'L':
            doViolations = !doViolations;
            break;
        case 'O':
            stripowner = !stripowner;
            grantas = !stripowner;
            break;
        case 'o':
	  if (doOrder == IdOrder) {
                doOrder = NameOrder;
		if (doTabOrder == IdOrder)
		  doTabOrder = NameOrder;
	  } else {
                doOrder = IdOrder;
		if (doTabOrder == NameOrder)
		  doTabOrder = IdOrder;
	  }
            break;
        case 'R':
            revauths = !revauths;
            break;
        case 'S':
            ids_server = !ids_server;
            break;
        case 'T':
            tabauths = !tabauths;
            break;
        case 'V':
            version = TRUE;
            break;
        case 's':
            summary = !summary;
            break;
        case 'v':
            if (optarg == (char *)NULL)
                verbose = !verbose;
            else
                verbose = atoi( optarg );
            break;
        case 'q':
            silent = !silent;
            break;
        case 'r':
            doreorg = 1;
            break;
        case 'a':
            SMICounts = !SMICounts;
            break;
        case 'l':
            loadflag = !loadflag;
            break;
        case 'm':
            dominimize = !dominimize;
            break;
        case 'e':
            tempint = atoi( optarg );
            if (tempint < 0 || tempint > 10000) {
                fprintf( stderr,
                         "EXTENT SIZE adjustment must be >=0 & <=10000.\n" );
                rexit(1);
            }
            AdjExt = ((float)tempint / 100.0);
            break;
        case 'n':
            tempint = atoi( optarg );
            if (tempint < 0 || tempint > 10000) {
                fprintf( stderr,
                         "NEXT SIZE adjustment must be >=0 & <=10000.\n" );
                rexit(1);
            }
            AdjNext = ((float)tempint / 100.0);
            break;
        case 'f':
            if (spflag)
                errflag++;
            else {
                spflag = TRUE;
                if (QuoteIdent) {
                    strncpy( ptempl, optarg, 1025 );
                    ptempl[1025] = (char)0;
                } else
                    strlower( ptempl, optarg, 1025 );
                ptempl[1025] = (char)0;
                if (strcmp( ptempl, "all" ) == 0)
                    strcpy( ptempl, "*" );
                doall = FALSE;
            }
            break;
        case 'd':
            if (dflag)
                errflag++;
            else{
                dflag++;
                if (QuoteIdent) {
                    strncpy( databasename, optarg, 128 );
                    databasename[128] = (char)0;
                } else
                    strlower( databasename, optarg, 128 );
            }
            break;
        case 't':
            if (tflag)
                errflag++;
            else{
                tflag = TRUE;
                if (QuoteIdent) {
                    strncpy( tablename, optarg, 128 );
                    tablename[128] = (char)0;
                } else
                    strlower( tablename, optarg, 128 );
                if (strcmp( tablename, "all" ) == 0)
                    strcpy( tablename, "*" );
                doall = FALSE;
            }
            break;
        case 'h':
            if (hflag)
                errflag++;
            else{
                hflag++;
                strncpy(ahost, optarg, 128);
                ahost[128] = (char)0;
            }
            break;
        case 0x100:
            /* Primary schema filename. */
            strcpy( filename, optarg );
            gotfile1 = TRUE;
            break;
        case 0x101:
            /* Secondary schema filename. */
            strcpy( filename2, optarg );
            gotfile2 = TRUE;
            break;
        case 0x102:
            /* User name for secure connection. */
            strcpy( username, optarg );
            gotusername = TRUE;
            break;
        case 0x103:
            /* Password for secure connection. */
            strcpy( passwd, optarg );
            gotpassword = TRUE;
            break;
        case 0x104:
            /* Enable record filtering for loads. */
            do_filtering = 1;
            break;
        case 0x105:
            /* Disable named partitions where implicitely named. */
            SimpleFrags = TRUE;
            break;
	case 0x107:
	    /* Eliminate EXTENT SIZE and NEXT EXTENT clauses. */
	    NoExtentClause = TRUE;
	    break;
	case 0x108:
	    /* Override all owners with the supplied name. */
	    UseSetOwner = TRUE;
	    replace_owners = TRUE;
	    snprintf( owner_override, sizeof owner_override, "%s", optarg );
	    break;
	case 0x109:
	    no_ags_procs = TRUE;
	    break;
	case 0x110:
	    sequence_only = TRUE;
	    if (optarg != (char *)0)
	      snprintf( sequence_str , sizeof sequence_str, "%s", optarg );
	    break;
	case 0x111:
	    wrap_at = atoi( optarg );
	    if (wrap_at == -1)
	        wrap_at = INT_MAX;
	    else if (wrap_at < 40)
	        wrap_at = 40;
	    break;
	case 0x112:
	    create_conditionally = !create_conditionally;
	    break;
	case 0x113:
	    suppress_directives = 1;
	    directives_only = 0;
	    break;
	case 0x114:
	    directives_only = 1;
	    suppress_directives = 0;
	    break;
	case 0x115:
	    do_myexport_scripts = TRUE;
	    do_myexport_express = FALSE;
	    break;
	case 0x116:
	    do_myexport_scripts = TRUE;
	    do_myexport_express = TRUE;
	    break;
        case 0x117:
            doreorg = 2;
            break;
	case 0x118:
	    no_blade_funcs = 1;
	    break;
	case 0x119:
	    forceautostats = !forceautostats;
	    break;
	case 0x11a:
	    snprintf( myexport_escape, sizeof myexport_escape, "%s", optarg );
	    break;
	case 0x120:
	    do_infrastructure = 1;
	    if (optarg != (char *)NULL) {
		if (strcmp( optarg, "sql" ) == 0)
		    infra_out = 1;
		else if (strcmp( optarg, "cmd" ) == 0)
		    infra_out = 0;
		else {
		    fprintf( stderr, "Invalid command type (%s)!\n", optarg );
		    usage( argv[0] );
		    exit(1);
		}
	    }
	    break;
	case 0x121:
	    do_drop_proc = 1;
	    break;
	case 0x122:
	    do_index_online = 1;
	    break;
	case 0x123:
	    squish_trig = 1;
	    break;
	case 0x124:
	    no_storage_opt = TRUE;
	    break;
	case 0x125:
	    doTabOrder = DependOrder;
	    break;
	case 0x126:
	    doDebug = 1;
	    break;
	case 0x127:
	    NoAuditTabs = 1;
	    snprintf( AuditMatch, sizeof AuditMatch, "%s", optarg );
	    break;
	case 0x128:
	    snprintf( myexport_delimiter, sizeof myexport_delimiter, "%s", optarg );
	    break;
	case 0x129:
	    NoValidateFK = 1;
	    break;
	case 0x12a:
	    doaggs = TRUE;
	    doall = FALSE;
	    break;
	case 0x12b:
	    minext = max( atoi( optarg ), minext );
	    break;
	case 0x12c:
	    minnext = max( atoi( optarg ), minnext );
	    break;
	case 0x12d:
	    KeepGenNames = 1;
	    break;
	case 0x12e:
	    NoValidateChk = 1;
	    break;
	case 0x12f:
	    haveProcFile = 1;
	    snprintf( ProcFileName, sizeof ProcFileName, "%s", optarg );
	    break;
	case 0x130:
	    haveViewFile = 1;
	    snprintf( ViewFileName, sizeof ViewFileName, "%s", optarg );
	    break;
	case 0x131:
	    haveIndexFile = 1;
	    snprintf( IndexFileName, sizeof IndexFileName, "%s", optarg );
	    break;
	case 0x132:
	    haveTrigFile = 1;
	    snprintf( TrigFileName, sizeof TrigFileName, "%s", optarg );
	    break;
	case 0x133:
	    haveConstrFile = 1;
	    snprintf( ConstrFileName, sizeof ConstrFileName, "%s", optarg );
	    break;
	case 0x134:
	    UseTableOwner = TRUE;
	    replace_owners = TRUE;
	    break;
	case 0x135:
	    UseDatabaseOwner = TRUE;
	    replace_owners = TRUE;
	    break;
	case 0x136:
	  if (QuoteIdent) {
	    strncpy( itabname, optarg, 128 );
	    tablename[128] = (char)0;
	  } else
	    strlower( itabname, optarg, 128 );
	  if (strcmp( itabname, "all" ) == 0)
	    strcpy( itabname, "*" );
	  HaveIRename = 1;
	  break;
	case 0x137:
	  Replace_Procs = TRUE;
	  break;
	case 0x138:
	  Replace_Views = TRUE;
	  break;
	case 0x139:
	  Replace_Triggers = TRUE;
	  break;
	case 0x140:
	  NoProcClean = TRUE;
	  break;
	case 0x141:
	    haveSynonymFile = 1;
	    snprintf( SynonymFileName, sizeof SynonymFileName, "%s", optarg );
	    break;
	case 0x142:
	    ViewsOnly = TRUE;
	    NoCasts = TRUE;
	    NoTypes = TRUE;
	    NoSequences = TRUE;
	    NoProcs = TRUE;
	    break;
	case 0x143:
	    usage( argv[0] );
	    exit( 0 );
	    break;
	case 0x144:
	    NoPrivSyns = TRUE;
	    break;
	case 0x145:
	    AlterChecks = TRUE;
	    break;
	case 0x146:
	    NoSetSessionAuth = 1;
	    break;
	case 0x147:
	    NoSetSessionAuth = 2;
	    break;
	case 0x156:
	    do_binary = TRUE;
	    // Intentional fallthrough to 0x148
	case 0x148:
	    do_export = TRUE;
	    break;
	case 0x157:
	    do_binary = TRUE;
	    // Intentional fallthrough to 0x149
	case 0x149:
	    do_import = TRUE;
	    break;
	case 0x150:
	    do_For_ER = TRUE;
	    break;
	case 0x151:
            if (optarg != (char *)NULL) {
		snprintf( no_data_tables, sizeof no_data_tables, ",%s,", optarg );
	    } else {
		strcpy( no_data_tables, ",tsinstancetable,");
	    }
	    tables_data_OK = FALSE;
	    break;
	case 0x152:
            if (optarg != (char *)NULL) {
		snprintf( no_data_accesses, sizeof no_data_accesses, ",%s,", optarg );
	    } else {
		strcpy( no_data_accesses, ",ts_rtl_vtam,ts_vtam," );
	    }
	    access_data_OK = FALSE;
	    break;
	case 0x153:
	    No_Triggers = TRUE;
	    break;
	case 0x154:
	    doTabOrder = MyexportOrder;
	    break;
	case 0x155:
	    do_myexport_scripts = TRUE;
	    do_myexport_express = TRUE;
	    do_binary = TRUE;
	    break;
	case 0x158:
	    include_sysprocs = TRUE;
	    break;
        case '?':
        default:
            errflag++;
            break;
        }
    }

    if (version) {
        strcpy( revision, Revision );
        printf( "\n%s: Features Version %s, Source %s!\n",
                argv[0],
                Version,
                strtok( revision, "$" ) );
        printf( "\tCopyright 1992-2022 by Art S. Kagel.\n"
                "\tLicense for private usage (including use by "
                "commercial enterprise \n"
                "\tfor its own needs) is granted.  "
                "Commercial and exploitation rights \n"
                "\tare reserved.\n\n" );
        rexit(0);
    }

    /* Check for consistent arguments.  Make them consistent where
       reasonable. */

    if (tflag == TRUE && HaveIRename == 1) {
        fprintf( stderr, "-t & --rename-auto-index are incompatible options!\n" );
        exit( 1 );
    }
    
    if (haveHowExtents)
        SMICounts = TRUE;

    if (dominimize)
        SMICounts = TRUE;

    if (AdjExt == 0.0)
        AdjExt = 1.0;

    if (SMICounts && conn5) {
        fprintf( stderr,
                 "-a and -5 are incompatible options, no SMI database in "
                 "version 5.xx servers\n" );
        fprintf( stderr, "Ignoring -a flag.\n" );
        SMICounts = FALSE;
    }

    if (UseSetOwner && stripowner) {
        fprintf( stderr,
                 "-O and --set-owner are incompatible options.\n" );
        fprintf( stderr, "Ignoring -O flag.\n" );
        stripowner = FALSE;
    }

    if (stripowner && UseTableOwner) {
        fprintf( stderr,
                 "-O and --use-table-owner are incompatible options.\n" );
        fprintf( stderr, "Ignoring -O flag.\n" );
        stripowner = FALSE;
    }

    if (UseSetOwner && UseTableOwner) {
        fprintf( stderr,
                 "--use-table-owner and --set-owner are incompatible options.\n" );
        fprintf( stderr, "Ignoring --use-table-owner flag.\n" );
        UseTableOwner = FALSE;
    }

    if (UseSetOwner && UseDatabaseOwner) {
        fprintf( stderr,
                 "--use-database-owner and --set-owner are incompatible options.\n" );
        fprintf( stderr, "Ignoring --use-table-owner flag.\n" );
        UseTableOwner = FALSE;
    }

    if (UseDatabaseOwner && UseTableOwner) {
        fprintf( stderr,
                 "--use-table-owner and --use-database-owner are incompatible options.\n" );
        fprintf( stderr, "Ignoring --use-table-owner flag.\n" );
        UseTableOwner = FALSE;
    }

    if (SMICounts && !ids_server) {
        fprintf( stderr,
                 "-a and -S are incompatible options, no SMI database in "
                 "Standard Engine servers\n" );
        fprintf( stderr, "Ignoring -a flag.\n" );
        SMICounts = FALSE;
    }

    if ((Have_c > 0 && (Have_i >= 0 || Have_I)) || (Have_I && Have_i >= 0)) {
        fprintf( stderr,
                "-c, -i, & -I are mutually exclusive!\n" );
        usage( argv[0] );
        rexit(1);
    }

    if (AdjNext != 0.0 && !ids_server) {
        fprintf( stderr,
                 "-n is incompatible with -S, SE servers do not support "
                 "extents.\n" );
        fprintf( stderr, "Ignoring -n flag and it's option.\n" );
        AdjNext = 0.0;
    }

    if (AdjExt != 1.0 && !ids_server) {
        fprintf( stderr,
                 "-e is incompatible with -S, SE servers do not support "
                 "extents.\n" );
        fprintf( stderr, "Ignoring -e flag and it's option.\n" );
        AdjExt = 1.0;
    }

    if ((errflag || dflag == 0) && do_infrastructure == 0) {
	fprintf( stderr, "No database specified.\n" );
        usage( argv[0] );
        rexit( 1 );
    }

    if (tflag && spflag && sequence_only) {
        fprintf( stderr, "--sequence, -t, and -p are mutually exclusive!\n" );
        usage( argv[0] );
        rexit(1);
    }

    if (tflag && spflag) {
        fprintf( stderr, "-t and -p are mutually exclusive!\n" );
        usage( argv[0] );
        rexit(1);
    }

    if (tflag && sequence_only) {
        fprintf( stderr, "-t and -sequence are mutually exclusive!\n" );
        usage( argv[0] );
        rexit(1);
    }

    if (sequence_only && spflag) {
        fprintf( stderr, "--sequence and -p are mutually exclusive!\n" );
        usage( argv[0] );
        rexit(1);
    }

    if (TypesOnly && NoTypes) {
        /* Not sensible.  Just ignore NoTypes. */
        NoTypes = FALSE;
    }

    if (TypesOnly && tflag) {
        /* Not sensible.  Squawk. */
        fprintf( stderr,
                 "-D and -T are mutually exclusive!\n" );
        rexit(1);
    }

    if (TypesOnly && spflag) {
        /* Not sensible.  Just ignore TypesOnly. */
        TypesOnly = FALSE;
        NoTypes = FALSE;
    }

    if ((!ids_server || spflag || conn5 || dorefs) && doreorg) {
        /* Reorg does not make sense for procedures and 5.xx has no ALTER
           FRAGMENT statement. */
        fprintf( stderr,
                 "-r and -%s are mutually exclusive!\n",
                 ids_server ? (spflag ? "f" : (conn5 ? "5" : "F") ) : "S" );
        rexit(1);
    }

    if (Have_c > 0) {
	fprintf(stderr, "in Have_c. Have_c = %d.\n", Have_c );
        if (Have_c == 1) {
            init_Serial = Serial_NoCurrent;
        } else {
            init_Serial = Serial_Current;
	}
    } else if (Have_I) {
        init_Serial = Serial_Default;
    } else if (Have_i > 0) {
        init_Serial = Serial_Value;
        init_Serial_Value = Have_i;
    } else if (init_Serial == Serial_Unset) {
        init_Serial = Serial_Current;
    }

    if (doreorg && (summary || loadflag)) {
        /* Silly, reorgs have no documentation needs and cannot be for
           dbimport. */
        summary = FALSE;
        loadflag = FALSE;
    }

    if (dorefs && !tflag ) {
        /* Silly, references are already being done for a database level request. */
        dorefs = FALSE;
    }

    if (dorefs && spflag ) {
        /* Silly, references have nothing to do with Stored Procedures. */
        dorefs = FALSE;
    }

    /* Default file for basic SQL create DDL output. */
    if (gotfile1) {
        fflush( stdout );
        outfile = fopen( filename, "w" );
        if (!outfile) {
            fprintf( stderr, "Cannot open primary output: %s.\n", filename );
            rexit(3);
        }
	/* If processed --primary ignore first filenames on the command line. */
	if (optind < argc)
	    optind++;
    } else if (optind < argc && !gotfile1) {
        strcpy( filename, argv[optind] );
        fflush( stdout );
        outfile = fopen( filename, "w" );
        if (!outfile) {
            fprintf( stderr, "Cannot open primary output: %s.\n", filename );
            rexit(3);
        }
        optind++;
        gotfile1 = TRUE;
    } else if (!gotfile1) {
	strcpy( filename, "stdout" );
    }
    outfile3 = outfile2 = outfile;
    TrigFile = ProcFile = ViewFile = IndexFile = ConstrFile = SynonymFile = outfile;

    /* Default file for SQL index DDL output. */ 
    if (gotfile2 && strcmp( filename, filename2 ) != 0) {
        fflush( stdout );
        TrigFile = ProcFile = ViewFile = IndexFile = ConstrFile = SynonymFile = outfile3 = outfile2 = fopen( filename2, "w" );
        if (!outfile2) {
            fprintf( stderr,
                     "Cannot open secondary output file: %s.\n",
                     filename2 );
            rexit(3);
        }
	/* If processed --secondary ignore remaining filenames on the command line. */
	if (optind < argc)
	    optind++;
    } else if (optind < argc && !gotfile2) {
        strcpy( filename2, argv[optind] );
	if (strcmp( filename, filename2 ) != 0) {
            fflush( stdout );
            TrigFile = ProcFile = ViewFile = IndexFile = ConstrFile = SynonymFile = outfile3 = outfile2 = fopen( filename2, "w" );
	    if (!outfile2) {
		fprintf( stderr,
			 "Cannot open secondary output file: %s.\n",
			 filename2 );
		rexit(3);
	    }
	}
        optind++;
    } else if (!gotfile2) {
	strcpy( filename2, "stdout" );
    }

    /* File for update statistics commands. */
    if (haveDistFile == 1) {
        outfile3 = fopen( filename3, "w" );
        if (!outfile3) {
            fprintf( stderr,
                     "Cannot open distributions output file: %s.\n",
                     filename3 );
            rexit(3);
        }
    }

    /* Privileges file. */
    if (haveGrantFile == 1) {
	if (haveDistFile == 1 && strcmp( filenameG, filename3 ) == 0) {
	    outfileG = outfile3;
	} else {
	    outfileG = fopen( filenameG, "w" );
	    if (!outfileG) {
		fprintf( stderr,
			 "Cannot open grants output file: %s.\n",
			 filenameG );
		rexit(3);
	    }
	}
    } else 
        outfileG = outfile2;

    /* File for procedures and functions. */
    if (haveProcFile == 1) {
	if (haveDistFile == 1 && strcmp( ProcFileName, filename3) == 0) {
	    ProcFile = outfile3;
	} else  if (haveGrantFile ==1 && strcmp( ProcFileName, filenameG) == 0) {
	    ProcFile = outfileG;
	} else {
	    ProcFile = fopen( ProcFileName, "w" );
	    if (!ProcFile) {
		fprintf( stderr,
			 "Cannot open procedures output file: %s.\n",
			 ProcFileName );
		rexit(3);
	    }
	}
    }

    /* File for views. */
    if (haveViewFile == 1) {
	if (haveProcFile == 1 && strcmp( ViewFileName, ProcFileName) == 0) {
	    ViewFile = ProcFile;
	} else  if (haveGrantFile ==1 && strcmp( ViewFileName, filenameG) == 0) {
	    ViewFile = outfileG;
	} else  if (haveDistFile ==1 && strcmp( ViewFileName, filename3) == 0) {
	    ViewFile = outfile3;
	} else {
	    ViewFile = fopen( ViewFileName, "w" );
	    if (!ViewFile) {
		fprintf( stderr,
			 "Cannot open views output file: %s.\n",
			 ViewFileName );
		rexit(3);
	    }
	}
    }

    /* File for Synonyms. */
    if (haveSynonymFile == 1) {
	if (haveViewFile == 1 && strcmp( SynonymFileName, ViewFileName) == 0) {
	    SynonymFile = ViewFile;
	} else  if (haveGrantFile ==1 && strcmp( SynonymFileName, filenameG) == 0) {
	    SynonymFile = outfileG;
	} else  if (haveDistFile ==1 && strcmp( SynonymFileName, filename3) == 0) {
	    SynonymFile = outfile3;
	} else {
	    SynonymFile = fopen( SynonymFileName, "w" );
	    if (!SynonymFile) {
		fprintf( stderr,
			 "Cannot open synonyms output file: %s.\n",
			 SynonymFileName );
		rexit(3);
	    }
	}
    }

    /* File for indexes. */
    if (haveIndexFile == 1) {
	/* If file path is the same as another file, use that handle. */
	if (haveProcFile == 1 && strcmp( IndexFileName, ProcFileName) == 0) {
	    IndexFile = ProcFile;
	} else  if (haveViewFile ==1 && strcmp( IndexFileName, ViewFileName) == 0) {
	    IndexFile = ViewFile;
	} else  if (haveDistFile ==1 && strcmp( IndexFileName, filename3) == 0) {
	    IndexFile = outfile3;
	} else  if (haveTrigFile ==1 && strcmp( IndexFileName, filenameG) == 0) {
	    IndexFile = outfileG;
	} else {
	    IndexFile = fopen( IndexFileName, "w" );
	    if (!IndexFile) {
		fprintf( stderr,
			 "Cannot open index output file: %s.\n",
			 IndexFileName );
		rexit(3);
	    }
	}
    }

    /* File for Triggers. */
    if (haveTrigFile == 1) {
	/* If file path is the same as another file, use that handle. */
	if (haveProcFile == 1 && strcmp( TrigFileName, ProcFileName) == 0) {
	    TrigFile = ProcFile;
	} else  if (haveViewFile ==1 && strcmp( TrigFileName, ViewFileName) == 0) {
	    TrigFile = ViewFile;
	} else  if (haveIndexFile ==1 && strcmp( TrigFileName, IndexFileName) == 0) {
	    TrigFile = IndexFile;
	} else  if (haveDistFile ==1 && strcmp( TrigFileName, filename3) == 0) {
	    TrigFile = outfile3;
	} else  if (haveGrantFile ==1 && strcmp( TrigFileName, filenameG) == 0) {
	    TrigFile = outfileG;
	} else {
	    TrigFile = fopen( TrigFileName, "w" );
	    if (!TrigFile) {
		fprintf( stderr,
			 "Cannot open trigger output file: %s.\n",
			 TrigFileName );
		rexit(3);
	    }
	}
    }

    /* File for Constraints. */
    if (haveConstrFile == 1) {
	/* If file path is the same as another file, use that handle. */
	if (haveProcFile == 1 && strcmp( ConstrFileName, ProcFileName) == 0) {
	    ConstrFile = ProcFile;
	} else  if (haveViewFile ==1 && strcmp( ConstrFileName, ViewFileName) == 0) {
	    ConstrFile = ViewFile;
	} else  if (haveIndexFile ==1 && strcmp( ConstrFileName, IndexFileName) == 0) {
	    ConstrFile = IndexFile;
	} else  if (haveTrigFile ==1 && strcmp( ConstrFileName, TrigFileName) == 0) {
	    ConstrFile = TrigFile;
	} else  if (haveGrantFile ==1 && strcmp( ConstrFileName, filenameG) == 0) {
	    ConstrFile = outfileG;
	} else  if (haveDistFile ==1 && strcmp( ConstrFileName, filename3) == 0) {
	    ConstrFile = outfile3;
	} else {
	    ConstrFile = fopen( ConstrFileName, "w" );
	    if (!ConstrFile) {
		fprintf( stderr,
			 "Cannot open constraint output file: %s.\n",
			 ConstrFileName );
		rexit(3);
	    }
	}
    }

    if (gotpassword && !gotusername) {
        char *user;
        if (user = getenv( "USER" )) {
            strncpy( username, user, sizeof username );
        } else if (user = getenv( "LOGNAME" ))
            strncpy( username, user, 1024 );
        else
            username[0] = (char)0;

        if (strlen( username ) == 0) {
            fprintf( stdout,
                     "Password supplied without supplied or available username!\n" );
            exit(1);
        }
        gotusername = TRUE;
    }
    if (gotusername && !gotpassword) {
        char *pass;
        if (pass = getenv( "PASSWD" ))
            strncpy( passwd, pass, sizeof passwd );
        else
            passwd[0] = (char)0;
        if (strlen( passwd ) == 0) {
            fprintf( stdout,
                     "Username supplied without supplied or available password!\n" );
            exit(1);
        }
	gotpassword = TRUE;
    }

    if (gotpassword) {
        if (passwd[0] == '@') {
	    if (read_pwd( passwd ) == 1) {
	        gotpassword = TRUE;
	    } else {
	        gotpassword = FALSE;
		fprintf( stderr, "Password file invalid!.\n" );
		exit(1);
	    }
	}
    }

    if (!silent) {
        if (HaveIRename == 1) {
	    fprintf( stdout,
	             "Writing RENAME INDEX statements for auto-generated indexes.\n" );
	} else if (do_infrastructure)
	    fprintf( stdout,
		     "Writing %s to rebuild instance to:       %s\n",
		     infra_out == 0 ? "commands" : "API SQL",
		     filename );
        else if (outfile == outfile2 && !haveProcFile && !haveViewFile && !haveIndexFile && !haveTrigFile &&
		 outfileG == outfile2 && outfile3 == outfile3)
	{
            if (outfile != stdout
                || (outfile == stdout && isatty( fileno( outfile ) )))
            {
		fprintf( stdout,
			 "Writing full schema DDL to:       %s\n",
			 filename );
            }
        } else if (!haveProcFile && !haveViewFile && !haveIndexFile && !haveTrigFile &&
		 outfileG == outfile2 && outfile3 == outfile3)
        {
            fprintf( stdout,
                     "Writing schema primary DDL to:    %s\n",
                     filename );
            fprintf( stdout,
                     "Writing schema secondary DDL to:  %s\n",
                     filename2 );
        } else if (haveProcFile || haveViewFile || haveIndexFile || haveTrigFile || haveConstrFile || outfileG != outfile2 ||
		   outfile3 != outfile2)
	{
	    fprintf( stdout,
		     "Writing DDL for TABLES and other primary objects to:   %s\n",
		     filename );
	    if (haveProcFile)
		fprintf( stdout,
			 "Writing PROCEDURE and FUNCTION DDL to:   %s\n",
			 ProcFileName );
	    if (haveViewFile)
		fprintf( stdout,
			 "Writing VIEW and SYNONYM DDL to:   %s\n",
			 ViewFileName );
	    if (haveIndexFile)
		fprintf( stdout,
			 "Writing INDEX DDl to:   %s\n",
			 IndexFileName );
	    if (haveTrigFile)
		fprintf( stdout,
			 "Writing TRIGGER DDL to:   %s\n",
			 TrigFileName );
	    if (haveConstrFile)
		fprintf( stdout,
			 "Writing CONSTRAINT DDL to:   %s\n",
			 ConstrFileName );
	    if (outfile3 != outfile2) {
		fprintf( stdout,
			 "Writing UPDATE STATISTICS DDL to: %s\n",
			 filename3 );
	    }
	    if (outfileG != outfile2) {
		fprintf( stdout,
			 "Writing GRANT DDL to: %s\n",
			 filenameG );
	    }
	    fprintf( stdout,
		     "Writing DDL for other support objects to:   %s\n",
		     filename2 );
	}
    }

    if (doall) {
        /* If doing database level schema report all procedures. */
        strcpy( ptempl, "*" );
    }

    if (!do_infrastructure)
	if(strcmp(databasename,"")==0 || ( tflag == 1 && strcmp(tablename,"")==0)){
	    usage( argv[0] );
	    rexit( 1 );
	}
    
    if ((at=strchr( databasename, '@' )) != (char *)0) {
        if (ahost[0] == (char)0) {
            *at++ = (char)0;
            strncpy( ahost, at, 129 );
        } else {
            fprintf( stderr, "Host specified in -h and -d!  Choose one.\n" );
            usage( argv[0] );
            rexit( 1 );
        }
    }

    if (LongKeyIndices && KeepGenNames) {
	fprintf( stderr,
		 "-K & --keep-generated-names are mutually exclusive options!\n" );
	usage( argv[0] );
	exit(1);
    }

    if ((do_export || do_import) && (do_myexport_scripts)) {
	fprintf( stderr,
		 "--myexport-scripts, --myexport-express, --myexport-binary, --export-scripts, \n"
		 "  and --import-scripts are mutually exclusive. --export-scripts and \n"
		 "  --import-scripts can be used together, but not with the others. Select one.\n" );
	usage( argv[0] );
	exit(1);
    }


    if (ahost[0] == (char)0) {
        char *servptr = getenv( "INFORMIXSERVER" );
        if (servptr) {
            strncpy( ahost, servptr, 129 );
            ahost[128] = (char)0;
        }
    }
    if (conn5 != 0)
        sprintf( host, "%s@%s", delimident( databasename ), ahost );
    else
        sprintf( host, "@%s", ahost );

    /* Verbose printout of options here: */
    if (verbose) {
	if (NoAuditTabs != 0) {
	    fprintf( stderr, "Audit tables, triggers, and procedures ignored.\n" );
	    fflush( stderr );
	}
	if (NoPrivSyns == TRUE) {
	    fprintf( stderr, "Private synonyms ignored.\n" );
	}
        if (NoExtentClause != 0) {
	    fprintf( stderr, "Extent clauses deselected.\n" );
	} else {
	    fprintf( stderr,
		     "Extent Adjustment: %f, Next size Adjustment: %f%s.\n",
		     AdjExt,
		     AdjNext,
		     AdjNext == 0 ? " using current next size" : " times calculated extent" );
	    fprintf( stderr,
		     "%s SMI Page counts.\n",
		     SMICounts ? "Using" : "Not using" );
	    fprintf( stderr,
		     "Max() and strlower() included as %s.\n",
		     inline_or_defs );
	}
	switch (doTabOrder) {
	case NameOrder:
	    fprintf( stderr,
		     "Ordering tables by name.\n" );
	    break;
	case IdOrder:
	    fprintf( stderr,
		     "Ordering tables by tabid.\n" );
	    break;
	case DependOrder:
	    fprintf( stderr,
		     "Ordering tables by dependencies.\n" );
	    break;
	case MyexportOrder:
	    fprintf( stderr,
		     "Ordering tables by inverse dependencies.\n" );
	    break;
	}
	switch (doOrder) {
	case NameOrder:
	    fprintf( stderr,
		     "Ordering other objects by name.\n" );
	    break;
	case IdOrder:
	    fprintf( stderr,
		     "Ordering other objects by tabid.\n" );
	    break;
	case DependOrder:
	    fprintf( stderr,
		     "Ordering other objects by dependencies.\n" );
	    break;
	}
	if (strlen( no_data_tables ) > 0)
	    fprintf( stderr,
		     "Not creating export/import scripts for tables in this list: \n"
		     "    %s\n", no_data_tables );
	if (strlen( no_data_accesses ) > 0)
	    fprintf( stderr,
		     "Not creating export/import scripts for tables created with the access methods \n"
		     "in this list: %s\n", no_data_accesses );
	if (No_Triggers)
	    fprintf( stderr, "Disabling SELECT triggers in export scripts.\n" );
    }
    
    if (sequence_str[0] == (char)0)
      snprintf( sequence_str , sizeof sequence_str, "%s", "*" );


    if (verbose != 0) {
	fflush( outfile );
        fprintf( stderr, "Connecting to host: %s.\n", host );
    }
    if (gotusername && gotpassword) {
        EXEC SQL CONNECT TO :host USER :username USING :passwd;
    } else {
        EXEC SQL CONNECT TO :host;
    }
    if (sqlca.sqlcode < 0) {
        if (gotusername && gotpassword) {
            fprintf( stderr,
                     "Cannot connect to host: %s as user %s using %s.  SQLcode: %d, ISAM: %d\n",
                     host,
                     username,
                     passwd,
                     sqlca.sqlcode,
                     sqlca.sqlerrd[1] );
        } else {
            fprintf( stderr,
                     "Cannot connect to host: %s.  SQLcode: %d, ISAM: %d\n",
                     host,
                     sqlca.sqlcode,
                     sqlca.sqlerrd[1] );
        }
        rexit(97);
    }

    memset( passwd, 0, sizeof passwd ); /* Wipe out the password to thwart peekers. */
    /* Make sure this is NOT an SE server.  Not used for anything right now! */
    is_OL_version = (sqlca.sqlwarn.sqlwarn3 == 'W');

    if (!conn5 && ids_server) {
        char dum[1000];
	int ret;

        /* Access sysmaster database */
        if (verbose) {
	    fflush( outfile );
            fprintf( stderr, "Opening Sysmaster database.\n" );
	}
        strcpy( database, "sysmaster" );
        EXEC SQL
            DATABASE :database;
        if ( sqlca.sqlcode ) {
            fprintf( stderr,
                     "Cannot access database (%s); error=%d, isam=%d\n",
                     database,
                     sqlca.sqlcode,
                     sqlca.sqlerrd[1] );
            rexit(99);
        }
	db_has_log = (sqlca.sqlwarn.sqlwarn1 == 'W');

	DatabaseConnected = 1;

        if (setexplain)
            EXEC SQL SET EXPLAIN ON;

	/* Checking features. */
	snprintf( statement,
		  sizeof statement,
		  "SELECT DBINFO( 'version', 'server-type' ), "
		  "       DBINFO( 'version', 'major' ), "
		  "       DBINFO( 'version', 'minor' ), "
		  "       DBINFO( 'version', 'level' ) "
		  "FROM \"informix\".systables "
		  "WHERE tabid = 1" );
	EXEC SQL PREPARE vers_s0 FROM :statement;
	if (sqlca.sqlcode == -728) {
	    /* Older versions do not support DBINFO( 'version', ... */
	    major_vers = 7;
	    minor_vers = 23;
	    level_vers[0] = (char)0;
	    snprintf( serv_type, sizeof serv_type, "Informix Dynamic Server" );
	} else if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Error preparing server version stmt: error=%d, isam=%d\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    rexit(99);
	} else {
	    EXEC SQL EXECUTE vers_s0 INTO :serv_type, :major_vers, :minor_vers, :level_vers;
	    if ( sqlca.sqlcode ) {
		fprintf( stderr,
			 "Error retrieving server version stmt: error=%d, isam=%d\n",
			 sqlca.sqlcode,
			 sqlca.sqlerrd[1] );
		rexit(99);
	    }
	}

	if (do_infrastructure) {
	    print_infrastructure( outfile, infra_out );
	    exit(0);
	}

	/* Check version dependent options and warn. */
	if (major_vers < 12 && NoValidateFK == 1) {
	    fprintf( stderr,
		     "--no-validate specified for a version %d source! This schema will only be valid\n"
		     "on a version 12+ target!\n\n",
		     major_vers );
	}
	ret = sscanf( level_vers, "%[CDFHTU]%d", dum, &lvl_num );
	if ((major_vers < 12 || (major_vers == 12 && minor_vers == 10 && lvl_num < 8)) && NoValidateChk) {
	    fprintf( stderr,
		     "--no-check specified for a version %d.%d.%s (%d) source! This schema will only be valid\n"
		     "on a version 12.10.xC8+ target!\n\n",
		     major_vers,
		     minor_vers,
		     level_vers,
		     lvl_num );
	}

        strcpy( database, databasename );
        /* Fetch system pagesize. */

        system_pagesize = getSystemPageSize();
	if (minext == 0)
	    minext = MINEXT;
	if (minnext == 0)
	    minnext = MINEXT;
	if (verbose) {
	    fprintf( stderr, "Minimum extents set to %dK.\n", minext );
	    fprintf( stderr, "Minimum next size set to %dK.\n", minnext );
	}

        /* Get dbspace information. */

        nspaces = getDbspaceInfo();

        /* Find out who owns the database for catalog queries in mode ANSI servers. */
        EXEC SQL SELECT owner INTO :dbowner FROM sysdatabases WHERE name = :database;
        if (sqlca.sqlcode == SQLNOTFOUND) {
            fprintf( stderr, "Database not found!\n" );
            rexit(99);
        } else if (sqlca.sqlcode) {
            fprintf( stderr,
                     "Error accessing database owner: error=%d, isam=%d\n",
                     sqlca.sqlcode,
                     sqlca.sqlerrd[1] );
            rexit(99);
        }
	if (UseDatabaseOwner || UseTableOwner) 
	    strcpy( owner_override, dbowner );

        /* Stuart Knight supplied the updated dbload support code in this
           section. -- ASK */
        if (summary || loadflag) {
            getIndexSummary();
        }
        /* End updated dbload support section. */
    } else {
        strcpy( database, databasename );
        sysdbspaces =
            (struct s_sysdbspaces *)malloc( 100 * sizeof (struct s_sysdbspaces) );
        if (!sysdbspaces) {
            fprintf( stderr, 
                     "Could not allocate memory for dbspace data. "
                     "Errno=%d.\n", 
                     errno );
            rexit(99);
        }
    }
    if (db_has_log) {
	EXEC SQL SET ISOLATION TO DIRTY READ;
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Setting isolation to dirty read failed! SQLCODE=%d, ISAM=%d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit(1);
	}
    }
    EXEC SQL SET LOCK MODE TO WAIT 60;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Setting lock mode to wait 60 failed! SQLCODE=%d, ISAM=%d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit(1);
    }

    if (setexplain)
        EXEC SQL SET EXPLAIN ON;

    /* Checking features. */
    snprintf( statement,
              sizeof statement,
              "SELECT DBINFO( 'version', 'server-type' ), "
              "       DBINFO( 'version', 'major' ), "
              "       DBINFO( 'version', 'minor' ), "
              "       DBINFO( 'version', 'level' ) "
              "FROM \"informix\".systables "
              "WHERE tabid = 1" );
    EXEC SQL PREPARE vers_s FROM :statement;
    if (sqlca.sqlcode == -728) {
      /* Older versions do not support DBINFO( 'version', ... */
        major_vers = 7;
        minor_vers = 23;
        level_vers[0] = (char)0;
        snprintf( serv_type, sizeof serv_type, "Informix Dynamic Server" );
    } else if (sqlca.sqlcode < 0) {
        fprintf( stderr,
                 "Error preparing server version stmt: error=%d, isam=%d\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1] );
        rexit(99);
    } else {
        EXEC SQL EXECUTE vers_s INTO :serv_type, :major_vers, :minor_vers, :level_vers;
        if ( sqlca.sqlcode ) {
            fprintf( stderr,
                     "Error retrieving server version stmt: error=%d, isam=%d\n",
                     sqlca.sqlcode,
                     sqlca.sqlerrd[1] );
            rexit(99);
        }
    }

    if (verbose) {
        fprintf( stderr,
                 "Server info: Type: %s, Version %d.%2.2d.%s\n",
                 serv_type,
                 major_vers,
                 minor_vers,
                 level_vers);
    }

    if ((do_myexport_scripts || do_export || do_import) &&
	(major_vers > 11 ||
	 (major_vers == 11 &&
	  (minor_vers > 50 ||
	   (minor_vers == 50 &&
	    (level_vers[0] > 'C' ||
	     (level_vers[0] == 'C' && level_vers[1] >= 6)))))))
    {
	/* Need EXTERNAL TABLES for this one, so disable if not available. */
	char *dir = (char *)NULL;

	if( (dir = getenv( "MYEXPORT_PATH" )) == (char *)NULL)
	    dir = ".";
    } else if (do_myexport_scripts || do_export || do_import) {
	if (verbose) {
	    fflush( outfile );
	    fprintf( stderr,
		     "External tables not available.  Disabling %s.\n",
		     do_export ? "--export-scripts" : (do_import ? "--import-scripts" : "--myexport-scripts") );
	}
	do_myexport_scripts = FALSE;
	do_myexport_express = FALSE;
	do_export = FALSE;
	do_import - FALSE;
    }

    /* If Myexport, check for any specific isolation level. */
    if ((do_myexport_scripts || do_export || do_import) && getenv( "MYEXPORT_ISOLATION" )) {
	sprintf( myexport_isolation, "%s", getenv( "MYEXPORT_ISOLATION" ) );
    }

    /* If Myexport, check that the escape sequence is a single character. */
    if ((do_myexport_scripts || do_export || do_import) && strlen( myexport_escape ) != 1 ) {
	fprintf(  stderr,
		  "The argument to --myexport-escape/--export-escape must be a single character >%s<!\n",
		 myexport_escape );
	usage( argv[0] );
	rexit( 1 );
    }

    /* If Myexport, check that the delimiter sequence is a single character. */
    if ((do_myexport_scripts || do_export || do_import) && strlen( myexport_delimiter ) != 1 ) {
	fprintf(  stderr,
		  "The argument to --myexport-delimiter/--export-delimiter must be a single character >%s<!\n",
		 myexport_delimiter );
	usage( argv[0] );
	rexit( 1 );
    }

    if ((doTabOrder == DependOrder || doTabOrder == MyexportOrder) &&
	!(major_vers > 11 ||
	  ( major_vers == 11 && minor_vers > 50) ||
	  ( major_vers == 11 && minor_vers == 50 &&
	    (level_vers[0] > 'C' || (level_vers[0] == 'C' && level_vers[1] >= '5'))))) {
	fprintf( stderr,
		 "Dependency ordering not supported by this server.  Version 11.50.xC5 or later required!\n" );
	usage( argv[0] );
	rexit( 1 );
    }

    if (strstr( serv_type, "Dynamic" ) != (char *)NULL)
        Dynamic = 1;
    else
        Dynamic = 0;

    AuditFlag = 64;
    if (major_vers >= 9) {
        Universal = 1;
        CrColFlag = 8;
        InHerFlag = 2;
        RawFlag = 0x00000010;
	CollectionFlag = 1024;
    } else {
        Universal = 0;
        CrColFlag = 2;
        InHerFlag = 0xFFFFFFFF;
        RawFlag = 0x00000008;
	CollectionFlag = 1024;
    }

    if (major_vers < 10) {
	suppress_directives = 1;
	if (directives_only) {
	    fprintf( stderr,
		     "External directives not supported by target server.  Ignoring --directives-only.\n" );
	    directives_only = 0;
	}
	if (do_index_online) {
	    fprintf( stderr,
		     "Online index build not supported by target server.  Ignoring --index-online.\n" );
	    do_index_online = 0;
	}
    }

    // ASK - chg'd to always get SMICounts 11/2023  //  if (major_vers >= 7 && SMICounts) {
    if (major_vers >= 7) {
        if (verbose) {
	    fflush( outfile );
            fprintf( stderr, "Collecting table space usage information.\n" );
	}
        /* Get count of matching tables. */
        ntables = getTableCount( database, tablename );

        /* Gather page use data from sysmaster for table(s). */

        pagecounts = (t_pagecounts *)malloc( (ntables + 1) * sizeof (t_pagecounts) );
        if (!pagecounts) {
            fprintf( stderr, 
                     "Could not allocate memory for table pagecounts. "
                     "Errno=%d.\n", 
                     errno );
            rexit(99);
        }
        memset( pagecounts, 0, (ntables + 1) * sizeof (t_pagecounts) );
        /* Leave an empty offset -1 record for sysmaster pseudotables. */
        /* Needed because these have no entries in sysptnhdr.         */
        /* It's a kludge, but there's no better way to handle it.      */
        pagecounts++;
        getSMICounts();
    }

    if (doreorg == 2 && (major_vers < 11 || (major_vers == 11 && minor_vers < 70))) {
	fprintf( stderr,
		 "Administration API function 'defragment' not available!\n" );
	exit(1);
    }

    if (LongKeyIndices && (!ids_server || major_vers < 7 || (major_vers == 7 && minor_vers < 30) || (major_vers == 9 && minor_vers < 21))) {
	fprintf( stderr,
		 "Informix %s version %d.%d does not support object names longer than 18 \n"
		 "characters. Ignoring -K.\n",
		 ids_server ? "IDS" : "SE",
		 major_vers,
		 minor_vers );
	LongKeyIndices = FALSE;
    }

    if (create_conditionally == TRUE && (major_vers < 11 || (major_vers == 11 && minor_vers < 70))) {
	if (verbose) {
	    fflush( outfile );
	    fprintf( stderr,
		     "Conditional object creation not supported by version %d.%d!  Ignoring.\n",
		     major_vers,
		     minor_vers );
	}
	create_conditionally = FALSE;
    }

    /* Does this server support external tables? */
    EXEC SQL
    SELECT count(*)
      INTO :has_external_tables
    FROM systables
    WHERE tabname = 'sysexternal';

    if (!conn5) {
        if (verbose) {
	    fflush( outfile );
            fprintf( stderr, "Opening database: %s.\n", database );
	}
        /* SELECT desired database. */
        EXEC SQL
            DATABASE :database;
        if ( sqlca.sqlcode ) {
            fprintf( stderr,
                     "Cannot access database (%s); error=%d, isam=%d\n",
                     database,
                     sqlca.sqlcode,
                     sqlca.sqlerrd[1] );
            rexit(99);
        }
        /* Make sure this is NOT an SE server.  Not used for anything right now! */
        is_OL_version = (sqlca.sqlwarn.sqlwarn3 == 'W');
	db_has_log = (sqlca.sqlwarn.sqlwarn1 == 'W');

	DatabaseConnected = 1;
	
	if (db_has_log) {
	    EXEC SQL SET ISOLATION TO DIRTY READ;
	    if (sqlca.sqlcode < 0) {
		fprintf( stderr,
			 "Setting isolation to dirty read failed! SQLCODE=%d, ISAM=%d.\n",
			 sqlca.sqlcode,
			 sqlca.sqlerrd[1] );
		exit(1);
	    }
	}
	EXEC SQL SET LOCK MODE TO WAIT 60;
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Setting lock mode to wait 60 failed! SQLCODE=%d, ISAM=%d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit(1);
	}

    } /* Otherwise database is already open. */

    if (verbose) {
	fflush( outfile );
	fprintf( stderr, "Database %s logging.\n", db_has_log ? "has" : "does not have" );
    }

    if (HaveIRename == 1) {
        /* Just generate renames and exit. */
        print_index_rename( IndexFile, itabname );
	exit( 0 );
    }

    if (directives_only) {
	print_directives( outfile, outfile2 );
	return 0;
    }

    if (tflag || doall) {

        if (verbose) {
	    fflush( outfile );
            fprintf( stderr, "Preparing Cursors.\n" );
	}

EXEC SQL IFNDEF vers_1150;
	/* If dependency-order specified, build the temp table for ordering. */
	if (doTabOrder == DependOrder || doTabOrder == MyexportOrder) {
	    EXEC SQL SELECT st.tabid ctabid, sr.ptabid ptabid, st.tabtype
		     FROM systables st, OUTER( sysreferences sr, sysconstraints sc)
		     WHERE  st.tabid = sc.tabid 
		       AND sc.constrid = sr.constrid
		     INTO TEMP hierarchy;
	    if (sqlca.sqlcode != 0) {
		fprintf( stderr, 
			 "Error building hierarchy map! Error: %d.\n",
			 sqlca.sqlcode );
		exit(1);
	    }
	    EXEC SQL DELETE FROM hierarchy 
		     WHERE ctabid = ptabid;
	    if (sqlca.sqlcode != 0) {
		fprintf( stderr, 
			 "Error cleaning loops from hierarchy map! Error: %d.\n",
			 sqlca.sqlcode );
		exit(1);
	    }

	    snprintf( tmpstr,
		      sizeof tmpstr,
		       "create temp table hierarchy1( ctabid int, ptabid int, tabtype char(1), hlevel int );" );
	    
	    EXEC SQL EXECUTE IMMEDIATE :tmpstr;
	    if (sqlca.sqlcode != 0) {
		fprintf( stderr, 
			 "Error building hierarchy1! Error: %d.\n",
			 sqlca.sqlcode );
		exit(1);
	    }

	    snprintf( tmpstr,
	    	      sizeof tmpstr,
		      "INSERT INTO hierarchy1 "
		      "SELECT ctabid, ptabid, tabtype, LEVEL as hlevel "
		      "FROM hierarchy "
		      "START WITH ptabid IS NULL "
		      "CONNECT BY NOCYCLE PRIOR ctabid = ptabid "
		      "ORDER SIBLINGS BY ctabid; " );

	    EXEC SQL EXECUTE IMMEDIATE :tmpstr;
	    if (sqlca.sqlcode != 0) {
		fprintf( stderr, 
			 "Error organizing hierarchy map! Error: %d.\n",
			 sqlca.sqlcode );
		fprintf( stderr, "Qry: %s\n", tmpstr );
		exit(1);
	    }
	    EXEC SQL DROP TABLE hierarchy;
	    EXEC SQL SELECT ctabid, ptabid, case when tabtype = 'V' then 999999999 else h1.hlevel end AS hlevel
		     FROM hierarchy1 AS h1
		     WHERE rowid = (
				    SELECT MAX(rowid)
				    FROM hierarchy1 AS h2
				    WHERE h2.ctabid = h1.ctabid
				        AND hlevel = (
					      SELECT MAX(hlevel)
					      FROM hierarchy1 AS h3
					      WHERE h3.ctabid = h2.ctabid
				    )
		     )
		     INTO TEMP hierarchy2;
	    if (sqlca.sqlcode != 0) {
		fprintf( stderr, 
			 "Error ordering hierarchy map! Error: %d.\n",
			 sqlca.sqlcode );
		exit(1);
	    }
	    EXEC SQL DROP TABLE hierarchy1;
	}
EXEC SQL ENDIF;

        /* Select for tables and columns. */
        statement[0] = (char)0;
        strcat( statement, "SELECT a.tabname, a.owner, a.tabid, " );
        if (ids_server) {
            strcat( statement, "a.partnum, a.fextsize, a.nextsize, " );
            strcat( statement,
                    "a.locklevel, a.ncols, b.colname, b.colno, b.coltype, " );
            strcat( statement, "b.collength, a.npused, a.tabtype, " );
	    if ((major_vers == 11 && minor_vers >= 50) || major_vers > 11)
		strcat( statement,
			"colmin, colmax, a.nrows, "
			"(SELECT count(*) "
			" FROM syscolumns csc"
			" WHERE csc.tabid = a.tabid"
			"   AND csc.colattr IN (0,128)), a.rowsize, " );
	    else
		strcat( statement,
			"colmin, colmax, a.nrows, "
			"(SELECT count(*) "
			" FROM syscolumns csc"
			" WHERE csc.tabid = a.tabid), a.rowsize, " );
            strcat( statement, "a.flags, ' ', ' ', " );
            if (Universal) {
                if (major_vers >= 10) {
		    strcat( statement, "a.pagesize, " );  /* Table's pagesize. */
                } else {
		    strcat( statement, "0, " );
                }
		if (major_vers > 11 || (major_vers == 11 && minor_vers >= 70)) {
		    strcat( statement, "a.statchange, a.statlevel, " );
		} else {
		    strcat( statement, " -1, ' ', " );
		}
		/* Fetch security policy details. */
		if (major_vers >= 11) {
		    strcat( statement, "a.secpolicyid, ss.secpolicyname, " );
		} else {
		    strcat( statement, "0, '' " );
		}
		strcat( statement, "b.extended_id, " );

		/* Fetch new attribute column (11.50+) for filtering. */
		if ((major_vers == 11 && minor_vers >= 50) || major_vers > 11) {
		    strcat( statement, "b.colattr, " );
		} else {
		    strcat( statement, "0, " );
		}
                strcat( statement, "am.am_owner, am.am_name, " );
                strcat( statement, "sx.owner, sx.name, " );
                strcat( statement, "p.owner, p.tabname " );
            } else {
                /* Forced SELECTION OF NULLS from a dummy outer join! */
                strcat( statement,
                        "0, 0, "
                        "n.owner, n.tabname, "
                        "n.owner, n.tabname, "
                        "n.owner, n.tabname " );
            }
        } else {
            strcat( statement, "0, 0, 0, ' ', " );
            strcat( statement, "ncols, b.colname, b.colno, b.coltype, " );
            strcat( statement, "b.collength, 0, a.tabtype, " );
            strcat( statement, "0, 0, nrows, ncols, rowsize, " );
            strcat( statement, "0, dirpath, audpath, 0, 0, " );
            strcat( statement,
                    "n.owner, n.tabname, "
                    "n.owner, n.tabname, "
                    "n.owner, n.tabname " );
        }
        if (do_filtering) {
            strcat( statement,
                    ", sv.targettid " );
        } else {
            strcat( statement, ", 0 " );
        }
	strcat( statement, ", CASE a.tabtype WHEN 'S' THEN 'Z' WHEN 'P' THEN 'Z' ELSE a.tabtype END as typeorder " );
        strcat( statement,
                "FROM \"informix\".systables a, "
                "outer \"informix\".syscolumns b, " );
        if (Universal) {
            strcat( statement,
                    " outer (\"informix\".sysinherits i, "
                    " \"informix\".systables p), "
                    " outer \"informix\".sysxtdtypes sx, "
                    " outer \"informix\".sysams am, "
		    " outer \"informix\".syssecpolicies ss " );
	    if (doTabOrder == DependOrder || doTabOrder == MyexportOrder) {
		strcat( statement,
			", hierarchy2 as h " );
	    }
        } else {
            strcat( statement,
                    " outer \"informix\".systables n " );
        }
        if (do_filtering) {
            strcat( statement, ", outer \"informix\".sysviolations sv " );
        }
        strcat( statement, "WHERE a.tabid = b.tabid " );
        strcat( statement, "AND a.tabtype != 'Q' " );   /* Ignore sequences for now. */
	if (ViewsOnly == 1) /* Only report Views. */
	  strcat( statement, "AND a.tabtype == 'V' " );
        strcat( statement, "AND a.tabname NOT MATCHES 'cdr_deltab_[0-9][0-9][0-9][0-9][0-9][0-9]*' " );
        if (!tflag || strcmp( tablename, "*" ) == 0)
            strcat( statement, "AND a.tabid >= 100 " );
	if (NoAuditTabs == 1) {
	    strcat( statement, "AND a.tabname NOT MATCHES '" );
	    strcat( statement, AuditMatch );
	    strcat( statement, "' " );
	}
        if (!conn5 && ((!tflag) || strcmp( tablename, "*" ) == 0)) {
            if ((major_vers == 7 && minor_vers >= 30) || (major_vers == 9 && minor_vers >= 20) || major_vers > 9) {
                strcat( statement, "AND a.tabid NOT IN (" );
                strcat( statement, "SELECT viotid FROM \"informix\".sysviolations ) " );
                strcat( statement, "AND a.tabid NOT IN (" );
                strcat( statement, "SELECT diatid FROM \"informix\".sysviolations ) " );
                /*  This is supposed to be supported but it gets -201 error in 7.31UD8
                strcat( statement, "AND a.tabid NOT IN (" );
                strcat( statement, "SELECT viotid FROM \"informix\".sysviolations UNION " );
                strcat( statement, "SELECT diatid FROM \"informix\".sysviolations ) " );
                */
            } else {
                strcat( statement, "AND a.tabid NOT IN (" );
                strcat( statement, "SELECT viotid FROM \"informix\".sysviolations ) " );
                strcat( statement, "AND a.tabid NOT IN (" );
                strcat( statement, "SELECT diatid FROM \"informix\".sysviolations ) " );
            }
        }
	if (no_blade_funcs == 1) {
	  strcat( statement, "AND a.tabname NOT IN ( " );
	  strcat( statement, "'sysblderrorlog','sysbldobjdepends','sysbldobjkinds','sysbldregistered'," );
	  strcat( statement, "'sysbldirequired','sysbldiprovided','sysbldobjects','bld_registered'," );
	  strcat( statement, "'bldi_required','bldi_provided' " );
	  strcat( statement, ") " );
	}
        strcat( statement, "AND a.tabname MATCHES ? " );
        if (doreorg)
            strcat( statement, "AND a.tabtype = 'T' AND b.colno = 1 " );
        if (Universal) {
            strcat( statement, "AND i.child = a.tabid " );
            strcat( statement, "AND i.parent = p.tabid " );
            strcat( statement, "AND a.type_xid = sx.extended_id " );
            strcat( statement, "AND a.am_id = am.am_id " );
	    strcat( statement, "AND a.secpolicyid = ss.secpolicyid " );
        } else {
            /* Force an invalid outer join to get NULLs for Universal columns */
            strcat( statement,
                    "AND a.tabid = n.tabid AND n.tabid = -999999999 " );
        }
        if (do_filtering) {
            strcat( statement, "AND sv.targettid = a.tabid " );
        }
        if (doTabOrder == NameOrder) {
            strcat( statement, "ORDER BY typeorder, a.tabname, b.colno; " );
	} else if (doTabOrder == DependOrder && doOrder == NameOrder) {
	    strcat( statement, "AND a.tabid = h.ctabid " );
            strcat( statement, "ORDER BY typeorder, h.hlevel, a.tabname, b.colno; " );
	} else if (doTabOrder == DependOrder) {
	    strcat( statement, "AND a.tabid = h.ctabid " );
            strcat( statement, "ORDER BY typeorder, h.hlevel, a.tabid, b.colno; " );
	} else if (doTabOrder == MyexportOrder) {
	    strcat( statement, "AND a.tabid = h.ctabid " );
            strcat( statement, "ORDER BY typeorder, h.hlevel DESC, a.tabid, b.colno; " );
	} else {
            strcat( statement, "ORDER BY typeorder, a.tabid, b.colno; " );
	}
	
	if (doDebug) {
	    fprintf( stderr, "Using statement:\n%s\n", statement );
	}
	
        EXEC SQL
            PREPARE stmt1 FROM :statement;
        if (sqlca.sqlcode) {
            fprintf( stderr,
                     "Cannot PREPARE select statement for table data, error: %d.\n",
                     sqlca.sqlcode );
            fprintf( stderr, "Error at character %d.\nStatement is:\n",
                     sqlca.sqlerrd[4] );
            fprintf( stderr, "%s\n\n", statement );
            fprintf( stderr,
                     "errp=%s, errm=%s\n\n",
                     sqlca.sqlerrp,
                     sqlca.sqlerrm );
            rexit(2);
        }
	
        EXEC SQL
            DECLARE a1 CURSOR FOR stmt1;
        if (sqlca.sqlcode ) {
            fprintf( stderr,
                     "Cannot declare a1 cursor; Error = %d\n", sqlca.sqlcode );
            rexit(1);
        }
	
	if (No_Triggers) {
	    strcpy( seltr_txt, 
		    "SELECT trigname FROM systriggers s WHERE s.tabid = ? AND s.event = 'S'; " );
	    EXEC SQL PREPARE seltrig_s FROM :seltr_txt;
	    if ( sqlca.sqlcode ) {
		fprintf( stderr,
			 "Cannot preparing statement seltrig_s; Error = %d\n",
			 sqlca.sqlcode );
		rexit(1);
	    }
	    
	    EXEC SQL DECLARE seltrig_c CURSOR FOR seltrig_s;
	    if ( sqlca.sqlcode ) {
		fprintf( stderr,
			 "Cannot declare cursor seltrig_c; Error = %d\n",
			 sqlca.sqlcode );
		rexit(1);
	    }
	}
	
        snprintf( statement4,
                  1024,
                  "SELECT tabid, colno, type, default "
                  "FROM \"informix\".sysdefaults "
                  "WHERE tabid = ? "
                  "  %s " 
                  "ORDER BY colno",
                  major_vers >= 9 ? "AND class = 'T' " : "" );
	
        EXEC SQL PREPARE def_s FROM :statement4;
        if (sqlca.sqlcode ) {
            fprintf( stderr,
                     "Cannot prepare def statement; Error = %d\n",
                     sqlca.sqlcode );
	    fprintf( stderr,
		     "def: %s;\n",
		     statement4 );
            rexit(1);
        }
	
        EXEC SQL DECLARE def CURSOR FOR def_s;
        if (sqlca.sqlcode ) {
            fprintf( stderr,
                     "Cannot declare def cursor; Error = %d\n", sqlca.sqlcode );
            rexit(1);
        }
	
	if (!NoSequences && sequence_only && (major_vers > 9 || (major_vers == 9 && minor_vers >= 40))) {
	    print_sequences( outfile2,
			     outfileG,
			     UseDatabaseOwner ? dbowner : (UseSetOwner ? owner_override : (char *)NULL ) );
	    return 0;
	}
	
        /* Prepare cursors needed below. */
        if (!tflag && !doreorg) {  /* If database level schema do user privileges. */
            /* Stuart Knight's extended dbload support code. -- ASK */
            if (loadflag) {
                fprintf( outfile,
                         "{ DATABASE %s delimiter %s }\n\n",
                         databasename,
			 myexport_delimiter );
            }
            /* End Stuart Knight's extended dbload support code. -- ASK */
	    if (verbose) {
		fflush( outfile );
		fprintf( stderr, "Processing user authorizations.\n" );
	    }
	    if (ViewsOnly == 0)
	        print_userauths( AuthsOnly ? outfileG : outfile2, outfileG, NoSetSessionAuth );

            if (Universal) {
                print_langauths( outfileG );
                if (!NoTypes) {
		    if (verbose) {
			fflush( outfile );
			fprintf( stderr, "Processing user defined types.\n" );
		    }
                    print_types( outfile, outfileG );
		}
                if (!NoCasts) {
		    if (verbose) {
			fflush( outfile );
			fprintf( stderr, "Processing user defined casts.\n" );
		    }
                    print_casts( outfile2 );
		}
                if (!NoSequences && (major_vers > 9 || (major_vers == 9 && minor_vers >= 40))) {
		    if (verbose) {
			fflush( outfile );
			fprintf( stderr, "Processing sequences.\n" );
		    }
                    print_sequences( outfile2,
				     outfileG,
				     UseDatabaseOwner ? dbowner : (UseSetOwner ? owner_override : (char *)NULL ) );
		}
		if (verbose) {
		    fflush( outfile );
		    fprintf( stderr, "Processing functions.\n" );
		}
		if (!NoProcs) {
		  /* Print all Access Method Support functions to the main output file. */
		  print_procedures( outfile, outfileG == outfile2 ? ProcFile : outfileG, ptempl, grantas, TRUE, TRUE );
		  /* Print all other Functions to the procedure file if any. */
		  print_procedures( ProcFile, outfileG == outfile2 ? ProcFile : outfileG, ptempl, grantas, TRUE, FALSE );
		}

		if (verbose) {
		    fflush( outfile );
		    fprintf( stderr, "Processing user defined access methods.\n" );
		}
                print_ams( outfile );
		if (verbose) {
		    fflush( outfile );
		    fprintf( stderr, "Processing user defined op classes.\n" );
		}
                print_opclasses( outfile );
            }
        }

        if (!conn5 && ids_server && !no_storage_opt) {
            char statement2[2048];

            /* Fragment privilege information. */
            snprintf( statement2,
                      sizeof statement2,
                    "SELECT fragtype, tabid, indexname, strategy, evalpos, "
		    "       nrows, npused, colno, flags, "
                    "       exprtext, dbspace, %s "
                    "FROM \"informix\".sysfragments "
                    "WHERE tabid = ? AND fragtype = 'T' "
                    "ORDER BY evalpos ",
                    major_vers >= 10 ? "partition" : "' '" );
            EXEC SQL PREPARE s_frags FROM :statement2;
            if (sqlca.sqlcode) {
                fprintf( stderr,
                         "Cannot PREPARE SELECT statement for table fragments, "
                         "error: %d.\n",
                         sqlca.sqlcode );
                fprintf( stderr, "Error at character %d.\nStatement is:\n",
                         sqlca.sqlerrd[4] );
                fprintf( stderr, "%s\n", statement2 );
                rexit(19);
            }

            fragments.exprtext.loc_loctype = LOCUSER;
            fragments.exprtext.loc_buffer = (char *)malloc( 4096 );
            if (!fragments.exprtext.loc_buffer) {
                fprintf( stderr, 
                         "Could not allocate memory for fragment expressions. "
                         "Errno=%d.\n", 
                         errno );
                rexit(99);
            }
            fragments.exprtext.loc_bufsize = 4096;
            fragments.exprtext.loc_write = writeit;
            fragments.exprtext.loc_open = openit;
            fragments.exprtext.loc_close = closeit;
            tblobbuff.bbsize = 2048;
            tblobbuff.buff = (char *)malloc( tblobbuff.bbsize );
            if (!tblobbuff.buff) {
                fprintf( stderr, 
                         "Could not allocate memory for tblob buffer. "
                         "Errno=%d.\n", 
                         errno );
                rexit(99);
            }
            tblobbuff.global_length = 0;

            EXEC SQL DECLARE frags CURSOR FOR s_frags;
            if (sqlca.sqlcode ) {
                fprintf( stderr,
                         "Cannot DECLARE frags CURSOR; Error = %d\n",
                         sqlca.sqlcode );
                rexit(20);
            }
        }

        if (verbose) {
	    fflush( outfile );
            fprintf( stderr, "Analyzing schema.\n" );
	}
        EXEC SQL
            OPEN a1 USING :tablename;
        if (sqlca.sqlcode == SQLNOTFOUND) {
            fprintf( stderr, "No table(s) matching \"%s\".\n", tablename );
            rexit(3);
        } else if ( sqlca.sqlcode ) {
            fprintf( stderr,
                     "Cannot open a1 cursor; Error = %d, ISAM = %d\n",
                     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
            rexit(1);
        }
	nrecs = 0;

        for(;;) {
            char tname[8], lname[130];

            fflush( outfile );
	    memset( &systables, 0, sizeof systables );
	    memset( &syscolumns, 0, sizeof syscolumns );

            EXEC SQL
                FETCH a1 INTO
                :systables.tabname,
                :systables.owner,
                :systables.tabid,
                :systables.partnum,
                :systables.fextsize,
                :systables.nextsize,
                :systables.locklevel,
		:real_ncols,
                :syscolumns.colname,
                :syscolumns.colno,
                :syscolumns.coltype,
                :syscolumns.collength,
                :systables.npused,
                :systables.tabtype,
                :syscolumns.colmin,
                :syscolumns.colmax,
                :systables.nrows,
                :systables.ncols,
                :systables.rowsize,
                :systables.flags,
                :systables.dirpath,
                :systables.audpath,
                :systables.pagesize,
		:systables.statchange indicator systables.statind,
		:systables.statlevel indicator systables.statlevelind,
		:systables.secpolicyid,
		:syssecpolicy,
                :syscolumns.extended_id,
		:syscolumns.colattr,
                :systables.am_owner,
                :systables.am_name indicator systables.am_ind,
                :systables.xtd_owner,
                :systables.xtd_name indicator systables.xtd_ind,
                :systables.par_owner,
                :systables.par_name indicator systables.par_ind,
                :systables.has_violations indicator systables.viol_ind,
		:taborder;
	    if (UseSetOwner) {
	        strcpy( original_owner, systables.owner );
		strcpy( systables.owner, owner_override );
	    } else if (UseDatabaseOwner) {
	        strcpy( original_owner, systables.owner );
		strcpy( systables.owner, dbowner );
	    } else if (UseTableOwner) {
	        strcpy( original_owner, systables.owner );
		strcpy( owner_override, systables.owner );
	    }
        
            if (sqlca.sqlcode && sqlca.sqlcode != SQLNOTFOUND) {
                fprintf( stderr,
                         "Error fetching a1 cursor;Error=%d\n",
                         sqlca.sqlcode);
		fprintf( stderr, "Error at character %d.\nStatement is:\n",
			 sqlca.sqlerrd[4] );
		fprintf( stderr, "%s\n\n", statement );
		fprintf( stderr,
			 "errp=%s, errm=%s\n\n",
			 sqlca.sqlerrp,
			 sqlca.sqlerrm );
		fprintf( stderr,
			 "    Could be processing tabid: %d, tabname: %s, colname: %s.\n",
			 systables.tabid,
			 systables.tabname,
			 syscolumns.colname );
		rexit(1);
            } else {
                if (sqlca.sqlcode == SQLNOTFOUND) {
		    if (!nrecs)
			break;
                    last_record = 1;
		}
        
		if (syscolumns.colattr == 3) {
		    replflags.vercols_flag++;
		} else if (syscolumns.colattr == 9) {
		    replflags.replcheck_flag++;
		} else if (syscolumns.colattr == 17 || syscolumns.colattr == 33 || syscolumns.colattr == 65) {
		    replflags.erkey_flag++;
		}

		nrecs++;
                if (!last_record && systables.tabtype[0] == 'V' && syscolumns.colno != 1)
                    continue;
		if (systables.tabid == prev.tabid && syscolumns.colno == prev_colno)
		    continue;
		prev_colno = syscolumns.colno;

                if (last_record || strcmp( prev.tabname, systables.tabname ) != 0
		    || (strcmp( prev.tabname, systables.tabname ) == 0 && strcmp( prev.owner, systables.owner ) != 0))
		{
                    for (; ncols < real_ncols; ncols++) {
                        if (!columns[ncols]) /* Only allocate once each! */
                            columns[ncols] = (char *)malloc( IDBYTES );
                        if (!columns[ncols]) {
                            fprintf( stderr, 
                                     "Could not allocate memory for columns. "
                                     "Errno=%d.\n", 
                                     errno );
                            rexit(99);
                        }
                    }
                    if (pass_no != 1) {
			if (prev.tabtype[0] == 'E') {
			    int ret;

			    fprintf( outfile, "\n) " );
			    ret = print_external( prev.tabid, outfile );
			    print_indexes( IndexFile, outfile3, SimpleFrags );
			} else if (prev.tabtype[0] == 'T') {
			    if ((do_myexport_scripts && export_OK) || do_export || do_import) {
				char impname[129], dimpname[151], *dir;

				if( (dir = getenv( "MYEXPORT_PATH" )) == (char *)NULL)
				    dir = getenv( "PWD" );

				if (do_myexport_scripts && export_OK) {
				    sprintf( tname, "%1.5s", prev.tabname );
				    while (strlen( tname ) < 5)
					strcat( tname, "_" );
				    snprintf( lname, sizeof lname, "%s%05d", tname, prev.tabid );
				} else if (do_import || do_export) {
				    snprintf( lname, sizeof lname, "%s", prev.tabname );
				}
				if (!stripowner)
				    snprintf( impname,
					      sizeof impname,
					      "%.119s%s",
					      prev.tabname,
					      do_import ? "_external" : "_myimport" );
				delimident2( impname, dimpname );

				/* NB - ASK - The v11.70 manual says that the ESCAPE option takes an escape character as
				   an argument.  However, doing that ALWAYS returns a syntax error!  For now, making
				   11.70 output look like 11.50 output which works just fine.
				*/
				if (do_import || (do_myexport_scripts && export_OK)) {
				    fprintf( outfileMI,
					     "\n) USING ( FORMAT \"%s\", \n"
					     "    DATAFILES ( \"disk:%s/%s.unl\" ), \n"
					     "    DELIMITER \"%s\"%s, %s ); \n",
					     do_binary ? "INFORMIX" : "DELIMITED",
					     dir,
					     lname,
					     myexport_delimiter,
					     (do_myexport_express ? ", EXPRESS" : ", DELUXE"),
					     ((major_vers >= 12) ? "ESCAPE ON" : ((minor_vers == 50) ? "ESCAPE " : "ESCAPE ")));
				    if (do_myexport_express)
					fprintf( outfileMI,
						 "ALTER TABLE %s TYPE( RAW );\n",
						 delimident( prev.tabname ) );
				    fprintf( outfileMI,
					     "INSERT INTO %s \n"
					     "    SELECT * FROM %s;\n",
					     delimident( prev.tabname ),
					     dimpname );
				    fprintf( outfileMI,
					     "DROP TABLE %s;\n",
					     dimpname );
				    if (do_myexport_express)
					fprintf( outfileMI,
						 "ALTER TABLE %s TYPE( STANDARD );\n",
						 delimident( prev.tabname ) );
				    if (strcmp( prev.tabname, "spatial_references" ) == 0) {
					/* Make sure that the spatial references cache is updated after the load. */
					fprintf( outfileMI, " --Make sure that the spatial references cache is updated after the load.\n" );
					fprintf( outfileMI, "EXECUTE FUNCTION se_metadatainit();\n" );
				    }
				    fclose( outfileMI );
				}                           
                                snprintf( impname,
                                          sizeof impname,
                                          "%.119s%s",
                                          prev.tabname,
					   do_import ? "_external" : "_myexport" );
                                delimident2( impname, dimpname );

				/* NB - ASK - The v11.70 manual says that the ESCAPE option takes an escape character as
				   an argument.  However, doing that ALWAYS returns a syntax error!  For now, making
				   11.70 output look like 11.50 output which works just fine.
				*/
				if (do_export || (do_myexport_scripts && export_OK)) {
				    fprintf( outfileME,
					     "\n) USING ( FORMAT \"%s\", \n"
					     "    DATAFILES ( \"disk:%s/%s.unl\" ), \n"
					     "    DELIMITER \"%s\"%s, %s ); \n",
					     do_binary ? "INFORMIX" : "DELIMITED",
					     dir,
					     lname,
					     myexport_delimiter,
					     (do_myexport_express ? ", EXPRESS" : ", DELUXE"),
					     ((major_vers >= 12) ? "ESCAPE ON" : ((minor_vers == 50) ? "ESCAPE " : "ESCAPE ")));
				    fprintf( outfileME,
					     "INSERT INTO %s \n"
					     "    SELECT * FROM %s;\n",
					     dimpname,
					     delimident( prev.tabname ) );
				    if (No_Triggers)
					if (strlen( sel_trig_list ) > 0)
					    fprintf( outfileME,
						     "SET TRIGGERS %s ENABLED;\n",
						     sel_trig_list );
				    fprintf( outfileME,
					     "DROP TABLE %s;\n",
					     dimpname );
				    fclose( outfileME );
				}
                            }

                            if (!doreorg) {
                                int ret;
                                if (!NoValidateChk && !AlterChecks)
				    ret = print_checks( outfile, 0 );
				else
				    ret = 1;
                                if (ret || (prev.xtd_ind < 0 || !Universal))
                                    fprintf( outfile, "\n)" );
                                else
                                    fprintf( outfile, "\n" );
                            } else
                                fprintf( outfile, "\n" );
                            if ((conn5 || gotdbs) && (prev.partnum/TBSFACT)) {
                                if (!conn5 && ids_server) {
                                    for (ix = 0;
                                         ix < nspaces &&
                                             sysdbspaces[ix].dbsnum<(prev.partnum/TBSFACT);
                                         ix++)
                                        ;
                                    if (sysdbspaces[ix].dbsnum != (prev.partnum/TBSFACT))
                                    {
                                        fprintf( stderr,
                                                 "Dbspace not found! Partnum: %x(%d) Table: %s.\n",
                                                 prev.partnum,
                                                 prev.partnum,
                                                 prev.tabname );
                                        rexit(98);
                                    }
                                } else {
                                    ix = (prev.partnum/TBSFACT)-1;
                                    if (sysdbspaces[ix].dbsnum != ix+1) {
                                        sprintf( sysdbspaces[ix].name,
                                                 "DbSpace#%d",
                                                 (prev.partnum/TBSFACT) );
                                        sysdbspaces[ix].dbsnum = ix + 1;
                                    }
                                }

                                if (systables.pagesize == 0)
                                    pagesize = system_pagesize;
                                else
                                    pagesize = prev.pagesize;
                                if (SMICounts) {
                                    pc = findcounts( database, prev.tabname );
                                    if (verbose > 1) {
                                        fprintf( stderr,
                                                 "Pagecounts[%d] {\n"
                                                 "    dbsname:  %s\n"
                                                 "    tabname:  %s\n"
                                                 "    nfrags:   %d\n"
                                                 "    pgsused:  %d\n"
                                                 "    pgsdata:  %d\n"
                                                 "    pgstotal: %d\n"
                                                 "    maxdata:  %d\n"
                                                 "    maxused:  %d\n",
                                                 pc,
                                                 pagecounts[pc].dbsname,
                                                 pagecounts[pc].tabname,
                                                 pagecounts[pc].nfrags,
                                                 pagecounts[pc].pgsused,
                                                 pagecounts[pc].pgsdata,
                                                 pagecounts[pc].pgstotal,
                                                 pagecounts[pc].maxdata,
                                                 pagecounts[pc].maxused );
                                    }
                                    fe = prev.fextsize > (pagecounts[pc].maxused) ?
                                        prev.fextsize : (pagecounts[pc].maxused);
                                    if (dominimize) {
					if (verbose > 1)
					    fprintf( stderr,
						     "In dominimize before minimizing fe = %d.\n",
						     fe );

					if (pagecounts[pc].pgsdata)
					    fe = (pagecounts[pc].pgsdata < fe) ?
						(int)((double)pagecounts[pc].pgsdata) : fe;
					else
					    fe = minext;
					if (verbose > 1) {
					    fprintf( stderr,
						     "In dominimize.  fe = %d\n"
						     "Pagecounts[%d] {\n"
						     "    dbsname:  %s\n"
						     "    tabname:  %s\n"
						     "    nfrags:   %d\n"
						     "    pgsused:  %d\n"
						     "    pgsdata:  %d\n"
						     "    pgstotal: %d\n"
						     "    maxdata:  %d\n"
						     "    maxused:  %d\n",
						     fe,
						     pc,
						     pagecounts[pc].dbsname,
						     pagecounts[pc].tabname,
						     pagecounts[pc].nfrags,
						     pagecounts[pc].pgsused,
						     pagecounts[pc].pgsdata,
						     pagecounts[pc].pgstotal,
						     pagecounts[pc].maxdata,
						     pagecounts[pc].maxused );
					}
 				    }
                                } else {
                                    fe = prev.fextsize;
                                }
                                if (verbose > 1 && !NoExtentClause && prev.am_ind < 0) {
                                    fprintf( stderr,
                                             "Line %d: Extent size adjustment %d ==> %d.\n",
                                             __LINE__,
                                             fe,
                                             max( (int)(AdjExt * (double)fe), minext ) );
                                }
                                fe = max( (int)(AdjExt * (double)fe), minext );
                                if (doreorg == 1) {
                                    if (prev.par_ind < 0) {
                                        fprintf( outfile,
                                                 " IN %s;\n",
                                                 ids_server
                                                 ? sysdbspaces[ix].name
                                                 : prev.dirpath );
                                    } else if (prev.flags == InHerFlag) {
                                        if (stripowner)
                                            fprintf( outfile,
                                                     " IN %s UNDER %s;\n",
                                                     ids_server
                                                     ? sysdbspaces[ix].name
                                                     : prev.dirpath,
                                                     prev.par_name );
                                        else
                                            fprintf( outfile,
                                                     " IN %s UNDER '%s'.%s;\n",
                                                     ids_server
                                                     ? sysdbspaces[ix].name
                                                     : prev.dirpath,
                                                     (UseTableOwner || UseDatabaseOwner) ? prev.owner : prev.par_owner,
                                                     prev.par_name );
                                    }
                                } else if (ids_server && !doreorg) {
				    if ((prev.flags & CrColFlag))
                                        fprintf( outfile, " WITH CRCOLS\n" );
                                    if (replflags.vercols_flag > 0)
                                        fprintf( outfile, " WITH VERCOLS\n" );
                                    if (replflags.replcheck_flag > 0 || do_For_ER == 1)
                                        fprintf( outfile, " WITH REPLCHECK\n" );
                                    if (replflags.erkey_flag > 0 || do_For_ER == 1)
                                        fprintf( outfile, " WITH ERKEY\n" );
                                    if ((prev.flags & AuditFlag))
                                        fprintf( outfile, " WITH AUDIT\n" );
				    if (syscolumns.colno == 1) {
					replflags.vercols_flag = 0;
					replflags.replcheck_flag = 0;
					replflags.erkey_flag = 0;
				    }

                                    if ((prev.flags & InHerFlag)) {
                                        if (stripowner)
                                            fprintf( outfile,
                                                     " UNDER %s ",
                                                     prev.par_name );
                                        else
                                            fprintf( outfile,
                                                     " UNDER '%s'.%s ",
                                                     (UseTableOwner || UseDatabaseOwner) ? prev.owner : prev.par_owner,
                                                     prev.par_name );
                                    }
				    if (prev.secpolicyid > 0) {
					fprintf( outfile,
						 " SECURITY POLICY %s\n",
						 prev_syssecpolicy );
				    }
				    if (!no_storage_opt)
					fprintf( outfile,
						 " IN %s",
						 sysdbspaces[ix].name );
                                    if (Universal)
                                        if (print_puts( outfile, CrColFlag ))
                                            fprintf( outfile, "\n " );
				    if (!NoExtentClause && prev.am_ind < 0) {
				        fprintf( outfile,
						 " EXTENT SIZE %d NEXT SIZE %d",
						 fe,
						 AdjNext > 0.0 ?
						 max((int)(0.505 + ((float)fe * AdjNext)), minnext ) :
						 prev.nextsize );
				    }
                                } else if (doreorg != 2)
                                    fprintf( outfile,
                                             " IN %s",
                                             prev.dirpath );
                            } else if (!(prev.partnum/TBSFACT)
                                       && ids_server
                                       && !conn5
				       && doreorg != 2)
                            {
				int nrows = 0, npused = 0, colno = 0, flags = 0;
				
                                if (prev.flags == InHerFlag &&
                                    prev.par_ind >= 0)
                                {
                                    if (stripowner)
                                        fprintf( outfile,
                                                 " UNDER %s",
                                                 prev.par_name );
                                    else
                                        fprintf( outfile,
                                                 " UNDER '%s'.%s",
                                                 (UseTableOwner || UseDatabaseOwner) ? prev.owner : prev.par_owner,
                                                 prev.par_name );
                                }
                                if (!doreorg) {
                                    if (prev.flags || replflags.vercols_flag || replflags.replcheck_flag || replflags.erkey_flag) {
                                        if ((prev.flags & 1))
                                            fprintf( outfile, "\n  WITH ROWIDS" );
                                        if ((prev.flags & CrColFlag))
                                            fprintf( outfile, "\n  WITH CRCOLS" );
					if (replflags.vercols_flag > 0)
					    fprintf( outfile, "\n  WITH VERCOLS" );
					if (replflags.replcheck_flag > 0 || do_For_ER == 1)
					    fprintf( outfile, "\n  WITH REPLCHECK" );
					if (replflags.erkey_flag > 0 || do_For_ER == 1)
					    fprintf( outfile, "\n  WITH ERKEY" );
					if ((prev.flags & AuditFlag))
					    fprintf( outfile, "\n  WITH AUDIT" );
                                        fprintf( outfile, "\n " );
                                    }
				    if (syscolumns.colno == 1) {
					replflags.vercols_flag = 0;
					replflags.replcheck_flag = 0;
					replflags.erkey_flag = 0;
				    }
                                }
                                blobbuff = &tblobbuff;
				if (!no_storage_opt) {
				    int have_rolling = 0;
				    
				    EXEC SQL OPEN frags USING :prev.tabid;
				    if (sqlca.sqlcode < 0) {
					fprintf( stderr,
						 "%d: Cannot open CURSOR frags.  Error = %d\n",
						 __LINE__,
						 sqlca.sqlcode );
					rexit(1);
				    }
				    while (sqlca.sqlcode == 0) {
					EXEC SQL FETCH frags INTO :fragments;
					if (sqlca.sqlcode < 0) {
					    fprintf( stderr,
						     "Cannot FETCH CURSOR frags.  Error = %d\n",
						     sqlca.sqlcode );
					    rexit(1);
					} else if (sqlca.sqlcode == SQLNOTFOUND)
					    break;
					tblobbuff.buff[tblobbuff.global_length] = (char)0;
					if (SimpleFrags && strcmp( fragments.dbspace, fragments.partition ) == 0)
					    PrintParts = FALSE;
					else
					    PrintParts = TRUE;
					if (fragments.strategy[0] == 'L') {
					    /* List Fragmentation, initial list column definition. */
					    if (fragments.evalpos < 0) {
						fprintf( outfile,
							 " FRAGMENT BY LIST (%s) \n",
							 tblobbuff.buff );
					    } else {
						/* List fragmentation, partition list. */
						if (fragments.evalpos > 0) fprintf( outfile, ", \n" );
						fprintf( outfile,
							 "  PARTITION %s %s IN %s",
							 fragments.partition,
							 tblobbuff.buff,
							 fragments.dbspace );
					    }
					    continue;
					} else if (fragments.strategy[0] == 'N') {
					    /* Range/Interval fragmentation, Interval Definition. */
					    if (fragments.evalpos < 0) {
						switch (fragments.evalpos) {
						case -4:
						    /* Vers 12.10+ Rolling Partitions. */
						    nrows = fragments.nrows;
						    npused = fragments.npused;
						    flags = fragments.flags;
						    colno = fragments.colno;
						    have_rolling = 1;
						    break;
						case -3:
						    fprintf( outfile,
							     " FRAGMENT BY RANGE( %s )\n",
							     tblobbuff.buff );
						    break;
						case -2:
						    fprintf( outfile,
							     "  INTERVAL( %s )",
							     tblobbuff.buff );
						    if (have_rolling) {
							if (colno > 0) {
							    fprintf( outfile,
								     "\n  ROLLING ( %d FRAGMENTS )",
								     colno );
							}
							if (nrows > 0) {
							    fprintf( outfile,
								     "\n  LIMIT TO %d %s",
								     nrows,
								     limit_unit( npused ) );
							    fprintf( outfile,
								     "\n  %s %s",
								     disposition( flags ),
								     which_disp( flags ) );
							} else {
							    fprintf( outfile,
								     "\n  %s",
								     disposition( flags ) );
							}
						    }
						    break;
						case -1:
						    if (((fragments.flags >> 15) % 2)  == 1)
							fprintf( outfile,
								 "\n  STORE IN ( %s() )\n",
								 tblobbuff.buff );
						    else
							fprintf( outfile,
								 "\n  STORE IN ( %s )\n",
								 tblobbuff.buff );
						    break;
						}
					    } else if (fragments.flags == 512) {
					        /* Ignore interval partitions. */
					        continue;
					    } else {
						if (fragments.evalpos > 0) fprintf( outfile, ", \n" );
						fprintf( outfile,
							 "  PARTITION %s %s IN %s",
							 fragments.partition,
							 tblobbuff.buff,
							 fragments.dbspace );
					    }
					    continue;
					} else if (fragments.strategy[0] == 'R') {
					    if (fragments.evalpos == 0) {
						if (major_vers >= 10 && PrintParts)
						    fprintf( outfile,
							     " FRAGMENT BY ROUND ROBIN \n\t" );
						else                                                
						    fprintf( outfile,
							     " FRAGMENT BY ROUND ROBIN IN \n\t" );
						if (major_vers >= 10 && PrintParts)
						    fprintf( outfile,
							     "PARTITION %s IN %s",
							     fragments.partition,
							     fragments.dbspace );
						else
						    fprintf( outfile,
							     "%s",
							     fragments.dbspace );
					    } else {
						fprintf( outfile, ", \n\t" );
						if (major_vers >= 10 && PrintParts)
						    fprintf( outfile,
							     "PARTITION %s IN %s",
							     fragments.partition,
							     fragments.dbspace );
						else
						    fprintf( outfile,
							     "%s",
							     fragments.dbspace );
					    }
					} else if (fragments.strategy[0] == 'E') {
					    /* Expression fragmentation. */
					    if (fragments.evalpos == 0) {
						fprintf( outfile,
							 " FRAGMENT BY EXPRESSION \n  " );
						if (major_vers >= 10 && PrintParts)
						    fprintf( outfile,
							     " PARTITION %s %s IN %s",
							     fragments.partition,
							     tblobbuff.buff,
							     fragments.dbspace );
						else {                                      
						    fprintf( outfile, "%s IN %s", tblobbuff.buff, fragments.dbspace );
						}
					    } else {
						fprintf( outfile, ", \n  " );
						if (major_vers >= 10 && PrintParts)
						    fprintf( outfile,
							     " PARTITION %s %s IN %s",
							     fragments.partition,
							     tblobbuff.buff,
							     fragments.dbspace );
						else {                                  
						    fprintf( outfile, "%s IN %s", tblobbuff.buff, fragments.dbspace );
						}
					    }
					}
				    }
				    EXEC SQL CLOSE frags;
				    if (doreorg) {
					fprintf( outfile, ";\n" );
				    }
				}
                                if (SMICounts) {
                                    pc = findcounts( database, prev.tabname );
                                    if (verbose > 1) {
                                        fprintf( stderr,
                                                 "Pagecounts[%d] {\n"
                                                 "    dbsname:  %s\n"
                                                 "    tabname:  %s\n"
                                                 "    nfrags:   %d\n"
                                                 "    pgsused:  %d\n"
                                                 "    pgsdata:  %d\n"
                                                 "    pgstotal: %d\n"
                                                 "    maxdata:  %d\n"
                                                 "    maxused:  %d\n",
                                                 pc,
                                                 pagecounts[pc].dbsname,
                                                 pagecounts[pc].tabname,
                                                 pagecounts[pc].nfrags,
                                                 pagecounts[pc].pgsused,
                                                 pagecounts[pc].pgsdata,
                                                 pagecounts[pc].pgstotal,
                                                 pagecounts[pc].maxdata,
                                                 pagecounts[pc].maxused );
                                    }
                                    fe = prev.fextsize > (pagecounts[pc].maxused) ?
                                        prev.fextsize : (pagecounts[pc].maxused);
                                    if (dominimize) {
					if (verbose > 1) {
					    fprintf( stderr,
						     "Pagecounts[%d] {\n"
						     "    dbsname:  %s\n"
						     "    tabname:  %s\n"
						     "    nfrags:   %d\n"
						     "    pgsused:  %d\n"
						     "    pgsdata:  %d\n"
						     "    pgstotal: %d\n"
						     "    maxdata:  %d\n"
						     "    maxused:  %d\n",
						     pc,
						     pagecounts[pc].dbsname,
						     pagecounts[pc].tabname,
						     pagecounts[pc].nfrags,
						     pagecounts[pc].pgsused,
						     pagecounts[pc].pgsdata,
						     pagecounts[pc].pgstotal,
						     pagecounts[pc].maxdata,
						     pagecounts[pc].maxused );
					}
					if (pagecounts[pc].pgsdata)
					    fe = (pagecounts[pc].pgsdata < fe) ?
						(int)((double)pagecounts[pc].pgsdata) : fe;
					else
					    fe = minext;
				    }
                                } else {
                                    fe = prev.fextsize;
                                }
                                if (verbose > 1 && !NoExtentClause && prev.am_ind < 0) {
                                    fprintf( stderr,
                                             "Line %d: Extent size adjustment %d ==> %d.\n",
                                             __LINE__,
                                             fe,
                                             max( (int)(AdjExt * (double)fe), minext ) );
                                }
                                fe = max( (int)(AdjExt * (double)fe), minext );
                                if (!doreorg) {
                                    if (Universal)
                                        print_puts( outfile, CrColFlag );
				    if (!NoExtentClause && prev.am_ind < 0) {
				        fprintf( outfile,
						 "\n  EXTENT SIZE %d NEXT SIZE %d",
						 fe,
						 AdjNext > 0.0 ?
						 max( (int)(0.505 + ((float)fe * AdjNext)), minnext ) :
						 prev.nextsize );
				    } else {
				        fprintf( outfile,
						 "\n " );
				    }
                                }
                            }
                            if (!doreorg) {
                                if (ids_server) {
				    if (((prev.flags >> 9) % 2) == 1)
					fprintf( outfile,
						 " COMPRESSED" );
                                    if (prev.am_ind >= 0) {
                                        if (stripowner)
                                            fprintf( outfile,
                                                     " USING %s",
                                                     prev.am_name );
                                        else
                                            fprintf( outfile,
                                                     " USING '%s'.%s",
                                                     prev.am_owner,
                                                     prev.am_name );
					EXEC SQL
					  SELECT am_param, am_space
					  INTO :systabamdata.am_param, :systabamdata.am_space
					  FROM systabamdata
					  WHERE tabid = :prev.tabid;
					if (sqlca.sqlcode != 100) {
					    fprintf( outfile,
						     "(%s)",
						     systabamdata.am_param );
					}
				    } else {
					fprintf( outfile,
						 " LOCK MODE %s",
						 prev.locklevel[0] == 'R'
						 ? "ROW" :
						 (prev.locklevel[0] == 'P'
						  ? "PAGE" : "PAGE, ROW") );
				    }
				    if (!(prev.flags & InHerFlag) && (prev.xtd_ind < 0 && Universal)) {
					if (prev.statind >= 0 && prev.statchange >= 0) {
					    fprintf( outfile,
						     " STATCHANGE %d\n",
						     prev.statchange );
					}
					if (prev.statlevelind >= 0
					    && (major_vers > 11 || (major_vers == 11 && minor_vers >= 70))) 
					{
					    switch (prev.statlevel[0]) {
					    case 'A':
						if (forceautostats)
						    fprintf( outfile,
							     " STATLEVEL AUTO\n" );
						break;
					    case 'F':
						fprintf( outfile,
							 " STATLEVEL FRAGMENT\n" );
						break;
					    case 'T':
						fprintf( outfile,
							 " STATLEVEL TABLE\n" );
						break;
					    }
					}
				    }
                                    fprintf( outfile, ";\n\n" );
                                    if (do_filtering && prev.viol_ind == -1) {
                                        fprintf( outfile2,
                                                 "START VIOLATIONS TABLE FOR %s;\n",
                                                 prev.tabname );
                                    }
                                } else {
                                    fprintf( outfile, ";\n\n" );
                                    if (prev.audpath[0] != (char)0 &&
                                        prev.audpath[0] != ' ') 
                                    {
                                        if (!stripowner)
                                            fprintf( outfile,
                                                     "CREATE AUDIT FOR "
                                                     "\"%s\".%s IN '%s';\n\n",
                                                     prev.owner,
                                                     delimident( prev.tabname ),
                                                     prev.dirpath );
                                        else
                                            fprintf( outfile,
                                                     "CREATE AUDIT FOR "
                                                     "%s IN '%s';\n\n",
                                                     delimident( prev.tabname ),
                                                     prev.dirpath );
                                    }
                                }
                                if (!loadflag) {
				    if (!NoExtentClause 
				        && SMICounts
					&& prev.fextsize < (pagecounts[pc].maxused)
					&& prev.am_ind < 0)
				    {
                                        fprintf( outfile,
                                                 "{\n\tCurrent first extent (%d) too small!\n"
                                                 "\tActual pages used by ",
                                                 prev.fextsize );
                                        if (pagecounts[pc].nfrags == 1)
                                            fprintf( outfile,
                                                     "table has been used.\n" );
                                        else
                                            fprintf( outfile,
                                                     "%s fragment has been used.\n",
                                                     HowName[HowExtents] );
                                    } else if (!NoExtentClause
					       && SMICounts
					       && prev.fextsize > 16
					       && pagecounts[pc].maxused
					       && prev.fextsize > (int)((double)(pagecounts[pc].maxused) * 1.25)
					       && prev.am_ind < 0)
                                    {
                                        if (dominimize && fe < prev.fextsize)
                                            fprintf( outfile,
                                                     "{\nCurrent first extent (%d) was too large!\n"
                                                     "\tActual pages used by table were substituted.\n",
                                                     prev.fextsize );
                                        else if (pagecounts[pc].nfrags <= 1)
                                            fprintf( outfile,
                                                     "{\nCurrent first extent (%d) too large!\n"
                                                     "\tActual pages used by table should be used "
                                                     "instead.\n",
                                                     prev.fextsize );
                                        else
                                            fprintf( outfile,
                                                     "{\nCurrent first extent (%d) too large!\n"
                                                     "\tActual pages used by %s fragment should be used "
                                                     "instead.\n",
                                                     prev.fextsize,
                                                     HowName[HowExtents] );
                                    } else if (!NoExtentClause && prev.am_ind < 0) {
                                        fprintf( outfile, "{\n" );
				    } else {  /* NoExtentClause with no other extent options. */
                                        fprintf( outfile, "\n" );
				    }
                                    if (!NoExtentClause && AdjExt != 1.0 && prev.am_ind < 0) {
                                        fprintf( outfile,
                                                 "\tEXTENT SIZE adjusted to %d%% of calculated EXTENT "
                                                 "SIZE, Current: %d.\n",
                                                 (int)((100.0 * AdjExt) + 0.505),
                                                 prev.fextsize );
                                    }
                                    if (!NoExtentClause && AdjNext > 0.0 && prev.am_ind < 0) {
                                        fprintf( outfile,
                                                 "\tNEXT SIZE adjusted to %d%% of %sEXTENT "
                                                 "SIZE, Current: %d.\n",
                                                 (int)((100.0 * AdjNext) + 0.505),
                                                 (AdjExt != 1.0 ? "adjusted " : ""),
                                                 prev.nextsize );
                                    }
                                    if (SMICounts && pc >= 0)
                                        if (pagecounts[pc].nfrags > 1)
                                            fprintf( outfile,
                                                     "\tTablespace header reports -"
                                                     "\n\t\tTotal:\t\t\tpgsused=%d, pgsdata=%d."
                                                     "\n\t\t%s fragment:"
                                                     "\tpgsused=%d, pgsdata=%d\n",
                                                     pagecounts[pc].pgsused,
                                                     pagecounts[pc].pgsdata,
                                                     HowName[HowExtents],
                                                     pagecounts[pc].maxused,
                                                     pagecounts[pc].maxdata );
                                        else
                                            fprintf( outfile,
                                                     "\tTablespace header reports -"
                                                     "\n\t\tTotal: \tpgsused=%d, pgsdata=%d.\n",
                                                     pagecounts[pc].pgsused,
                                                     pagecounts[pc].pgsdata );
                                    if (prev.tabid > 99 && pc >= 0) {
				        if (!NoExtentClause && prev.am_ind < 0) {
					    fprintf( outfile,
						     "\tPlease review extent sizing and adjust to allow "
						     "for growth.\n}\n\n" );
					}
                                    } else if (pc < 0) {
                                        fprintf( outfile,
                                                 "\tSMI pseudo-table.\n}\n\n" );
                                    } else if (prev.tabid <= 99) {
                                        fprintf( outfile,
                                                 "\tSystem catalog table.\n}\n\n" );
				    }
                                }
                                if (doViolations)
                                    print_violations( outfile );
                                if (prev.tabtype[0] == 'T' && revauths)
                                    if (!stripowner)
				      if (grantas)
                                        fprintf( outfileG,
                                                 "REVOKE ALL ON \"%s\".%s "
                                                 "FROM public AS \"%s\";\n",
                                                 prev.owner,
                                                 delimident( prev.tabname ),
						 prev.owner );
				      else
                                        fprintf( outfileG,
                                                 "REVOKE ALL ON \"%s\".%s "
                                                 "FROM public;\n",
                                                 prev.owner,
                                                 delimident( prev.tabname ) );
                                    else if (grantas)
                                        fprintf( outfileG,
                                                 "REVOKE ALL ON %s FROM "
                                                 "public AS \"%s\";\n",
                                                 delimident( prev.tabname ),
						 prev.owner );
				    else
                                        fprintf( outfileG,
                                                 "REVOKE ALL ON %s FROM "
                                                 "public;\n",
                                                 delimident( prev.tabname ) );
                                if (ids_server) {
                                    if (prev.xtd_ind >= 0 && Universal) {
                                        print_alters( outfile, pc );
                                    }
                                    print_indexes( IndexFile, outfile3, SimpleFrags );
                                } else
                                    print_seindexes( IndexFile, outfile3 );
                                if (doDistrib) {
                                    print_distributions( outfile3,
                                                         prev.tabid,
                                                         prev.tabname,
                                                         columns );
				}
                                if (do_filtering) {
                                    fprintf( outfile2,
                                             "SET INDEXES FOR %s FILTERING WITHOUT ERROR;\n\n",
                                             prev.tabname );
                                }
				print_auths( prev.tabid,
					     prev.tabname,
					     outfileG,
					     prev.owner,
					     tabauths,
					     colauths,
					     grantas );
                            }
                        }
                    } else {
                        pass_no = 0;
                    }
		    memset( &prev, 0, sizeof prev );
                    memcpy( &prev, &systables, sizeof prev );
		    strcpy( prev_owner, original_owner );
		    strcpy( prev_syssecpolicy, syssecpolicy );

		    if (systables.tabtype[0] == 'E')
		      is_external_table = 1;
		    else
		      is_external_table = 0;
                    if (!last_record) {
                        switch (systables.tabtype[0]) {
                        case 'T': {
			    if (verbose) {
				fflush( outfile );
				fprintf( stderr,
					 "    Processing %stable %s.\n",
					 (systables.flags & RawFlag) ? "raw " : "",
					 systables.tabname );
			    }
                            if (!doreorg) {
                                if (summary || loadflag) {
                                    /* Get index key length and actual SMI nrows. */
                                    obtainidxsize( database,
                                                   systables.tabname,
                                                   systables.tabid,
                                                   &indexlength,
                                                   &systables.nrows );
                                    sprintf( tname, "%1.5s", systables.tabname );
                                    while (strlen( tname ) < 5)
                                        strcat( tname, "_" );
					sprintf( lname, "%s%05d", tname, systables.tabid );
				    if (!stripowner)
					fprintf( outfile,
						 "{ TABLE \"%s\".%s row size = %d number of columns = %d ",
						 systables.owner,
						 delimident( systables.tabname ),
						 systables.rowsize,
						 systables.ncols );
				    else
					fprintf( outfile,
						 "{ TABLE %s row size = %d number of columns = %d ",
						 delimident( systables.tabname ),
						 systables.rowsize,
						 systables.ncols );
                                    fprintf( outfile,
                                             "index size = %d }\n", indexlength );

                                }
                                if (loadflag) {
                                    fprintf( outfile,
                                             "{ unload file name = %s.unl number of rows = %lld }\n\n",
                                             lname, (long long)systables.nrows );
                                }
                            }
                            if (SMICounts) {
                                pc = findcounts( database, systables.tabname );
                                if (verbose > 1) {
                                    fprintf( stderr,
                                             "Pagecounts[%d] {\n"
                                             "    dbsname:  %s\n"
                                             "    tabname:  %s\n"
                                             "    nfrags:   %d\n"
                                             "    pgsused:  %d\n"
                                             "    pgsdata:  %d\n"
                                             "    pgstotal: %d\n"
                                             "    maxdata:  %d\n"
                                             "    maxused:  %d\n",
                                             pc,
                                             pagecounts[pc].dbsname,
                                             pagecounts[pc].tabname,
                                             pagecounts[pc].nfrags,
                                             pagecounts[pc].pgsused,
                                             pagecounts[pc].pgsdata,
                                             pagecounts[pc].pgstotal,
                                             pagecounts[pc].maxdata,
                                             pagecounts[pc].maxused );
                                }
                                fe = systables.fextsize > (pagecounts[pc].maxused) ?
                                    systables.fextsize : (pagecounts[pc].maxused);
                                if (dominimize) {
				    if (verbose > 1) {
					fprintf( stderr,
						 "Pagecounts[%d] {\n"
						 "    dbsname:  %s\n"
						 "    tabname:  %s\n"
						 "    nfrags:   %d\n"
						 "    pgsused:  %d\n"
						 "    pgsdata:  %d\n"
						 "    pgstotal: %d\n"
						 "    maxdata:  %d\n"
						 "    maxused:  %d\n",
						 pc,
						 pagecounts[pc].dbsname,
						 pagecounts[pc].tabname,
						 pagecounts[pc].nfrags,
						 pagecounts[pc].pgsused,
						 pagecounts[pc].pgsdata,
						 pagecounts[pc].pgstotal,
						 pagecounts[pc].maxdata,
						 pagecounts[pc].maxused );
				    }
				    if (pagecounts[pc].pgsdata)
					fe = (pagecounts[pc].pgsdata < fe) ?
					    (int)((double)pagecounts[pc].pgsdata) : fe;
				    else
					fe = minext;
				}
                            } else {
                                fe = systables.fextsize;
                            }
                            if (!NoExtentClause && verbose > 1 && prev.am_ind < 0) {
                                fprintf( stderr,
                                         "Line %d: Extent size adjustment %d ==> %d.\n",
                                         __LINE__,
                                         fe,
                                         max( (int)(AdjExt * (double)fe), minext ) );
                            }
                            fe = max( (int)(AdjExt * (double)fe), minext );
                            if (!stripowner) {
				char ddbase[135], dname[135];

                                if (doreorg == 1) {
                                    fprintf( outfile,
                                             "\nALTER TABLE \"%s\".%s MODIFY NEXT SIZE %d;\n",
                                             systables.owner,
                                             delimident( systables.tabname ),
                                             AdjNext > 0.0 ? max( ((int)(0.505 + ((float)fe * AdjNext))), minnext ) : systables.nextsize );
                                    fprintf( outfile,
                                             "ALTER FRAGMENT ON TABLE \"%s\".%s INIT",
                                             systables.owner, 
                                             delimident( systables.tabname ) );
				} else if (doreorg == 2) {
				    delimident2( databasename, ddbase );
				    delimident2( systables.tabname, dname );
				    fprintf( outfile,
					     "execute function task( 'defragment', '%s:\"%s\".%s' );",
					     ddbase,
					     systables.owner,
					     dname );
                                } else
                                    fprintf( outfile,
                                             "CREATE %sTABLE%s \"%s\".%s ",
                                             (systables.flags & RawFlag) ? "RAW " : (systables.flags & CollectionFlag ? "COLLECTION " : ""),
					     create_conditionally ? " IF NOT EXISTS" : "",
                                             systables.owner, 
                                             delimident( systables.tabname ) );
                            } else {
				char ddbase[135], dname[135];

                                if (doreorg == 1) {
                                    fprintf( outfile,
                                             "\nALTER TABLE %s MODIFY NEXT SIZE %d;\n",
                                             delimident( systables.tabname ),
                                             AdjNext > 0.0 ? max( ((int)(0.505 + ((float)fe * AdjNext))), minnext ) : systables.nextsize );
                                    fprintf( outfile,
                                             "ALTER FRAGMENT ON TABLE %s INIT",
                                             delimident( systables.tabname ) );
				} else if (doreorg == 2) {
				    delimident2( databasename, ddbase );
				    delimident2( systables.tabname, dname );
				    fprintf( outfile,
					     "execute function task( 'defragment', '%s:\"%s\".%s' );",
					     ddbase,
					     systables.owner,
					     dname );
                                } else
                                    fprintf( outfile,
                                              "CREATE %sTABLE%s %s ",
                                             (systables.flags & RawFlag) ? "RAW " : (systables.flags & CollectionFlag ? "COLLECTION " : ""),
					     create_conditionally ? " IF NOT EXISTS" : "",
                                             delimident( systables.tabname ) );
                            }

			    OK = 0;
			    if (!access_data_OK && strlen(systables.am_name) > 0) {
				snprintf( am_match, sizeof am_match, "*,%s,*", systables.am_name );
				OK += ((fnmatch( am_match, no_data_accesses, 0 ) == FNM_NOMATCH) ? 1 : 0);
			    } else {
				am_match[0] = (char)0;
				OK++;
			    }

			    if (!tables_data_OK) {
				snprintf( tbl_match, sizeof tbl_match, "*,%s,*", systables.tabname );
				OK += ((fnmatch( tbl_match, no_data_tables, 0 ) == FNM_NOMATCH) ? 1 : 0);
			    } else {
				tbl_match[0] = (char)0;
				OK++;
			    }

			    if (OK == 2) {
				export_OK = 1;
			    } else {
				export_OK = 0;
			    }

			    if ((do_myexport_scripts && export_OK) || do_export || do_import) {
				char impname[129], dimpname[151];
				char *dir = (char *)NULL;
				
				if( (dir = getenv( "MYEXPORT_PATH" )) == (char *)NULL)
				    dir = getenv( "PWD" );

				snprintf( impname,
					  sizeof impname,
					  "%.119s%s",
					  systables.tabname,
					   do_import ? "_external" : "_myimport" );
				delimident2( impname, dimpname );

				if (do_import || (do_myexport_scripts && export_OK)) {
				    snprintf( filenameMI, 
					      sizeof filenameMI, 
					      "%s/%s.%s%s.sql", 
					      dir, 
					      databasename, 
					      systables.tabname,
					      do_import ? ".import" : ".myimport" );
				    outfileMI = fopen( filenameMI, "w" );
				    if (!outfileMI) {
					fprintf( stderr,
						 "Cannot open myimport script file: %s. Errno: %d.\n",
						 filenameMI,
						 errno );
					rexit(3);
				    }
				}

				/* Import script. */
				if (do_import || (do_myexport_scripts && export_OK)) {
				    if (major_vers == 12 || (major_vers == 11 && minor_vers == 70)) {
					fprintf( outfileMI, "SET ENVIRONMENT DEFAULTESCCHAR '%s';\n", myexport_escape );
				    }

				    fprintf( outfileMI,
					     "CREATE EXTERNAL TABLE IF NOT EXISTS %s ( \n",
					     dimpname );
				}

				snprintf( impname,
					  sizeof impname,
					  "%.119s%s",
					  systables.tabname,
					   do_export ? "_external" : "_myexport" );
				delimident2( impname, dimpname );

				/* Export script */
				if (do_export || (do_myexport_scripts && export_OK)) {
				    snprintf( filenameME, 
					      sizeof filenameME, 
					      "%s/%s.%s%s.sql", 
					      dir, 
					      databasename, 
					      systables.tabname,
					      do_export ? ".export" : ".myexport" );

				    outfileME = fopen( filenameME, "w" );
				    if (!outfileME) {
					fprintf( stderr,
						 "Cannot open myexport script file: %s. Errno: %d.\n",
						 filenameME,
						 errno );
					rexit(3);
				    }
				    
				    /* Export script. */
				    if (major_vers == 12 || (major_vers == 11 && minor_vers == 70)) {
					fprintf( outfileME, "SET ENVIRONMENT DEFAULTESCCHAR '%s';\n", myexport_escape );
				    }
				    
				    if (strlen( myexport_isolation ) > 0) {
					fprintf( outfileME,
						 "SET ISOLATION %s;\n",
						 myexport_isolation );
				    }
				    if (No_Triggers) {
					sel_trig_list[0] = (char)0;
					EXEC SQL OPEN seltrig_c USING :systables.tabid;
					if (sqlca.sqlcode != 0) {
					    fprintf( stderr, "Cannot open seltrig_c. Error: %d.\n", sqlca.sqlcode );
					    rexit( 1 );
					}
					while (TRUE) {
					    EXEC SQL FETCH seltrig_c INTO :trgname;
					    if (sqlca.sqlcode == SQLNOTFOUND)
						break;
					    else if (sqlca.sqlcode != 0) {
						fprintf( stderr, "Cannot fetch seltrig_c. Error: %d.\n", sqlca.sqlcode );
						rexit( 1 );
					    }
					    if (strlen( sel_trig_list ) > 0)
						strcat( sel_trig_list, "," );
					    strcat( sel_trig_list, trgname );
					}
					if (strlen( sel_trig_list ) > 0)
					    fprintf( outfileME,
						     "SET TRIGGERS %s DISABLED;\n",
						     sel_trig_list );
				    }
				    fprintf( outfileME,
					     "CREATE EXTERNAL TABLE IF NOT EXISTS %s ( \n",
					     dimpname );
				}
			    }
                        }
                        break;
                        case 'P':
			  if (NoPrivSyns == TRUE) {
			        continue;
			  }
                        case 'S':
			    if (verbose) {
				fflush( outfile );
				fprintf( stderr, "    Processing synonym %s.\n", systables.tabname );
			    }
                            if (!doreorg)
                                print_synonyms( SynonymFile, outfileG == outfile2 ? SynonymFile : outfileG, tabauths, colauths, grantas );
                            continue;  /* Skip to next table, we're done. */
                        case 'V':
			    if (verbose) {
				fflush( outfile );
				fprintf( stderr, "    Processing view %s.\n", systables.tabname );
			    }
                            if (!doreorg) {
                                print_views( systables.tabid,
                                             systables.tabname,
                                             systables.owner,
                                             ViewFile,
                                             outfileG == outfile2 ? ViewFile : outfileG,
                                             tabauths,
                                             colauths,
                                             grantas );
				print_triggers( ((ViewsOnly || haveViewFile) ? ViewFile : TrigFile), systables.tabtype, systables.tabid );
				print_auths( prev.tabid,
					     prev.tabname,
					     outfileG == outfile ? ViewFile : outfileG,
					     prev.owner,
					     tabauths,
					     colauths,
					     grantas );
			    }
                            continue;  /* Skip to next table, we're done. */
			case 'E':
			    if (verbose) {
				fflush( outfile );
				fprintf( stderr, "    Processing external table %s.\n", systables.tabname );
			    }
			    if (!stripowner)
			        fprintf( outfile,
				         "CREATE EXTERNAL TABLE%s \"%s\".%s ",
					 create_conditionally ? " IF NOT EXISTS" : "",
				         systables.owner,
				         delimident( systables.tabname ) );
			    else
			        fprintf( outfile,
				         "CREATE EXTERNAL TABLE%s %s ",
					 create_conditionally ? " IF NOT EXISTS" : "",
				         delimident( systables.tabname ) );
			    break;
                        }
                    }

                    if (def_open) {
                        EXEC SQL close def;
                        def_open = 0;
                    }
                    EXEC SQL OPEN def USING :systables.tabid;
                    if ( sqlca.sqlcode ) {
                        fprintf( stderr,
                                 "Cannot open def cursor; Error = %d\n",
                                 sqlca.sqlcode );
                        rexit(1);
                    } else { /* Prime the pump! */
                        def_open = 1;
                        EXEC SQL
                            FETCH def
                            INTO :sysdefaults.tabid,
                            :sysdefaults.colno,
                            :sysdefaults.type,
                            :sysdefaults.deflt;
                        if (sqlca.sqlcode < 0) {
                            fprintf( stderr,
                                     "Cannot make initial FETCH cursor def. "
                                     "Err: %d, ISAM: %d\n",
                                     sqlca.sqlcode,
                                     sqlca.sqlerrd[1] );
                            rexit(18);
                        }
                    }
                } else if (systables.tabtype[0] == 'V')
                    continue; /* Ignore view columns after the first. */
                if (last_record)
                    break;
            }

	    if (!doreorg) {
                if (systables.xtd_ind >= 0 && Universal) {
                    /* Only do this once. */
                    if (syscolumns.colno == 1)
                        if (stripowner)
                            fprintf( outfile,
                                     " OF TYPE %s ",
                                     systables.xtd_name );
                        else
                            fprintf( outfile,
                                     " OF TYPE '%s'.%s ",
                                     UseDatabaseOwner ? systables.owner : systables.xtd_owner,
                                     systables.xtd_name );
		    if ((do_myexport_scripts && export_OK) || do_export || do_import) {
			if (syscolumns.colno != 1 ) {
			    if ((do_myexport_scripts && export_OK) || do_import)
				fprintf( outfileMI, ",\n" );
			    if ((do_myexport_scripts && export_OK) || do_export)
				fprintf( outfileME, ",\n" );
			}
			if ((do_myexport_scripts && export_OK) || do_import) {
			    fprintf( outfileMI, "\t%s ", syscolumns.colname );
			    print_coltype( 0,
					   outfileMI,
					   systables.tabid,
					   systables.tabname,
					   systables.owner,
					   syscolumns,
					   pc );
			}
			if ((do_myexport_scripts && export_OK) || do_export) {
			    fprintf( outfileME, "\t%s ", syscolumns.colname );
			    print_coltype( 0,
					   outfileME,
					   systables.tabid,
					   systables.tabname,
					   systables.owner,
					   syscolumns,
					   pc );
			}
		    }
                } else {
		    if (syscolumns.colattr == 0 || syscolumns.colattr == 128) {
			/* Skip printout for hidden columns:
			   colattr == 3 :: VERCOLS
			   colattr == 9 :: REPLCHECK
			   colattr == 17, 33, or 65 :: ERKEY
			*/
			if (syscolumns.colno == 1) {
			    fprintf( outfile, "(\n" );
			} else {
			    fprintf( outfile, ",\n" );
			    if (((do_myexport_scripts && export_OK) || do_export || do_import) && systables.tabtype[0] == 'T') {
				if (do_import || (do_myexport_scripts && export_OK))
				    fprintf( outfileMI, ",\n" );
				if (do_export || (do_myexport_scripts && export_OK))
				    fprintf( outfileME, ",\n" );
			    }
			}
			while (sysdefaults.colno < syscolumns.colno) {
			    EXEC SQL
				FETCH def
				INTO :sysdefaults.tabid,
				:sysdefaults.colno,
				:sysdefaults.type,
				:sysdefaults.deflt;
			    if (sqlca.sqlcode != 0) {
				break;
			    }
			}
			if (sqlca.sqlcode == SQLNOTFOUND) {
			    sysdefaults.tabid = INT_MAX;
			    sysdefaults.colno = SHRT_MAX;
			} else if (sqlca.sqlcode < 0) {
			    fprintf( stderr,
				     "Cannot FETCH cursor def. Err: %d, ISAM: %d\n",
				     sqlca.sqlcode,
				     sqlca.sqlerrd[1] );
			    rexit(18);
			}
			if (systables.tabtype[0] == 'V')
			    continue;
			allows_nulls = !(syscolumns.coltype & SQLNONULL);
			fprintf( outfile, "\t%s ", syscolumns.colname );
			hold_ret = print_coltype( 0,
						  outfile,
						  systables.tabid,
						  systables.tabname,
						  systables.owner,
						  syscolumns,
						  pc );
			if (type_ret == 0) type_ret = hold_ret;
			if (sysdefaults.tabid == systables.tabid &&
			    sysdefaults.colno == syscolumns.colno)
			    print_defaults( outfile, syscolumns, sysdefaults );
			if (!allows_nulls)
			    fprintf( outfile, " NOT NULL" );

			if (((do_myexport_scripts && export_OK) || do_export || do_import) && systables.tabtype[0] == 'T') {
			    if (do_import || (do_myexport_scripts && export_OK)) {
				fprintf( outfileMI, "\t%s ", syscolumns.colname );
				print_coltype( 0,
					       outfileMI,
					       systables.tabid,
					       systables.tabname,
					       systables.owner,
					       syscolumns,
					       pc );
			    }
			    if (do_export || (do_myexport_scripts && export_OK)) {
				fprintf( outfileME, "\t%s ", syscolumns.colname );
				print_coltype( 0,
					       outfileME,
					       systables.tabid,
					       systables.tabname,
					       systables.owner,
					       syscolumns,
					       pc );
			    }
			}
		    }		}
                /* Copy for indexes. */
                strcpy( columns[syscolumns.colno - 1], syscolumns.colname );
            }
        }
	if (ViewsOnly == 1)
	    return 0;

        if (!doreorg) {
	    if (verbose) {
		fflush( outfile );
		fprintf( stderr, "Processing constraints.\n" );
	    }
            if (ids_server) {
		if (NoValidateChk || AlterChecks)
		    print_checks( ConstrFile, 1 );
                print_constraints( ConstrFile, tflag );
                if (dorefs) {
		    if (verbose) {
			fflush( outfile );
			fprintf( stderr, "Processing references to selected table(s).\n" );
		    }
                    print_references( ConstrFile, tflag );
                    print_allviewrefs( ConstrFile, outfileG, tabauths, colauths, grantas );
                }
            } else { 
                print_seconstraints( ConstrFile, tflag );
                if (dorefs) {
		    if (verbose) {
			fflush( outfile );
			fprintf( stderr, "Processing references to selected table(s).\n" );
		    }
                    print_sereferences( ConstrFile, tflag );
                    print_allviewrefs( ConstrFile, outfileG, tabauths, colauths, grantas );
                }
            }
        }
    } else if (spflag) {
        if (Universal) {
            print_langauths( outfileG );
            if (!NoTypes)
                print_types( outfile, outfileG );
            print_procedures( ProcFile, outfileG == outfile ? ProcFile : outfileG, ptempl, grantas, TRUE, FALSE );
        }
    } else if (TypesOnly) {
        print_types( outfile, outfileG );
    }
    /* Now do aggregates and procedures. */
    if (!doreorg && !TypesOnly) {
        if ((doall || doaggs) && Universal) {
	    if (verbose) {
		fflush( outfile );
		fprintf( stderr, "Processing user defined aggregates.\n" );
	    }
            print_aggregates( outfile2 );
	}
        if (spflag || doall) {
	    if (verbose) {
		fflush( outfile );
		fprintf( stderr, "Processing procedures.\n" );
	    }
            print_procedures( ProcFile, outfileG == outfile ? ProcFile : outfileG, ptempl, grantas, FALSE, FALSE );
	}
        if (tflag || doall) {
	    if (verbose) {
		fflush( outfile );
		fprintf( stderr, "Processing triggers. \n" );
	    }
            print_triggers( TrigFile, "T", 0 );
	}
    }

    /* Do External Directives. */
    if (!suppress_directives && !tflag && !doreorg) {
	if (verbose) {
	    fflush( outfile );
	    fprintf( stderr, "Processing external directives. \n" );
	}
	print_directives( outfile, outfile2 );
    }

    fputs( "\n", outfile );

    fflush( outfile );
    if (verbose) {
	fflush( outfile );
        fprintf( stderr, "Finished.\n" );
    }

    return type_ret;
}

void rexit(int stat)
{
    if (stat != 99 && DatabaseConnected == 1)
        EXEC SQL
            CLOSE DATABASE;
    exit(stat);
}

char *strtoupper( char *out, char *in )
{
    while (*in) {
        *out = toupper( *in );
        out++;
        in++;
    }
    *out = (char)0;

    return out;
}

int findcounts( char *db, char *tb )
{
    int i;

    for (i=0; i<ntables; i++)
        if (strcmp( db, pagecounts[i].dbsname ) == 0
            && strcmp( tb, pagecounts[i].tabname ) == 0)
            return i;
    if (strcmp( db, "sysmaster" ) != 0) {
        fprintf( stderr, "Internal pagecount error(%s:%s).\n", db, tb );
        rexit( 10 );
    }

    return -1;
}

char *delimident( char *ident )
{
    static char dident[150];

    sprintf( dident,
             "%s%s%s",
             QuoteIdent ? "\"" : "",
             ident,
             QuoteIdent ? "\"" : "" );

    return dident;
}

char *delimident2( char *ident, char *dident )
{
    sprintf( dident,
             "%s%s%s",
             QuoteIdent ? "\"" : "",
             ident,
             QuoteIdent ? "\"" : "" );

    return dident;
}

int read_pwd( char *pwd_param )
{
    char pwd_file[4097];
    struct stat stats;
    FILE *f_pwd;
    int len;
    uid_t uid;

    uid = getuid();

    strncpy( pwd_file, &pwd_param[1], sizeof pwd_file );
    if (stat( pwd_file, &stats ) != 0) {
        return 0;
    }

    if (stats.st_uid != uid) {
        return 0;
    }

    if (S_ISREG(stats.st_mode) == 0 || (stats.st_mode & 00777) != 00600) {
        return 0;
    }

    f_pwd = fopen( pwd_file, "r" );
    if (fgets( pwd_param, sizeof pwd_file, f_pwd ) == (char *)NULL) {
        return 0;
    }

    if ((len = strlen( pwd_param )) > 0)
	pwd_param[len-1] = (char)0;

    fclose( f_pwd );

    return 1;
}
