/* myschema.h - "C" version of header file myschema.ec
*/
/*
   RCS Header:
   -----------
   $Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/myschema.h,v 2.45 2020/05/10 18:42:36 art Exp $

   RCS Log:
   --------
   $Log: myschema.h,v $
   Revision 2.45  2020/05/10 18:42:36  art
   Fixed the dbspace information statement and structure.

   Revision 2.44  2018/02/26 23:39:57  art
   Changed npused to double to reflect new catalog tables.

   Revision 2.43  2015/05/07 13:36:35  art
   Added code to print out VTI Access Method parameter lists for virtual tables.

   Revision 2.42  2013/12/29 20:03:19  art
   Added support for v12.10 ROLLING fragments DETACH and DISCARD keywords.
   Support for DETACH (ANY|INTERVAL FIRST|INTERVAL ONLY) no in there yet
   because the documented syntax doesn't work!

   Revision 2.41  2013/11/20 15:18:06  art
   Added #define of atoll() to strtoll() for HPUX users in the header file.

   Revision 2.40  2013/06/20 17:47:32  art
   Looks like sysviews.viewtext was expanded from 64 bytes to 256 bytes in V12! Fixed.

   Revision 2.39  2013/01/24 00:03:49  art
   Added Centaurus feature support.

   Revision 2.38  2012/08/08 21:35:58  art
   Fixed constraints printouts to support owners.  Full support for -O and --set-owner implemented.

   Revision 2.37  2012/01/31 19:00:57  art
   Better fix for big nrows problem.

   Revision 2.36  2012/01/31 17:54:21  art
   Fixed problem with obtaining row counts for tables with more than 2^31 rows.

   Revision 2.35  2011/07/18 21:14:02  art
   Added isproc column to sysprocedures structure.

   Revision 2.34  2011/04/29 13:35:51  art
   Fixed problems handling SERIAL8 type column initialization.
   Added support for STATLEVEL and STATCHANGE in CREATE TABLE statements.
   Added support for SECURITY POLICY in CREATE TABLE statements.  Output of policy
   definitions still not supported.
   Added option --force_autostats to force the output of STATLEVEL AUTO which is
   the default and so not printed normally.

   Revision 2.33  2010/11/17 20:49:40  art
   Added support for WITH REPLCHECK, WITH ERKEY, and WITH AUDIT clauses in CREATE TABLE.

   Revision 2.32  2010/11/08 20:40:31  art
   Added new --create_conditionally feature to optionally add " IF NOT EXISTS " clauses to c
   create object statements.  Feature supported by 11.70.

   Revision 2.31  2010/05/04 23:23:32  art
   Fixed sequence output for IDS 10.00 and 11.10 (oops).
   Implemented new option --sequence, --sequence=matches-spec to print only sequence
   definitions.

   Revision 2.30  2010/03/24 16:10:13  art
   Modified sequence code to acquire the current/last sequence value from
   sysactptnhdr in sysmaster if the database has logging (no remote join
   to sysmaster from a non-logged database).

   Revision 2.29  2010/03/04 14:41:01  art
   A new option to ignore instrumentation routines added by AGS Server Studio
   needed to debug stored procedures in SS.

   Revision 2.28  2010/01/06 13:55:18  art
   Added new option --set-owner to override all object owners with a specified owner name.

   Revision 2.27  2009/11/23 18:47:05  art
   Added support for EXTERNAL tables (11.50xc6+).
   Improved 64bit compiler support.  Removed warnings.

   Revision 2.26  2009/06/03 18:25:25  art
   Removed stdint.h.  Not needed and not available everywhere.

   Revision 2.25  2009/03/20 14:09:22  art
   Fixed extent calculations to take account for differing page sizes, even
   between fragments of the same table.

   Revision 2.24  2009/03/15 23:02:21  art
   y
   y

   Revision 2.23  2009/02/15 21:42:41  art
   Fixed SERIAL, SERIAL8, and BIGSERIAL handling.  BIGINT and BIGSERIAL handling added.

   Revision 2.22  2009/01/14 17:34:34  root
   Added support for VERCOLS.

   Revision 2.21  2009/01/01 17:58:25  art
   Removed RCSWhat ident string which was BB specific.

...

   Revision 1.1  1999/09/08 14:57:39  kagel
   Initial revision


*/

#if defined( __hpux )
#define atoll( t ) strtoll( (t), (char **)NULL, 10 )
#endif

#if !defined( myschemaHEADER )
#define myschemaHEADER 1
static const char myschemaRCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/myschema.h,v 2.45 2020/05/10 18:42:36 art Exp $";

#if !defined( FALSE )
#define FALSE (0)
#define TRUE (!FALSE)
#endif

#define string char
#define fixchar char

#define IDBYTES (129)
#define UNAMELEN (33)

/* DEFINITIONS SHARED BY MAINLINE AND DEPENDENT CODE. */

typedef struct s_pagecounts {
    string dbsname[IDBYTES];
    string tabname[IDBYTES];
    int serialv;
    long long serialv8;
    long long serialvBig;
    int nfrags;
    int pgsused;
    int pgsdata;
    int pgstotal;
    int maxdata;
    int maxused;
} t_pagecounts;

typedef struct s_idxcounts {
    int keylen, detlen;
    double nrows;
    string dbsname[IDBYTES],
	tabname[IDBYTES];
} t_idxcounts;

typedef struct S_BLOBBUFF {
    char *buff;
    int bbsize, global_length;
} t_blobbuff;


#if defined(NEED_SQL_TYPES)
/* Definitions only needed in .ec files. */
 
typedef struct SYSROLEAUTH_S {
        string rolename[UNAMELEN];
        string grantee[UNAMELEN];
        string is_grantable[2];
} sysroleauth_t;

typedef struct systables_s { 
    string tabname[IDBYTES], owner[UNAMELEN], dirpath[65], audpath[65];
    int tabid;
    int partnum, fextsize, nextsize;
    char locklevel[2];
    double npused;
    char tabtype[2];
    double nrows;
    int ncols;
    int rowsize;
    int flags;
    int has_violations;
    int vercols_flag;
    int replcheck_flag;
    int erkey_flag;
    int pagesize;
    int secpolicyid;
    short statchange;
    char statlevel[2];
    char am_owner[UNAMELEN], am_name[IDBYTES];
    char xtd_owner[UNAMELEN], xtd_name[IDBYTES];
    char par_owner[UNAMELEN], par_name[IDBYTES];
    short am_ind, xtd_ind, par_ind, statind, statlevelind;
} systables_t;

typedef struct syscolumns_s {
    int tabid;
    string colname[IDBYTES];
    short colno;
    short coltype;
    short collength;
    int colmin;
    int colmax;
    int extended_id;
} syscolumns_t;

typedef struct sysindexes_s {
    char idxname[IDBYTES];
    char owner[UNAMELEN];
    int tabid;
    char idxtype[2], clustered[2];
    short part[16];
    int op_class_id[16];
    int amid, funcid;
    int indexattr;
} sysindexes_t;

typedef struct sysconstraints_s {
    int constrid;
    string constrname[IDBYTES];
    int tabid;
    string constrtype[2],
	idxname[IDBYTES];
    string owner[UNAMELEN];
} sysconstraints_t;

typedef struct SYSCOLDEPEND_S {
    int constrid;
    int tabid;
    short colno;
} syscoldepend_t;

typedef struct SYSCHECKS_S {
    int constrid;
    string type[2];
    short seqno;
    string checktext[33];
} syschecks_t;
 
typedef struct sysdefaults_s {
    int tabid;
    short colno;
    string type[2];
    string deflt[257];
} sysdefaults_t;

typedef struct SYSFRAGMENTS_S {
    char fragtype[2];
    int tabid;
    char indexname[IDBYTES];
    char strategy[2];
    int evalpos;
    double nrows;
    double npused;
    int colno;
    int flags;
    loc_t exprtext;
    string dbspace[IDBYTES];
} sysfragments_t;

typedef struct iSYSFRAGMENTS_S {
    char fragtype[2];
    int tabid;
    char indexname[IDBYTES];
    char strategy[2];
    int evalpos;
    loc_t exprtext;
    string dbspace[IDBYTES];
} isysfragments_t;

typedef struct s_sysdbspaces {
    int dbsnum;
    char name[IDBYTES];
    char owner[UNAMELEN];
    int pagesize, fchunk, nchunks;
    int is_mirrored, is_blobspace, is_sbspace, is_temp;
    int flags;
} sysdbspaces_t;

typedef struct {
    string servername[IDBYTES], dbname[IDBYTES], owner[UNAMELEN], tabname[IDBYTES], ltabname[IDBYTES];
    int btabid;
} syssyntable_t;

typedef struct {
    char viewtext[257];
} sysviews_t;

typedef struct SYSPROCEDURES_S {
    string procname[IDBYTES];
    string owner[UNAMELEN];
    int procid;  /* SERIAL */
    string mode[2];
    int retsize;
    int symsize;
    int datasize;
    int codesize;
    int numargs;
    char isproc[2];
} sysprocedures_t;
 
typedef struct SYSPROCBODY_S {
    int procid;
    string datakey[2];
    int seqno;
    fixchar data[256];
    char fill;
} sysprocbody_t;
 
typedef struct SYSPROCAUTH_S {
    string grantor[UNAMELEN];
    string grantee[UNAMELEN];
    int procid;
    string procauth[2];
} sysprocauth_t;

typedef struct SYSTRIGBODY_S {
    int trigid;
    string datakey[2];
    int seqno;
    string data[257];
} systrigbody_t;


/* DEFINITIONS ONLY FOR DEPENDENT MODULES. */

extern systables_t systables, prev;
extern syscolumns_t syscolumns;
extern sysindexes_t sysindexes;
extern sysconstraints_t sysconstraints;
extern syscoldepend_t syscoldepend;
extern syschecks_t syschecks;
extern sysdefaults_t sysdefaults;
extern sysfragments_t fragments;
extern sysfragments_t ifragments;
extern short isconstraint;
extern sysdbspaces_t *sysdbspaces;
extern syssyntable_t syssyntable;
extern sysviews_t sysviews;
#endif

extern char databasename[301], host[201], statement[2048], statement2[2048],
    statement3[1024];
extern char database[IDBYTES], tablename[IDBYTES], dbspace[IDBYTES], alien[10];
extern char *bigtxt;
extern long bigtxtsz;
extern t_pagecounts *pagecounts;
extern t_idxcounts *idxcounts;
extern long ntables, alltables, gotdbs,  TBSFACT;
extern int is_OL_version, is_external_table, has_external_tables;
extern int major_vers, minor_vers;
extern int conn5, stripowner, ids_server, verbose, silent, doDistrib;
extern int ncols, ncols2, NoKeyIndices;
extern int Dynamic;
extern int Universal;
extern t_blobbuff iblobbuff, tblobbuff, *blobbuff;
extern char *a_space;
extern char dtparts[7][10];
extern char *columns[32767],  *columns2[32767];
extern int replace_owners, no_ags_procs, db_has_log, sequence_only;
extern char owner_override[33], original_owner[33];
#endif

