# $Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/myschema.mk,v 2.55 2020/11/20 12:20:37 art Exp $
.SUFFIXES:
.SUFFIXES: .ec,v .c,v .h,v .eh,v .ec .c .eh .h .o 
SHELL=/usr/bin/ksh
VPATH=.:./RCS

LDFLAGS=			# none needed on most systems
# LDFLAGS=-static		# static libraries are not compatible under GCC v5 or later
# LDFLAGS=-xarch=v9		# For Solaris 64bit versions 
# LDFLAGS=-q64 			# For AIX 64 bit mode using XLC
# LDFLAGS=-q64 -static		# For AIX 64 bit mode & static link using XLC
# LDFLAGS=-maix64		# For AIX 64 bit mode using GCC

CFLAGS=$(CDFLAGS)
# CFLAGS=$(CDFLAGS) -Ae		# For HP-32bit executables
# CFLAGS=$(CDFLAGS) -Ae +DD64	# For HP-64bit executables

ECFLAGS=$(CFLAGS)               # Works for ESQL/C 3.50 and later
# ECFLAGS=$(CFLAGS) -EDNOINT8=1   # For ESQL/C 7.xx, CSDK earlier than 3.10, and C4GL 7.xx.
# ECFLAGS=$(CFLAGS) -EDbigint=int # For CSDK earlier than 3.50

# CDFLAGS=-g
CDFLAGS=-O3             # Mostly only for GCC
# CDFLAGS=-O2 -qmaxmem=32768    # For AIX using XLC to improve optimization.
# CDFLAGS=-O2 -qmaxmem=32768 -q64    # For AIX 64bit using XLC
# CDFLAGS=-O3 -maix64	# For AIX 64bit using GCC
# CDFLAGS=-O            # For HP and other systems that do not have -O3
# CDFLAGS=-Onolimit	# For myschema.ec only - the HP compiler complains that 
#			# it does not have enough memory to optimize & suggests
#			# this option.
# CDFLAGS=-xarch=v9     # For Solaris 64bit versions 

CC=gcc
# CC=gcc-4.8		# ESQL/C libraries are not compatible with -static under GCC v5 or later
# CC=cc                 # For HP and others without GNU C
# CC=xlc		# For AIX's native compiler

ESQL=esql
# ESQL=c4gl # If you have 4gl r7.xx but no ESQL/C

INFORMIXC=$(CC)

AR2_ARGS=-c -g		# For ar2 to produce a clean System V style archive
#AR2_ARGS=-c -d		# For ar2 to produce a clean BSD style archive
#AR2_ARGS=-c -i		# For ar2 to produce a clean ar2 style archive
AR_ARGS=-rcv		# For ar to produce a clean local style archive

# To use a local version of ar either comment out the next 2 lines or change.
# AR=ar2
# AR_ARGS=$(AR2_ARGS)
AR=ar

##############################################################################
# Objects list
##############################################################################

OBJS=	myschema.o blobfuncs.o cleancat.o format_triggers.o getsysdata.o \
	obtainidxsize.o print_auths.o print_checks.o print_constraints.o \
	print_defaults.o print_indexes.o print_procedures.o print_triggers.o \
	print_violations.o print_synonyms.o print_views.o usage.o \
	print_seconstraints.o print_seindexes.o print_distributions.o \
	print_userauths.o tokens.o return_extended.o print_casts.o \
	print_types.o print_coltype.o print_aggregates.o print_alters.o \
	print_puts.o print_sequences.o print_opclasses.o print_ams.o \
	getopt.o print_external.o print_directives.o print_infrastructure.o \
	return_coltypes.o print_index_rename.o

SRC = blobfuncs.ec print_checks.ec print_seindexes.ec format_triggers.ec \
	print_constraints.ec print_synonyms.ec getsysdata.ec \
	print_defaults.ec print_triggers.ec myschema.ec print_indexes.ec \
	print_userauths.ec obtainidxsize.ec print_procedures.ec print_views.ec \
	print_auths.ec print_seconstraints.ec  print_violations.ec cleancat.c  \
	print_distributions.ec tokens.c tokens.h return_extended.ec \
	print_casts.ec print_types.ec print_coltype.ec print_aggregates.ec \
	print_sequences.ec print_opclasses.ec print_ams.ec \
	print_infrastructure.ec return_coltypes.ec print_index_rename.ec \
	getopt.c getopt.h prototypes.h print_external.ec print_directives.ec \
	print_alters.ec print_puts.ec usage.c myschema.h myschema.eh \
	myschema.mk myschema.mk.norcs myschema.mk.norcs.nognumake myschema.README

##############################################################################
# Default rules
##############################################################################

.ec.o:
ifneq (,$(findstring c4gl, $(ESQL)))
	${INFORMIXDIR}/lib/tools/i4glc3 -EDNOINT8=1 $<
	INFORMIXC=$(INFORMIXC) $(ESQL) $(ECFLAGS) -c $*.c
else
	INFORMIXC=$(INFORMIXC) $(ESQL) $(ECFLAGS) -c $<
endif

.c.o:
	INFORMIXC=$(INFORMIXC) $(ESQL) $(ECFLAGS) -c $<

.h,v.h:
	co $<

.eh,v.eh:
	co $<

.c,v.c:
	co $<

.ec,v.ec:
	co $<


##############################################################################
# Target dependencies
##############################################################################

myschema: myschema.mk $(OBJS)
	INFORMIXC=$(INFORMIXC) $(ESQL) $(LDFLAGS) -o myschema $(OBJS)

myschema.static: myschema.mk $(OBJS)
	INFORMIXC=$(INFORMIXC) $(ESQL) $(LDFLAGS) -o myschema.static -static $(OBJS)

myschema.source.ar: $(SRC)
	$(AR) $(AR_ARGS) $@ $^

copy: $(SRC)
	cp -p $(SRC) $(TARGET)

clean:
	rm -f *.o
	for file in *.ec; do rm -f $${file%%ec}c; done
	rm -f myschema

force:
	-rm -f *.o
	for file in *.ec; do rm -f $${file%%ec}c; done
	$_ myschema

rcsinit:
	ci -NVers_6_27 -k -u -t-'Import of initial myschema source.' $(SRC)

rcsupdate: $(SRC)
	ci -NVers_6_27 -k -u -m'Import of features release 6.27' $(SRC)

##############################################################################
# Object dependencies
##############################################################################

myschema.o: myschema.ec myschema.eh prototypes.h getopt.h

getopt.o: getopt.h

blobfuncs.o: blobfuncs.ec myschema.eh prototypes.h

format_triggers.o: format_triggers.ec myschema.eh prototypes.h

getsysdata.o: getsysdata.ec myschema.eh prototypes.h

obtainidxsize.o: obtainidxsize.ec myschema.eh prototypes.h

print_auths.o: print_auths.ec myschema.eh prototypes.h

print_userauths.o: print_userauths.ec myschema.eh prototypes.h

print_checks.o: print_checks.ec myschema.eh prototypes.h

print_constraints.o: print_constraints.ec myschema.eh prototypes.h

print_seconstraints.o: print_seconstraints.ec myschema.eh prototypes.h

print_defaults.o: print_defaults.ec myschema.eh prototypes.h

print_distributions.o: print_distributions.ec myschema.eh prototypes.h

print_indexes.o: print_indexes.ec myschema.eh prototypes.h

print_seindexes.o: print_seindexes.ec myschema.eh prototypes.h

print_procedures.o: print_procedures.ec myschema.eh prototypes.h
	INFORMIXC=$(INFORMIXC) esql -c print_procedures.ec

print_synonyms.o: print_synonyms.ec myschema.eh prototypes.h

print_triggers.o: print_triggers.ec myschema.eh prototypes.h

print_views.o: print_views.ec myschema.eh prototypes.h tokens.h

print_casts.o: print_casts.ec myschema.eh prototypes.h 

print_types.o: print_types.ec myschema.eh prototypes.h

print_coltype.o: print_coltype.ec myschema.eh prototypes.h

return_extended.o: return_extended.ec myschema.eh prototypes.h

print_violations.o: print_violations.ec myschema.eh prototypes.h

print_aggregates.o: print_aggregates.ec myschema.eh prototypes.h

print_sequences.o: print_sequences.ec myschema.eh prototypes.h

print_opclasses.o: print_opclasses.ec myschema.eh prototypes.h

print_ams.o: print_ams.ec myschema.eh prototypes.h

print_alters.o: print_alters.ec myschema.eh prototypes.h

print_puts.o: print_puts.ec myschema.eh prototypes.h

print_directives.o: print_directives.ec myschema.eh prototypes.h

print_infrastructure.o: print_infrastructure.ec myschema.eh prototypes.h

return_coltypes.o: return_coltypes.ec myschema.eh prototypes.h

print_index_rename.o: print_index_rename.ec myschema.eh

usage.o: usage.c

cleancat.o: cleancat.c

tokens.o: tokens.c

##############################################################################
# Source dependencies
##############################################################################

blobfuncs.ec: blobfuncs.ec,v

format_triggers.ec: format_triggers.ec,v

getsysdata.ec: getsysdata.ec,v

myschema.ec: myschema.ec,v

getopt.c: getopt.c,v

obtainidxsize.ec: obtainidxsize.ec,v

print_auths.ec: print_auths.ec,v

print_userauths.ec: print_userauths.ec,v

print_checks.ec: print_checks.ec,v

print_constraints.ec: print_constraints.ec,v

print_seconstraints.ec: print_seconstraints.ec,v

print_defaults.ec: print_defaults.ec,v

print_distributions.ec: print_distributions.ec,v

print_indexes.ec: print_indexes.ec,v

print_seindexes.ec: print_seindexes.ec,v

print_procedures.ec: print_procedures.ec,v

print_synonyms.ec: print_synonyms.ec,v

print_triggers.ec: print_triggers.ec,v

print_views.ec: print_views.ec,v

print_casts.ec: print_casts.ec,v

print_types.ec: print_types.ec,v

print_coltype.ec: print_coltype.ec,v

return_extended.ec: return_extended.ec,v

print_violations.ec: print_violations.ec,v

print_aggregates.ec: print_aggregates.ec,v

print_sequences.ec: print_sequences.ec,v

print_opclasses.ec: print_opclasses.ec,v

print_ams.ec: print_ams.ec,v

print_alters.ec: print_alters.ec,v

print_puts.ec: print_puts.ec,v

print_infrastructure.ec: print_infrastructure.ec,v

return_coltypes.ec: return_coltypes.ec,v

usage.c: usage.c,v

cleancat.c: cleancat.c,v

tokens.c: tokens.c,v

tokens.h: tokens.h,v

myschema.h: myschema.h,v

myschema.eh: myschema.eh,v

print_directives.ec: print_directives.ec,v

