/* obtainidxsize.ec - Tally up all of the indexes, applying the formula to 
	each, returning one amount for each table passed.
*/
/*
   RCS Header:
   -----------
   $Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/obtainidxsize.ec,v 2.7 2012/01/31 19:00:57 art Exp $

   RCS Log:
   --------
   $Log: obtainidxsize.ec,v $
   Revision 2.7  2012/01/31 19:00:57  art
   Better fix for big nrows problem.

   Revision 2.6  2012/01/31 17:54:21  art
   Fixed problem with obtaining row counts for tables with more than 2^31 rows.

   Revision 2.5  2009/02/11 18:30:34  art
   Fixed problem with LVARCHARS on AIX-64bit.

   Revision 2.4  2008/12/30 17:32:38  art
   Removed BB specific ident string RCSWhat

   Revision 2.3  2004/09/14 17:24:48  kagel
   Fixing ANSI database access.

   Revision 2.2  1999/11/03 20:42:14  kagel
   Had to make identifier and username length changes to support IDS.2000
   again.  Some strange interaction between my makefile and using emacs to
   check-out of RCS lost those changes.

   Revision 2.1  1999/09/29 16:43:17  kagel
   Establishing current  source version id for initial working version of
   newly modularized source.  This is Feature Version 6.1 Source revision
   2.1.

   Revision 1.2  1999/09/22 20:49:47  kagel
   Moved statement preparation and cursor declaration in here.

   Revision 1.1  1999/09/01 19:11:37  kagel
   Initial revision


*/

static const char RCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/obtainidxsize.ec,v 2.7 2012/01/31 19:00:57 art Exp $";
static const char RCSCompile[] = "@(#)$cc: " __FILE__ " " __DATE__ " " __TIME__ " $";

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>

EXEC SQL INCLUDE sqlhdr;
EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;
EXEC SQL INCLUDE datetime;

EXEC SQL INCLUDE myschema.eh;

EXEC SQL INCLUDE prototypes.h;

static int prepped = 0;

int obtainidxsize( char *db, char *tb, int tabno, int *len, double *rows )
{

/* Tally up all of the indexes, applying the formula to each,
   returning one amount for each table passed  */

    int i, j, OK = 0, res;
    EXEC SQL BEGIN DECLARE SECTION;
    int tabid = tabno;
    string idxname[129];
    EXEC SQL END DECLARE SECTION;

    if (conn5 || !ids_server) {
	/* Version 5.xx and SE servers have no SMI maybe code other
           calculation later. */
	*len = 0;
	return 1;
    } else {
	/* Cursor to get key lengths of detached indexes. */
	if (!prepped) {
	    sprintf( statement3, 
		     "SELECT UNIQUE idxname "
		     "FROM \"informix\".sysindexes si, \"informix\".sysfragments sf "
		     "WHERE si.tabid = ? AND si.idxname = sf.indexname "
		     "  AND si.tabid = sf.tabid AND fragtype = 'I' "
		     "  AND sf.partn = ("
		     "SELECT min(partn) "
		     "FROM \"informix\".sysfragments f "
		     "WHERE f.tabid = si.tabid AND si.idxname = f.indexname) "
		     "ORDER BY 1" );
	    EXEC SQL PREPARE s_detach FROM :statement3;
	    if (sqlca.sqlcode) {
		fprintf( stderr,
			 "Cannot PREPARE statement s_detach, error: %d.\n",
			 sqlca.sqlcode );
		fprintf( stderr, "Error at character %d.\nStatement is:\n",
			 sqlca.sqlerrd[4] );
		rexit(18);
	    }
	    EXEC SQL DECLARE c_detach CURSOR FOR s_detach;
	    if (sqlca.sqlcode ) {
		fprintf( stderr,
			 "Cannot declare cursor c_detach; Error = %d\n",
			 sqlca.sqlcode );
		rexit(17);
	    }
	    prepped = 1;
	}
	/* This loop works for attached indexes only. */
	for (i=0; i<alltables; i++) {
	    if (strcmp( tb, idxcounts[i].tabname ) == 0) {
		*len = idxcounts[i].keylen;
		*rows = idxcounts[i].nrows;
		OK = 1;
		break;
	    }
	}
	if (OK) {
	    /* Detached index loop. */
	    EXEC SQL OPEN c_detach USING :tabid;
	    if (sqlca.sqlcode < 0) {
		fprintf( stderr,
			 "Error openning CURSOR c_detach.  SQLCode=%d, ISAM=%d.\n",
			 sqlca.sqlcode,
			 sqlca.sqlerrd[1] );
		rexit(22);
	    }
	    EXEC SQL FETCH c_detach INTO :idxname;
	    if (sqlca.sqlcode < 0) {
		fprintf( stderr,
			 "Error openning CURSOR c_detach.  SQLCode=%d, ISAM=%d.\n",
			 sqlca.sqlcode,
			 sqlca.sqlerrd[1] );
		rexit(22);
	    } else if (sqlca.sqlcode == SQLNOTFOUND)
		return OK;
	    j = 0;
	    for (;j < alltables;) {
		if ((res = strcmp( idxcounts[j].tabname, idxname)) > 0) {
		    EXEC SQL FETCH c_detach INTO :idxname;
		    if (sqlca.sqlcode < 0) {
			fprintf( stderr,
				 "Error openning CURSOR c_detach.  SQLCode=%d, ISAM=%d.\n",
				 sqlca.sqlcode,
				 sqlca.sqlerrd[1] );
			rexit(22);
		    } else if (sqlca.sqlcode == SQLNOTFOUND)
			break;
		} else if (res < 0) {
		    j++;
		} else if (res == 0) {
		    *len += idxcounts[j].detlen;
		    j++;
		}
	    }
	    EXEC SQL CLOSE c_detach;
	}
	return OK;
    }
}


