/* ***************************************************************************
 * print_ams.ec - Implements the CREATE ACCESS METHOD DDL statements for user
 *		      defined access methods.
 * ***************************************************************************
 *
 */
/*
  RCS Header:
  -----------
  $Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_ams.ec,v 1.7 2019/03/15 14:48:19 art Exp $


  RCS Log:
  --------
  $Log: print_ams.ec,v $
  Revision 1.7  2019/03/15 14:48:19  art
  Added support for CREATE OR REPLACE routines with --replace-procs option (v14.10 feature).

  Revision 1.6  2010/11/08 20:40:31  art
  Added new --create_conditionally feature to optionally add " IF NOT EXISTS " clauses to c
  create object statements.  Feature supported by 11.70.

  Revision 1.5  2009/07/19 14:47:03  art
  Missed removing am_truncate for older engines in one spot.

  Revision 1.4  2009/07/14 19:06:24  art
  Finished fix.

  Revision 1.3  2009/07/14 18:53:35  art
  Fixed version specific problems with the pages size calcs.
  Fixed version specific problem with handling user defined system access
  methods.

  Revision 1.2  2009/06/02 02:21:40  art
  Added support for Access Methods and Op Classes.

  Revision 1.1  2009/06/02 01:48:05  art
  Initial revision


*/

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <errno.h>

EXEC SQL INCLUDE sqlhdr;
EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;

EXEC SQL INCLUDE myschema.eh;

EXEC SQL INCLUDE prototypes.h;
 
#ifndef lint
static const char RCSIdent[] = "$Id: print_ams.ec,v 1.7 2019/03/15 14:48:19 art Exp $";
static const char RCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_ams.ec,v 1.7 2019/03/15 14:48:19 art Exp $";
static const char RCSCompile[] = "@(#)$cc: " __FILE__ " " __DATE__ " " __TIME__ " $";
#endif

EXEC SQL BEGIN DECLARE SECTION;

typedef struct SYSAMS_S {
	string am_name[129];
	string am_owner[33];
	int am_id;
	string am_type[2];
	string am_sptype[4];
	int am_defopclass;
	string am_defopclass_name[129];
	int am_keyscan;
	int am_unique;
	int am_cluster;
	int am_rowids;
	int am_readwrite;
	int am_parallel;
	float am_costfactor;
	int am_create;
	string am_create_name[129];
	int am_drop;
	string am_drop_name[129];
	int am_open;
	string am_open_name[129];
	int am_close;
	string am_close_name[129];
	int am_insert;
	string am_insert_name[129];
	int am_delete;
	string am_delete_name[129];
	int am_update;
	string am_update_name[129];
	int am_stats;
	string am_stats_name[129];
	int am_scancost;
	string am_scancost_name[129];
	int am_check;
	string am_check_name[129];
	int am_beginscan;
	string am_beginscan_name[129];
	int am_endscan;
	string am_endscan_name[129];
	int am_rescan;
	string am_rescan_name[129];
	int am_getnext;
	string am_getnext_name[129];
	int am_getbyid;
	string am_getbyid_name[129];
	int am_build;
	string am_build_name[129];
	int am_init;
	string am_init_name[129];
	int am_truncate;
	string am_truncate_name[129];
} sysams_t;

typedef struct SYSAMS_INT_S {
	short am_name;
	short am_owner;
	short am_id;
	short am_type;
	short am_sptype;
	short am_defopclass;
	short am_defopclass_name;
	short am_keyscan;
	short am_unique;
	short am_cluster;
	short am_rowids;
	short am_readwrite;
	short am_parallel;
	short am_costfactor;
	short am_create;
	short am_create_name;
	short am_drop;
	short am_drop_name;
	short am_open;
	short am_open_name;
	short am_close;
	short am_close_name;
	short am_insert;
	short am_insert_name;
	short am_delete;
	short am_delete_name;
	short am_update;
	short am_update_name;
	short am_stats;
	short am_stats_name;
	short am_scancost;
	short am_scancost_name;
	short am_check;
	short am_check_name;
	short am_beginscan;
	short am_beginscan_name;
	short am_endscan;
	short am_endscan_name;
	short am_rescan;
	short am_rescan_name;
	short am_getnext;
	short am_getnext_name;
	short am_getbyid;
	short am_getbyid_name;
	short am_build;
	short am_build_name;
	short am_init;
	short am_init_name;
	short am_truncate;
	short am_truncate_name;
} sysams_ind_t;

EXEC SQL END DECLARE SECTION;

void print_ams( FILE *outfile )
{
    EXEC SQL BEGIN DECLARE SECTION;
    sysams_t sysams;
    sysams_ind_t sysams_ind;
    string statement[65536];
    EXEC SQL END DECLARE SECTION;
    int len;
    char copy[2049];
    char *s, *t;
	
    snprintf( statement,
	      sizeof statement,
	      "SELECT "
	      "am_name, "
	      "am_owner, "
	      "am_id, "
	      "am_type, "
	      "am_sptype, "
	      "am_defopclass, "
	      "opclassname, "
	      "am_keyscan, "
	      "am_unique, "
	      "am_cluster, "
	      "am_rowids, "
	      "am_readwrite, "
	      "am_parallel, "
	      "am_costfactor, "
	      "am_create, "
	      "p7.procname AS am_create_name, "
	      "am_drop, "
	      "p8.procname AS am_drop_name, "
	      "am_open, "
	      "p9.procname AS am_open_name, "
	      "am_close, "
	      "p10.procname AS am_close_name, "
	      "am_insert, "
	      "p11.procname AS am_insert_name, "
	      "am_delete, "
	      "p12.procname AS am_delete_name, "
	      "am_update, "
	      "p13.procname AS am_update_name, "
	      "am_stats, "
	      "p14.procname AS am_stats_name, "
	      "am_scancost, "
	      "p15.procname AS am_scancost_name, "
	      "am_check, "
	      "p16.procname AS am_check_name, "
	      "am_beginscan, "
	      "p17.procname AS am_beginscan_name, "
	      "am_endscan, "
	      "p18.procname AS am_endscan_name, "
	      "am_rescan, "
	      "p19.procname AS am_rescan_name, "
	      "am_getnext, "
	      "p20.procname AS am_getnext_name, "
	      "am_getbyid, "
	      "p21.procname AS am_getbyid_name, "
	      "am_build, "
	      "p22.procname AS am_build_name, "
	      "am_init, "
	      "p23.procname AS am_init_name, %s"
	      "FROM (((((((((((((((((((sysams "
	      "LEFT OUTER JOIN sysopclasses AS p1 "
	      "ON sysams.am_defopclass = p1.opclassid) "
	      "LEFT OUTER JOIN sysprocedures AS p7 "
	      "ON sysams.am_create = p7.procid) "
	      "LEFT OUTER JOIN sysprocedures AS p8 "
	      "ON sysams.am_drop = p8.procid) "
	      "LEFT OUTER JOIN sysprocedures AS p9 "
	      "ON sysams.am_open = p9.procid) "
	      "LEFT OUTER JOIN sysprocedures AS p10 "
	      "ON sysams.am_close = p10.procid) "
	      "LEFT OUTER JOIN sysprocedures AS p11 "
	      "ON sysams.am_insert = p11.procid) "
	      "LEFT OUTER JOIN sysprocedures AS p12 "
	      "ON sysams.am_delete = p12.procid) "
	      "LEFT OUTER JOIN sysprocedures AS p13 "
	      "ON sysams.am_update = p13.procid) "
	      "LEFT OUTER JOIN sysprocedures AS p14 "
	      "ON sysams.am_stats = p14.procid) "
	      "LEFT OUTER JOIN sysprocedures AS p15 "
	      "ON sysams.am_scancost = p15.procid) "
	      "LEFT OUTER JOIN sysprocedures AS p16 "
	      "ON sysams.am_check = p16.procid) "
	      "LEFT OUTER JOIN sysprocedures AS p17 "
	      "ON sysams.am_beginscan = p17.procid) "
	      "LEFT OUTER JOIN sysprocedures AS p18 "
	      "ON sysams.am_endscan = p18.procid) "
	      "LEFT OUTER JOIN sysprocedures AS p19 "
	      "ON sysams.am_rescan = p19.procid) "
	      "LEFT OUTER JOIN sysprocedures AS p20 "
	      "ON sysams.am_getnext = p20.procid) "
	      "LEFT OUTER JOIN sysprocedures AS p21 "
	      "ON sysams.am_getbyid = p21.procid) "
	      "LEFT OUTER JOIN sysprocedures AS p22 "
	      "ON sysams.am_build = p22.procid) "
	      "LEFT OUTER JOIN sysprocedures AS p23 "
	      "ON sysams.am_init = p23.procid) %s ) "
	      "WHERE am_id > 6 and am_sptype != 'D' {* Filter out system "
	      "access methods *} "
	      "ORDER BY 3; ",
	      major_vers >= 10 ?
	      "am_truncate, p24.procname AS am_truncate_name " :
	      "0, '' ",
	      major_vers >= 10 ?
	      "LEFT OUTER JOIN sysprocedures AS p24 "
	      "ON sysams.am_truncate = p24.procid " : "" );

    EXEC SQL PREPARE am_stmt FROM :statement;

    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error preparing statement for ams_c!  SQL: %d  ISAM: %d\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	fprintf( stderr, "  Msg: %s.\n", sqlca.sqlerrm );
	exit( 1 );
    }

    EXEC SQL DECLARE ams_c CURSOR FOR am_stmt;

    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error declaring cursor ams_c!  SQL: %d  ISAM: %d\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	fprintf( stderr, "  Msg: %s.\n", sqlca.sqlerrm );
	exit( 1 );
    }

    EXEC SQL OPEN ams_c;
    if (sqlca.sqlcode == SQLNOTFOUND)
	return;
    else if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error opening cursor ams_c!  SQL: %d  ISAM: %d\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	fprintf( stderr, "  Msg: %s.\n", sqlca.sqlerrm );
	exit( 1 );
    }

    for (;;) {
	int first = 1;

	EXEC SQL FETCH ams_c INTO :sysams indicator :sysams_ind;
	if (sqlca.sqlcode == SQLNOTFOUND)
	    break;
	else if (sqlca.sqlcode != 0) {
	    fprintf( stderr,
		     "Error fetching cursor ams_c!  SQL: %d  ISAM: %d\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit( 1 );
	}

	if (stripowner) {
	    fprintf( outfile,
		     "CREATE %s ACCESS_METHOD%s %s (",
		     sysams.am_type[0] == 'P' ? "PRIMARY" : "SECONDARY",
		     create_conditionally ? " IF NOT EXISTS" : "",
		     sysams.am_name );
	} else {
	    fprintf( outfile,
		     "CREATE %s ACCESS_METHOD%s \"%s\".%s (",
		     sysams.am_type[0] == 'P' ? "PRIMARY" : "SECONDARY",
		     create_conditionally ? " IF NOT EXISTS" : "",
		     sysams.am_owner,
		     sysams.am_name );
	}
	
	/* Value Settings */
	if (sysams_ind.am_sptype != -1) {
	    /* Has sptype defined. */
	    fprintf( outfile,
		     "%s\n\tam_sptype = '%s'",
		     first ? "" : ", ",
		     sysams.am_sptype );
	    first = 0;
	}

	if (sysams_ind.am_defopclass_name != -1 && sysams.am_defopclass != 0) {
	    /* Has default op class. */
	    fprintf( outfile,
		     "%s\n\tam_defopclass = %s",
		     first ? "" : ", ",
		     sysams.am_defopclass_name );
	    first = 0;
	}

	if (sysams_ind.am_costfactor != -1 && sysams.am_costfactor != 0.0) {
	    /* Has cost factor set. */
	    fprintf( outfile,
		     "%s\n\tam_costfactor = %f",
		     first ? "" : ", ",
		     sysams.am_costfactor );
	    first = 0;
	}


	/* Flag Settings */
	if (sysams_ind.am_keyscan != -1 &&sysams.am_keyscan == 1) {
	    /* Has keyscan feature. */
	    fprintf( outfile,
		     "%s\n\tam_keyscan",
		     first ? "" : ", " );
	    first = 0;
	}

	if (sysams_ind.am_unique != -1 &&sysams.am_unique == 1) {
	    /* Has unique feature. */
	    fprintf( outfile,
		     "%s\n\tam_unique",
		     first ? "" : ", " );
	    first = 0;
	}

	if (sysams_ind.am_cluster != -1 &&sysams.am_cluster == 1) {
	    /* Has cluster feature. */
	    fprintf( outfile,
		     "%s\n\tam_cluster",
		     first ? "" : ", " );
	    first = 0;
	}

	if (sysams_ind.am_rowids != -1 &&sysams.am_rowids == 1) {
	    /* Has rowids feature. */
	    fprintf( outfile,
		     "%s\n\tam_rowids",
		     first ? "" : ", " );
	    first = 0;
	}

	if (sysams_ind.am_readwrite != -1 &&sysams.am_readwrite == 1) {
	    /* Has readwrite feature. */
	    fprintf( outfile,
		     "%s\n\tam_readwrite",
		     first ? "" : ", " );
	    first = 0;
	}

	if (sysams_ind.am_parallel != -1 &&sysams.am_parallel == 1) {
	    /* Has parallel feature. */
	    fprintf( outfile,
		     "%s\n\tam_parallel",
		     first ? "" : ", " );
	    first = 0;
	}

	/* Function Settings */
	if (sysams_ind.am_create_name != -1 && sysams.am_create != 0) {
	    /* Has am_create function. */
	    fprintf( outfile,
		     "%s\n\tam_create = %s",
		     first ? "" : ", ",
		     sysams.am_create_name );
	    first = 0;
	}

	if (sysams_ind.am_drop_name != -1 && sysams.am_drop != 0) {
	    /* Has am_drop function. */
	    fprintf( outfile,
		     "%s\n\tam_drop = %s",
		     first ? "" : ", ",
		     sysams.am_drop_name );
	    first = 0;
	}

	if (sysams_ind.am_open_name != -1 && sysams.am_open != 0) {
	    /* Has am_open function. */
	    fprintf( outfile,
		     "%s\n\tam_open = %s",
		     first ? "" : ", ",
		     sysams.am_open_name );
	    first = 0;
	}

	if (sysams_ind.am_close_name != -1 && sysams.am_close != 0) {
	    /* Has am_close function. */
	    fprintf( outfile,
		     "%s\n\tam_close = %s",
		     first ? "" : ", ",
		     sysams.am_close_name );
	    first = 0;
	}

	if (sysams_ind.am_insert_name != -1 && sysams.am_insert != 0) {
	    /* Has am_insert function. */
	    fprintf( outfile,
		     "%s\n\tam_insert = %s",
		     first ? "" : ", ",
		     sysams.am_insert_name );
	    first = 0;
	}

	if (sysams_ind.am_delete_name != -1 && sysams.am_delete != 0) {
	    /* Has am_delete function. */
	    fprintf( outfile,
		     "%s\n\tam_delete = %s",
		     first ? "" : ", ",
		     sysams.am_delete_name );
	    first = 0;
	}

	if (sysams_ind.am_update_name != -1 && sysams.am_update != 0) {
	    /* Has am_update function. */
	    fprintf( outfile,
		     "%s\n\tam_update = %s",
		     first ? "" : ", ",
		     sysams.am_update_name );
	    first = 0;
	}

	if (sysams_ind.am_stats_name != -1 && sysams.am_stats != 0) {
	    /* Has am_stats function. */
	    fprintf( outfile,
		     "%s\n\tam_stats = %s",
		     first ? "" : ", ",
		     sysams.am_stats_name );
	    first = 0;
	}

	if (sysams_ind.am_scancost_name != -1 && sysams.am_scancost != 0) {
	    /* Has am_scancost function. */
	    fprintf( outfile,
		     "%s\n\tam_scancost = %s",
		     first ? "" : ", ",
		     sysams.am_scancost_name );
	    first = 0;
	}

	if (sysams_ind.am_check_name != -1 && sysams.am_check != 0) {
	    /* Has am_check function. */
	    fprintf( outfile,
		     "%s\n\tam_check = %s",
		     first ? "" : ", ",
		     sysams.am_check_name );
	    first = 0;
	}

	if (sysams_ind.am_beginscan_name != -1 && sysams.am_beginscan != 0) {
	    /* Has am_beginscan function. */
	    fprintf( outfile,
		     "%s\n\tam_beginscan = %s",
		     first ? "" : ", ",
		     sysams.am_beginscan_name );
	    first = 0;
	}

	if (sysams_ind.am_endscan_name != -1 && sysams.am_endscan != 0) {
	    /* Has am_endscan function. */
	    fprintf( outfile,
		     "%s\n\tam_endscan = %s",
		     first ? "" : ", ",
		     sysams.am_endscan_name );
	    first = 0;
	}

	if (sysams_ind.am_rescan_name != -1 && sysams.am_rescan != 0) {
	    /* Has am_rescan function. */
	    fprintf( outfile,
		     "%s\n\tam_rescan = %s",
		     first ? "" : ", ",
		     sysams.am_rescan_name );
	    first = 0;
	}

	if (sysams_ind.am_getnext_name != -1 && sysams.am_getnext != 0) {
	    /* Has am_getnext function. */
	    fprintf( outfile,
		     "%s\n\tam_getnext = %s",
		     first ? "" : ", ",
		     sysams.am_getnext_name );
	    first = 0;
	}

	if (sysams_ind.am_getbyid_name != -1 && sysams.am_getbyid != 0) {
	    /* Has am_getbyid function. */
	    fprintf( outfile,
		     "%s\n\tam_getbyid = %s",
		     first ? "" : ", ",
		     sysams.am_getbyid_name );
	    first = 0;
	}

	if (sysams_ind.am_build_name != -1 && sysams.am_build != 0) {
	    /* Has am_build function. */
	    fprintf( outfile,
		     "%s\n\tam_build = %s",
		     first ? "" : ", ",
		     sysams.am_build_name );
	    first = 0;
	}

	if (sysams_ind.am_init_name != -1 && sysams.am_init != 0) {
	    /* Has am_init function. */
	    fprintf( outfile,
		     "%s\n\tam_init = %s",
		     first ? "" : ", ",
		     sysams.am_init_name );
	    first = 0;
	}

	if (sysams_ind.am_truncate_name != -1 && sysams.am_truncate != 0) {
	    /* Has am_truncate function. */
	    fprintf( outfile,
		     "%s\n\tam_truncate = %s",
		     first ? "" : ", ",
		     sysams.am_truncate_name );
	    first = 0;
	}

	fprintf( outfile, " );\n" );
    }
    EXEC SQL CLOSE ams_c;
    EXEC SQL FREE ams_c;
	
}
