/* print_auths.ec - print out table level authorizations.
*/
/*
   RCS Header:
   -----------
   $Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_auths.ec,v 2.23 2018/06/22 16:30:05 art Exp $

   RCS Log:
   --------
   $Log: print_auths.ec,v $
   Revision 2.23  2018/06/22 16:30:05  art
   New options: --use-database-owner, --use-table-owner, --constraint-file
   Fixed some problems with the partial schema file output.
   Expanded "owner"s to all objects.

   Revision 2.22  2014/03/18 15:24:05  art
   Added support for GRANT SETSESSIONAUTH.  FYI: Not supported by dbschema!

   Revision 2.21  2010/01/06 13:55:18  art
   Added new option --set-owner to override all object owners with a specified owner name.

   Revision 2.20  2009/11/23 18:47:05  art
   Added support for EXTERNAL tables (11.50xc6+).
   Improved 64bit compiler support.  Removed warnings.

   Revision 2.19  2009/10/18 17:23:16  art
   Fixed collection type handling.  Thanks to Bruce Simms for pointing out this one.
   Fixed summary comment to include table owner if not suppressed ala dbschema.
   Removed unneeded function prototypes.

   Revision 2.18  2009/02/11 18:30:34  art
   Fixed problem with LVARCHARS on AIX-64bit.

   Revision 2.17  2009/01/01 17:58:25  art
   Removed RCSWhat ident string which was BB specific.

   Revision 2.16  2004/09/14 19:38:45  kagel
   Extended referencing view code to recurse to catch indirect references.
   through other views.
   Fixed missing view owners if owner names not being stripped.

   Revision 2.15  2004/01/13 22:38:49  kagel
   Case, make pretty.

# Revision 2.14  2003/09/30  21:28:36  21:28:36  kagel 
# Added undocumented -o option to support mydbdiff.  Changed some default
# sort orders so the output will be more consistent when tables are altered.
# Increased the Features version to 6.25 to recognize the changes.
# 
   Revision 2.13  2003/05/15 19:43:08  kagel
   Fixed bug in column auths causing each column auth to be repeated for every
   table authorized user.

   Revision 2.12  2002/12/13 15:26:35  kagel
   Fixed problem with handling blank grantor when grantor == grantee in columns
   and function/procedure auths.  Previously fixed in table auths.

   Revision 2.11  2002/11/21 21:16:14  kagel
   Added DELIMIDENT support.

   Revision 2.10  2002/11/20 21:47:54  kagel
   Added DELIMIDENT support, RAW table support.

   Revision 2.9  2002/10/29 20:24:15  kagel
   Completed same compression changes for column and fragment priveleges.

   Revision 2.8  2002/10/29 19:35:14  kagel
   Print_auths now combines table authorizations into one line for those
   granted WITH GRANT OPTION and one for those without.  This will noticably
   reduce the volume of output.  Another improvement over dbschema.  Thanks to
   Jonathan Leffler for the suggestion.

   Revision 2.7  2002/08/12 20:51:24  kagel
   Fixed empty 'AS' clause if table owner granting to itself.

   Revision 2.6  2000/09/13 13:18:54  kagel
   Added quotes to user and role names to support upper and mixed case
   names.

   Revision 2.5  2000/08/25 19:19:09  kagel
   A false start.  Works correctly now.

   Revision 2.4  2000/08/25 19:01:02  kagel
   Fixed privilege code.  Was assuming if 'r' privilege (references) present
   then  ALL are present and output 'ALL' instead.  Also if ALL no column
   privileges were being generated.

   Revision 2.3  1999/11/03 19:08:57  kagel
   Fixed updated identifier lengths.

   Revision 2.2  1999/11/01 21:39:02  kagel
   Fixed tabauth variables to hold 9 bytes instead of 8 bytes of flags.

   Revision 2.1  1999/09/29 16:42:24  kagel
   Establishing current  source version id for initial working version of
   newly modularized source.  This is Feature Version 6.1 Source revision
   2.1.

   Revision 1.2  1999/09/22 21:06:50  kagel
   Moved statement preparation and cursor declaration in here.

   Revision 1.1  1999/09/01 19:42:43  kagel
   Initial revision


*/

static const char RCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_auths.ec,v 2.23 2018/06/22 16:30:05 art Exp $";
static const char RCSCompile[] = "@(#)$cc: " __FILE__ " " __DATE__ " " __TIME__ " $";

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>

EXEC SQL INCLUDE sqlhdr;
EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;
EXEC SQL INCLUDE datetime;

EXEC SQL INCLUDE myschema.eh;

EXEC SQL INCLUDE prototypes.h;

static int prepped = 0;

void print_auths( int tabid,
		  char *tabname,
		  FILE *outfile,
		  char *owner,
		  int tabauths,
		  int colauths,
		  int grantas )
{
    int j, k;
    char grnt[20], ASclaus[20], cASclaus[20], fASclaus[20];
    char mainstr[1024];
    char tmp1[150], tmp2[150];
    int is_stuff;

    EXEC SQL BEGIN DECLARE SECTION;
    string grantor[UNAMELEN], grantee[UNAMELEN], tabauth[10];
    string cgrantor[UNAMELEN], cgrantee[UNAMELEN], colauth[4];
    string fgrantor[UNAMELEN], fgrantee[UNAMELEN], fragauth[7];
    string colname[IDBYTES], fragment[IDBYTES];
    int tbid;
    EXEC SQL END DECLARE SECTION;

    /* Select for table privilege information. */
    if (!prepped) {
	EXEC SQL PREPARE s_auth FROM 
	    "SELECT grantor, grantee, tabauth FROM \"informix\".systabauth WHERE tabid = ? ORDER BY 1, 2";
	if (sqlca.sqlcode) {
	    fprintf( stderr,
		     "Cannot PREPARE select statement for table privileges, "
		     "error: %d.\n",
		     sqlca.sqlcode );
	    fprintf( stderr, "Error at character %d.\nStatement is:\n",
		     sqlca.sqlerrd[4] );
	    fprintf( stderr, "%s", statement );
	    rexit(19);
	}
	EXEC SQL DECLARE auth CURSOR FOR s_auth;
	if (sqlca.sqlcode ) {
	    fprintf( stderr,
		     "Cannot declare auth cursor; Error = %d\n",
		     sqlca.sqlcode );
	    rexit(20);
	}

	/* Column privilege information. */
	EXEC SQL PREPARE s_cauth FROM
	    "SELECT grantor, grantee, colname, colauth FROM \"informix\".syscolauth sca, \"informix\".syscolumns sc WHERE sc.tabid = ? AND sca.grantee = ? AND sc.tabid = sca.tabid and sc.colno = sca.colno ORDER BY 1,2,3";
	if (sqlca.sqlcode) {
	    fprintf( stderr,
		     "Cannot PREPARE select statement for column privileges, "
		     "error: %d.\n",
		     sqlca.sqlcode );
	    fprintf( stderr, "Error at character %d.  Statement is:\n",
		     sqlca.sqlerrd[4] );
	    fprintf( stderr,
		     "SELECT grantor, grantee, colname, colauth "
		     "FROM \"informix\".syscolauth sca, \"informix\".syscolumns sc "
		     "WHERE sc.tabid = ? AND sca.grantee = ? "
		     "AND sc.tabid = sca.tabid and sc.colno = sca.colno "
		     "ORDER BY 2,1,3" );
	    rexit(19);
	}
	EXEC SQL DECLARE cauth CURSOR FOR s_cauth;
	if (sqlca.sqlcode ) {
	    fprintf( stderr,
		     "Cannot declare cauth cursor; Error = %d\n",
		     sqlca.sqlcode );
	    rexit(20);
	}

	if (!conn5 && ids_server) {
	    /* Fragment privilege information. */
	    EXEC SQL PREPARE s_fauth FROM
		"SELECT grantor, grantee, fragment, fragauth FROM \"informix\".sysfragauth sfa WHERE sfa.tabid = ? ORDER BY 1, 2, 3 ";
	    if (sqlca.sqlcode) {
		fprintf( stderr,
			 "Cannot PREPARE select statement for fragment privileges, "
			 "error: %d.\n",
			 sqlca.sqlcode );
		fprintf( stderr, "Error at character %d.  Statement is:\n",
			 sqlca.sqlerrd[4] );
		fprintf( stderr,
			 "SELECT grantor, grantee, fragment, fragauth "
			 "FROM \"informix\".sysfragauth sfa "
			 "WHERE sfa.tabid = ? ORDER BY 1, 2, 3 " );
		rexit(19);
	    }
	    EXEC SQL DECLARE fauth CURSOR FOR s_fauth;
	    if (sqlca.sqlcode ) {
		fprintf( stderr,
			 "Cannot declare fauth cursor; Error = %d\n",
			 sqlca.sqlcode );
		rexit(20);
	    }
	}
	prepped = 1;
    }

    tbid = tabid;
    EXEC SQL OPEN auth USING :tbid;
    if (sqlca.sqlcode) {
	fprintf( stderr,
		 "Error opening cursor auth, Error = %d\n",
		 sqlca.sqlcode );
	rexit(14);
    }

    EXEC SQL FETCH auth INTO :grantor, :grantee, :tabauth;
    while (!sqlca.sqlcode) {
	if (grantor[0] == ' ' || strlen(grantor) == 0)
	    strcpy( grantor, grantee );
	if (replace_owners) {
	    strcpy( grantor, owner_override );
	}
	if (grantas)
	    sprintf( ASclaus, " AS \"%s\"", grantor );
	else
	    ASclaus[0] = (char)0;
	if ((strncmp(tabauth, "su*idxar", 8) == 0) ||
	    (strncmp(tabauth, "SU*IDXAR", 8) == 0) ||
	    (strncmp(tabauth, "su-idxar", 8) == 0) ||
	    (strncmp(tabauth, "SU-IDXAR", 8) == 0))
	{
	    if (tabauth[0] == 'S')
		strcpy( grnt, " WITH GRANT OPTION" );
	    else
		grnt[0] = (char)0;
	    if (tabauths)
	        if (stripowner)
		    fprintf( outfile,
			     "GRANT ALL ON %s TO \"%s\"%s%s;\n",
			     delimident( tabname ),
			     grantee,
			     grnt,
			     ASclaus );
		else
		    fprintf( outfile,
			     "GRANT ALL ON \"%s\".%s TO \"%s\"%s%s;\n",
			     owner,
			     delimident( tabname ),
			     grantee,
			     grnt,
			     ASclaus );
	} else {
	    is_stuff = 0;
	    mainstr[0] = (char)0;
	    for (j = 0; j < 8; j++) {
		grnt[0] = (char)0;
		switch (tabauth[j]) {
		case 'r':
		    if (tabauths)
			strcat( mainstr, "REFERENCES, " );
		    break;
		case 's':
		    if (tabauths)
			strcat( mainstr, "SELECT, " );
		    break;
		case 'u':
		    if (tabauths)
			strcat( mainstr, "UPDATE, " );
		    break;
		case 'i':
		    if (tabauths)
			strcat( mainstr, "INSERT, " );
		    break;
		case 'd':
		    if (tabauths)
			strcat( mainstr, "DELETE, " );
		    break;
		case 'x':
		    if (tabauths)
			strcat( mainstr, "INDEX, " );
		    break;
		case 'a':
		    if (tabauths)
			strcat( mainstr, "ALTER, " );
		    break;
		}
	    }
	    is_stuff = strlen( mainstr );
	    if (is_stuff) {
		is_stuff -= 2;
		mainstr[is_stuff] = (char)0;
	    }
	    if (tabauths && is_stuff)
	        if (stripowner)
		    fprintf( outfile, "GRANT %s ON %s TO \"%s\"%s;\n",
			     mainstr,
			     delimident( tabname ),
			     grantee,
			     ASclaus );
		else
		    fprintf( outfile, "GRANT %s ON \"%s\".%s TO \"%s\"%s;\n",
			     mainstr,
			     owner,
			     delimident( tabname ),
			     grantee,
			     ASclaus );
	    is_stuff = 0;
	    mainstr[0] = (char)0;
	    for (j = 0; j < 8; j++) {
		grnt[0] = (char)0;
		switch (tabauth[j]) {
		case 'R':
		    if (tabauths)
			strcat( mainstr, "REFERENCES, " );
		    break;
		case 'S':
		    if (tabauths)
			strcat( mainstr, "SELECT, " );
		    break;
		case 'U':
		    if (tabauths)
			strcat( mainstr, "UPDATE, " );
		    break;
		case 'I':
		    if (tabauths)
			strcat( mainstr, "INSERT, " );
		    break;
		case 'D':
		    if (tabauths)
			strcat( mainstr, "DELETE, " );
		    break;
		case 'X':
		    if (tabauths)
			strcat( mainstr, "INDEX, " );
		    break;
		case 'A':
		    if (tabauths)
			strcat( mainstr, "ALTER, " );
		    break;
		}
	    }
	    is_stuff = strlen( mainstr );
	    if (is_stuff) {
		is_stuff -= 2;
		mainstr[is_stuff] = (char)0;
	    }
	    mainstr[is_stuff] = (char)0;
	    if (tabauths && is_stuff)
	        if (stripowner)
		    fprintf( outfile,
			     "GRANT %s ON %s TO \"%s\" WITH GRANT OPTION%s;\n",
			     mainstr,
			     delimident( tabname ),
			     grantee,
			     ASclaus );
		else
		    fprintf( outfile,
			     "GRANT %s ON \"%s\".%s TO \"%s\" WITH GRANT OPTION%s;\n",
			     mainstr,
			     owner,
			     delimident( tabname ),
			     grantee,
			     ASclaus );
	}
	if (tabauth[2] == '*' && colauths) {
	    EXEC SQL OPEN cauth USING :tbid, :grantee;
	    if (sqlca.sqlcode) {
		fprintf( stderr,
			 "Error opening cursor cauth, Error = %d\n",
			 sqlca.sqlcode );
		rexit(14);
	    }

	    EXEC SQL FETCH cauth
		INTO :cgrantor, :cgrantee, :colname, :colauth;
	    while (!sqlca.sqlcode) {
		if (cgrantor[0] == ' ' || strlen(cgrantor) == 0)
		    strcpy( cgrantor, cgrantee );
		if (replace_owners) {
		    strcpy( cgrantor, owner_override );
		}
		if (grantas)
		    sprintf( cASclaus, " AS \"%s\"", cgrantor );
		else
		    cASclaus[0] = (char)0;
		is_stuff = 0;
		mainstr[0] = (char)0;
		for (k=0;k<3;k++) {
		    grnt[0] = (char)0;
		    switch( colauth[k] ) {
		    case 's':
			strcat( mainstr, "SELECT, " );
			break;
		    case 'u':
			strcat( mainstr, "UPDATE, " );
			break;
		    case 'r':
			strcat( mainstr, "REFERENCES, " );
			break;
		    default:
			break;
		    }
		}
		is_stuff = strlen( mainstr );
		if (is_stuff) {
		    is_stuff -= 2;
		    mainstr[is_stuff] = (char)0;
		}
		if (is_stuff)
		    if (stripowner)
		      fprintf( outfile,
			       "GRANT %s (%s) ON %s TO \"%s\"%s;\n",
			       mainstr,
			       delimident2( colname, tmp1 ),
			       delimident2( tabname, tmp2 ),
			       cgrantee,
			       cASclaus );
		    else
		      fprintf( outfile,
			       "GRANT %s (%s) ON \"%s\".%s TO \"%s\"%s;\n",
			       mainstr,
			       delimident2( colname, tmp1 ),
			       owner,
			       delimident2( tabname, tmp2 ),
			       cgrantee,
			       cASclaus );
		is_stuff = 0;
		mainstr[0] = (char)0;
		for (k=0;k<3;k++) {
		    grnt[0] = (char)0;
		    switch( colauth[k] ) {
		    case 'S':
			strcat( mainstr, "SELECT, " );
			break;
		    case 'U':
			strcat( mainstr, "UPDATE, " );
			break;
		    case 'R':
			strcat( mainstr, "REFERENCES, " );
			break;
		    default:
			break;
		    }
		}
		is_stuff = strlen( mainstr );
		if (is_stuff) {
		    is_stuff -= 2;
		    mainstr[is_stuff] = (char)0;
		}
		if (is_stuff)
		    if (stripowner)
		      fprintf( outfile,
			       "GRANT %s (%s) ON %s TO \"%s\" "
			       "WITH GRANT OPTION%s;\n",
			       mainstr,
			       delimident2( colname, tmp1 ),
			       delimident2( tabname, tmp2 ),
			       cgrantee,
			       cASclaus );
		    else
		      fprintf( outfile,
			       "GRANT %s (%s) ON \"%s\".%s TO \"%s\" "
			       "WITH GRANT OPTION%s;\n",
			       mainstr,
			       delimident2( colname, tmp1 ),
			       owner,
			       delimident2( tabname, tmp2 ),
			       cgrantee,
			       cASclaus );
		EXEC SQL FETCH cauth
		    INTO :cgrantor, :cgrantee, :colname, :colauth;
	    }
	    if (sqlca.sqlcode && sqlca.sqlcode != SQLNOTFOUND) {
		fprintf( stderr,
			 "Error fetching cursor cauth; Error = %d\n",
			 sqlca.sqlcode );
		rexit(14);
	    }
	    EXEC SQL CLOSE cauth;
	}
	EXEC SQL FETCH auth INTO :grantor, :grantee, :tabauth;
    }
    if (sqlca.sqlcode && sqlca.sqlcode != SQLNOTFOUND) {
	fprintf( stderr,
		 "Error fetching cursor auth; Error = %d\n",
		 sqlca.sqlcode );
	rexit(14);
    }
    if (tabauths && !conn5 && ids_server) {

	EXEC SQL OPEN fauth USING :tbid;
	if (sqlca.sqlcode) {
	    fprintf( stderr,
		     "Error opening cursor fauth, Error = %d\n",
		     sqlca.sqlcode );
	    rexit(14);
	}
	
	EXEC SQL FETCH fauth
	    INTO :fgrantor, :fgrantee, :fragment, :fragauth;
	while (!sqlca.sqlcode) {
	    if (fgrantor[0] == ' ' || strlen(fgrantor) == 0)
		strcpy( fgrantor, fgrantee );
	    if (replace_owners) {
	        strcpy( fgrantor, owner_override );
	    }
	    if (grantas)
	        sprintf( fASclaus, " AS \"%s\"", fgrantor );
	    else
		fASclaus[0] = (char)0;
	    is_stuff = 0;
	    mainstr[0] = (char)0;
	    for (k=0;k<5;k++) {
		grnt[0] = (char)0;
		switch( fragauth[k] ) {
		case 'i':
		    strcat( mainstr, "INSERT, " );
		    break;
		case 'u':
		    strcat( mainstr, "UPDATE, " );
		    break;
		case 'd':
		    strcat( mainstr, "DELETE, " );
		    break;
		default:
		    break;
		}
	    }
	    is_stuff = strlen( mainstr );
	    if (is_stuff) {
		is_stuff -= 2;
		mainstr[is_stuff] = (char)0;
	    }
	    if (is_stuff)
	        if (stripowner)
		  fprintf( outfile,
			   "GRANT FRAGMENT %s ON %s (%s) TO "
			   "\"%s\"%s;\n",
			   mainstr,
			   delimident( tabname ),
			   fragment,
			   fgrantee,
			   fASclaus );
		else
		  fprintf( outfile,
			   "GRANT FRAGMENT %s ON \"%s\".%s (%s) TO "
			   "\"%s\"%s;\n",
			   mainstr,
			   owner,
			   delimident( tabname ),
			   fragment,
			   fgrantee,
			   fASclaus );
	    is_stuff = 0;
	    mainstr[0] = (char)0;
	    for (k=0;k<5;k++) {
		grnt[0] = (char)0;
		switch( fragauth[k] ) {
		case 'I':
		    strcat( mainstr, "INSERT, " );
		    break;
		case 'U':
		    strcat( mainstr, "UPDATE, " );
		    break;
		case 'D':
		    strcat( mainstr, "DELETE, " );
		    break;
		default:
		    break;
		}
	    }
	    is_stuff = strlen( mainstr );
	    if (is_stuff) {
		is_stuff -= 2;
		mainstr[is_stuff] = (char)0;
	    }
	    if (is_stuff)
	        if (stripowner)
		  fprintf( outfile,
			   "GRANT FRAGMENT %s ON %s (%s) TO "
			   "\"%s\" WITH GRANT OPTION%s;\n",
			   mainstr,
			   delimident( tabname ),
			   fragment,
			   fgrantee,
			   fASclaus );
		else
		  fprintf( outfile,
			   "GRANT FRAGMENT %s ON \"%s\".%s (%s) TO "
			   "\"%s\" WITH GRANT OPTION%s;\n",
			   mainstr,
			   owner,
			   delimident( tabname ),
			   fragment,
			   fgrantee,
			   fASclaus );
	    EXEC SQL FETCH fauth
		INTO :fgrantor, :fgrantee, :fragment, :fragauth;
	}
	EXEC SQL CLOSE fauth;
    }
    fprintf( outfile, "\n" );
    EXEC SQL CLOSE auth;
}


