/* print_coltype.ec - 
 */
/*
  RCS Header:
  -----------
  $Header: /home/art/GoogleDrive/Latest/myschema.d/RCS/print_coltype.ec,v 1.49 2023/11/16 12:39:46 art Exp $

  RCS Log:
  --------
  Revision 1.7  2003/10/29 16:13:12  kagel
  Improved backward compatibility by IFDEFing out new ANSI host parameter
  style parameter list in favor of old $parameter/K&R style list when NOINT8
  is defined.

  Revision 1.6  2002/10/03 15:52:33  kagel
  Fixed calculation of varchar precision.

  Revision 1.5  2001/10/31 21:39:09  kagel
  Added a parameter for syscolumns instead of using the global version.  This
  allows print_alters() to call this function also for specific columns after
  a table is finished processing.

  Revision 1.4  2001/10/30 14:38:54  kagel
  Cosmetic changes.

  Revision 1.3  2001/10/29 18:37:25  kagel
  Changes implementing support for collection types.

  Revision 1.2  2001/09/25 17:52:19  kagel
  Added support for ROW type columns.

  Revision 1.1  2001/09/25 14:24:55  kagel
  Initial revision


*/

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>

EXEC SQL INCLUDE sqlhdr;
EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;
EXEC SQL INCLUDE datetime;
EXEC SQL INCLUDE int8;
EXEC SQL INCLUDE ifxtypes.h;
EXEC SQL INCLUDE myschema.eh;

EXEC SQL INCLUDE prototypes.h;

static const char RCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest/myschema.d/RCS/print_coltype.ec,v 1.49 2023/11/16 12:39:46 art Exp $";
static const char RCSCompile[] = "@(#)$cc: " __FILE__ " " __DATE__ " " __TIME__ " $";

/* EXTERNS */
extern int SMICounts, NoCurrent;
extern FILE *outfileMI;
extern FILE *outfileME;
extern int QuoteIdent;

int print_coltype(
		   EXEC SQL IFNDEF NOINT8;
		   int is_rowtype,
		   FILE *outfile,
		   EXEC SQL BEGIN DECLARE SECTION;
		   PARAMETER long tabid,
		   PARAMETER char *tabname,
		   PARAMETER char *owner,
		   PARAMETER syscolumns_t syscolumns,
		   EXEC SQL END DECLARE SECTION;
		   int pc
		   EXEC SQL ELSE;
		   is_rowtype,
		   outfile,
		   tabid,
		   tabname,
		   owner,
		   syscolumns,
		   pc
		   EXEC SQL ENDIF;
		   )
EXEC SQL IFDEF NOINT8;
int is_rowtype;
FILE *outfile;
$parameter long tabid;
$parameter char *tabname;
$parameter char *owner;
$parameter syscolumns_t syscolumns;
int pc;
EXEC SQL ENDIF;
{
  char part1[30], part2[30], typeowner[UNAMELEN], typename[65536];
  int p1, p2, pt1, pt2, fr1, fr2;
  int ext;
  char mode[2];
  int is_bson = 0;

  EXEC SQL BEGIN DECLARE SECTION;
  static int first = 1;
  static char statement[2048];
  static int SQL_LOGICAL_CHAR;
  EXEC SQL END DECLARE SECTION;

  if (first == 1) {
      EXEC SQL SELECT (MOD( flags, 3 ) + 1) INTO :SQL_LOGICAL_CHAR FROM 'informix'.systables WHERE tabname = ' VERSION';
      // fprintf( stderr, "Forcing SQL_LOGICAL_CHAR = %d\n", SQL_LOGICAL_CHAR );
      first = 0;
  }
  if (syscolumns.extended_id != 0 ) {
    ext = return_extended( syscolumns.extended_id, typeowner, typename, mode );
    if (syscolumns.extended_id == 26) is_bson = 99;
  } else {
    ext = 0;
  }
  
  switch ((syscolumns.coltype % 256)) {
  case SQLCHAR:
    if (syscolumns.extended_id == 0) 
      fprintf( outfile, "CHAR(%d)", syscolumns.collength / SQL_LOGICAL_CHAR);
    else {
      if (!stripowner ) {
	if (ext == 0) {
	  if ((syscolumns.coltype % 256) == 22 || (major_vers == 9 && minor_vers <= 30))
	    fprintf( outfile, "\"%s\".%s",
		     typeowner,
		     typename );
	  else
	    switch (syscolumns.extended_id) {
	    case 15:
	    case 16:
	    case 17:
	    case 24:
	    case 25:
	    case 26:
	      fprintf( outfile, "\"%s\".%s",
		       typeowner,
		       typename );
	      break;
	    default:
	      fprintf( outfile, "\"%s\".%s(%d)",
		       typeowner,
		       typename,
		       syscolumns.collength );
	      break;
	    }
	} else
	  fprintf( outfile, "\"%s\".%s", typeowner, typename );
      } else {
	if (ext == 0) {
	  if ((syscolumns.coltype % 256) == 22 || (major_vers == 9 && minor_vers <= 30))
	    fprintf( outfile, "%s", typename );
	  else
	    switch (syscolumns.extended_id) {
	    case 15:
	    case 16:
	    case 17:
	    case 24:
	    case 25:
	    case 26:
	      fprintf( outfile, "%s",
		       typename );
	      break;
	    default:
	      fprintf( outfile, "%s(%d)",
		       typename,
		       syscolumns.collength );
	      break;
	    }
	} else
	  fprintf( outfile, "%s", typename );
      }
    }
    break;
  case SQLNCHAR:
    if (ext == 0)
      fprintf( outfile, "NCHAR(%d)", syscolumns.collength / SQL_LOGICAL_CHAR );
    else
      fprintf( outfile, "\"%s\".%s", typeowner, typename );
	
    break;
    EXEC SQL IFNDEF NOINT8;
  case SQLUDTFIXED:
  case SQLUDTVAR:
  case SQLROW:
    {
      if (!stripowner ) {
	if (ext == 0 && mode[0] != 'S')
	  if ((syscolumns.coltype % 256) == 22 || (major_vers == 9 && minor_vers <= 30))
	    fprintf( outfile, "\"%s\".%s",
		     typeowner,
		     typename );
	  else
	    switch (syscolumns.extended_id) {
	    case 15:
	    case 16:
	    case 17:
	    case 24:
	    case 25:
	    case 26:
	      fprintf( outfile, "\"%s\".%s",
		       typeowner,
		       typename );
	      break;
	    case 1:
		// fprintf( stderr, "At case 1: line %d.\n", __LINE__ );
		fprintf( outfile, "\"%s\".%s(%d)",
			 typeowner,
			 typename,
			 syscolumns.collength / SQL_LOGICAL_CHAR );
		break;
	    default:
	      fprintf( outfile, "\"%s\".%s(%d)",
		       typeowner,
		       typename,
		       syscolumns.collength );
	      break;
	    }
	else
	  fprintf( outfile, "\"%s\".%s", typeowner, typename );
      } else {
	if (ext == 0 && mode[0] != 'S')
	  if ((syscolumns.coltype % 256) == 22 || (major_vers == 9 && minor_vers <= 30))
	    fprintf( outfile, "%s", typename );
	  else
	    switch (syscolumns.extended_id) {
	    case 15:
	    case 16:
	    case 17:
	    case 24:
	    case 25:
	    case 26:
	      fprintf( outfile, "%s",
		       typename );
	      break;
	    case 1:
		// fprintf( stderr, "At case 1: line %d.\n", __LINE__ );
		fprintf( outfile, "\"%s\".%s(%d)",
			 typeowner,
			 typename,
			 syscolumns.collength / SQL_LOGICAL_CHAR );
		break;
	    default:
	      fprintf( outfile, "%s(%d)",
		       typename,
		       syscolumns.collength );
	      break;
	    }
	else
	  fprintf( outfile, "%s", typename );
      }
    }
    break;
  case SQLSET:
  case SQLMULTISET:
  case SQLLIST:
    {
      if (ext == 0)
	if (major_vers == 9 && minor_vers <= 30)
	  fprintf( outfile, "%s", typename );
	else
	  fprintf( outfile,
		   "%s(%d)",
		   typename,
		   syscolumns.collength );
      else
	fprintf( outfile, "%s", typename );
    }
    break;
    EXEC SQL ENDIF;
  case SQLSMINT:
    if (ext == 0)
      fprintf( outfile, "SMALLINT" );
    else
      fprintf( outfile, "\"%s\".%s", typeowner, typename );
    break;
  case SQLINT:
    if (ext == 0)
      fprintf( outfile, "INTEGER" );
    else
      fprintf( outfile, "\"%s\".%s", typeowner, typename );
    break;
  case SQLFLOAT:
    if (ext == 0)
      fprintf( outfile, "FLOAT" );
    else
      fprintf( outfile, "\"%s\".%s", typeowner, typename );
    break;
  case SQLSMFLOAT:
    if (ext == 0)
      fprintf( outfile, "SMALLFLOAT" );
    else
      fprintf( outfile, "\"%s\".%s", typeowner, typename );
    break;
    EXEC SQL IFNDEF NOINT8;
  case SQLINT8:
    if (ext == 0)
      fprintf( outfile, "INT8" );
    else
      fprintf( outfile, "\"%s\".%s", typeowner, typename );
    break;
    EXEC SQL ENDIF;

#if defined( SQLINFXBIGINT )
  case SQLINFXBIGINT:
    if (ext == 0)
      fprintf( outfile, "BIGINT" );
    else
      fprintf( outfile, "\"%s\".%s", typeowner, typename );
    break;
#endif
  case SQLDECIMAL:
    if (ext == 0) {
      if ((syscolumns.collength % 256) == 255)
	/* Floating point. */
	fprintf( outfile,
		 "DECIMAL(%d)",
		 syscolumns.collength / 256 );
      else
	/* Fixed point. */
	fprintf( outfile,
		 "DECIMAL(%d,%.1d)",
		 syscolumns.collength / 256,
		 syscolumns.collength % 256 );
    } else
      fprintf( outfile, "\"%s\".%s", typeowner, typename );
    break;
  case SQLSERIAL: {
    EXEC SQL BEGIN DECLARE SECTION;
    int maxmum;
    short i_maxmum;
    EXEC SQL END DECLARE SECTION;

    if (is_rowtype || is_external_table) {
      fprintf( outfile, "SERIAL" );
    } else if (SMICounts && init_Serial == Serial_NoCurrent) {
      fprintf( outfile,
	       "SERIAL(%d)",
	       pagecounts[pc].serialv );
    } else if (init_Serial == Serial_Default) {
	if (SMICounts) {
	    fprintf( outfile, "SERIAL {Curr: %d}", pagecounts[pc].serialv );
	} else {
	    fprintf( outfile, "SERIAL" );
	}
    } else if (init_Serial == Serial_Value) {
      if (SMICounts)
	fprintf( outfile,
		 "SERIAL(%d) {Curr: %d}",
		 init_Serial_Value,
		 pagecounts[pc].serialv );
      else
	fprintf( outfile,
		 "SERIAL(%d)",
		 init_Serial_Value );
    } else { /* Serial_Current: Use maximum existing value if valid. */
      strcpy( statement, "SELECT max(" );
      strcat( statement, syscolumns.colname );
      strcat( statement, ") FROM " );
      if (strlen( original_owner ) > 0) {
	strcat( statement, "\"" );
	strcat( statement, original_owner );
	strcat( statement, "\".\"" );
      } else if (QuoteIdent == 1)
	  strcat( statement, "\"" );
      strcat( statement, tabname );
      if (QuoteIdent == 1)
	  strcat( statement, "\"" );
      EXEC SQL PREPARE serl FROM :statement;
      if (sqlca.sqlcode) {
	fprintf( stderr,
		 "Cannot PREPARE SELECT statement for current "
		 "serial number, error: %d.\n",
		 sqlca.sqlcode );
	fprintf( stderr, "Error at character %d.\nStatement is:\n",
		 sqlca.sqlerrd[4] );
	fprintf( stderr, "%s\n", statement );
	rexit(16);
      }
      EXEC SQL EXECUTE serl INTO :maxmum indicator :i_maxmum;
      if (sqlca.sqlcode) {
	/* ASK - making use of the existing serialv instead of abort! 2023-11-15 
	fprintf( stderr,
		 "Cannot EXECUTE statement serl for %s.%s, Err: %d, ISAM: %d.\n",
		 systables.tabname,
		 syscolumns.colname,
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1]);
	fprintf( stderr, "Serl: %s;\n", statement );
	rexit(16);
	*/
	maxmum = pagecounts[pc].serialv;
	i_maxmum = 0;
      }
      EXEC SQL FREE serl;
      if (i_maxmum == -1)
	maxmum = 0;
      if (maxmum < 0 || maxmum == INT_MAX) {
	if (SMICounts)
	  fprintf( outfile,
		   "SERIAL(%d) {Max value out of range: %d.  Current: %d.}",
		   pagecounts[pc].serialv,
		   maxmum,
		   pagecounts[pc].serialv );
	else
	  fprintf( outfile,
		   "SERIAL {Max value out of range: %d.}",
		   maxmum );
      } else if (SMICounts)
	fprintf( outfile,
		 "SERIAL(%d) {Curr: %d}",
		 maxmum + 1,
		 pagecounts[pc].serialv );
      else 
	fprintf( outfile,
		 "SERIAL(%d)",
		 maxmum + 1 );
    }
  }
    break;
    EXEC SQL IFNDEF NOINT8;
#if defined(SQLBIGSERIAL)
  case SQLBIGSERIAL: {
    EXEC SQL BEGIN DECLARE SECTION;
    ifx_int8_t maxmum8;
    char maxstr[30];
    short i_maxmum;
    EXEC SQL END DECLARE SECTION;
    long long maxmum;

    if (is_rowtype || is_external_table)
      fprintf( outfile, "BIGSERIAL" );
    else if (SMICounts && init_Serial == Serial_NoCurrent)
      fprintf( outfile,
	       "SERIAL(%lld)",
	       (long long int)pagecounts[pc].serialvBig );
    else if (init_Serial == Serial_Default) {
      if (SMICounts)
	fprintf( outfile,
		 "BIGSERIAL {Curr: %lld}",
		 (long long int)pagecounts[pc].serialvBig );
      else
	fprintf( outfile, "BIGSERIAL" );
    } else if (init_Serial == Serial_Value) {
      if (SMICounts)
	fprintf( outfile,
		 "BIGSERIAL(%lld) {Curr: %lld}",
		 (long long int)init_Serial_Value,
		 (long long int)pagecounts[pc].serialvBig );
      else
	fprintf( outfile,
		 "BIGSERIAL(%lld)",
		 (long long int)init_Serial_Value );
    } else { /* Serial_Current: Use maximum existing value if valid. */
      strcpy( statement, "SELECT max(" );
      strcat( statement, syscolumns.colname );
      strcat( statement, ")::int8 FROM " );
      if (strlen( original_owner ) > 0) {
	strcat( statement, "\"" );
	strcat( statement, original_owner );
	strcat( statement, "\"." );
      }
      strcat( statement, tabname );
      EXEC SQL PREPARE bigserl FROM :statement;
      if (sqlca.sqlcode) {
	fprintf( stderr,
		 "Cannot PREPARE SELECT statement for current "
		 "serial number, error: %d.\n",
		 sqlca.sqlcode );
	fprintf( stderr, "Error at character %d.\nStatement is:\n",
		 sqlca.sqlerrd[4] );
	fprintf( stderr, "%s\n", statement );
	rexit(16);
      }
      EXEC SQL EXECUTE bigserl INTO :maxmum8 indicator :i_maxmum;
      if (sqlca.sqlcode) {
	fprintf( stderr,
		 "Cannot EXECUTE statement bigserl for %s.%s, Err: %d, ISAM: %d.\n",
		 systables.tabname,
		 syscolumns.colname,
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1]);
	fprintf( stderr, "Error at character %d.\nStatement is:\n",
		 sqlca.sqlerrd[4] );
	fprintf( stderr, "%s\n", statement );
	rexit(16);
      }
      EXEC SQL FREE bigserl;
      ifx_int8toasc( &maxmum8, maxstr, sizeof maxstr );
      maxmum = atoll( maxstr );
      if (i_maxmum == (short)-1)
	maxmum = 0LL;
      if ((long long int)maxmum < 0LL || (long long int)maxmum == 9223372036854775807LL) {
	if (SMICounts)
	  fprintf( outfile,
		   "BIGSERIAL(%lld) {Max value out of range: %lld.  Current: %lld.}",
		   (long long int)pagecounts[pc].serialvBig,
		   (long long int)maxmum,
		   (long long int)pagecounts[pc].serialvBig );
	else
	  fprintf( outfile,
		   "BIGSERIAL(1) {Max value out of range: %lld.}",
		   (long long int)maxmum );
      } else if (SMICounts)
	fprintf( outfile,
		 "BIGSERIAL(%lld) {Curr: %lld}",
		 (long long int)(maxmum + 1LL),
		 (long long int)pagecounts[pc].serialvBig );
      else 
	fprintf( outfile,
		 "BIGSERIAL(%lld)",
		 (long long int)(maxmum + 1LL) );
    }
  }
    break;
#endif
  case SQLSERIAL8: {
    EXEC SQL BEGIN DECLARE SECTION;
    ifx_int8_t maxmum8;
    char maxstr[30];
    bigint bigmaxmum;
    short i_maxmum;
    EXEC SQL END DECLARE SECTION;
    int err, idx;
    char printstr[128];

    if (is_rowtype || is_external_table)
      fprintf( outfile, "SERIAL8" );
    else if (SMICounts && init_Serial == Serial_NoCurrent)
      fprintf( outfile,
	       "SERIAL8(%lld)",
	       (long long int)pagecounts[pc].serialv8 );
    else if (init_Serial == Serial_Default) {
      if (SMICounts)
	fprintf( outfile,
		 "SERIAL8 {Curr: %lld}",
		 (long long int)pagecounts[pc].serialv8 );
      else
	fprintf( outfile, "SERIAL8" );
    } else if (init_Serial == Serial_Value) {
      if (SMICounts)
	fprintf( outfile,
		 "SERIAL8(%lld) {Curr: %lld}",
		 (long long int)init_Serial_Value,
		 (long long int)pagecounts[pc].serialv8 );
      else
	fprintf( outfile,
		 "SERIAL8(%d)",
		 init_Serial_Value );
    } else { /* Serial_Current: Use maximum existing value if valid. */
      strcpy( statement, "SELECT max(" );
      strcat( statement, syscolumns.colname );
      strcat( statement, ") FROM " );
      if (strlen( original_owner ) > 0) {
	strcat( statement, "\"" );
	strcat( statement, original_owner );
	strcat( statement, "\"." );
      }
      strcat( statement, tabname );
      EXEC SQL PREPARE serl8 FROM :statement;
      if (sqlca.sqlcode) {
	fprintf( stderr,
		 "Cannot PREPARE SELECT statement for current "
		 "serial8 number, error: %d.\n",
		 sqlca.sqlcode );
	fprintf( stderr, "Error at character %d.\nStatement is:\n",
		 sqlca.sqlerrd[4] );
	fprintf( stderr, "%s\n", statement );
	rexit(16);
      }
      EXEC SQL EXECUTE serl8 INTO :maxmum8 indicator :i_maxmum;
      if (sqlca.sqlcode) {
	fprintf( stderr,
		 "Cannot EXECUTE statement serl8 for %s.%s, Err: %d, ISAM: %d.\n",
		 systables.tabname,
		 syscolumns.colname,
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1]);
	fprintf( stderr, "Error at character %d.\nStatement is:\n",
		 sqlca.sqlerrd[4] );
	fprintf( stderr, "%s\n", statement );
	rexit(16);
      }
      EXEC SQL FREE serl8;
      ifx_int8toasc( &maxmum8, maxstr, sizeof maxstr );
      bigmaxmum = atoll( maxstr );
      if (i_maxmum == -1)
	bigmaxmum = 0LL;
      if ((long long int)bigmaxmum < 0LL || (long long int)bigmaxmum == 9223372036854775807LL) {
	if (SMICounts)
	  fprintf( outfile,
		   "SERIAL8(%lld) {Max value out of range: %lld.  Current: %lld.}",
		   (long long int)pagecounts[pc].serialv8,
		   (long long int)bigmaxmum,
		   (long long int)pagecounts[pc].serialv8 );
	else
	  fprintf( outfile,
		   "SERIAL8(1) {Max value out of range: %lld.}",
		   (long long int)bigmaxmum );
      } else if (SMICounts)
	fprintf( outfile,
		 "SERIAL8(%lld) {Curr: %lld}",
		 (long long int)(bigmaxmum + 1LL),
		 (long long int)pagecounts[pc].serialv8 );
      else 
	fprintf( outfile,
		 "SERIAL8(%lld)",
		 (long long int)(bigmaxmum + 1LL) );
    }
  }
    break;
    EXEC SQL ENDIF;
  case SQLDATE:
    if (ext == 0)
      fprintf( outfile, "DATE" );
    else
      fprintf( outfile, "\"%s\".%s", typeowner, typename );
    break;
  case SQLMONEY:
    if (ext == 0)
      fprintf( outfile,
	       "MONEY(%d,%.1d)",
	       syscolumns.collength / 256,
	       syscolumns.collength % 256 );
    else
      fprintf( outfile, "\"%s\".%s", typeowner, typename );
    break;
  case SQLDTIME:
    if (ext == 0) {
      pt1 = (((syscolumns.collength / 16) % 16) <= 10) ?
	(((syscolumns.collength / 16) % 16) / 2) : 6;
      strcpy( part1, dtparts[pt1] );
      pt2 = ((syscolumns.collength % 16) <= 10) ?
	((syscolumns.collength % 16) / 2) : 6;
      if (pt2 == 6) {
	fr2 = (syscolumns.collength % 16) - 10;
	sprintf( part2, "%s(%d)", dtparts[pt2], fr2 );
      } else
	strcpy( part2, dtparts[pt2] );
      fprintf( outfile,
	       "DATETIME %s TO %s",
	       part1,
	       part2 );
    } else {
      fprintf( outfile, "%s", typename );
    }
    break;
  case SQLINTERVAL:
    if (ext == 0) {
      p1 = (syscolumns.collength / 16) % 16;
      pt1 = (p1 <= 10) ? (p1 / 2) : 6;
      fr2 = (syscolumns.collength % 16) - 10;
      p2 = syscolumns.collength % 16;
      pt2 = (p2 <= 10) ? (p2 / 2) : 6;
      if (pt1 != 6) {
	fr1 = (syscolumns.collength / 256) + p1 - p2;
	sprintf( part1, "%s(%d)", dtparts[pt1], fr1 );
      } else
	strcpy( part1, dtparts[pt1] );
      if (pt2 == 6) {
	sprintf( part2, "%s(%d)", dtparts[pt2], fr2 );
      } else
	strcpy( part2, dtparts[pt2] );
      fprintf( outfile,
	       "INTERVAL %s TO %s",
	       part1,
	       part2 );
    } else {
      fprintf( outfile, "%s", typename );
    }
    break;
  case SQLVCHAR:
    if (syscolumns.extended_id == 0)
      fprintf( outfile,
	       "VARCHAR(%u,%.1u)",
	       ((unsigned)syscolumns.collength % (unsigned)256) / SQL_LOGICAL_CHAR,
	       ((unsigned)syscolumns.collength / (unsigned)256) % (unsigned)256 );
    else {
      if (!stripowner ) {
	if ((syscolumns.coltype % 256) == 22 || (major_vers == 9 && minor_vers <= 30))
	  fprintf( outfile, "\"%s\".%s",
		   typeowner,
		   typename );
	else
	  switch (syscolumns.extended_id) {
	  case 15:
	  case 16:
	  case 17:
	  case 24:
	  case 25:
	  case 26:
	    fprintf( outfile, "\"%s\".%s",
		     typeowner,
		     typename );
	    break;
	    case 1:
		// fprintf( stderr, "At case 1: line %d.\n", __LINE__ );
		fprintf( outfile, "\"%s\".%s(%d)",
			 typeowner,
			 typename,
			 syscolumns.collength / SQL_LOGICAL_CHAR );
		break;
	  default:
	    fprintf( outfile, "\"%s\".%s(%d)",
		     typeowner,
		     typename,
		     syscolumns.collength );
	    break;
	  }
      } else {
	if (ext == 0)
	  if ((syscolumns.coltype % 256) == 22 || (major_vers == 9 && minor_vers <= 30))
	    fprintf( outfile, "%s", typename );
	  else
	    switch (syscolumns.extended_id) {
	    case 15:
	    case 16:
	    case 17:
	    case 24:
	    case 25:
	    case 26:
	      fprintf( outfile, "%s",
		       typename );
	      break;
	    case 40:
	    case 1:
		fprintf( outfile, "%s(%d)",
			 typename,
			 ((unsigned)syscolumns.collength) / SQL_LOGICAL_CHAR );
			 break;
	    default:
	      fprintf( outfile, "%s(%d)",
		       typename,
		       syscolumns.collength );
	      break;
	    }
	else
	  fprintf( outfile, "%s", typename );
      }
    }
    break;
  case SQLNVCHAR:
    if (ext == 0)
      fprintf( outfile,
	       "NVARCHAR(%u,%.1u)",
	       ((unsigned)syscolumns.collength % (unsigned)256) / SQL_LOGICAL_CHAR,
	       ((unsigned)syscolumns.collength / (unsigned)256) % (unsigned)256 );
    else {
      fprintf( outfile, "%s", typename );
    }
    break;
  case SQLBYTES:
    {
      EXEC SQL BEGIN DECLARE SECTION;
      string spacename[129];
      EXEC SQL END DECLARE SECTION;
      if ( syscolumns.colmin == -1 || outfile == outfileME || outfile == outfileMI )
	strcpy( spacename, "TABLE" );
      else
	EXEC SQL
	  SELECT spacename
	  INTO :spacename
	  FROM "informix".sysblobs
	  WHERE tabid = :tabid 
	  AND colno = :syscolumns.colno;
      fprintf( outfile,
	       "BYTE IN %s",
	       spacename );
    }
    break;
  case SQLTEXT:
    {
      EXEC SQL BEGIN DECLARE SECTION;
      string spacename[129];
      EXEC SQL END DECLARE SECTION;
      if ( syscolumns.colmin == -1 || outfile == outfileME || outfile == outfileMI )
	strcpy( spacename, "TABLE" );
      else
	EXEC SQL
	  SELECT spacename
	  INTO :spacename
	  FROM "informix".sysblobs
	  WHERE tabid = :tabid 
	  AND colno = :syscolumns.colno;
      fprintf( outfile,
	       "TEXT IN %s",
	       spacename );
    }
    break;

  default:
    fprintf( outfile, "UNKNOWN TYPE(%d, %d)",
	     syscolumns.coltype,
	     syscolumns.collength );
    break;
  }
  if (has_external_tables)
    probe_external( tabid, syscolumns.colno, outfile );

  return is_bson;
}
