/* print_constraints.ec - Print all constraints associated with a table.
 */
/*
  RCS Header:
  -----------
  $Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_constraints.ec,v 2.36 2021/07/14 18:41:23 art Exp $

  RCS Log:
  --------
  $Log: print_constraints.ec,v $
  Revision 2.36  2021/07/14 18:41:23  art
  Removed old code that is noop.

  Revision 2.35  2020/11/25 12:24:06  art
  Extended --no-blade-procs/--no-blade-funcs to also exclude the resources added by the 
  blade management routines when any blade is registered.

  Revision 2.34  2019/09/10 16:42:03  art
  Moved UDT CREATE TYPE statements to the main file, GRANTS on TYPES to the grant file.
  Cleaned format for DISTINCT TYPEs.

  Revision 2.33  2019/04/08 13:47:24  art
  Added AS "owner" to REVOKE statements.
  Added the owner of the referenced table in FOREIGN KEY constraints.

  Revision 2.32  2019/04/03 02:46:56  art
  Fixed -F references printout.

  Revision 2.31  2018/06/22 16:30:05  art
  New options: --use-database-owner, --use-table-owner, --constraint-file
  Fixed some problems with the partial schema file output.
  Expanded "owner"s to all objects.

  Revision 2.30  2016/12/13 23:51:19  art
  Added support for unencrypted storage spaces when encrypted storage is enabled (12.10.xC8+).
  Added new option --no-check to output CHECK constraints with the NOVALIDATE option for
  migrating databases (effective only on a populated table created without CHECK constraints.

  Revision 2.29  2016/11/30 07:08:12  art
  Fixed problem with unique constraints.

  Revision 2.28  2016/11/28 00:37:38  art
  Added new option, --keep-generated-names, to create constraints with the names auto-generated
  in the existing schema. The default is to not explicitely name constraints that have generated
  names.

*/

static const char RCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_constraints.ec,v 2.36 2021/07/14 18:41:23 art Exp $";
static const char RCSCompile[] = "@(#)$cc: " __FILE__ " " __DATE__ " " __TIME__ " $";

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <errno.h>

EXEC SQL INCLUDE sqlhdr;
EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;
EXEC SQL INCLUDE datetime;

EXEC SQL INCLUDE myschema.eh;
EXEC SQL INCLUDE prototypes.h;


void print_constraints( FILE *outfile, int tflag ) 
{
    int i, a, b;
    int uknum = 1, fknum = 1;

    EXEC SQL BEGIN DECLARE SECTION;
    string reftable[129], refowner[33], delrule[2];
    long primconstr, ptabid, rtabid;
    char localstr1[2048], localstr2[2048];
    char state_prim[2], state_psc[2], state_psc2[2];
    char stmt_cols[4096];
    EXEC SQL END DECLARE SECTION;


    if ((major_vers == 11 && minor_vers >= 50) || major_vers > 11) {
	snprintf( stmt_cols,
		  sizeof stmt_cols,
		  "SELECT colno, colname "
		  "FROM \"informix\".syscolumns "
		  "WHERE tabid = ? AND colattr NOT IN (17, 33, 65)"
		  "ORDER BY colno " );
    } else {
	snprintf( stmt_cols,
		  sizeof stmt_cols,
		  "SELECT colno, colname "
		  "FROM \"informix\".syscolumns "
		  "WHERE tabid = ? "
		  "ORDER BY colno " );
    }
    EXEC SQL PREPARE s_cols FROM :stmt_cols;
    if (sqlca.sqlcode) {
        fprintf( stderr,
                 "Cannot PREPARE select statement s_cols, error: %d.\n",
                 sqlca.sqlcode );
        fprintf( stderr, "Error at character %d.\n",
                 sqlca.sqlerrd[4] );
        rexit(18);
    }
    EXEC SQL DECLARE cols CURSOR FOR s_cols;
    if (sqlca.sqlcode < 0) {
        fprintf( stderr,
                 "Error DECLARing CURSOR cols: Err=%d, ISAM=%d\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1] );
        rexit( 21 );
    }

    strcpy( alien, "alien" );
    snprintf( localstr1,
              sizeof localstr1,
              "SELECT st.tabname, st.owner, rt.tabname, rt.owner, sr.primary, sr.ptabid,  "
              "      sr.delrule, sc.constrid, sc.constrname, sc.constrtype, sc.owner, "
              "      si.idxname, si.tabid, si.part1, si.part2, si.part3,  "
              "      si.part4, si.part5, si.part6, si.part7, si.part8,  "
              "      si.part9, si.part10, si.part11, si.part12, si.part13,  "
              "      si.part14, si.part15, si.part16, rc.tabid, os.state, os2.state "
              "FROM \"informix\".systables st, \"informix\".sysconstraints sc, "
              "     \"informix\".sysindexes si, \"informix\".sysreferences sr, "
              "     \"informix\".systables rt, \"informix\".sysconstraints rc, "
              "     \"informix\".sysobjstate os, \"informix\".sysobjstate os2 "
              "WHERE st.tabid = sc.tabid "
              "  AND st.tabtype != 'Q' "
              "  AND st.tabname NOT MATCHES 'cdr_deltab_[0-9][0-9][0-9][0-9][0-9][0-9]*' "
              "  AND rt.tabid = sr.ptabid "
              "  AND sc.constrid = sr.constrid "
              "  AND sc.tabid = si.tabid "
              "  AND sc.idxname = si.idxname "
              "  AND sc.constrtype = 'R' "
              "  AND st.tabname MATCHES ? "
              "  AND os.tabid = st.tabid AND os.name = sc.constrname AND os.objtype = 'C' "
	      "  AND os2.tabid = st.tabid AND os2.name = si.idxname AND os2.objtype = 'I' "
              "  AND sr.primary = rc.constrid " );
    if (doOrder == IdOrder)
        strcat( localstr1, "ORDER BY si.tabid, sc.constrid " );
    else
        strcat( localstr1, "ORDER BY si.tabid, sc.constrname " );

    EXEC SQL PREPARE psc_1 FROM :localstr1;
    if (sqlca.sqlcode < 0) {
        fprintf( stderr,
                 "Error PREPARING statement for alien: Err=%d, ISAM=%d\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1] );
	fprintf(stderr, "Alien:\n%s;\n", localstr1 );
        rexit( 21 );
    }

    EXEC SQL DECLARE :alien CURSOR FOR psc_1;
    if (sqlca.sqlcode < 0) {
        fprintf( stderr,
                 "Error DECLARing CURSOR alien: Err=%d, ISAM=%d\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1] );
        rexit( 21 );
    }

    snprintf( localstr2,
              sizeof localstr2,
              "SELECT tabname, st.owner, constrid, constrname, constrtype, sc.owner, "
              "      si.idxname, si.tabid, si.part1, si.part2, si.part3, si.part4, "
              "      si.part5, si.part6, si.part7, si.part8, si.part9, si.part10, "
              "      si.part11, si.part12, si.part13, si.part14, "
              "      si.part15, si.part16, state "
              "FROM \"informix\".systables st, \"informix\".sysconstraints sc, "
              "     \"informix\".sysindexes si, \"informix\".sysobjstate os "
              "WHERE st.tabid = sc.tabid "
              "  AND st.tabtype != 'Q' "
              "  AND st.tabname NOT MATCHES 'cdr_deltab_[0-9][0-9][0-9][0-9][0-9][0-9]*' "
              "  AND sc.tabid = si.tabid "
              "  AND sc.idxname = si.idxname "
              "  AND sc.constrtype = 'U' "
              "  AND os.tabid = st.tabid AND os.name = sc.constrname AND os.objtype = 'C' "
              "  AND st.tabname MATCHES ? " );
    if (doOrder == IdOrder)
        strcat( localstr2, "ORDER BY si.tabid, constrid " );
    else
        strcat( localstr2, "ORDER BY si.tabid, constrname " );
    EXEC SQL PREPARE psc_2 FROM :localstr2;
    if (sqlca.sqlcode < 0) {
        fprintf( stderr,
                 "Error PREPARING statement for uniq: Err=%d, ISAM=%d\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1] );
        rexit( 21 );
    }

    EXEC SQL DECLARE uniq CURSOR FOR psc_2;
    if (sqlca.sqlcode < 0) {
        fprintf( stderr,
                 "Error DECLARing CURSOR uniq: Err=%d, ISAM=%d\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1] );
        rexit( 21 );
    }

    strcpy( statement2,
            "SELECT tabname, st.owner, constrid, constrname, constrtype, sc.owner, \n"
            "     si.idxname, si.tabid, si.part1, si.part2, si.part3, \n"
            "     si.part4, si.part5, si.part6, si.part7, si.part8, \n"
            "     si.part9, si.part10, si.part11, si.part12, si.part13, \n"
            "     si.part14, si.part15, si.part16, state \n"
            "FROM \"informix\".systables st, \n"
            "     \"informix\".sysconstraints sc, \n"
            "     \"informix\".sysindexes si, \n"
            "     \"informix\".sysobjstate os \n"
            "WHERE st.tabid = sc.tabid \n"
            "  AND sc.tabid = si.tabid \n"
            "  AND st.tabtype != 'Q' "
            "  AND st.tabname NOT MATCHES 'cdr_deltab_[0-9][0-9][0-9][0-9][0-9][0-9]*' "
            "  AND sc.idxname = si.idxname \n"
            "  AND sc.constrtype = 'P' \n"
            "  AND os.tabid = st.tabid AND os.name = sc.constrname AND os.objtype = 'C' \n"
            "  AND st.tabname MATCHES ? \n" );
    if (no_blade_funcs == 1) {
        strcat( statement2, "AND st.tabname NOT IN ( " );
	strcat( statement2, "'sysblderrorlog','sysbldobjdepends','sysbldobjkinds','sysbldregistered'," );
	strcat( statement2, "'sysbldirequired','sysbldiprovided','sysbldobjects','bld_registered'," );
	strcat( statement2, "'bldi_required','bldi_provided' " );
	strcat( statement2, ") " );
    }
    if (!tflag || strcmp( tablename, "*" ) == 0)
        strcat( statement2, "  AND st.tabid >= 100 " );
    if (doOrder == IdOrder)
        strcat( statement2, "ORDER BY si.tabid, constrid; " );
    else
        strcat( statement2, "ORDER BY si.tabid, constrname; " );
    EXEC SQL PREPARE s_prim FROM :statement2;
    if (sqlca.sqlcode) {
        fprintf( stderr,
                 "Cannot PREPARE select statement s_prim, "
                 "error: %d, ISAM=%d.\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1] );
        fprintf( stderr, "Error at character %d.\n",
                 sqlca.sqlerrd[4] );
        fprintf( stderr, "%s\n\n", statement2 );
        rexit(18);
    }
            
    EXEC SQL DECLARE prim CURSOR FOR s_prim;
    if (sqlca.sqlcode < 0) {
        fprintf( stderr,
                 "Error DECLARing CURSOR prim: Err=%d, ISAM=%d\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1] );
        rexit( 21 );
    }

    /* Primary Key Constraints. */
    EXEC SQL OPEN prim USING :tablename;
    if (sqlca.sqlcode < 0) {
        fprintf( stderr,
                 "Error OPENing CURSOR prim: Err=%d, ISAM=%d\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1] );
        rexit( 22 );
    }
    for (;;) {
        EXEC SQL FETCH prim
            INTO
            :systables.tabname,
	    :systables.owner,
            :sysconstraints.constrid,
            :sysconstraints.constrname,
            :sysconstraints.constrtype,
	    :sysconstraints.owner,
            :sysindexes.idxname,
            :sysindexes.tabid,
            :sysindexes.part[0],
            :sysindexes.part[1],
            :sysindexes.part[2],
            :sysindexes.part[3],
            :sysindexes.part[4],
            :sysindexes.part[5],
            :sysindexes.part[6],
            :sysindexes.part[7],
            :sysindexes.part[8],
            :sysindexes.part[9],
            :sysindexes.part[10],
            :sysindexes.part[11],
            :sysindexes.part[12],
            :sysindexes.part[13],
            :sysindexes.part[14],
            :sysindexes.part[15],
            :state_prim;
        if (sqlca.sqlcode == SQLNOTFOUND) {
            break;
        } else if (sqlca.sqlcode < 0) {
            fprintf( stderr,
                     "Error FETCHing CURSOR prim: Err=%d, ISAM=%d\n",
                     sqlca.sqlcode,
                     sqlca.sqlerrd[1] );
            rexit( 21 );
        }

        EXEC SQL OPEN cols USING :sysindexes.tabid;
        if (sqlca.sqlcode < 0) {
            fprintf( stderr,
                     "Error OPENing CURSOR cols: Err=%d, ISAM=%d\n",
                     sqlca.sqlcode,
                     sqlca.sqlerrd[1] );
            rexit( 22 );
        }
        for (i=0;;i++) {
            EXEC SQL FETCH cols INTO :syscolumns.colno, :syscolumns.colname;
            if (sqlca.sqlcode == SQLNOTFOUND)
                break;
            else if (sqlca.sqlcode < 0) {
                fprintf( stderr,
                         "Error FETCHing CURSOR cols: Err=%d, ISAM=%d\n",
                         sqlca.sqlcode,
                         sqlca.sqlerrd[1] );
            }
            if (ncols <= i) {
                columns[i] = (char *)malloc( IDBYTES );
                if (!columns[i]) {
                    fprintf( stderr,
                             "Cannot allocate memory for column constraint "
                             "data.  Errno=%d.\n",
                             errno );
                    rexit(99);
                }
                ncols++;
            }
            strcpy( columns[i], syscolumns.colname );
        }
        EXEC SQL CLOSE cols;

	if (stripowner)
	  fprintf( outfile,
		   "ALTER TABLE %s ADD CONSTRAINT PRIMARY KEY (\n",
		   systables.tabname );
	else
	  fprintf( outfile,
		   "ALTER TABLE \"%s\".%s ADD CONSTRAINT PRIMARY KEY (\n",
		   UseDatabaseOwner ? dbowner : (UseSetOwner ? owner_override : systables.owner),
		   systables.tabname );

        for (i=0; i<16 && sysindexes.part[i]; i++) {
            if (i)
                fprintf( outfile, ",\n" );
            fprintf( outfile,
                     "\t%s",
                     columns[abs(sysindexes.part[i]) - 1] );
        }
        fprintf( outfile, "\n)" );

        if (sscanf( sysconstraints.constrname, "u%d_%d", &a, &b ) == 2) 
	    if (LongKeyIndices && stripowner)
		fprintf( outfile, " CONSTRAINT %s_pk", systables.tabname );
	    else if (LongKeyIndices)
		fprintf( outfile,
			 " CONSTRAINT \"%s\".%s_pk",
			 UseDatabaseOwner ? dbowner : (UseTableOwner ? systables.owner : (UseSetOwner ? owner_override : sysconstraints.owner)),
			 systables.tabname );
	    else if (KeepGenNames)
		if (stripowner)
		    fprintf( outfile, " CONSTRAINT %s", sysconstraints.constrname );
		else
		    fprintf( outfile,
			     " CONSTRAINT \"%s\".%s",
			     UseDatabaseOwner ? dbowner : (UseTableOwner ? systables.owner : (UseSetOwner ? owner_override : sysconstraints.owner)),
			     sysconstraints.constrname );
	    else
		fprintf( outfile, " " );
        else if (stripowner)
            fprintf( outfile, " CONSTRAINT %s", sysconstraints.constrname );
	else
            fprintf( outfile,
		     " CONSTRAINT \"%s\".%s",
		     UseDatabaseOwner ? dbowner : (UseTableOwner ? systables.owner : (UseSetOwner ? owner_override : sysconstraints.owner)),
		     sysconstraints.constrname );

        if (do_filtering) {
            print_objectstate( outfile, "D", "E" ); /* If filtering, create disabled, then filter. */
        } else {
            print_objectstate( outfile, state_prim, "E" );
        }
	fprintf( outfile, ";\n\n" );

        if (do_filtering) {
	    if (stripowner)
		fprintf( outfile,
			 "SET CONSTRAINTS %s FILTERING WITHOUT ERROR;\n\n",
			 sysconstraints.constrname );
	    else
		fprintf( outfile,
			 "SET CONSTRAINTS \"%s\".%s FILTERING WITHOUT ERROR;\n\n",
			 UseDatabaseOwner ? dbowner : (UseTableOwner ? systables.owner : (UseSetOwner ? owner_override : sysconstraints.owner)),
			 sysconstraints.constrname );
            fprintf( outfile,
                     "SELECT '%s', COUNT(*) as num_violations FROM %s_vio;\n",
                     systables.tabname,
                     systables.tabname );
        }

    }
    EXEC SQL CLOSE prim;

    /* Unique Constraints. */
    EXEC SQL OPEN uniq USING :tablename;
    if (sqlca.sqlcode < 0) {
        fprintf( stderr,
                 "Error OPENing CURSOR uniq: Err=%d, ISAM=%d\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1] );
        rexit( 22 );
    }
    for (;;) {
        EXEC SQL FETCH uniq
            INTO
            :systables.tabname,
	    :systables.owner,
            :sysconstraints.constrid,
            :sysconstraints.constrname,
            :sysconstraints.constrtype,
	    :sysconstraints.owner,
            :sysindexes.idxname,
            :sysindexes.tabid,
            :sysindexes.part[0],
            :sysindexes.part[1],
            :sysindexes.part[2],
            :sysindexes.part[3],
            :sysindexes.part[4],
            :sysindexes.part[5],
            :sysindexes.part[6],
            :sysindexes.part[7],
            :sysindexes.part[8],
            :sysindexes.part[9],
            :sysindexes.part[10],
            :sysindexes.part[11],
            :sysindexes.part[12],
            :sysindexes.part[13],
            :sysindexes.part[14],
            :sysindexes.part[15],
            :state_psc2;
        if (sqlca.sqlcode == SQLNOTFOUND) {
            break;
        } else if (sqlca.sqlcode < 0) {
            fprintf( stderr,
                     "Error FETCHing CURSOR uniq: Err=%d, ISAM=%d\n",
                     sqlca.sqlcode,
                     sqlca.sqlerrd[1] );
            rexit( 21 );
        }

        EXEC SQL OPEN cols USING :sysindexes.tabid;
        if (sqlca.sqlcode < 0) {
            fprintf( stderr,
                     "Error OPENing CURSOR cols: Err=%d, ISAM=%d\n",
                     sqlca.sqlcode,
                     sqlca.sqlerrd[1] );
            rexit( 22 );
        }

        for (i=0;;i++) {
            EXEC SQL FETCH cols INTO :syscolumns.colno, :syscolumns.colname;
            if (sqlca.sqlcode == SQLNOTFOUND)
                break;
            else if (sqlca.sqlcode < 0) {
                fprintf( stderr,
                         "Error FETCHing CURSOR cols: Err=%d, ISAM=%d\n",
                         sqlca.sqlcode,
                         sqlca.sqlerrd[1] );
            }
            if (ncols <= i) {
                columns[i] = (char *)malloc( IDBYTES );
                if (!columns[i]) {
                    fprintf( stderr,
                             "Cannot allocate memory for column constraint "
                             "data(2).  Errno=%d.\n",
                             errno );
                    rexit(99);
                }
                ncols++;
            }
            strcpy( columns[i], syscolumns.colname );
        }
        EXEC SQL CLOSE cols;
	if (sysindexes.part[0] <= i) {
	    /* Otherwise this is the WITH ERKEY constraint which should not be output. */
	    if (stripowner)
	      fprintf( outfile,
		       "ALTER TABLE %s ADD CONSTRAINT UNIQUE (\n",
		       systables.tabname );
	    else
	      fprintf( outfile,
		       "ALTER TABLE \"%s\".%s ADD CONSTRAINT UNIQUE (\n",
		       UseDatabaseOwner ? dbowner : (UseSetOwner ? owner_override : systables.owner),
		       systables.tabname );
	    for (i=0; i<16 && sysindexes.part[i]; i++) {
		if (i)
		    fprintf( outfile, ",\n" );
		fprintf( outfile,
			 "\t%s",
			 columns[abs(sysindexes.part[i]) - 1] );
	    }
	    fprintf( outfile, "\n)" );

	    if (sscanf( sysconstraints.constrname, "u%d_%d", &a, &b ) == 2)
		if (LongKeyIndices && stripowner)
		    fprintf( outfile, " CONSTRAINT %s_uk%.3d", systables.tabname, uknum++ );
		else if (LongKeyIndices)
		    fprintf( outfile,
			     " CONSTRAINT \"%s\".%s_uk%.3d",
			     UseDatabaseOwner ? dbowner : (UseTableOwner ? systables.owner : (UseSetOwner ? owner_override : sysconstraints.owner)),
			     systables.tabname,
			     uknum++ );
		else if (KeepGenNames)
		    if (stripowner)
			fprintf( outfile, " CONSTRAINT %s", sysconstraints.constrname );
		    else
			fprintf( outfile,
				 " CONSTRAINT \"%s\".%s",
				 UseDatabaseOwner ? dbowner : (UseTableOwner ? systables.owner : (UseSetOwner ? owner_override : sysconstraints.owner)),
				 sysconstraints.constrname );
		else
		    fprintf( outfile, " " );
	    else if (stripowner)
		fprintf( outfile, " CONSTRAINT %s", sysconstraints.constrname );
	    else
		fprintf( outfile,
			 " CONSTRAINT \"%s\".%s",
			 UseDatabaseOwner ? dbowner : (UseTableOwner ? systables.owner : (UseSetOwner ? owner_override : sysconstraints.owner)),
			 sysconstraints.constrname );
	    if (do_filtering) {
		print_objectstate( outfile, "D", "E" ); /* If filtering, create disabled, then filter. */
	    } else {
		print_objectstate( outfile, state_psc2, "E" );
	    }
	    fprintf( outfile, ";\n\n" );

	    if (do_filtering) {
		if (stripowner)
		    fprintf( outfile,
			     "SET CONSTRAINTS %s FILTERING WITHOUT ERROR;\n\n",
			     sysconstraints.constrname );
		else
		    fprintf( outfile,
			     "SET CONSTRAINTS \"%s\".%s FILTERING WITHOUT ERROR;\n\n",
			     UseDatabaseOwner ? dbowner : (UseTableOwner ? systables.owner : (UseSetOwner ? owner_override : sysconstraints.owner)),
			     sysconstraints.constrname );
		fprintf( outfile,
			 "SELECT '%s', COUNT(*) as num_violations FROM %s_vio;\n",
			 systables.tabname,
			 systables.tabname );
	    }
	}
    }
    EXEC SQL CLOSE uniq;

    /* Foreign Key Constraints. */
    EXEC SQL OPEN :alien using :tablename;
    if (sqlca.sqlcode < 0) {
        fprintf( stderr,
                 "Error OPENing CURSOR alien: Err=%d, ISAM=%d\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1] );
        rexit( 22 );
    }
    for (;;) {
        EXEC SQL FETCH :alien
            INTO
            :systables.tabname,
	    :systables.owner,
            :reftable,
	    :refowner,
            :primconstr,
            :ptabid,
            :delrule,
            :sysconstraints.constrid,
            :sysconstraints.constrname,
            :sysconstraints.constrtype,
	    :sysconstraints.owner,
            :sysindexes.idxname,
            :sysindexes.tabid,
            :sysindexes.part[0],
            :sysindexes.part[1],
            :sysindexes.part[2],
            :sysindexes.part[3],
            :sysindexes.part[4],
            :sysindexes.part[5],
            :sysindexes.part[6],
            :sysindexes.part[7],
            :sysindexes.part[8],
            :sysindexes.part[9],
            :sysindexes.part[10],
            :sysindexes.part[11],
            :sysindexes.part[12],
            :sysindexes.part[13],
            :sysindexes.part[14],
            :sysindexes.part[15],
            :rtabid,
            :state_psc,
	    :state_psc2;
        if (sqlca.sqlcode == SQLNOTFOUND) {
            break;
        } else if (sqlca.sqlcode < 0) {
            fprintf( stderr,
                     "Error FETCHing CURSOR alien: Err=%d, ISAM=%d\n",
                     sqlca.sqlcode,
                     sqlca.sqlerrd[1] );
            rexit( 21 );
        }

        EXEC SQL OPEN cols USING :sysindexes.tabid;
        if (sqlca.sqlcode < 0) {
            fprintf( stderr,
                     "Error OPENing CURSOR cols: Err=%d, ISAM=%d\n",
                     sqlca.sqlcode,
                     sqlca.sqlerrd[1] );
            rexit( 22 );
        }
        for (i=0;;i++) {
            EXEC SQL FETCH cols INTO :syscolumns.colno, :syscolumns.colname;
            if (sqlca.sqlcode == SQLNOTFOUND)
                break;
            else if (sqlca.sqlcode < 0) {
                fprintf( stderr,
                         "Error FETCHing CURSOR cols: Err=%d, ISAM=%d\n",
                         sqlca.sqlcode,
                         sqlca.sqlerrd[1] );
            }
            if (ncols <= i) {
                columns[i] = (char *)malloc( IDBYTES );
                if (!columns[i]) {
                    fprintf( stderr,
                             "Cannot allocate memory for column constraint "
                             "data(3).  Errno=%d.\n",
                             errno );
                    rexit(99);
                }
                ncols++;
            }
            strcpy( columns[i], syscolumns.colname );
        }
        EXEC SQL CLOSE cols;

	if (stripowner)
	  fprintf( outfile,
		   "ALTER TABLE %s ADD CONSTRAINT FOREIGN KEY (\n",
		   systables.tabname );
	else
	  fprintf( outfile,
		   "ALTER TABLE \"%s\".%s ADD CONSTRAINT FOREIGN KEY (\n",
		   UseDatabaseOwner ? dbowner : (UseSetOwner ? owner_override : systables.owner),
		   systables.tabname );
        for (i=0; i<16 && sysindexes.part[i]; i++) {
            if (i)
                fprintf( outfile, ",\n" );
            fprintf( outfile,
                     "\t%s",
                     columns[abs(sysindexes.part[i]) - 1] );
        }
        if (rtabid == ptabid) {
	    if (stripowner)
	        fprintf( outfile, "\n) REFERENCES %s (\n", reftable );
	    else
	        fprintf( outfile, 
			 "\n) REFERENCES \"%s\".%s (\n",
			 UseDatabaseOwner ? dbowner : (UseSetOwner ? owner_override : refowner ),
			 reftable );
            EXEC SQL
                SELECT part1, part2, part3, part4, part5, part6, part7, part8
                , part9, part10, part11, part12, part13, part14, part15, part16
                INTO
                :sysindexes.part[0],
                :sysindexes.part[1],
                :sysindexes.part[2],
                :sysindexes.part[3],
                :sysindexes.part[4],
                :sysindexes.part[5],
                :sysindexes.part[6],
                :sysindexes.part[7],
                :sysindexes.part[8],
                :sysindexes.part[9],
                :sysindexes.part[10],
                :sysindexes.part[11],
                :sysindexes.part[12],
                :sysindexes.part[13],
                :sysindexes.part[14],
                :sysindexes.part[15]
                FROM "informix".sysindexes si, "informix".sysconstraints sc
                WHERE si.tabid = sc.tabid
                AND si.idxname = sc.idxname
                AND sc.constrid = :primconstr;

            EXEC SQL OPEN cols USING :ptabid;
            if (sqlca.sqlcode < 0) {
                fprintf( stderr,
                         "Error OPENing CURSOR cols2: Err=%d, ISAM=%d\n",
                         sqlca.sqlcode,
                         sqlca.sqlerrd[1] );
                rexit( 22 );
            }
            for (i=0;;i++) {
                EXEC SQL FETCH cols INTO :syscolumns.colno, :syscolumns.colname;
                if (sqlca.sqlcode == SQLNOTFOUND)
                    break;
                else if (sqlca.sqlcode < 0) {
                    fprintf( stderr,
                             "Error FETCHing CURSOR cols2: Err=%d, ISAM=%d\n",
                             sqlca.sqlcode,
                             sqlca.sqlerrd[1] );
                }
                if (ncols2 <= i) {
                    columns2[i] = (char *)malloc( IDBYTES );
                    if (!columns2[i]) {
                        fprintf( stderr,
                                 "Cannot allocate memory for column constraint "
                                 "data(4).  Errno=%d.\n",
                                 errno );
                        rexit(99);
                    }
                    ncols2++;
                }
                strcpy( columns2[i], syscolumns.colname );
            }
            EXEC SQL CLOSE cols;
        
            for (i=0; i<16 && sysindexes.part[i]; i++) {
                if (i)
                    fprintf( outfile, ",\n" );
                fprintf( outfile,
                         "\t%s",
                         columns2[abs(sysindexes.part[i]) - 1] );
            }
            fprintf( outfile, "\n)" );
        } else {
            fprintf( outfile, "\n) REFERENCES %s", reftable );
        }
        if (delrule[0] == 'C')
            fprintf( outfile, " ON DELETE CASCADE" );

        if (sscanf( sysconstraints.constrname, "r%d_%d", &a, &b ) == 2)
	    if (LongKeyIndices && stripowner)
		fprintf( outfile, " CONSTRAINT %s_uk%.3d", systables.tabname, fknum++ );
	    else if (LongKeyIndices)
		fprintf( outfile,
			 " CONSTRAINT \"%s\".%s_uk%.3d",
			 UseDatabaseOwner ? dbowner : (UseTableOwner ? systables.owner : (UseSetOwner ? owner_override : sysconstraints.owner)),
			 systables.tabname,
			 fknum++ );
	    else if (KeepGenNames)
		if (stripowner)
		    fprintf( outfile, " CONSTRAINT %s", sysconstraints.constrname );
		else
		    fprintf( outfile,
			     " CONSTRAINT \"%s\".%s",
			     UseDatabaseOwner ? dbowner : (UseTableOwner ? systables.owner : (UseSetOwner ? owner_override : sysconstraints.owner)),
			     sysconstraints.constrname );
	    else
		fprintf( outfile, " " );
        else if (stripowner)
            fprintf( outfile, " CONSTRAINT %s", sysconstraints.constrname );
	else
            fprintf( outfile,
		     " CONSTRAINT \"%s\".%s",
		     UseDatabaseOwner ? dbowner : (UseTableOwner ? systables.owner : (UseSetOwner ? owner_override : sysconstraints.owner)),
		     sysconstraints.constrname );
        if (do_filtering) {
            print_objectstate( outfile, "D", state_psc2 ); /* If filtering, create disabled, then filter. */
	} else if (NoValidateFK) {
	    print_objectstate( outfile, "V", state_psc2 );
        } else {
            print_objectstate( outfile, state_psc, state_psc2 );
        }
	fprintf( outfile, ";\n\n" );

        if (do_filtering) {
	    if (stripowner)
		fprintf( outfile,
			 "SET CONSTRAINTS %s FILTERING WITHOUT ERROR;\n\n",
			 sysconstraints.constrname );
	    else
		fprintf( outfile,
			 "SET CONSTRAINTS \"%s\".%s FILTERING WITHOUT ERROR;\n\n",
			 UseDatabaseOwner ? dbowner : (UseTableOwner ? systables.owner : (UseSetOwner ? owner_override : sysconstraints.owner)),
			 sysconstraints.constrname );
            fprintf( outfile,
                     "SELECT '%s', COUNT(*) as num_violations FROM %s_vio;\n",
                     systables.tabname,
                     systables.tabname );
        }
    }
    EXEC SQL CLOSE :alien;
}

void print_objectstate( FILE *outfile, char *state, char *idxstate )
{
    switch (state[0]) {
      case 'D':
	  fprintf( outfile, " DISABLED" );
	  break;
      case 'F':
	  fprintf( outfile, " FILTERING WITHOUT ERRORS" );
	  break;
      case 'G':
	  fprintf( outfile, " FILTERING WITH ERRORS" );
	  break;
      case 'E':
	  break;
    }
    if (idxstate[0] == 'D')
	fprintf( outfile, " INDEX DISABLED" );
}

void print_references( FILE *outfile, int tflag ) 
{
    int i, a, b;
    int uknum = 1, fknum = 1;

    EXEC SQL BEGIN DECLARE SECTION;
    string reftable[129], delrule[2], state[2], state2[2];
    long primconstr, ptabid, rtabid;
    EXEC SQL END DECLARE SECTION;


    EXEC SQL DECLARE c_refs CURSOR FOR
        SELECT st.tabname, st.owner, rt.tabname, sr.primary, sr.ptabid, 
	sr.delrule, sc.constrid, sc.constrname, sc.constrtype, sc.owner,
        si.idxname, si.tabid, si.part1, si.part2, si.part3, 
        si.part4, si.part5, si.part6, si.part7, si.part8, 
        si.part9, si.part10, si.part11, si.part12, si.part13, 
        si.part14, si.part15, si.part16, rc.tabid, os.state, os2.state
        FROM "informix".systables st, "informix".sysconstraints sc,
        "informix".sysindexes si, "informix".sysreferences sr,
        "informix".systables rt, "informix".sysconstraints rc,
        "informix".sysobjstate os, "informix".sysobjstate os2
        WHERE st.tabid = sc.tabid
        AND rt.tabid = sr.ptabid
        AND sc.constrid = sr.constrid
        AND sc.tabid = si.tabid
        AND sc.idxname = si.idxname
        AND sc.constrtype = 'R'
        AND rt.tabname MATCHES :tablename
        AND sr.primary = rc.constrid
        AND os.tabid = st.tabid AND os.name = sc.constrname AND os.objtype = 'C'
        AND os2.tabid = si.tabid AND os2.name = si.idxname AND os2.objtype = 'I'
        ORDER BY si.tabid, constrid;
    if (sqlca.sqlcode < 0) {
        fprintf( stderr,
                 "Error DECLARing CURSOR c_refs: Err=%d, ISAM=%d\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1] );
        rexit( 21 );
    }

    /* Foreign Key Constraints. */
    EXEC SQL OPEN c_refs;
    if (sqlca.sqlcode < 0) {
        fprintf( stderr,
                 "Error OPENing CURSOR c_refs: Err=%d, ISAM=%d\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1] );
        rexit( 22 );
    }
    for (;;) {
        EXEC SQL FETCH c_refs
            INTO
            :systables.tabname,
	    :systables.owner,
            :reftable,
            :primconstr,
            :ptabid,
            :delrule,
            :sysconstraints.constrid,
            :sysconstraints.constrname,
            :sysconstraints.constrtype,
	    :sysconstraints.owner,
            :sysindexes.idxname,
            :sysindexes.tabid,
            :sysindexes.part[0],
            :sysindexes.part[1],
            :sysindexes.part[2],
            :sysindexes.part[3],
            :sysindexes.part[4],
            :sysindexes.part[5],
            :sysindexes.part[6],
            :sysindexes.part[7],
            :sysindexes.part[8],
            :sysindexes.part[9],
            :sysindexes.part[10],
            :sysindexes.part[11],
            :sysindexes.part[12],
            :sysindexes.part[13],
            :sysindexes.part[14],
            :sysindexes.part[15],
            :rtabid,
            :state,
	    :state2;
        if (sqlca.sqlcode == SQLNOTFOUND) {
            break;
        } else if (sqlca.sqlcode < 0) {
            fprintf( stderr,
                     "Error FETCHing CURSOR c_refs: Err=%d, ISAM=%d, Ref: %s, %s\n",
                     sqlca.sqlcode,
                     sqlca.sqlerrd[1],
		     sqlca.sqlerrm,
		     sqlca.sqlerrp );
            rexit( 21 );
        }


        EXEC SQL OPEN cols USING :sysindexes.tabid;
        if (sqlca.sqlcode < 0) {
            fprintf( stderr,
                     "Error OPENing CURSOR cols: Err=%d, ISAM=%d\n",
                     sqlca.sqlcode,
                     sqlca.sqlerrd[1] );
            rexit( 22 );
        }
        for (i=0;;i++) {
            EXEC SQL FETCH cols INTO :syscolumns.colno, :syscolumns.colname;
            if (sqlca.sqlcode == SQLNOTFOUND)
                break;
            else if (sqlca.sqlcode < 0) {
                fprintf( stderr,
                         "Error FETCHing CURSOR cols: Err=%d, ISAM=%d\n",
                         sqlca.sqlcode,
                         sqlca.sqlerrd[1] );
            }
            if (ncols2 <= i) {
                columns2[i] = (char *)malloc( IDBYTES );
                if (!columns2[i]) {
                    fprintf( stderr,
                             "Cannot allocate memory for column constraint "
                             "data(5).  Errno=%d.\n",
                             errno );
                    rexit(99);
                }
                ncols2++;
            }
            strcpy( columns2[i], syscolumns.colname );
        }
        EXEC SQL CLOSE cols;

	if (stripowner)
	  fprintf( outfile,
		   "ALTER TABLE %s ADD CONSTRAINT FOREIGN KEY (\n",
		   systables.tabname );
	else
	  fprintf( outfile,
		   "ALTER TABLE \"%s\".%s ADD CONSTRAINT FOREIGN KEY (\n",
		   UseDatabaseOwner ? dbowner : (UseTableOwner ? systables.owner : (UseSetOwner ? owner_override : sysconstraints.owner)),
		   systables.tabname );
        for (i=0; i<16 && sysindexes.part[i]; i++) {
            if (i)
                fprintf( outfile, ",\n" );
            fprintf( outfile,
                     "\t%s",
                     columns2[abs(sysindexes.part[i]) - 1] );
        }
        if (rtabid == ptabid) {
            fprintf( outfile, "\n) REFERENCES %s (\n", reftable );
            EXEC SQL
                SELECT part1, part2, part3, part4, part5, part6, part7, part8
                , part9, part10, part11, part12, part13, part14, part15, part16
                INTO
                :sysindexes.part[0],
                :sysindexes.part[1],
                :sysindexes.part[2],
                :sysindexes.part[3],
                :sysindexes.part[4],
                :sysindexes.part[5],
                :sysindexes.part[6],
                :sysindexes.part[7],
                :sysindexes.part[8],
                :sysindexes.part[9],
                :sysindexes.part[10],
                :sysindexes.part[11],
                :sysindexes.part[12],
                :sysindexes.part[13],
                :sysindexes.part[14],
                :sysindexes.part[15]
                FROM "informix".sysindexes si, "informix".sysconstraints sc
                WHERE si.tabid = :ptabid
                AND si.idxname = sc.idxname
                AND sc.constrid = :primconstr;

            EXEC SQL OPEN cols USING :ptabid;
            if (sqlca.sqlcode < 0) {
                fprintf( stderr,
                         "Error OPENing CURSOR cols2: Err=%d, ISAM=%d\n",
                         sqlca.sqlcode,
                         sqlca.sqlerrd[1] );
                rexit( 22 );
            }
            for (i=0;;i++) {
                EXEC SQL FETCH cols INTO :syscolumns.colno, :syscolumns.colname;
                if (sqlca.sqlcode == SQLNOTFOUND)
                    break;
                else if (sqlca.sqlcode < 0) {
                    fprintf( stderr,
                             "Error FETCHing CURSOR cols2: Err=%d, ISAM=%d\n",
                             sqlca.sqlcode,
                             sqlca.sqlerrd[1] );
                }
                if (ncols <= i) {
                    columns[i] = (char *)malloc( IDBYTES );
                    if (!columns[i]) {
                        fprintf( stderr,
                                 "Cannot allocate memory for column constraint "
                                 "data(6).  Errno=%d.\n",
                                 errno );
                        rexit(99);
                    }
                    ncols++;
                }
                strcpy( columns[i], syscolumns.colname );
            }
            EXEC SQL CLOSE cols;
        
            for (i=0; i<16 && sysindexes.part[i]; i++) {
                if (i)
                    fprintf( outfile, ",\n" );
                fprintf( outfile,
                         "\t%s",
                         columns[abs(sysindexes.part[i]) - 1] );
            }
            fprintf( outfile, "\n)" );
        } else {
            fprintf( outfile, "\n) REFERENCES %s", reftable );
        }
        if (delrule[0] == 'C')
            fprintf( outfile, " ON DELETE CASCADE" );

        if (sscanf( sysconstraints.constrname, "r%d_%d", &a, &b ) == 2)
	    if (LongKeyIndices && stripowner)
		fprintf( outfile, " CONSTRAINT %s_uk%.3d", systables.tabname, uknum );
	    else if (LongKeyIndices)
		fprintf( outfile,
			 " CONSTRAINT \"%s\".%s_uk%.3d",
			 UseDatabaseOwner ? dbowner : (UseTableOwner ? systables.owner : (UseSetOwner ? owner_override : sysconstraints.owner)),
			 systables.tabname,
			 uknum );
	    else
		fprintf( outfile, " " );
        else if (stripowner)
            fprintf( outfile, " CONSTRAINT %s", sysconstraints.constrname );
	else
            fprintf( outfile,
		     " CONSTRAINT \"%s\".%s",
		     UseDatabaseOwner ? dbowner : (UseTableOwner ? systables.owner : (UseSetOwner ? owner_override : sysconstraints.owner)),
		     sysconstraints.constrname );
        print_objectstate( outfile, state, state2 );
	fprintf( outfile, ";\n\n" );

    }
    EXEC SQL CLOSE c_refs;
}
