/* print_defaults.ec - print out default constraints.
*/
/*
   RCS Header:
   -----------
   $Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_defaults.ec,v 2.14 2011/05/13 14:47:02 art Exp $

   RCS Log:
   --------
   $Log: print_defaults.ec,v $
   Revision 2.14  2011/05/13 14:47:02  art
   Added a #if defined() to use of SQLINFXBIGINT to better support ESQL/C 3.10 users.
   Thanks to Kenneth Penza for pointing the oversight out.

   Revision 2.13  2010/12/17 15:41:19  art
   Support for BIGINT type defaults was missing.  Fixed.  Thanks to Dusam Simic for
   catching this one and reporting it.

   Revision 2.12  2009/09/09 15:47:31  art
   Fixed problem with zero length defaults for character types.
   Was generating a single space character as default, should
   be an empty string as these are different for variable
   char type columns.

   Revision 2.11  2009/02/11 18:30:34  art
   Fixed problem with LVARCHARS on AIX-64bit.

   Revision 2.10  2009/01/01 17:58:25  art
   Removed RCSWhat ident string which was BB specific.

   Revision 2.9  2008/12/30 18:05:02  art
   Fixed resolution display for default clauses if type INTERVAL.
   Copied corrected code from print_coltype.ec

   Revision 2.8  2004/02/23 17:43:46  kagel
   *** empty log message ***

   Revision 2.7  2004/02/23 17:37:50  kagel
   Added support for defaults in LVARCHAR and other UDT type columns treated as
   strings

   Revision 2.6  2001/11/01 13:39:09  kagel
   Fixed default output for a CHAR column default of a blank string.  Was
   outputting an empty string, now like dbschema it outputs a single space
   string.

   Revision 2.5  2001/10/30 14:11:46  kagel
   Added defaults for INT8 and BOOLEAN.

   Revision 2.4  2001/10/29 18:37:30  kagel
   Changes implementing support for collection types.

   Revision 2.3  2001/10/08 13:27:07  kagel
   Fixed SWITCH statement test to eliminate the NOT NULL flag.

   Revision 2.2  2001/09/10 18:33:18  kagel
   Object support added.

   Revision 2.1  1999/09/01 19:35:48  kagel
   Establishing current  source version id for initial working version of
   newly modularized source.  This is Feature Version 6.1 Source revision
   2.1.

   Revision 1.1  1999/09/01 19:35:48  kagel
   Initial revision


*/

static const char RCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_defaults.ec,v 2.14 2011/05/13 14:47:02 art Exp $";
static const char RCSCompile[] = "@(#)$cc: " __FILE__ " " __DATE__ " " __TIME__ " $";

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>

EXEC SQL INCLUDE sqlhdr;
EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;
EXEC SQL INCLUDE datetime;

EXEC SQL INCLUDE myschema.eh;

EXEC SQL INCLUDE prototypes.h;


void print_defaults( FILE *outfile,
		     syscolumns_t syscolumns,
		     sysdefaults_t sysdefaults )
{
    int fr1, fr2, pt1, pt2;
    int p1, p2;
    char *ptr, part1[30], part2[30];
    
    switch (sysdefaults.type[0]) {
    case 'N':
	fprintf( outfile, "\n\t    DEFAULT NULL" );
	break;
    case 'C':
	/* Don't forget to Check for 'not null'*/
	if ((syscolumns.coltype & 0xff) == SQLDTIME) { 
	    pt1 = (((syscolumns.collength / 16) % 16) <= 10) ?
		(((syscolumns.collength / 16) % 16) / 2) : 6;
	    strcpy( part1, dtparts[pt1] );
	    pt2 = ((syscolumns.collength % 16) <= 10) ?
		((syscolumns.collength % 16) / 2) : 6;
	    if (pt2 == 6) {
		fr2 = (syscolumns.collength % 16) - 10;
		sprintf( part2, "%s(%d)", dtparts[pt2], fr2 );
	    } else
		strcpy( part2, dtparts[pt2] );
	    fprintf( outfile,
		     "\n\t    DEFAULT CURRENT %s TO %s",
		     part1,
		     part2 );
	} else
	    fprintf( outfile, "\n\t    DEFAULT CURRENT" );
	break;
    case 'T':
	fprintf( outfile, "\n\t    DEFAULT TODAY" );
	break;
    case 'U':
	fprintf( outfile, "\n\t    DEFAULT USER" );
	break;
    case 'S':
	fprintf( outfile, "\n\t    DEFAULT SITENAME" );
	break;
    case 'L':
	switch ( syscolumns.coltype % 256 ) {
	case SQLCHAR:
	case SQLVCHAR:
	case SQLNCHAR:
	case SQLNVCHAR:
EXEC SQL IFNDEF NOINT8;
	case SQLUDTVAR:
EXEC SQL ENDIF;
	    if (strlen( sysdefaults.deflt ))
		fprintf( outfile, "\n\t    DEFAULT '%s'", sysdefaults.deflt );
	    else
		fprintf( outfile, "\n\t    DEFAULT ''" );
	    break;
	case SQLSMINT:
	case SQLINT:
	case SQLSMFLOAT:
	case SQLFLOAT:
	case SQLMONEY:
	case SQLDECIMAL:
EXEC SQL IFNDEF NOINT8;
#if defined (SQLINFXBIGINT)
	case SQLINFXBIGINT:
#endif
	case SQLINT8:
	case SQLBOOL:
EXEC SQL ENDIF;
	    ptr = strtok( sysdefaults.deflt, " " ); /* Skip BASE 6 representation */
	    ptr += strlen( ptr ) + 1;	/* Get string representation. */
	    fprintf( outfile, "\n\t    DEFAULT %s", ptr );
	    break;
	case SQLDATE:
	    ptr = strtok( sysdefaults.deflt, " " ); /* Skip BASE 6 representation */
	    ptr += strlen( ptr ) + 1;	/* Get string representation. */
	    fprintf( outfile, "\n\t    DEFAULT '%s'", ptr );
	    break;
	case SQLDTIME:
	    ptr = strtok( sysdefaults.deflt, " " ); /* Skip BASE 6 representation */
	    ptr += strlen( ptr ) + 1;	/* Get string representation. */
	    pt1 = (((syscolumns.collength / 16) % 16) <= 10) ?
		(((syscolumns.collength / 16) % 16) / 2) : 6;
	    strcpy( part1, dtparts[pt1] );
	    pt2 = ((syscolumns.collength % 16) <= 10) ?
		((syscolumns.collength % 16) / 2) : 6;
	    if (pt2 == 6) {
		fr2 = (syscolumns.collength % 16) - 10;
		sprintf( part2, "%s(%d)", dtparts[pt2], fr2 );
	    } else
		strcpy( part2, dtparts[pt2] );
	    fprintf( outfile,
		     "\n\t    DEFAULT DATETIME (%s) %s TO %s",
		     ptr,
		     part1,
		     part2 );
	    break;
	case SQLINTERVAL:
	    ptr = strtok( sysdefaults.deflt, " " ); /* Skip BASE 6 representation */
	    ptr += strlen( ptr ) + 1;	/* Get string representation. */
	    p1 = (syscolumns.collength / 16) % 16;
	    pt1 = (p1 <= 10) ? (p1 / 2) : 6;
	    fr2 = (syscolumns.collength % 16) - 10;
	    p2 = syscolumns.collength % 16;
	    pt2 = (p2 <= 10) ? (p2 / 2) : 6;
	    if (pt1 != 6) {
	        fr1 = (syscolumns.collength / 256) + p1 - p2;
		sprintf( part1, "%s(%d)", dtparts[pt1], fr1 );
	    } else
		sprintf( part1, "%s", dtparts[pt1] );
	    pt2 = ((syscolumns.collength % 16) <= 10) ?
		((syscolumns.collength % 16) / 2) : 6;
	    if (pt2 == 6) {
		sprintf( part2, "%s(%d)", dtparts[pt2], fr2 );
	    } else
		strcpy( part2, dtparts[pt2] );
	    fprintf( outfile,
		     "\n\t    DEFAULT INTERVAL (%s) %s TO %s",
		     ptr,
		     part1,
		     part2 );
	    break;
EXEC SQL IFNDEF NOINT8;
	case SQLUDTFIXED:
	    if (syscolumns.extended_id == 5) {
		/* Boolean. */
		fprintf( outfile,
			 "\n\t    DEFAULT '%s'",
			 sysdefaults.deflt );
		break;
	    }/* else fall through */
	/* ROW literal */
	case SQLROW:
	  fprintf( outfile, "\n\t    DEFAULT \'%s\'", sysdefaults.deflt );
	  break;	
EXEC SQL ENDIF;
	default:
	    fprintf( stderr,
		     "Invalid datatype (%d) for DEFAULT clause!\n",
		     syscolumns.coltype );
	    rexit(17);
	}
	break;
    default:
	fprintf( stderr, "Invalid DEFAULT TYPE: %s.\n", sysdefaults.type );
	rexit(20);
    }
}

