/* print_infrastructure.ec - Print out commands to duplicate storage and logging.
 */
/*
  RCS Header:
  -----------
  $Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_infrastructure.ec,v 1.26 2023/08/30 18:45:16 art Exp $

  RCS Log:
  --------
  $Log: print_infrastructure.ec,v $
  Revision 1.26  2023/08/30 18:45:16  art
  Improved handling of encrypted/unencrypted chunks.

  Revision 1.25  2023/08/30 18:22:18  art
  Added support for logged/unlogged default sbspaces.
  Added support for access time logging/nologging default sbspaces.

  Revision 1.24  2020/11/04 22:28:20  art
  Fixed plogdbspace API function call.

  Revision 1.23  2020/06/02 17:30:13  art
  Fixed empty call to dbaccess in "cmd" shell style output where there are not 
  storage pool entries.

  Revision 1.22  2019/12/04 16:56:04  art
  Fixed extendable output for --infrastructure.

  Revision 1.21  2019/09/10 20:17:08  art
  Grouped storagepool add API commands into a single dbaccess session
  for efficiency when --infrastructure=cmd

  Revision 1.20  2019/09/10 17:52:25  art
  Added support for storagepool entries.

  Revision 1.19  2019/03/27 17:11:26  art
  Changed infrastructuer printout to not generate code to set plog dbspaces extendable
  as they are extendable by default.

  Revision 1.18  2018/12/31 22:04:29  art
  Making changes to get myschema to compile/link under GCC v7+

  Revision 1.17  2018/01/04 17:02:07  art
  Updated copyright. Fixed handling of PLOG dbspaces.

  Revision 1.16  2017/12/26 16:42:19  art
  Fixed several printf's

  Revision 1.15  2017/05/03 21:08:19  art
  Fixed handling of temp smartblob spaces.

  Revision 1.14  2016/12/13 23:51:19  art
  Added support for unencrypted storage spaces when encrypted storage is enabled (12.10.xC8+).
  Added new option --no-check to output CHECK constraints with the NOVALIDATE option for
  migrating databases (effective only on a populated table created without CHECK constraints.

  Revision 1.13  2013/06/20 17:47:32  art
  Fixed problem with chunk ordering when a chunk was dropped and another added later.

  Revision 1.1  2013/04/29 19:33:18  art
  Initial revision

  Revision 1.12  2013/03/19 23:30:44  art
  Fixed chunks extendable feature.

  Revision 1.11  2013/03/14 04:11:28  art
  Added support for marking chunks extendable in the infrastructure script if
  the source server's chunk is so marked.  This is not supported by dbschema.

  Revision 1.10  2012/11/12 14:50:04  art
  Finally fixed chunk calculations in --infrastructure

  Revision 1.9  2012/08/26 20:42:26  art
  Fixed calculations for infrastructure printout.

  Revision 1.8  2012/08/26 18:31:03  art
  Typos.

  Revision 1.7  2012/08/26 18:16:53  art
  Fixed chunk size calculations.  Larger chunks were experiencing 32bit wordwrap.

  Revision 1.6  2011/12/27 15:35:30  art
  Added support for External Spaces to --infrastructure.  Handled SQL API version better than
  dbschema, shelling out to onspaces rather than just putting the suggestion in the output.

  Revision 1.5  2011/06/29 21:55:40  art
  Fixed --infrastructure output for logical log relocation.

  Revision 1.4  2011/06/29 13:42:49  art
  Added diagnostic.

  Revision 1.3  2011/06/28 15:55:24  art
  Fixed output to support versions earlier than 10.00 that only support a fixed page size.

  Revision 1.2  2011/06/28 15:26:06  art
  Fixed problem compiling print_infrastructure.ec with versions of ESQL/C earlier than 3.50.

  Revision 1.1  2011/06/21 21:50:44  art
  Initial revision


*/
static const char RCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_infrastructure.ec,v 1.26 2023/08/30 18:45:16 art Exp $";
static const char RCSCompile[] = "@(#)$cc: " __FILE__ " " __DATE__ " " __TIME__ " $";

#if defined( _AIX ) /* Needed as AIX defines a locale control structure loc_t */
#if !defined( _H_LOCALEDEF )
#define _H_LOCALEDEF
#endif
#endif

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <errno.h>

EXEC SQL INCLUDE sqlhdr;
EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;
EXEC SQL INCLUDE datetime;
EXEC SQL INCLUDE locator;

EXEC SQL INCLUDE myschema.eh;
EXEC SQL INCLUDE prototypes.h;

EXEC SQL BEGIN DECLARE SECTION;

typedef struct SYSDBSPACENAMES_S {
    int dbsnum;
    string name[129];
    string owner[33];
    int pagesize;
    int fchunk;
    int nchunks;
    double create_size;
    double extend_size;
    int is_mirrored;
    int is_blobspace;
    int is_sbspace;
    int is_temp;
    int is_encrypted;
    int flags;
    short chunk_chknum;
    short chunk_dbsnum;
    short chunk_nxchknum;
    int chunk_pagesize;
    int chunk_chksize;
    EXEC SQL ifdef SQLINFXBIGINT;
    bigint chunk_offset;
    EXEC SQL else;
    int chunk_offset;
    EXEC SQL endif;
    int chunk_nfree;
    int chunk_mdsize;
    int chunk_udsize;
    int chunk_udfree;
    int chunk_is_offline;
    int chunk_is_recovering;
    int chunk_is_blobchunk;
    int chunk_is_sbchunk;
    int chunk_is_inconsistent;
    int chunk_is_extendable;
    int chunk_flags;
    string chunk_fname[257];
    string chunk_mfname[257];
    EXEC SQL ifdef SQLINFXBIGINT;
    bigint chunk_moffset;
    EXEC SQL else;
    int chunk_moffset;
    EXEC SQL endif;
    int chunk_mis_offline;
    int chunk_mis_recovering;
    int chunk_mflags;
    int sbs_log_on;
    int sbs_log_off;
    int sbs_acctime_on;
    int sbs_acctime_off;
} sysdbspacenames_t;
typedef struct SYSDBSIND_S {
    short dbsnum;
    short name;
    short owner;
    short pagesize;
    short fchunk;
    short nchunks;
    short create_size;
    short extend_size;
    short is_mirrored;
    short is_blobspace;
    short is_sbspace;
    short is_temp;
    short flags;
    short chunk_chknum;
    short chunk_dbsnum;
    short chunk_nxchknum;
    short chunk_pagesize;
    short chunk_chksize;
    short chunk_offset;
    short chunk_nfree;
    short chunk_mdsize;
    short chunk_udsize;
    short chunk_udfree;
    short chunk_is_offline;
    short chunk_is_recovering;
    short chunk_is_blobchunk;
    short chunk_is_sbchunk;
    short chunk_is_inconsistent;
    short chunk_is_extendable;
    short chunk_flags;
    short chunk_fname;
    short chunk_mfname;
    short chunk_moffset;
    short chunk_mis_offline;
    short chunk_mis_recovering;
    short chunk_mflags;
} sysdbsind_t;
typedef struct {
  string path[256];
  bigint offset;
  bigint totalsz;
  bigint chunksz;
  int priority;
} storagepool_t;
EXEC SQL END DECLARE SECTION;

typedef enum { oncmd=0, viasql=1 }  out_type;

void print_infrastructure( FILE *outfile, int which_way )
{
    EXEC SQL BEGIN DECLARE SECTION;
    sysdbspacenames_t dbspace;
    sysdbsind_t indicators;
    int fext, next, partnum;
    string statement[4096];
    string pname[129];
    int psize, ppage;
    int lnumber, lchunk, lsize;
    int e_id;
    string e_name[129], e_location[257];
    storagepool_t pool;
    EXEC SQL END DECLARE SECTION;
    char cmmt[100];
    char dbtyp[100];
    char extstr[100];
    char pagstr[100];
    long long chk_offset, chk_size, mir_offset;
    int system_pagesize; 
    int lstdbs = 0, plog_done = 0, is_plog = 0, new_chunk_no = 0;
    char unencrypt_cmd[15];
    char unencrypt_sql[15];

    if (major_vers < 12 || (major_vers == 12 && minor_vers == 10 && lvl_num < 8)) {
	snprintf( unencrypt_cmd, sizeof unencrypt_cmd, "%s", " " );
	snprintf( unencrypt_sql, sizeof unencrypt_sql, "%s", " " );
    } else {
	snprintf( unencrypt_cmd, sizeof unencrypt_cmd, "%s", "-u " );
	snprintf( unencrypt_sql, sizeof unencrypt_sql, "%s", "unencrypted " );
    }
    system_pagesize = getSystemPageSize();
    snprintf( statement,
	      sizeof statement,
	      "SELECT d.dbsnum, d.name, d.owner, d.pagesize, d.fchunk, d.nchunks, "
	      "      %s"
	      "      bitval(d.flags ,'0x2' ) AS is_mirrored, "
	      "      bitval(d.flags ,'0x10' ) AS is_blobspace, "
	      "      bitval(d.flags ,'0x8000' ) AS is_sbspace, "
	      "      bitval(d.flags ,'0x2000' ) AS is_temp, "
	      "      bitval(d.flags ,'0x10000000' ) AS is_encrypted, "
	      "      d.flags, "
	      "      c.chknum, c.dbsnum, c.nxchknum, c.pagesize, c.chksize, c.offset, c.nfree, c.mdsize, "
	      "      c.udsize, c.udfree, c.is_offline, c.is_recovering, c.is_blobchunk, c.is_sbchunk, "
	      "      c.is_inconsistent, %s, c.flags, c.fname, c.mfname, c.moffset, "
	      "      c.mis_offline, c.mis_recovering, c.mflags, "
	      "      bitval(d.sbcflag, '0x00000001' ) as log_on, "
	      "      bitval(d.sbcflag, '0x00000002' ) as log_off, "
	      "      bitval(d.sbcflag, '0x00000008' ) as acctime_on, "
	      "      bitval(d.sbcflag, '0x00000010' ) as acctime_off "
	      "FROM sysdbstab d, syschunks c "
	      "WHERE d.dbsnum = c.dbsnum "
	      "ORDER BY d.dbsnum, c.chknum",
	      (major_vers > 11 || (major_vers == 11 && minor_vers >= 70)) ? 
	      "d.create_size, d.extend_size, " : "0.0, 0.0, ",
	      (major_vers > 11 || (major_vers == 11 && minor_vers >= 70)) ? 
	      "c.is_extendable" : "0" );


    EXEC SQL PREPARE p_dbspaces_s FROM :statement;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error preparing statement p_dbspaces for dbspaces. SQL: %d, ISAM: %d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	fprintf( stderr,
		 "Errm: %s.\n",
		 sqlca.sqlerrm );
	exit(1);
    }

    EXEC SQL DECLARE p_dbspaces CURSOR FOR p_dbspaces_s;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error declaring cursor p_dbspaces for dbspaces. SQL: %d, ISAM: %d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit(1);
    }
 
    EXEC SQL OPEN p_dbspaces;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error opening cursor p_dbspaces for dbspaces. SQL: %d, ISAM: %d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit(1);
    }

    for(;;) {
        dbtyp[0] = (char)0;
	cmmt[0] = (char)0;
	is_plog = 0;
	EXEC SQL FETCH p_dbspaces INTO :dbspace INDICATOR :indicators;
	
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Error fetching cursor for dbspaces. SQL: %d, ISAM: %d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit(1);
	} else if (sqlca.sqlcode == SQLNOTFOUND) {
	    break;
	}
	new_chunk_no++;
	if ((out_type)which_way == oncmd) {
	    /* Output onspaces commands. */
	tsblobkludge:
	    if (dbspace.dbsnum != lstdbs) {
		/* First chunk in the dbspace.  Issue create command. */ 
		extstr[0] = (char)0;
		strcpy( dbtyp, "-c " );
		if (dbspace.dbsnum == 1) {
		    /* Get the partition table extent for the dbspace. */
		    strcpy( cmmt, "# " );
		} else {
		    cmmt[0] = (char)0;
		}
		if (dbspace.is_blobspace) {
		    strcat( dbtyp, "-b" );
		    snprintf( pagstr, sizeof pagstr, "-g %lld ", (long long)(dbspace.pagesize / (system_pagesize * 1024)) );
		} else if (dbspace.is_sbspace && dbspace.is_temp) {
		    /* Temp smart blobspace */
		    strcat( dbtyp, "-S -t" );
		    pagstr[0] = (char)0;
		    if ((out_type)which_way != oncmd)
			strcpy( cmmt, "! " );
		} else if (dbspace.is_sbspace) {
		    /* Ordinary smart blobspace */
		    strcat( dbtyp, "-S" );
		    pagstr[0] = (char)0;
		} else if (dbspace.is_temp) {
		    strcat( dbtyp, "-t -d" );
		    if (major_vers >= 10)
			snprintf( pagstr, sizeof pagstr, "-k %lld ", (long long)(dbspace.pagesize / 1024) );
		    else
			pagstr[0] = (char)0;
		} else if ((dbspace.flags & 0x01000000) != 0) {
		    /* Physical log dbspace! */
		    strcat( dbtyp, "-P" );
		    pagstr[0] = (char)0; /* Page size for Physical Log dbspaces is always default size. */
		    is_plog = 1;
		    plog_done = 1;
		} else {
		    strcat( dbtyp, "-d" );
		    if (major_vers >= 10)
			snprintf( pagstr, sizeof pagstr, "-k %lld ", (long long)(dbspace.pagesize / 1024) );
		    else
			pagstr[0] = (char)0;
		}
		if (dbspace.is_blobspace == 0 && dbspace.is_sbspace == 0 &&
		    dbspace.is_temp == 0 &&
		    (dbspace.flags & 0x01000000) == 0)
		{
		    partnum = ((dbspace.dbsnum << 20) + 1);
		    EXEC SQL SELECT fextsiz, nextsiz
			INTO :fext, :next
			FROM sysptnhdr
			WHERE partnum = :partnum;
		    snprintf( extstr, sizeof extstr, "-ef %d -en %d ", fext * system_pagesize, next * system_pagesize );
		}
	    } else {
		extstr[0] = (char)0;
		strcpy( dbtyp, "-a " );
		pagstr[0] = (char)0;
	    }
	    if (dbspace.is_sbspace) {
		if (dbspace.chunk_mdsize == 0)
		    snprintf( extstr,
			      sizeof extstr,
			      "-U " );
		else
		    snprintf( extstr,
			      sizeof extstr,
			      "-Ms %lld ",
			      ((long long)dbspace.chunk_mdsize / (long long)system_pagesize) );
	    }
	    if (dbspace.is_blobspace) {
		chk_offset = (long long)dbspace.chunk_offset * (long long)system_pagesize;
		chk_size   = (long long)dbspace.chunk_chksize * (long long)system_pagesize;
		if (dbspace.is_mirrored && indicators.chunk_mfname >= 0)
		    mir_offset = (long long)dbspace.chunk_moffset * (long long)system_pagesize;
	    } else {
		chk_offset = (long long)dbspace.chunk_offset * (long long)system_pagesize;
		chk_size   = (long long)dbspace.chunk_chksize * (long long)system_pagesize;
		if (dbspace.is_mirrored && indicators.chunk_mfname >= 0)
		    mir_offset = (long long)dbspace.chunk_moffset * (long long)system_pagesize;
	    }
	    if (dbspace.is_mirrored && indicators.chunk_mfname >= 0)
		/* Chunk is mirrored. */
		fprintf( outfile,
			 "%sonspaces %s %s -p %s -o %lld -s %lld -m %s %lld %s%s%s\n",
			 cmmt,
			 dbtyp,
			 dbspace.name,
			 dbspace.chunk_fname,
			 chk_offset,
			 chk_size,
			 dbspace.chunk_mfname,
			 mir_offset,
			 pagstr,
			 extstr,
			 (dbspace.is_encrypted == 0) ? unencrypt_cmd : "" );
	    else
		fprintf( outfile,
			 "%sonspaces %s %s -p %s -o %lld -s %lld %s%s%s\n",
			 cmmt,
			 dbtyp,
			 dbspace.name,
			 dbspace.chunk_fname,
			 chk_offset,
			 chk_size,
			 pagstr,
			 extstr,
			 (dbspace.is_encrypted == 0) ? unencrypt_cmd : "" );
	    if (dbspace.is_sbspace && dbspace.sbs_log_on && dbspace.dbsnum != lstdbs) {
		fprintf( outfile,
			 "onspaces -ch %s -Df LOGGING=ON\n", dbspace.name );
	    } else if (dbspace.is_sbspace && dbspace.sbs_log_on && dbspace.dbsnum != lstdbs) {
		fprintf( outfile,
			 "onspaces -ch %s -Df LOGGING=OFF\n", dbspace.name );
	    }
	    if (dbspace.is_sbspace && dbspace.sbs_acctime_on && dbspace.dbsnum != lstdbs) {
		fprintf( outfile,
			 "onspaces -ch %s -Df ACCESSTIME=ON\n", dbspace.name );
	    } else if (dbspace.is_sbspace && dbspace.sbs_acctime_off && dbspace.dbsnum != lstdbs) {
		fprintf( outfile,
			 "onspaces -ch %s -Df ACCESSTIME=OFF\n", dbspace.name );
	    }
	    lstdbs = dbspace.dbsnum;
	} else {
	    /* Output SQL API commands. */
	    if (dbspace.dbsnum != lstdbs) {
		/* First chunk in the dbspace.  Issue create command. */ 
		if (dbspace.dbsnum == 1) {
		    /* Get the partition table extents for the dbspace. */
		    strcpy( cmmt, "-- " );
		} else {
		    cmmt[0] = (char)0;
		}
		if (dbspace.is_blobspace) {
		    snprintf( dbtyp, sizeof dbtyp, "create %sblobspace", (dbspace.is_encrypted == 0) ? unencrypt_sql : "" );
		    snprintf( pagstr, sizeof pagstr, ", '%lld'", (long long)(dbspace.pagesize / (system_pagesize * 1024)) );
		} else if (dbspace.is_sbspace && dbspace.is_temp) {
		    /* Temp smart blobspace. Can't be handled by an API command need to shell out to onspaces. */
		    fprintf( outfile, "-- SQL API command for temp smart blobspaces not implemented in IDS yet.\n-- Using a shell out to onspaces.\n" );
		    goto tsblobkludge;
		} else if (dbspace.is_sbspace) {
		    snprintf( dbtyp, sizeof dbtyp, "create %ssbspace", (dbspace.is_encrypted == 0) ? unencrypt_sql : "" );
		    pagstr[0] = (char)0;
		} else if  (dbspace.is_temp) {
		    snprintf( dbtyp, sizeof dbtyp, "create %stempdbspace", (dbspace.is_encrypted == 0) ? unencrypt_sql : "" );
		    snprintf( pagstr, sizeof pagstr, ", '%lld'", (long long)(dbspace.pagesize / 1024) );
		} else if ((dbspace.flags & 0x01000000) != 0) {
		    snprintf( dbtyp, sizeof dbtyp, "create %splogspace", (dbspace.is_encrypted == 0) ? unencrypt_sql : "" );
		    snprintf( pagstr, sizeof pagstr, ", '%lld'", (long long)(dbspace.pagesize / 1024) );
		    is_plog = 1;
		    plog_done = 1;
		} else {
		    snprintf( dbtyp, sizeof dbtyp, "create %sdbspace", (dbspace.is_encrypted == 0) ? unencrypt_sql : "" );
		    snprintf( pagstr, sizeof pagstr, ", '%lld'", (long long)(dbspace.pagesize / 1024) );
		}
		if (dbspace.is_blobspace == 0 && dbspace.is_sbspace == 0) {
		    partnum = ((dbspace.dbsnum << 20) + 1);
		    EXEC SQL SELECT fextsiz, nextsiz
			INTO :fext, :next
			FROM sysptnhdr
			WHERE partnum = :partnum;
		    snprintf( extstr, sizeof extstr, ", '%d', '%d' ", fext * system_pagesize, next * system_pagesize );
		} else
		    extstr[0] = (char)0;
	    } else {
		extstr[0] = (char)0;
		strcpy( dbtyp, "add chunk" );
		pagstr[0] = (char)0;
	    }

	    if (dbspace.is_blobspace) {
		chk_offset = (long long)dbspace.chunk_offset * (long long)system_pagesize;
		chk_size   = (long long)dbspace.chunk_chksize * (long long)system_pagesize;
		if (dbspace.is_mirrored && indicators.chunk_mfname >= 0)
		    mir_offset = (long long)dbspace.chunk_moffset * (long long)system_pagesize;
	    } else {
		chk_offset = (long long)dbspace.chunk_offset * (long long)system_pagesize;
		chk_size   = (long long)dbspace.chunk_chksize * (long long)system_pagesize;
		if (dbspace.is_mirrored && indicators.chunk_mfname >= 0)
		    mir_offset = (long long)dbspace.chunk_moffset * (long long)system_pagesize;
	    }
	    if (dbspace.fchunk == dbspace.chunk_chknum) {
		if ((dbspace.flags & 0x01000000) != 0) {
		  /* PLOG dbspace handler. */
		    if (dbspace.is_mirrored) {
			fprintf( outfile,
				 "%sEXECUTE FUNCTION TASK( '%s', '%s', '%s' '%lld', '%lld', '%s', '%lld' );\n",
				 cmmt,
				 dbtyp,
				 dbspace.name,
				 dbspace.chunk_fname,
				 chk_size,
				 chk_offset,
				 dbspace.chunk_mfname,
				 mir_offset );
		    } else {
			fprintf( outfile,
				 "%sEXECUTE FUNCTION TASK( '%s', '%s', '%s' '%lld', '%lld' );\n",
				 cmmt,
				 dbtyp,
				 dbspace.name,
				 dbspace.chunk_fname,
				 chk_size,
				 chk_offset );
		    }
		} else {
		    fprintf( outfile,
			     "%sEXECUTE FUNCTION TASK( '%s', '%s', '%s' '%lld', '%lld'%s%s );\n",
			     cmmt,
			     dbtyp,
			     dbspace.name,
			     dbspace.chunk_fname,
			     chk_size,
			     chk_offset,
			     pagstr,
			     extstr );
		    if (dbspace.is_mirrored && indicators.chunk_mfname >= 0) {
			/* Chunk is mirrored. */
			fprintf( outfile,
				 "EXECUTE FUNCTION TASK( 'add mirror', '%s', '%s', '%lld', '%s', '%lld' );\n",
				 dbspace.name,
				 dbspace.chunk_fname,
				 chk_offset,
				 dbspace.chunk_mfname,
				 mir_offset );
			if (dbspace.fchunk == dbspace.chunk_chknum) /* First chunk.  Enable mirroring. */
			    fprintf( outfile,
				     "EXECUTE FUNCTION TASK( 'start mirroring', '%s' );\n",
				     dbspace.name );
		    }
		}
	    } else {
		if (dbspace.is_mirrored && indicators.chunk_mfname >= 0) {
		    fprintf( outfile,
			     "EXECUTE FUNCTION TASK( '%s', '%s', '%s' '%lld', '%lld', '%s', '%lld'%s%s );\n",
			     dbtyp,
			     dbspace.name,
			     dbspace.chunk_fname,
			     chk_size,
			     chk_offset,
			     dbspace.chunk_mfname,
			     mir_offset,
			     pagstr,
			     extstr );
		} else {
		    fprintf( outfile,
			     "EXECUTE FUNCTION TASK( '%s', '%s', '%s' '%lld', '%lld'%s%s );\n",
			     dbtyp,
			     dbspace.name,
			     dbspace.chunk_fname,
			     chk_size,
			     chk_offset,
			     pagstr,
			     extstr );
		}
	    }

	    if (dbspace.is_sbspace && dbspace.sbs_log_on && dbspace.dbsnum != lstdbs) {
		fprintf( outfile,
			 "EXECUTE FUNCTION TASK( 'set sbspace logging on', '%s' );\n", dbspace.name );
	    } else if (dbspace.is_sbspace && dbspace.sbs_log_on && dbspace.dbsnum != lstdbs) {
		fprintf( outfile,
			 "EXECUTE FUNCTION TASK( 'set sbspace logging off', '%s' );\n", dbspace.name );
	    }
	    if (dbspace.is_sbspace && dbspace.sbs_acctime_on && dbspace.dbsnum != lstdbs) {
		fprintf( outfile,
			 "EXECUTE FUNCTION TASK( 'set sbspace accesstime on', '%s' );\n", dbspace.name );
	    } else if (dbspace.is_sbspace && dbspace.sbs_acctime_off && dbspace.dbsnum != lstdbs) {
		fprintf( outfile,
			 "EXECUTE FUNCTION TASK( 'set sbspace accesstime off', '%s' );\n", dbspace.name );
	    }
	    lstdbs = dbspace.dbsnum;
	}

	if (dbspace.chunk_is_extendable && is_plog == 0) {
	    if ((out_type)which_way != oncmd) {
		fprintf( outfile,
			 "EXECUTE FUNCTION TASK( 'modify chunk extendable', '%d' );\n",
			 new_chunk_no );
	    } else {
		fprintf( outfile,
			 "dbaccess sysadmin - <<EOF\n"
			 "    EXECUTE FUNCTION TASK( 'modify chunk extendable', '%d' );\n"
			 "EOF\n",
			 new_chunk_no );
	    }
	}
    }


    /* Output any EXTSpaces. */
    EXEC SQL DECLARE ext_curs CURSOR FOR
	SELECT id, name, locatio
	FROM sysextspaces
	ORDER BY id;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error declaring cursor ext_curs for External spaces. SQL: %d, ISAM: %d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit(1);
    }

    EXEC SQL OPEN ext_curs;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error opening cursor ext_curs for External spaces. SQL: %d, ISAM: %d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit(1);
    }
    for (;;) {
	EXEC SQL FETCH ext_curs INTO :e_id, :e_name, :e_location;
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Error fetching cursor for External dbspaces. SQL: %d, ISAM: %d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit(1);
	} else if (sqlca.sqlcode == SQLNOTFOUND) {
	    break;
	}
	if ((out_type)which_way == oncmd) {
	    fprintf( outfile,
		     "onspaces -c -x %s -l %s\n",
		     e_name,
		     e_location );
	} else {
	    fprintf( outfile,
		     "-- SQL API command for external spaces not implemented in IDS yet.  \n-- Using a shell out to onspaces.\n" );
	    fprintf( outfile,
		     "! onspaces -c -x %s -l %s\n",
		     e_name,
		     e_location );
	}
    }

    /* Process storagepool entries in sysadmin. */
    if (major_vers > 11 || (major_vers == 11 && minor_vers >= 70)) {
      EXEC SQL DECLARE strpool_c CURSOR FOR
      SELECT path, beg_offset, (end_offset - end_offset), chunk_size, priority
	FROM sysadmin:storagepool
	WHERE status = "Active"
	ORDER BY entry_id;
      if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Error declaring cursor for storagepool. SQL: %d, ISAM: %d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit(1);
      }
      EXEC SQL OPEN strpool_c;
      if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Error opening cursor for storagepool. SQL: %d, ISAM: %d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit(1);
      }
      for (;;) {
	int first = 1;

	EXEC SQL FETCH strpool_c INTO :pool;
	if (sqlca.sqlcode < 0) {
	  fprintf( stderr,
		   "Error fetching cursor for storagepool. SQL: %d, ISAM: %d.\n",
		   sqlca.sqlcode,
		   sqlca.sqlerrd[1] );
	  exit(1);
	} else if (sqlca.sqlcode == SQLNOTFOUND) {
	  if ((out_type)which_way == oncmd && first > 1)
	    fprintf( outfile,"EOF\n" );
	  break;
	}
	if ((out_type)which_way != oncmd) {
	  fprintf( outfile,
		   "EXECUTE FUNCTION TASK( 'storagepool add', '%s', '%ld', '%ld', '%ld', '%d' );\n",
		   pool.path,
		   pool.offset,
		   pool.totalsz,
		   pool.chunksz,
		   pool.priority);
	} else {
	  if (first == 1) {
	    fprintf( outfile,"dbaccess sysadmin - <<EOF\n" );
	    first++;
	  }
	  fprintf( outfile,
		   "    EXECUTE FUNCTION TASK( 'storagepool add', '%s', '%ld', '%ld', '%ld', '%d' );\n",
		   pool.path,
		   pool.offset,
		   pool.totalsz,
		   pool.chunksz,
		   pool.priority );
	}
      }
    }

    /* Process Physical Log if PLOG dbspace not used above. */
    if (plog_done == 0) {
	/* Move and resize physical log. */
	EXEC SQL SELECT d.name, p.pl_physize, d.pagesize
	    INTO :pname, :psize, :ppage
	    FROM sysplog p, syschunks c, sysdbspaces d
	    WHERE p.pl_chunk = c.chknum
	    AND c.dbsnum = d.dbsnum;
	if ((out_type)which_way == oncmd) {
	    fprintf( outfile,
		     "onparams -p -s %lld -d %s -y \n",
		     (long long)(psize * (ppage / 1024)),
		     pname );
	} else {
	    fprintf( outfile,
		     "EXECUTE FUNCTION TASK( 'alter plog', '%s', '%lld' );\n",
		     pname,
		     (long long)(psize * (ppage / 1024)));
	}
    }

    /* Move and modify the number and size of the logical logs. */
    if ((out_type)which_way == oncmd)
	fprintf( outfile,
		 "dbaccess sysmaster - <<EOF\n"
		 "    UNLOAD TO 'myschema.tmp.data' DELIMITER ' '\n"
		 "    SELECT number FROM syslogfil;\n"
		 "EOF\n" );
    else
	fprintf( outfile,
		 "SELECT l.number FROM syslogfil INTO TEMP myschema_tmp;\n" );
    
    EXEC SQL DECLARE llog_c CURSOR FOR
	SELECT l.number, d.name, l.size, d.pagesize
	FROM syslogfil l, syschunks c, sysdbspaces d
	WHERE l.chunk = c.chknum
	  AND c.dbsnum = d.dbsnum
	ORDER BY number;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error declaring cursor llog_c for logical logs. SQL: %d, ISAM: %d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit(1);
    }

    EXEC SQL OPEN llog_c;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error opening cursor llog_c. SQL: %d, ISAM: %d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit(1);
    }

    for (;;) {
	EXEC SQL FETCH llog_c INTO :lnumber, :pname, :lsize, :ppage;
	if (sqlca.sqlcode > 0)
	    break;
	else if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Error fetching cursor llog_c. SQL: %d, ISAM: %d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit(1);
	}
	if ((out_type)which_way == oncmd) {
	    fprintf( outfile,
		     "onparams -a -d %s -s %lld \n",
		     pname,
		     (long long)(lsize * (ppage / 1024)) );
	} else {
	    fprintf( outfile,
		     "EXECUTE FUNCTION TASK ('add log', '%s', '%lld' );\n",
		     pname,
		     (long long)(lsize * (ppage / 1024)) );
	}
	
    }
    if ((out_type)which_way == oncmd) {
	fprintf( outfile, "onmode -c\n" );
	fprintf( outfile,
		 "for lnum in $(cat myschema.tmp.data); do\n"
		 "    onmode -l; onmode -c;\n"
		 "    onparams -d -l $lnum -y \n"
		 "done\n" );
	fprintf( outfile, "rm myschema.tmp.data\n" );
    } else {
	fprintf( outfile, "EXECUTE FUNCTION TASK( 'checkpoint' );\n" );
	fprintf( outfile,
		 "SELECT TASK( 'onmode', 'l' ) FROM myschema_tmp;\n" );
	fprintf( outfile, "EXECUTE FUNCTION TASK( 'checkpoint' );\n" );
	fprintf( outfile,
		 "SELECT TASK( 'drop log', number ) FROM myschema_tmp;\n" );
	fprintf( outfile, "DROP TABLE myschema_tmp;\n" );
    }
}
