/* ***************************************************************************
 * print_opclass.ec - Implements the CREATE OP CLASS DDL statements for user
 *		      defined secondary access methods.
 * ***************************************************************************
 *
 */
/*
  RCS Header:
  -----------
  $Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_opclasses.ec,v 1.4 2019/03/15 14:48:19 art Exp $


  RCS Log:
  --------
  $Log: print_opclasses.ec,v $
  Revision 1.4  2019/03/15 14:48:19  art
  Added support for CREATE OR REPLACE routines with --replace-procs option (v14.10 feature).

  Revision 1.3  2011/05/31 14:35:02  art
  Fixed typo in the output of OP Class Strategy clauses.

  Revision 1.2  2010/11/08 20:40:31  art
  Added new --create_conditionally feature to optionally add " IF NOT EXISTS " clauses to c
  create object statements.  Feature supported by 11.70.

  Revision 1.1  2009/06/02 01:48:05  art
  Initial revision


*/

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <errno.h>

EXEC SQL INCLUDE sqlhdr;
EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;

EXEC SQL INCLUDE myschema.eh;

EXEC SQL INCLUDE prototypes.h;
 
#ifndef lint
static const char RCSIdent[] = "$Id: print_opclasses.ec,v 1.4 2019/03/15 14:48:19 art Exp $";
static const char RCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_opclasses.ec,v 1.4 2019/03/15 14:48:19 art Exp $";
static const char RCSCompile[] = "@(#)$cc: " __FILE__ " " __DATE__ " " __TIME__ " $";
#endif

EXEC SQL BEGIN DECLARE SECTION;

typedef struct SYSOPCLASSES_S {
    string opclassname[129];
    string owner[33];
    int amid;
    string am_name[129];
    string am_owner[33];
    int opclassid;  /* SERIAL */
    string ops[2049];  /* LVARCHAR */
    string support[2049];  /* LVARCHAR */
} sysopclasses_t;
EXEC SQL END DECLARE SECTION;

#if !defined( __GNUC__ )
#define inline 
#endif

static inline __attribute__((always_inline)) void clean_ops( char *in )
{
    char *s = in;

    while (*s) {
	if (*s == ';')
	    *s = ',';
	s++;
    }
    s--;
    if (*s == ',')
	*s = (char)0;
} 

void print_opclasses( FILE *outfile )
{
    EXEC SQL BEGIN DECLARE SECTION;
    sysopclasses_t sysopclass;
    EXEC SQL END DECLARE SECTION;
    int len;
    char copy[2049];
    char *s, *t;

    EXEC SQL DECLARE opclass_c CURSOR FOR
	SELECT opclassname, owner, amid, am_name, am_owner,
	       opclassid, ops, support
	FROM sysopclasses so, sysams sa
	WHERE so.amid = sa.am_id
	  AND opclassid > 2 /* Filter out system defined op classes */
	ORDER BY amid, owner, opclassid;

    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error declaring cursor opclass_c!  SQL: %d  ISAM: %d\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit( 1 );
    }

    EXEC SQL OPEN opclass_c;
    if (sqlca.sqlcode == SQLNOTFOUND)
	return;
    else if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error opening cursor opclass_c!  SQL: %d  ISAM: %d\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit( 1 );
    }

    for (;;) {
	EXEC SQL FETCH opclass_c INTO :sysopclass;
	if (sqlca.sqlcode == SQLNOTFOUND)
	    break;
	else if (sqlca.sqlcode != 0) {
	    fprintf( stderr,
		     "Error fetching cursor opclass_c!  SQL: %d  ISAM: %d\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit( 1 );
	}

	if (stripowner) {
	    fprintf( outfile,
		     "CREATE OPCLASS%s %s FOR %s\n",
		     create_conditionally ? " IF NOT EXISTS" : "",
		     sysopclass.opclassname,
		     sysopclass.am_name );
	} else {
	    fprintf( outfile,
		     "CREATE OPCLASS%s \"%s\".%s FOR \"%s\".%s\n",
		     create_conditionally ? " IF NOT EXISTS" : "",
		     sysopclass.owner,
		     sysopclass.opclassname,
		     sysopclass.am_owner,
		     sysopclass.am_name );
	}
	clean_ops( sysopclass.ops );
	fprintf( outfile,
		 "STRATEGIES ( %s )\n",
		 sysopclass.ops );
	clean_ops( sysopclass.support );
	fprintf( outfile,
		 "SUPPORT ( %s );\n\n",
		 sysopclass.support );
    }
    EXEC SQL CLOSE opclass_c;
    EXEC SQL FREE opclass_c;
	
}
