/* print_procedures.ec - Print out stored procedure body.
*/
/*
   RCS Header:
   -----------
   $Header: /home/art/GoogleDrive/Latest/myschema.d/RCS/print_procedures.ec,v 2.79 2023/11/15 15:51:45 art Exp $

   RCS Log:
   --------
   $Log: print_procedures.ec,v $
   Revision 2.79  2023/11/15 15:51:45  art
   Formatting again.

   Revision 2.78  2023/11/15 15:40:52  art
   formatting

   Revision 2.77  2023/11/15 15:35:38  art
   Previous broke OUT parameters. Fixed.

   Revision 2.76  2023/11/15 15:12:46  art
   Fixed routine parameter lists in GRANT and REVOKE statements that contain complex parameters.
   Thanks to Davorin Kremenjas for reporting the issue.

   Revision 2.75  2022/12/20 14:00:01  art
   Fixed issued with GRANT and REVOKE EXECUTE privs for routines 
   that have collection type arguments (LIST, SET, MULTISET, etc.)

   Revision 2.74  2022/11/17 13:15:34  art
   trace again.

   Revision 2.73  2022/11/17 13:13:44  art
   Removed trace.

   Revision 2.72  2022/11/06 18:58:55  art
   Fixed problem with REVOKE and GRANT statements for 
   procedures with IN, OUT, and INOUT parameters.

   Revision 2.70  2020/11/15 22:40:43  art
   Fixed an initialization bug in stored procedure handling. Thanks to Jerome Outhier
   for pointing it out and suggesting a fix!

   Revision 2.69  2020/05/10 17:39:21  art
   Fixed problem with support functions for Datablade supplied Secondary Access Methods
   were being written to the procedure file. However, the Access Method definitions 
   required those functions to already exist. Changed so that functions mentioned in 
   sysams are written to the main file along with the AMS definition. Thanks to Kenneth
   Penza for pointing out the issue.

   Revision 2.68  2020/05/08 16:38:35  art
   Fixed problem handling databases with multi-byte characters.

   Revision 2.67  2019/04/23 18:25:49  art
   Improved handling of comments embedded in the proc create lines.

   Revision 2.66  2019/04/09 20:03:37  art
   Improved stored procedure handling. Stripping out IF NOT EXISTS from the procedure
   source (apparently the engine saves that. New options --no-proc-cleaning to disable
   the scrubbing and rewriting of the CREATE PROCEDURE/FUNCTION statement line because
   that sometimes causes issues.

   Revision 2.65  2019/03/15 14:48:19  art
   Added support for CREATE OR REPLACE routines with --replace-procs option (v14.10 feature).

   Revision 2.64  2017/08/17 02:33:29  artka
   Summary: Fixed printf missing tag. Changed cmmt_copy() to static inline

   Revision 2.63  2015/11/24 17:09:46  art
   Fixed routine owner handling where there are "--" comments embedded in the parameter list.
   Fixed the column list included in UPDATE STATISTICS LOW when -u is passed.
   Fixed an issue with the DROP statement not printing ROW type parameters when --drop-proc is passed.
   Thanks to David Link and Snorri Bergmann for pointing these out.

   Revision 2.62  2014/11/04 15:33:04  art
   Fixed handling of very long procedure argument lists.

   Revision 2.61  2014/07/29 02:58:04  art
   Eliminated audit triggers and procedures if --no_audit_tabs is passed.

   Revision 2.60  2013/12/24 17:30:14  art
   Fixed some compile time warnings on HPUX.

   Revision 2.59  2013/03/19 23:12:54  art
   Fixed handling of attached/IN TABLE indexes. Fixed a problem when
   -O and --drop-proc are specified together causing the DROP routine
   statement to have no argument list. Added missing support for
   REFERENCES BYTE/TEXT arguments in DROP routine statements when
   --drop-proc is specified.

   Revision 2.58  2013/03/04 18:09:13  art
   Fixed syntax error in REVOKE EXECUTE when -O is enabled.

   Revision 2.57  2012/11/28 12:45:02  art
   Fixed uninitialized variables. Fixed possible memory overrun in print_procedures.ec.
   Fixed handling of BIGSERIAL initialization in releases prior to 11.50, again.

   Revision 2.56  2012/11/28 12:05:31  art
   Fixed problems in print_procedures.ec with uninitialized vars and possible string overrun.

   Revision 2.55  2011/07/22 19:05:10  art
   Fixed grantee quoting.

   Revision 2.54  2011/07/21 21:12:35  art
   Mostly fixed.  Cores sometimes of outfile is a redirect.

   Revision 2.53  2011/07/21 19:03:15  art
   Fixed GRANT/REVOKE typo.

   Revision 2.52  2011/07/21 16:29:22  art
   Fixed argument list processing for the REVOKE execute.

   Revision 2.51  2011/07/21 16:13:55  art
   Added REVOKE...AS "owner" support to the REVOKE execute...

   Revision 2.50  2011/07/21 16:04:54  art
   Fixed problems with embedded comments again from TALX.
   Imporved --drop-proc processing.
   Added REVOKE EXECUTE... FROM "PUBLIC"; to remove default priveleges from the routine added at creation.
   Fixed SET/LIST/MULTISET argument handling in DROP...

   Revision 2.49  2011/07/01 21:10:38  art
   fixes to make overloaded functions work with --drop-proc.

   Revision 2.48  2011/07/01 16:32:43  art
   Added new feature --drop-proc to optionally insert a drop procedure statement before each create.
   Thanks to Bruce Simms for suggesting and funding this one.

   Revision 2.47  2011/04/06 03:32:55  art
   Added new feature --no-blade-funcs/--no-blade-procs to eliminate stored
   routines associated with IBM Datablades from the output.

   Revision 2.46  2010/11/08 20:40:31  art
   Added new --create_conditionally feature to optionally add " IF NOT EXISTS " clauses to c
   create object statements.  Feature supported by 11.70.

   Revision 2.45  2010/03/17 14:49:40  art
   This addresses 90% of the odd procedures with embedded comments before and within the CREATE statement.

   Revision 2.44  2010/03/05 00:53:43  art
   More fixes.

   Revision 2.43  2010/03/04 18:18:41  art
   typo

   Revision 2.42  2010/03/04 18:10:28  art
   Fixed handling of "funny" comments AGAIN!

   Revision 2.41  2010/03/04 14:49:37  art
   typo

   Revision 2.40  2010/03/04 14:41:01  art
   A new option to ignore instrumentation routines added by AGS Server Studio
   needed to debug stored procedures in SS.

   Revision 2.39  2010/03/02 20:25:52  art
   Fixed again.

   Revision 2.38  2010/03/02 14:21:15  art
   Fixed problem handling procedures added to the DB by AGS Server Studio's
   SPL debugger.  They have a copyright comment between CREATE and the
   procedure name.

   Revision 2.37  2010/01/06 13:55:18  art
   Added new option --set-owner to override all object owners with a specified owner name.

   Revision 2.36  2009/12/10 17:58:06  art
   Fixed case of CREATE PROCEDURE arguments.

   Revision 2.35  2009/12/10 17:53:56  art
   Fixed problem with garbage in CREATE PROCEDURE lines.

   Revision 2.34  2009/11/23 18:47:05  art
   Added support for EXTERNAL tables (11.50xc6+).
   Improved 64bit compiler support.  Removed warnings.


   Revision 2.33  2009/10/02 06:24:56  art
   Added new option -g to place user privileges into a separate file.
   Improved trigger parser.
   Made indexes and constraint processing recognize and report object state (ie disabled,
   filtering, etc.).
   Improved --filtering option again.

   Revision 2.32  2009/09/13 19:58:57  art
   Mostly cosmetic.  Modified so the warning about not finding the CREATE in the first
   text record is only printed under verbose mode.

   Revision 2.31  2009/09/09 16:18:49  art
   Fixed a variable type.  Internal fix affecting old compilers only.

   Revision 2.30  2009/09/09 16:17:40  art
   *** empty log message ***

   Revision 2.29  2009/02/11 18:30:34  art
   Fixed problem with LVARCHARS on AIX-64bit.

   Revision 2.28  2009/01/01 17:58:25  art
   Removed RCSWhat ident string which was BB specific.

   Revision 2.27  2008/12/30 17:43:39  art
   Fixed missing blank dummy field in select from sysprocbody (thought I fixed that long ago!)

   Revision 2.26  2008/06/13 15:51:21  art
   Stripped trailing spaces from ALTER ROUTINE and added semi-colon.

   Revision 2.25  2008/06/13 15:23:38  art
   Removed carriage return characters.

   Revision 2.24  2008/06/13 14:25:52  art
   Added support for procedure ALTERs.

   Revision 2.23  2006/12/29 13:52:32  kagel
   Fixed output for DBA procedures/functions. Thanks to Robert Sosnowski for
    pointing this one out.

   Revision 2.22  2006/11/09 15:36:15  kagel
   Fixed mismatch between CREATE and final END in create procedure/function
   output.  Now does not replace 'procedure' with 'function' in the create
   line for UDRs that return something in 9.xx+.

   Revision 2.21  2005/08/19 13:03:07  kagel
   Oops.  Forgot to append the tail.

   Revision 2.20  2005/08/19 12:57:53  kagel
   typo

   Revision 2.19  2005/08/19 12:55:41  kagel
   Added stripowner (-O) processing to stored procedures.

   Revision 2.18  2004/10/25 23:57:03  kagel
   Fixed problem with not printing functions when -f <procspec> is specified.

   Revision 2.17  2004/10/21 14:34:21  kagel
   Added keyword PROCEDURE and argument lists to GRANT EXECUTE statements for
   9.xx procedures.

   Revision 2.16  2004/09/20 17:57:07  kagel
   Fixed new -F referring VIEWS feature to eliminate duplicated views when
   a view references multiple tables requested in -t argument.  Also fixed a
   bug in print_procedures that oddly only caused trouble in mode ANSI DBs.
   Made some formatting changes to the output to tweak the looks a bit, minor.

   Revision 2.15  2004/09/14 17:18:24  kagel
   Fixing ANSI database access.

   Revision 2.14  2003/10/29 17:41:19  kagel
   Typo

   Revision 2.13  2003/10/29 15:47:31  kagel
   Improved backward compatibility by IFDEFing use of lvarchar host vars.

   Revision 2.12  2003/09/30 21:10:29  kagel
   Added undocumented -o option to support mydbdiff.  Changed some default
   sort orders so the output will be more consistent when tables are altered.
   Increased the Features version to 6.25 to recognize the changes.

   Revision 2.11  2002/12/03 14:01:59  kagel
   Added quotes to usernames in grants in print_procedures().

   Revision 2.10  2001/09/28 16:53:44  kagel
   Fixed retro problem with 5/6/7/SE servers due to new code to 9.xx.

   Revision 2.9  2001/09/26 17:46:55  kagel
   Corrected 'GRANT EXECUTE' statements for FUNCTIONS to include the FUNCTION
   keyword and argument list types.

   Revision 2.8  2001/09/14 17:40:19  kagel
   Object support added.

   Revision 2.7  2001/05/11 17:43:25  kagel
   Added yet another mode that is used by system procedures and functions.

   Revision 2.6  2001/05/08 14:19:17  kagel
   Bug found by Ravi Krishna.

   Revision 2.5  2001/05/08 14:19:17  kagel
   Added mode filters to eliminate built-in 9.xx procedures and functions.

   Revision 2.4  2000/08/15 20:30:50  kagel
   Oops forgot The Thingamabob that does the job

   Revision 2.3  2000/08/15 20:24:58  kagel
   Fixed bug causing procedures with DOCUMENT clauses to lose their trailing
   semi-colon.  This is an Informix bug, not myschema.

   Revision 2.2  2000/08/03 21:03:30  kagel
   Fixed bug in procedure body text printout.  Was passing text as format
   string which could cause problems if the procedure text included any '%'
   characters.  Added "%s" format and now pass the text as third parameter.

# Revision 2.1  1999/09/29  16:42:04  kagel
# Establishing current  source version id for initial working version of
# newly modularized source.  This is Feature Version 6.1 Source revision
# 2.1.
#
   Revision 1.1  1999/09/01 19:30:43  kagel
   Initial revision


*/

static const char RCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest/myschema.d/RCS/print_procedures.ec,v 2.79 2023/11/15 15:51:45 art Exp $";
static const char RCSCompile[] = "@(#)$cc: " __FILE__ " " __DATE__ " " __TIME__ " $";

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <errno.h>

EXEC SQL INCLUDE sqlhdr;
EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;
EXEC SQL INCLUDE datetime;

EXEC SQL INCLUDE myschema.eh;

EXEC SQL INCLUDE prototypes.h;

#if !defined( __GNUC__ )
#define inline 
#endif

static inline __attribute__((always_inline)) void cmmt_copy( char *targ, char *src, int len );

extern int NoAuditTabs;

void print_procedures( FILE *outfile, FILE *outfileG, char *ptempl, int grantas, int funcs, int doams )
{
    EXEC SQL BEGIN DECLARE SECTION;
    sysprocedures_t sysprocedures;
    sysprocbody_t sysprocbody;
    sysprocauth_t sysprocauth;
    char *templ = ptempl;
    char stmt[2049];
    string c_pauth[20];
    string params[32001];
    EXEC SQL END DECLARE SECTION;
    int nosemi = 0;
    int procfound;
    char argstr[5121];
    char pre_comment[300];
    char emb_comment[300];
    char arglist[60000];
    char *tail;
    char *procwords[] = {
      "-bad header-",
      "PROCEDURE",
      "FUNCTION",
      "ROUTINE" 
    };


    if (funcs && !Universal)
	return;

    if (funcs) {
	EXEC SQL CREATE TEMP TABLE ams_procs( proc_id int );
	if (sqlca.sqlcode < 0) {
	  fprintf( stderr,
		   "Error creating ams_procs temp table.  "
		   "SQLCode=%d, ISAM=%d.\n",
		   sqlca.sqlcode,
		   sqlca.sqlerrd[1] );
	  rexit(88);
	}
	EXEC SQL insert into ams_procs SELECT am_create FROM sysams WHERE am_type = "S" ;
	EXEC SQL insert into ams_procs SELECT am_drop FROM sysams WHERE am_type = "S" ;
	EXEC SQL insert into ams_procs SELECT am_open FROM sysams WHERE am_type = "S" ;
	EXEC SQL insert into ams_procs SELECT am_close FROM sysams WHERE am_type = "S" ;
	EXEC SQL insert into ams_procs SELECT am_insert FROM sysams WHERE am_type = "S" ;
	EXEC SQL insert into ams_procs SELECT am_delete FROM sysams WHERE am_type = "S" ;
	EXEC SQL insert into ams_procs SELECT am_update FROM sysams WHERE am_type = "S" ;
	EXEC SQL insert into ams_procs SELECT am_stats FROM sysams WHERE am_type = "S" ;
	EXEC SQL insert into ams_procs SELECT am_scancost FROM sysams WHERE am_type = "S" ;
	EXEC SQL insert into ams_procs SELECT am_check FROM sysams WHERE am_type = "S" ;
	EXEC SQL insert into ams_procs SELECT am_beginscan FROM sysams WHERE am_type = "S" ;
	EXEC SQL insert into ams_procs SELECT am_endscan FROM sysams WHERE am_type = "S" ;
	EXEC SQL insert into ams_procs SELECT am_rescan FROM sysams WHERE am_type = "S" ;
	EXEC SQL insert into ams_procs SELECT am_getnext FROM sysams WHERE am_type = "S" ;
	EXEC SQL insert into ams_procs SELECT am_getbyid FROM sysams WHERE am_type = "S" ;
	EXEC SQL insert into ams_procs SELECT am_build FROM sysams WHERE am_type = "S" ;
	EXEC SQL insert into ams_procs SELECT am_init FROM sysams WHERE am_type = "S" ;
	EXEC SQL insert into ams_procs SELECT am_truncate FROM sysams WHERE am_type = "S" ;
	
	sprintf( stmt,
		 "SELECT  "
		 "       procname, "
		 "       owner, "
		 "       procid, "
		 "       mode, "
		 "       retsize, "
		 "       symsize, "
		 "       datasize, "
		 "       codesize, "
		 "       numargs, "
		 "       %s, "
		 "       %s "
		 "FROM \"informix\".sysprocedures as sp "
		 "WHERE procname MATCHES ? "
		 " %s "
		 " %s "
		 " %s "
		 " %s "
		 " %s "
		 " %s ",
		 (Universal) ? "isproc" : "'t'",
		 (Universal) ? "paramtypes::lvarchar(4096)" : "''::varchar",
		 (Universal) ? "  AND isproc = 'f' " : "",
		 (no_blade_funcs) ? " AND (externalname NOT MATCHES '*/extend/*' OR externalname IS NULL) " : "",
		 (NoAuditTabs) ? " AND procname NOT MATCHES '*_[diu]trgprc'" : "",
		 (doams) ? " AND procid in (SELECT DISTINCT proc_id FROM ams_procs where proc_id > 0) "
		         : " AND procid NOT in (SELECT DISTINCT proc_id FROM ams_procs where proc_id > 0) ",
		 (no_ags_procs) ? "  AND procname NOT MATCHES 'deb_*' AND procname NOT MATCHES 'ssd_pr*'  " : "",
		 (include_sysprocs == TRUE) ? " " : "  AND mode NOT IN ('P', 'd', 'r', 'o') ");
    } else {
	  sprintf( stmt,
		 "SELECT  "
		 "       procname, "
		 "       owner, "
		 "       procid, "
		 "       mode, "
		 "       retsize, "
		 "       symsize, "
		 "       datasize, "
		 "       codesize, "
		 "       numargs, "
		 "       %s, "
		 "       %s "
		 "FROM \"informix\".sysprocedures "
		 "WHERE procname MATCHES ? "
		 " %s "
		 " %s "
		 " %s "
		 " %s "
		 "  AND mode NOT IN ('P', 'p', 'd', 'r', 'o', 't') ",
		 (Universal) ? "isproc" : "'t'",
		 (Universal) ? "paramtypes" : "''",
		 (Universal) ? "  AND isproc = 't' " : "",
		 (no_blade_funcs) ? " AND (externalname NOT MATCHES '*/extend/*' OR externalname IS NULL) " : "",
		 (NoAuditTabs) ? " AND procname NOT MATCHES '*_[diu]trgprc'" : "",
		 (no_ags_procs) ? "  AND procname NOT MATCHES 'deb_*' " : "" );
    }
    if (doOrder == NameOrder)
	strcat( stmt, "ORDER BY procname;\n " );
    else
	strcat( stmt, "ORDER BY procid;\n " );

    EXEC SQL PREPARE proc_stmt FROM :stmt;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Cannot PREPARE statement for Stored Procedure header.  "
		 "SQLCode=%d, ISAM=%d, pos = %d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1],
		 sqlca.sqlerrd[4] );
	fprintf( stderr,
		 "%s",
		 stmt );
	rexit(88);
    }

    fprintf( outfile, "\n" );
    EXEC SQL DECLARE proc_curs CURSOR FOR proc_stmt;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Cannot declare cursors for Stored Procedures.  "
		 "SQLCode=%d, ISAM=%d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	rexit(88);
    }
    EXEC SQL PREPARE s_ptext FROM
	"SELECT *, ' ' FROM \"informix\".sysprocbody \
	 WHERE procid = ? AND datakey = 'T' \
	 ORDER BY 1,2,3";
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Cannot prepare statement for Stored Procedure text.  "
		 "SQLCode=%d, ISAM=%d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	rexit(88);
    }

    EXEC SQL DECLARE c_ptext CURSOR FOR s_ptext;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Cannot declare cursor for Stored Procedure text.  "
		 "SQLCode=%d, ISAM=%d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	rexit(88);
    }

    EXEC SQL PREPARE s_atext FROM 
	"SELECT *, ' ' FROM \"informix\".sysprocbody \
	 WHERE procid = ? AND datakey = 'A' \
	 ORDER BY 1,2,3";
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Cannot prepare statement for Stored Procedure ALTERs.  "
		 "SQLCode=%d, ISAM=%d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	rexit(88);
    }

    EXEC SQL DECLARE c_atext CURSOR FOR s_atext;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Cannot declare cursor for Stored Procedure ALTERs.  "
		 "SQLCode=%d, ISAM=%d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	rexit(88);
    }


    EXEC SQL OPEN proc_curs USING :templ;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Cannot open cursor for Stored Procedures.  "
		 "SQLCode=%d, ISAM=%d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	rexit(88);
    }

    EXEC SQL PREPARE s_pauth2 FROM
	"SELECT pa.* \
	     FROM \"informix\".sysprocauth pa	\
	     WHERE pa.procid = ?";
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Cannot prepare statement for Stored Procedure privileges.  "
		 "SQLCode=%d, ISAM=%d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	rexit(88);
    }

    strcpy( c_pauth, "c_pauth" );

    EXEC SQL DECLARE :c_pauth CURSOR FOR s_pauth2;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Cannot declare cursor for Stored Procedure privileges.  "
		 "SQLCode=%d, ISAM=%d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	rexit(88);
    }

    for (;;) {
	EXEC SQL FETCH proc_curs INTO :sysprocedures, :params;
        if (replace_owners) {
	    strcpy( sysprocedures.owner, owner_override );
	}
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Cannot FETCH proc_curs.  SQLCode=%d, ISAM=%d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    rexit(87);
	} else if (sqlca.sqlcode == SQLNOTFOUND) {
	    break;
	}
	/* Make an argument list for when it is needed. */
	if (funcs || Universal) {
	    char parts[5224], *apart, del[6];
	    int first = 1;

	    strcpy( argstr, "( " );
	    strcpy( parts, params );
	    apart = strtok( parts," ,)(" );
	    while (apart) {
		del[0] = params[ (apart+strlen(apart) - parts) ];
		if (params[ (apart+strlen(apart) - parts)] == ')' &&
		    params[ (apart+strlen(apart) + 1 - parts)] == ',') {
		    del[1] = ',';
		    del[2] = (char)0;
		} else
		    del[1] = (char)0;
		if (strcmp( apart, "in" ) == 0 ||
		    strcmp( apart, "out" ) == 0 ||
		    strcmp( apart, "inout" ) == 0)
		{
		    // NOOP -- ASK
		} else {
		    if ( strncmp( apart, "not", 3 ) != 0 &&
			 strncmp( apart, "null", 4 ) != 0 &&
			 strncmp( apart, "year", 4 ) != 0 &&
			 strncmp( apart, "month", 5 ) != 0 &&
			 strncmp( apart, "hour", 4 ) != 0 &&
			 strncmp( apart, "day", 3 ) != 0 &&
			 strncmp( apart, "minute", 6 ) != 0 &&
			 strncmp( apart, "second", 6 ) != 0 &&
			 strncmp( apart, "fraction", 8 ) != 0 &&
			 strncmp( apart, "to", 2 ) != 0 &&
			 first == 0)
		    first = 0;
		    strcat(argstr, apart);
		    strcat( argstr, del );
		    if (strcmp( del, " " ) != 0)
			strcat( argstr, " " );
		}
		apart = strtok( (char *)NULL," ,)(" );
	    }
	    strcat( argstr, ")" );
    } else
	    argstr[0] = (char)0;

	if ((strstr( sysprocedures.procname, "deb_" ) == sysprocedures.procname
	     || strstr( sysprocedures.procname, "ssd_pr" ) == sysprocedures.procname)
	    && !no_ags_procs)
	  fprintf( stderr,
		   "Warning: This database may contain AGS Server Studio SPL Debugger procedures \n"
		   "which should be removed before the generated schema can be used successfully to\n"
		   "create a new database!  Remove debugging from all procedures using Server \n"
		   "Studio or add --no-ags-procs.  The latter, however, will not remove debugging\n"
		   "instrumentation from your SPL routines that have been debugged!\n" );

	if (do_drop_proc) {
	    return_coltypes( sysprocedures.procid, arglist );
	    if (stripowner) {
		fprintf( outfile,
			 "DROP %s%s %s %s;\n",
			 sysprocedures.isproc[0] == 't' ? "PROCEDURE" : "FUNCTION",
			 create_conditionally ? " IF EXISTS" : "",
			 sysprocedures.procname,
			 arglist );
	    } else {
		fprintf( outfile,
			 "DROP %s%s \"%s\".%s %s;\n",
			 sysprocedures.isproc[0] == 't' ? "PROCEDURE" : "FUNCTION",
			 create_conditionally ? " IF EXISTS" : "",
			 sysprocedures.owner,
			 sysprocedures.procname,
			 arglist );
	    }
	}

	EXEC SQL OPEN c_ptext USING :sysprocedures.procid;
	if (sqlca.sqlcode < 0) {

	    fprintf( stderr,
		     "Cannot open cursor for Stored Procedure text.  "
		     "SQLCode=%d, ISAM=%d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    rexit(88);
	}
    
	/* Process the routine source text. */
	procfound = 0;
	for (;;) {
	    EXEC SQL FETCH c_ptext INTO :sysprocbody;
	    if (sqlca.sqlcode < 0) {
		fprintf( stderr,
			 "Cannot FETCH c_ptext.  SQLCode=%d, ISAM=%d.\n",
			 sqlca.sqlcode,
			 sqlca.sqlerrd[1] );
		rexit(87);
	    } else if (sqlca.sqlcode == SQLNOTFOUND) {
		break;
	    }
	    /* Force a trailing NULL after 'data' for fprintf(). Can't use STRING
	     * type for 'data' as it strips any trailing hard spaces. */
	    sysprocbody.fill = (char)0;

	    if (sysprocbody.seqno == 1) {
	        procfound = find_proc( sysprocbody.data, &tail, (char *)pre_comment, (char *)emb_comment );
		if ((procfound != 0 && tail != (char *)NULL) && !NoProcClean) {
		    if (stripowner) {
		        fprintf( outfile,
				 "%sCREATE %s%s%s%s %s %s %s",
				 pre_comment,
				 (Replace_Procs == TRUE ? "OR REPLACE " : ""),
				 (sysprocedures.mode[0] == 'D' ? "DBA " : ""),
				 procwords[procfound],
				 create_conditionally ? " IF NOT EXISTS" : "",
				 emb_comment,
				 sysprocedures.procname,
				 tail );
		    } else {
		        fprintf( outfile,
				 "%sCREATE %s%s%s%s %s \"%s\".%s %s",
				 pre_comment,
				 (Replace_Procs == TRUE ? "OR REPLACE " : ""),
				 (sysprocedures.mode[0] == 'D' ? "DBA " : ""),
				 procwords[procfound],
				 create_conditionally ? " IF NOT EXISTS" : "",
				 emb_comment,
				 sysprocedures.owner,
				 sysprocedures.procname,
				 tail );
		    }
		} else {
		    /* Hmm.  The 'create' and opening parenthesis should both be
		     * in the first record.  ASK: This can only happen when the
		     * CREATE is preceded by a large inline comment or a large
		     * comment is embedded within the CREATE... itself (AGS does
		     * the latter for the procs it creates for its SPL
		     * debugger).  Have to just print what's in sysprocbody
		     * without editing for owner.  Only warn if verbose is on.
		     */
		    if (verbose)
			fprintf( stderr,
				 "Complete SPL create not found in seqno 1!  Proc: %s.\n",
				 sysprocedures.procname );
		    fprintf( outfile, "%s", sysprocbody.data );
		}
	    } else {
		fprintf( outfile, "%s", sysprocbody.data );
	    }
	    fflush( outfile );
	    nosemi = find_semi( sysprocbody.data ); 
	}

	if (nosemi)
	    fprintf( outfile, ";\n\n" );
	else
	    fprintf( outfile, "\n\n" );
	fflush( outfile );

	EXEC SQL OPEN c_atext USING :sysprocedures.procid;
	if (sqlca.sqlcode < 0) {

	    fprintf( stderr,
		     "Cannot open cursor for Stored Procedure ALTERs.  "
		     "SQLCode=%d, ISAM=%d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    rexit(88);
	}

	/* Process ALTERs of the procedure's options.  Mostly for external functions. */
	for (;;) {
	    EXEC SQL FETCH c_atext INTO :sysprocbody;
	    if (sqlca.sqlcode < 0) {
		fprintf( stderr,
			 "Cannot FETCH c_atext.  SQLCode=%d, ISAM=%d.\n",
			 sqlca.sqlcode,
			 sqlca.sqlerrd[1] );
		rexit(87);
	    } else if (sqlca.sqlcode == SQLNOTFOUND) {
		break;
	    }
	    /* Force a trailing NULL after 'data' for fprintf(). */
	    sysprocbody.fill = (char)0;
	    stripstr( sysprocbody.data, 256 );

	    fprintf( outfile, "%s;\n", sysprocbody.data );
	}

	EXEC SQL OPEN :c_pauth USING :sysprocedures.procid;
	if (sqlca.sqlcode < 0) {

	    fprintf( stderr,
		     "Cannot open cursor for Stored Procedure privileges.  "
		     "SQLCode=%d, ISAM=%d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    rexit(88);
	}
	/* Revoke public priveleges here, they will be added back below if they exist. */

	if (!grantas)
	    if (!stripowner)
		fprintf( outfileG,
			 "REVOKE EXECUTE ON %s\"%s\".%s%s FROM \"public\";\n",
			 (sysprocedures.isproc[0] == 'f') ? "FUNCTION " : "PROCEDURE ",
			 sysprocedures.owner,
			 sysprocedures.procname,
			 argstr );
	    else
		fprintf( outfileG,
			 "REVOKE EXECUTE ON %s%s%s FROM \"public\";\n",
			 (sysprocedures.isproc[0] == 'f') ? "FUNCTION " : "PROCEDURE ",
			 sysprocedures.procname,
			 argstr );
	else
	    if (!stripowner)
		fprintf( outfileG,
			 "REVOKE EXECUTE ON %s\"%s\".%s%s FROM \"public\" AS \"%s\";\n",
			 (sysprocedures.isproc[0] == 'f') ? "FUNCTION " : "PROCEDURE ",
			 sysprocedures.owner,
			 sysprocedures.procname,
			 argstr,
			 sysprocedures.owner );
	    else
		fprintf( outfileG,
			 "REVOKE EXECUTE ON %s%s%s FROM \"public\" AS \"%s\";\n",
			 (sysprocedures.isproc[0] == 'f') ? "FUNCTION " : "PROCEDURE ",
			 sysprocedures.procname,
			 argstr,
			 sysprocedures.owner );

	for (;;) {
	    EXEC SQL FETCH :c_pauth INTO :sysprocauth;
	    if (replace_owners) {
	        strcpy( sysprocauth.grantor, owner_override );
	    }
	    if (sqlca.sqlcode < 0) {
		fprintf( stderr,
			 "Cannot FETCH :c_pauth.  SQLCode=%d, ISAM=%d.\n",
			 sqlca.sqlcode,
			 sqlca.sqlerrd[1] );
		rexit(87);
	    } else if (sqlca.sqlcode == SQLNOTFOUND) {
		break;
	    }

	    if (grantas && !stripowner)
		fprintf( outfileG,
			 "GRANT EXECUTE ON %s\"%s\".%s%s TO \"%s\" AS \"%s\";\n",
			 (sysprocedures.isproc[0] == 'f') ? "FUNCTION " : "PROCEDURE ",
			 sysprocedures.owner,
			 sysprocedures.procname,
			 argstr,
			 sysprocauth.grantee,
			 sysprocauth.grantor );
	    else if (grantas)
		fprintf( outfileG,
			 "GRANT EXECUTE ON %s%s%s TO \"%s\" AS \"%s\";\n",
			 (sysprocedures.isproc[0] == 'f') ? "FUNCTION " : "PROCEDURE ",
			 sysprocedures.procname,
			 argstr,
			 sysprocauth.grantee,
			 sysprocauth.grantor );
	    else if (!stripowner)
		fprintf( outfileG,
			 "GRANT EXECUTE ON %s\"%s\".%s%s TO \"%s\";\n",
			 (sysprocedures.isproc[0] == 'f') ? "FUNCTION " : "PROCEDURE ",
			 sysprocedures.owner,
			 sysprocedures.procname,
			 argstr,
			 sysprocauth.grantee );
	    else
		fprintf( outfileG,
			 "GRANT EXECUTE ON %s%s%s TO \"%s\";\n",
			 (sysprocedures.isproc[0] == 'f') ? "FUNCTION " : "PROCEDURE ",
			 sysprocedures.procname,
			 argstr,
			 sysprocauth.grantee );
	}

	fprintf( outfile, "\n" );
	fflush( outfile );
    }
    if (funcs) {
      EXEC SQL DROP TABLE ams_procs;
    }

}

static inline __attribute__((always_inline)) void cmmt_copy( char *targ, char *src, int len )
{
    strncpy( targ, src, len );
    targ[len] = (char)0;
}

/* Look for trailing semicolon */
int find_semi( char *string )
{
    char *str;

    str = string + strlen( string ) - 1;
    while (str >= string && (*str == ' ' || *str == '\n' || *str == '\r'))
	str--;
    if (*str == ';')
	return 0;
    else
	return 1;
}

int find_create( char *str, int *loc )
{
    int ret, ix, ix2, len;
    char *strloc, *cmmtloc, *cmmtend, *t;
    int clen;

    static char *temp;
    static size_t templen = 0;

    len = strlen( str );

    if (len > templen) {
	templen = len;
	temp = realloc( temp, (templen + 1) );
	if (temp == (char *)NULL) {
	    fprintf( stderr, "Out of memory in find_create(). Errno = %d\n", errno );
	    exit(1);
	}
    }

    /* Create an all UPPER case version of the string for searching. */
    for (ix=0, ix2=0; ix<len; ix++) {
	temp[ix2++] = toupper( str[ix] );
    }

    t = temp;
 strp_cmmt:

    /* Trying to strip out leading comments. */
    cmmtloc = cmmtend = (char *)NULL;
    if ((cmmtloc = strstr( t, "--" )) != (char *)NULL) {
      cmmtend = strstr( cmmtloc, "\n" );
      clen = 1;
    } else if ((cmmtloc = strstr( t, "{" )) != (char *)NULL) {
      cmmtend = strstr( cmmtloc, "}" );
      clen = 1;
    } else if ((cmmtloc = strstr( t, "/*" )) != (char *)NULL) {
      cmmtend = strstr( cmmtloc, "*/" );
      clen = 2;
    }
    //    if (cmmtloc != (char *)NULL && cmmtend == (char *)NULL) {
    //      /* Found comment start but not end. */
    //      cmmtend = cmmtloc + strlen( cmmtloc );
    //    }

    strloc = strstr( t, "CREATE PROCEDURE" );
    if (strloc == (char *)NULL) {
        strloc = strstr( t, "CREATE FUNCTION" );
	if (strloc == (char *)NULL) {
            strloc = strstr( t, "CREATE ROUTINE" );
	    if (strloc == (char *)NULL) {
	      strloc = strstr( t, "CREATE DBA PROCEDURE" );
	      if (strloc == (char *)NULL) {
		strloc = strstr( t, "CREATE DBA FUNCTION" );
		if (strloc == (char *)NULL) {
		  strloc = strstr( t, "CREATE DBA ROUTINE" );
		  if (strloc == (char *)NULL) {
		    return 0;
		  } else
		    ret = 3;
		} else
		  ret = 2;
	      } else 
		ret = 1;
            } else
	      ret = 3;
        } else
	  ret = 2;
    } else
      ret = 1;
    if (strloc != (char *)NULL) {
      if (cmmtloc != (char *)NULL && cmmtloc < strloc) {
	if (cmmtend == (char *)NULL) {
	  *loc = 0;
	  return 0;	  
	} else if (cmmtend > strloc) {
	  t = cmmtend + clen;
	  goto strp_cmmt;
	}
      }
      *loc = (int)((long long)strloc - (long long)t);
    } else {
      *loc = 0;
      return 0;
    }
    return ret;
}


/* Look for end of procedure. 
   Return:	0 - not found or non-standard
		1 - found 'procedure'
		2 - found 'function'
		3 - found 'routine'
  Also, parenloc returns the location of the opening parenthesis
  for the CREATE statement if found.
*/
int find_proc( char *str, char **parenloc, char *pre_comment, char *emb_comment )
{
    char temp[300], *strloc, *cmmtloc, *cmmtend, *pl;
    int len = strlen( str );
    int ix, ix2, comlen = 0, ret = 0, found = 0, in = 0;
    long loc;
    int pos2 = 0;
    char cmt = (char)0;
    char *cmmtstrt = (char *)0;
     
    emb_comment[0] = (char)0;
    pre_comment[0] = (char)0;

    /* Strip leading comments. */
    ix = 0;
    ret = find_create( &(str[ix]), &pos2);
    for (; ix<len; ix++) {
	if (!in && str[ix] == '{') {
	    found = -1;
	    in = 1;
	    if (ret && pos2 < ix) {
		/* Previously found CREATE started before this comment.  Eureka! */
	        found = ix;
	        cmmtstrt = &str[ix];
		cmmtend = strstr( cmmtstrt, "}" );
		if (cmmtend == (char *)NULL)
		    cmmtend = cmmtstrt + strlen( cmmtstrt );
		// ix = pos2;
		break;
	    }
	    ret = find_create( &(str[ix+1]), &pos2);
	} else if (!in && str[ix] == '-' && str[ix+1] == '-') {
	    found = -1;
	    in = 2;
	    if (ret && pos2 < ix) {
		/* Previously found CREATE started before this comment.  Eureka! */
	        found = ix;
	        cmmtstrt = &str[ix];
		cmmtend = strstr( cmmtstrt, "\n" );
		if (cmmtend == (char *)NULL)
		    cmmtend = cmmtstrt + strlen( cmmtstrt );
		// ix = pos2;
		break;
	    }
	    ret = find_create( &(str[ix+1]), &pos2);
	} else if (in == 1 && str[ix] == '}') {
	    /* End of brace comment. */
	    found = ++ix;
	    in = 0;
	    if (pos2 < ix) {
		/* CREATE was within this comment. Ignore it. */
		pos2 = 0;
		ret = 0;
	    }
	} else if (in == 2 && str[ix] == '\n') {
	    /* End of dash comment. */
	    found = ++ix;
	    in = 0;
	    if (pos2 < ix) {
		/* CREATE was within this comment. Ignore it. */
		pos2 = 0;
		ret = 0;
	    }
	}
    }

    if (found && cmmtstrt && cmmtstrt > &str[pos2]) {
        /* Comment found, but it starts after CREATE... */
        cmt = *cmmtstrt; /* Save lead comment character. */
    }

    if (found < 0) { /* No lead comment or last comment not complete. */
	pos2 = ix = 0;

	ret = find_create( &(str[ix]), &pos2 );
    } else if (ret == 0) { /* CREATE is after last comment in the block. */
	ix=found;
	ret = find_create( &(str[ix]), &pos2 );
    }

    /* Search of the opening parenthesis. */
    if (cmmtstrt && cmmtstrt < &str[pos2]) {
        /* Comment before CREATE. Look after comment. */
	cmmt_copy( pre_comment, str, ix );
	*parenloc = (pl = strstr( &(str[pos2]), "(" ));
    } else {
        /* No comments or comment follows CREATE. */
        *parenloc = (pl = strstr( &(str[pos2]), "(" ));
	if (*parenloc > cmmtstrt && *parenloc < cmmtend) {
	  /* Paren is within the comment. Search after. */
	  *parenloc = (pl = strstr( &(cmmtend[1]), "(" ));
	}
    }

    return ret;
}
