/* print_seindexes.ec - Print out indexes for a table for an SE engine.
*/
/*
   RCS Header:
   -----------
   $Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_seindexes.ec,v 2.13 2015/11/24 17:09:46 art Exp $

   RCS Log:
   --------
   $Log: print_seindexes.ec,v $
   Revision 2.13  2015/11/24 17:09:46  art
   Fixed routine owner handling where there are "--" comments embedded in the parameter list.
   Fixed the column list included in UPDATE STATISTICS LOW when -u is passed.
   Fixed an issue with the DROP statement not printing ROW type parameters when --drop-proc is passed.
   Thanks to David Link and Snorri Bergmann for pointing these out.

   Revision 2.12  2009/11/23 18:47:05  art
   Added support for EXTERNAL tables (11.50xc6+).
   Improved 64bit compiler support.  Removed warnings.

   Revision 2.11  2009/02/11 18:30:34  art
   Fixed problem with LVARCHARS on AIX-64bit.

   Revision 2.10  2009/01/01 17:58:25  art
   Removed RCSWhat ident string which was BB specific.

   Revision 2.9  2004/08/18 18:14:54  kagel
   Ported -k feature to SE schemas.

   Revision 2.8  2000/07/11 17:12:07  kagel
   Added support for FILLFACTOR using an environment variable and new -m
   flag to force using minimal initial extent when current extent is too
   large rather than just printing a warning.

   Revision 2.7  2000/04/14 16:50:01  kagel
   Fixed a bug which caused all indexes to be skipped (== s/b != in test).
   Fixed ORDER BY clause to a better order though the actual output should not
   change.

   Revision 2.6  2000/04/12 18:29:44  kagel
   Fixed problem affecting implicitely named constraint indexes which are
   referenced by more than one constraint.  This caused multiple indexes with
   different explicit names to be generated with identical keys.  Indexes are
   now output in the following order: any clustered index first followed by
   other indexes in order by indexname.

   Revision 2.5  1999/10/26 22:21:56  kagel
   Fixed minor problems.

   Revision 2.4  1999/10/26 21:44:08  kagel
   typo revisited.

   Revision 2.3  1999/10/26 21:41:55  kagel
   typo.

   Revision 2.2  1999/10/26 21:24:37  kagel
   Added new feature to duplicate current levels of UPDATE STATISTICS.  By
   default written to filename3, optionally to a separate file.
   Modified function to output the LOW stats statements for each index.

   Revision 2.1  1999/09/22 21:10:41  kagel
   Establishing current  source version id for initial working version of
   newly modularized source.  This is Feature Version 6.1 Source revision
   2.1.

   Revision 1.1  1999/09/22 21:10:41  kagel
   Initial revision


*/

static const char RCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_seindexes.ec,v 2.13 2015/11/24 17:09:46 art Exp $";
static const char RCSCompile[] = "@(#)$cc: " __FILE__ " " __DATE__ " " __TIME__ " $";

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>

EXEC SQL INCLUDE sqlhdr;
EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;
EXEC SQL INCLUDE datetime;

EXEC SQL INCLUDE myschema.eh;

EXEC SQL INCLUDE prototypes.h;


static int FillFactor = 0;
static int prepped = 0;

void print_seindexes( FILE *outfile, FILE *distfile )
{
    int i, j;
    char typetext[60], lastidx[IDBYTES];
    char *p;
    static char statement[2048];


    /* Determine user's preferred FILLFACTOR. */
    if (!FillFactor) {
	p = getenv( "FILLFACTOR" );
	if (p)
	    FillFactor = atol( p );
    }
	
    /* Select for index information. */
    if (!prepped) {
	strcpy( statement, "SELECT i.idxname, i.tabid, idxtype, clustered, " );
	strcat( statement, "      part1, part2, part3, part4, part5, part6, " );
	strcat( statement, "      part7, part8, " );
	strcat( statement, "      constrid, constrname, c.tabid, " );
	strcat( statement, "      constrtype, c.idxname" );
	strcat( statement, "  FROM \"informix\".sysindexes i, outer \"informix\".sysconstraints c " );
	strcat( statement, "  WHERE i.tabid = ? " );
	if (NoKeyIndices)
	    strcat( statement, "    AND i.idxname[1,1] != ' ' " );
	strcat( statement, "    AND i.tabid = c.tabid AND i.idxname = c.idxname" );
	strcat( statement, "  ORDER BY 2, 4 DESC, 1;" );
	EXEC SQL PREPARE s_seidx from :statement;
	if (sqlca.sqlcode) {
	    fprintf( stderr,
		     "Cannot PREPARE select statement for indexes, error: %d.\n",
		     sqlca.sqlcode );
	    fprintf( stderr, "Error at character %d.\nStatement is:\n",
		     sqlca.sqlerrd[4] );
	    fprintf( stderr, "%s", statement );
	    rexit(19);
	}
	EXEC SQL DECLARE seidx CURSOR FOR s_seidx;
	if (sqlca.sqlcode ) {
	    fprintf( stderr,
		     "Cannot declare idx cursor; Error = %d\n", sqlca.sqlcode );
	    rexit(20);
	}
	prepped = 1;
    }
    lastidx[0] = (char)0;
    EXEC SQL OPEN seidx USING :prev.tabid;
    if (sqlca.sqlcode) {
	fprintf( stderr,
		 "Error opening cursor idx; Error = %d\n",
		 sqlca.sqlcode );
	rexit(14);
    }
    memset( &sysindexes, 0, sizeof sysindexes );
    EXEC SQL
	FETCH seidx INTO
	:sysindexes.idxname,
	:sysindexes.tabid,
	:sysindexes.idxtype,
	:sysindexes.clustered,
	:sysindexes.part[0],
	:sysindexes.part[1],
	:sysindexes.part[2],
	:sysindexes.part[3],
	:sysindexes.part[4],
	:sysindexes.part[5],
	:sysindexes.part[6],
	:sysindexes.part[7],
	:sysconstraints.constrid indicator isconstraint,
	:sysconstraints.constrname,
	:sysconstraints.tabid,
	:sysconstraints.constrtype,
	:sysconstraints.idxname;

    while (!sqlca.sqlcode) {
	if (strcmp( sysindexes.idxname, lastidx ) != 0) {
	    strcpy( lastidx, sysindexes.idxname );
	    if (isconstraint == 0 && sysindexes.idxname[0] == ' ') {
		for (p = sysindexes.idxname; *p == ' '; p++)
		    *p = sysconstraints.constrtype[0];
		strtok( sysindexes.idxname, " " );
		fprintf( outfile,
			 "-- Index < %s> is a constraint index.  "
			 "Creating normal index: ",
			 p );
		fprintf( outfile, "%s.\n", sysindexes.idxname );
	    } else
		strtok( sysindexes.idxname, " " );
	    typetext[0] = ' ';
	    typetext[1] = (char)0;
	    if (sysindexes.idxtype[0] == 'U')
		strcat( typetext, "UNIQUE " );
	    if (sysindexes.clustered[0] == 'C')
		strcat( typetext, "CLUSTER " );
	    fprintf( outfile,
		     "CREATE%sINDEX %s ON %s (\n",
		     typetext,
		     sysindexes.idxname,
		     prev.tabname );
	    for (i=0; i<8 && sysindexes.part[i]; i++) {
		j = abs(sysindexes.part[i]) - 1;
		if (i)
		    fprintf( outfile, ",\n" );
		fprintf( outfile,
			 "\t%s %s",
			 columns[j],
			 sysindexes.part[i] > 0 ? "ASC" : "DESC" );
	    }
	    if (FillFactor)
		fprintf( outfile, "\n) FILLFACTOR %d;\n", FillFactor );
	    else
		fprintf( outfile, "\n);\n" );
	    if (doDistrib) { 
		fprintf( distfile,
			 "UPDATE STATISTICS%s FOR TABLE %s ( \n",
			 conn5 ? "" : " LOW",
			 prev.tabname );
		for (i=0; i<8 && sysindexes.part[i]; i++) {
		    j = abs(sysindexes.part[i]) - 1;
		    if (i)
			fprintf( distfile, ",\n" );
		    fprintf( distfile,
			     "\t%s",
			     columns[j] );
		}
		fprintf( distfile, " );\n" );
	    }
	    fprintf( outfile, "\n" );
	}
	memset( &sysindexes, 0, sizeof sysindexes );
	EXEC SQL
	    FETCH seidx INTO
	    :sysindexes.idxname,
	    :sysindexes.tabid,
	    :sysindexes.idxtype,
	    :sysindexes.clustered,
	    :sysindexes.part[0],
	    :sysindexes.part[1],
	    :sysindexes.part[2],
	    :sysindexes.part[3],
	    :sysindexes.part[4],
	    :sysindexes.part[5],
	    :sysindexes.part[6],
	    :sysindexes.part[7],
	    :sysconstraints.constrid indicator isconstraint,
	    :sysconstraints.constrname,
	    :sysconstraints.tabid,
	    :sysconstraints.constrtype,
	    :sysconstraints.idxname;
    }
    if (sqlca.sqlcode && sqlca.sqlcode != SQLNOTFOUND) {
	fprintf( stderr,
		 "Error fetching index information. Error = %d\n",
		 sqlca.sqlcode );
	rexit(14);
    }
    
    EXEC SQL CLOSE idx;
}
