/* print_sequences.ec - Print CREATE SEQUENCE statements to a schema.
 * 
*/
/*
   RCS Header:
   -----------
   $Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_sequences.ec,v 1.23 2022/06/08 16:24:11 art Exp $

   RCS Log:
   --------
   $Log: print_sequences.ec,v $
   Revision 1.23  2022/06/08 16:24:11  art
   Removed "START AT ..." from sequences that have a minimum value greater than 1.
   The initial value is set in an ALTER statement following the CREATE SEQUENCE anyway.

   Revision 1.22  2019/11/18 22:57:47  art
   Fixed REVOKE ... AS <owner> to conform to -O -A & --set-owner.
   Added AS <owner> to REVOKEs for sequences.

   Revision 1.21  2018/06/22 16:30:05  art
   New options: --use-database-owner, --use-table-owner, --constraint-file
   Fixed some problems with the partial schema file output.
   Expanded "owner"s to all objects.

   Revision 1.20  2015/07/07 15:10:00  art
   Added ORDER BY clauses for more consistent output for sequences, types, and casts.

   Revision 1.19  2012/11/12 16:39:03  art
   Added support for NOMINVALUE & NOMAXVALUE.

   Revision 1.18  2012/11/12 15:53:16  art
   *** empty log message ***

   Revision 1.17  2012/11/12 15:45:00  art
   *** empty log message ***

   Revision 1.16  2012/11/12 15:39:26  art
   Fixed handling of NOMINVALUE & NOMAXVALUE in sequence definition.

   Revision 1.15  2011/04/15 13:08:57  art
   Fixed bug in print_sequences no properly handling sequences with no cache.
   Thanks to Cesar Martins again.

   Revision 1.14  2010/11/08 20:40:31  art
   Added new --create_conditionally feature to optionally add " IF NOT EXISTS " clauses to c
   create object statements.  Feature supported by 11.70.

   Revision 1.13  2010/08/25 13:52:14  art
   Fixed several compile time warnings.  Thanks to John Adamski for identifying them.

   Revision 1.12  2010/05/04 23:23:32  art
   Fixed sequence output for IDS 10.00 and 11.10 (oops).
   Implemented new option --sequence, --sequence=matches-spec to print only sequence
   definitions.

   Revision 1.11  2010/05/04 20:12:34  art
   Back to using nextval as dbschema does.  Using the serial8 in sysactptnhdr isn't reliable.
   It's not always updated.

   Revision 1.10  2010/03/24 16:10:13  art
   Modified sequence code to acquire the current/last sequence value from
   sysactptnhdr in sysmaster if the database has logging (no remote join
   to sysmaster from a non-logged database).

   Revision 1.9  2009/10/18 17:23:16  art
   Fixed collection type handling.  Thanks to Bruce Simms for pointing out this one.
   Fixed summary comment to include table owner if not suppressed ala dbschema.
   Removed unneeded function prototypes.

   Revision 1.8  2009/10/06 19:16:40  art
   Fixed privilege output from print_synonyms and print_sequences to conform to
   -g option.

   Revision 1.7  2009/06/01 13:52:11  art
   Cleanup.

   Revision 1.6  2009/02/11 18:30:34  art
   Fixed problem with LVARCHARS on AIX-64bit.

   Revision 1.5  2005/05/03 12:57:58  kagel
   Currval only works within a session after nextval has already been called.
   Changed back to using nextval.

   Revision 1.4  2005/03/08 15:56:05  kagel
   Changed <sequence>.nextval to <sequence>.currval.
   This will keep myschema from incrementing the sequence value.

   Revision 1.3  2004/09/14 20:23:11  kagel
   Extended referencing view code to recurse to catch indirect references.
   through other views.
   Fixed missing view owners if owner names not being stripped.

   Revision 1.2  2004/09/14 17:17:06  kagel
   Fixing ANSI database access.

   Revision 1.1  2004/07/07 18:06:57  kagel
   Initial revision


*/

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <errno.h>

EXEC SQL INCLUDE sqlhdr;
EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;

EXEC SQL INCLUDE myschema.eh;

EXEC SQL INCLUDE prototypes.h;
 
#ifndef lint
static const char RCSIdent[] = "$Id: print_sequences.ec,v 1.23 2022/06/08 16:24:11 art Exp $";
static const char RCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_sequences.ec,v 1.23 2022/06/08 16:24:11 art Exp $";
static const char RCSCompile[] = "@(#)$cc: " __FILE__ " " __DATE__ " " __TIME__ " $";
#endif

/* Extern data */
extern int grantas;

void stripstr( char *str, int len ) {
    char *p = (str + len - 1);
    while (*p == ' ' && p >= str) {
	*p = (char)0;
	p--;
    }
}

void print_sequences( FILE *outfile, FILE *outfileG, char *DbOwner )
{
    EXEC SQL IFNDEF NOINT8;
    EXEC SQL BEGIN DECLARE SECTION;
    struct syssequences_s {
	string tabname[IDBYTES];
	string owner[UNAMELEN];
	int seqid;
	int tabid;
	int8 start_val;
	int8 inc_val;
	int8 min_val;
	int8 max_val;
	short cycle;
	int8 restart_val;
	int cache;
	short order;
    } syssequence;
    struct seq_ind_s {
	short tabname;
	short owner;
	short seqid;
	short tabid;
	short start_val;
	short inc_val;
	short min_val;
	short max_val;
	short cycle;
	short restart_val;
	short cache;
	short order;
    } seq_ind;
    int8 restart_value;
    char restart_sql[500];
    char statement[1024];
    EXEC SQL END DECLARE SECTION;
    char i8str1[101], i8str2[101], i8str3[101];
    char cachestr[101];
    long long min_val, max_val, inc_val;
    long long restart_val;

    snprintf( statement,
	      sizeof statement,
	      "SELECT st.tabname, st.owner, ss.seqid, ss.tabid, ss.start_val, "
	      "       ss.inc_val, ss.min_val, ss.max_val, ss.cycle, "
	      "       ss.restart_val, ss.cache, ss.order "
	      "FROM \"informix\".systables st, \"informix\".syssequences ss "
	      "WHERE ss.tabid = st.tabid "
	      "   AND st.tabtype = 'Q' "
	      "   AND st.tabname MATCHES ? %s;",
	      doOrder == NameOrder ? "ORDER BY 1" : "ORDER BY st.tabid" );
    EXEC SQL PREPARE seq_s FROM :statement;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error PREPARING statement for sequence definition. Error=%d, ISAM=%d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit(1);
    }

    EXEC SQL DECLARE seq_c CURSOR FOR seq_s;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error DECLARING cursor for sequence definition. Error=%d, ISAM=%d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit(1);
    }

    EXEC SQL OPEN seq_c USING :sequence_str;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error OPENING cursor for sequence definition. Error=%d, ISAM=%d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	exit(1);
    }
    for(;;) {
	EXEC SQL FETCH seq_c INTO :syssequence INDICATOR seq_ind;
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Error FETCHING cursor for sequence definition. Error=%d, ISAM=%d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit(1);
	} else if (sqlca.sqlcode == SQLNOTFOUND)
	    break;
	if (!stripowner)
	    fprintf( outfile,
		     "CREATE SEQUENCE%s \"%s\".%s \n",
		     create_conditionally ? " IF NOT EXISTS" : "",
		     DbOwner ? DbOwner : syssequence.owner,
		     syssequence.tabname );
	else
	    fprintf( outfile,
		     "CREATE SEQUENCE%s %s \n",
		     create_conditionally ? " IF NOT EXISTS" : "",
		     syssequence.tabname );
	ifx_int8toasc( &syssequence.inc_val, i8str1, 100 );
	i8str1[100] = (char)0;
	stripstr( i8str1, 100 );
	inc_val = atoll( i8str1 );
	if (seq_ind.start_val == (short)-1) {
	    memset( i8str2, 0, sizeof i8str2 );
	} else {	    
	    ifx_int8toasc( &syssequence.start_val, i8str2, 100 );
	}
	i8str2[100] = (char)0;
	stripstr( i8str2, 100 );
	if (seq_ind.start_val == (short)-1) {
	    fprintf( outfile,
		     "\tINCREMENT BY %s \n",
		     i8str1 );
	} else {
	    fprintf( outfile,
		     "\tINCREMENT BY %s START WITH %s\n",
		     i8str1,
		     i8str2 );
	}
		 
	ifx_int8toasc( &syssequence.min_val, i8str3, 100 );
	i8str3[100] = (char)0;
	stripstr( i8str3, 100 );
	min_val = atoll( i8str3 );
        if (inc_val > 0LL && min_val == 1LL) {
	    fprintf( outfile, "\tNOMINVALUE " );
        } else if (inc_val < 0 && min_val == -9223372036854775807LL) {
	    fprintf( outfile, "\tNOMINVALUE " );
	} else {
	    fprintf( outfile, "\tMINVALUE %s ", i8str3 );
	}
	ifx_int8toasc( &syssequence.max_val, i8str2, 100 );
	i8str2[100] = (char)0;
	stripstr( i8str2, 100 );
	max_val = atoll( i8str2 );
        if (inc_val > 0LL && max_val == 9223372036854775807LL) {
	    fprintf( outfile, "NOMAXVALUE \n" );
        } else if (inc_val < 0 && max_val == -1LL) {
	    fprintf( outfile, "NOMAXVALUE \n" );
	} else {
	    fprintf( outfile, "MAXVALUE %s\n", i8str2 );
	}
	if (syssequence.cache)
	    snprintf( cachestr,
		      sizeof cachestr,
		      "CACHE %d",
		      syssequence.cache );
	else
	    snprintf( cachestr,
		      sizeof cachestr,
		      "NOCACHE" );
	fprintf( outfile,
		 "\t%s %s %s;\n",
		 ((syssequence.cycle == 1) ? "CYCLE" : "NOCYCLE"),
		 cachestr,
		 ((syssequence.order == 1) ? "ORDER" : "NOORDER") );

	if (!stripowner)
	    snprintf( restart_sql,
		      sizeof restart_sql,
		      "SELECT \"%s\".%s.nextval FROM \"informix\".systables "
		      "WHERE tabid = 1;",
		      syssequence.owner,
		      syssequence.tabname );
	else
	    snprintf( restart_sql,
		      sizeof restart_sql,
		      "SELECT %s.nextval FROM \"informix\".systables "
		      "WHERE tabid = 1;",
		      syssequence.tabname );

	EXEC SQL PREPARE restart_s FROM :restart_sql;
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Error PREPARING stmt for sequence restart value. Error=%d, ISAM=%d, pos: %d, msg: %s.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1],
		     sqlca.sqlerrd[4],
		     sqlca.sqlerrm );
	    fprintf( stderr, "Txt: %s;\n", restart_sql );
	    exit(1);
	}
	EXEC SQL EXECUTE restart_s INTO :restart_value;
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Error EXECUTING stmt for sequence restart value. Error=%d, ISAM=%d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit(1);
	}
	ifx_int8toasc( &restart_value, i8str1, 100 );
	i8str1[100] = (char)0;
	stripstr( i8str1, 100 );
	restart_val = atoll( i8str1 );

	if (!stripowner) {
	    fprintf( outfile,
		     "ALTER SEQUENCE \"%s\".%s RESTART WITH %s;\n\n",
		     DbOwner ? DbOwner : syssequence.owner,
		     syssequence.tabname,
		     i8str1 );
	    if (grantas)
	      fprintf( outfileG,
		       "REVOKE ALL ON \"%s\".%s FROM public AS \"%s\";\n\n",
		       DbOwner ? DbOwner : syssequence.owner,
		       syssequence.tabname,
		       DbOwner ? DbOwner : syssequence.owner );
	    else
	      fprintf( outfileG,
		       "REVOKE ALL ON \"%s\".%s FROM public;\n\n",
		       DbOwner ? DbOwner : syssequence.owner,
		       syssequence.tabname );
	} else {
	    fprintf( outfile,
		     "ALTER SEQUENCE %s RESTART WITH %s;\n\n",
		     syssequence.tabname,
		     i8str1 );
	    fprintf( outfileG,
		     "REVOKE ALL ON %s FROM public;\n\n",
		     syssequence.tabname );
	}

	print_auths( syssequence.tabid,
		     syssequence.tabname,
		     outfileG,
		     DbOwner ? DbOwner : syssequence.owner,
		     1,
		     0,
		     grantas );
    }
    EXEC SQL CLOSE seq_c;
    EXEC SQL FREE seq_c;

    EXEC SQL ELSE;
    /* Sequences/int8 not supported by current ESQL/C compiler!  Nothing to do. */
    return;
    EXEC SQL ENDIF;
}
