/* print_synonyms.ec - Print definitions of synonyms.
*/
/*
   RCS Header:
   -----------
   $Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_synonyms.ec,v 2.14 2018/06/22 16:30:05 art Exp $

   RCS Log:
   --------
   $Log: print_synonyms.ec,v $
   Revision 2.14  2018/06/22 16:30:05  art
   New options: --use-database-owner, --use-table-owner, --constraint-file
   Fixed some problems with the partial schema file output.
   Expanded "owner"s to all objects.

   Revision 2.13  2010/11/08 20:40:31  art
   Added new --create_conditionally feature to optionally add " IF NOT EXISTS " clauses to c
   create object statements.  Feature supported by 11.70.

   Revision 2.12  2010/01/06 13:55:18  art
   Added new option --set-owner to override all object owners with a specified owner name.

   Revision 2.11  2009/10/06 19:16:40  art
   Fixed privilege output from print_synonyms and print_sequences to conform to
   -g option.

   Revision 2.10  2009/02/11 18:30:34  art
   Fixed problem with LVARCHARS on AIX-64bit.

   Revision 2.9  2009/01/01 17:58:25  art
   Removed RCSWhat ident string which was BB specific.

   Revision 2.8  2004/09/14 20:21:59  kagel
   Extended referencing view code to recurse to catch indirect references.
   through other views.
   Fixed missing view owners if owner names not being stripped.

   Revision 2.7  2004/07/07 18:44:38  kagel
   Added support for sequences.
   Changed PRIVATE synonym handling to not strip owner names.
    Fixed a bug that caused myschema to exit when encountering a sequence
    record in systables.
   Added print_sequences.ec to the makefiles.
   Fixed a typo in myschema.h.^C

   Revision 2.6  1999/12/30 16:07:27  kagel
   Changed explicit SELECT to a CURSOR declared once.

   Revision 2.5  1999/12/16 17:53:35  kagel
   Fixed typo.

   Revision 2.4  1999/12/16 16:56:15  kagel
   Added PUBLIC/PRIVATE keyword where appropriate.

   Revision 2.3  1999/12/01 21:37:01  kagel
   Fixed log message to 2.2.

   Revision 2.2  1999/12/01 20:51:14  kagel
   Problem with synonyms pointing to a foreign database on the same server
   when run against a 7.3x ESQL due to some change in the ESQL libraries.
   The change relates to an all spaces char column fetched into a type
   STRING field, 7.2x returned a single space while 7.3x (9.xx) returns a
   zero length string.  While I was in there I restructured the code so that
   local synonyms (pointing to a table in the same database) are no longer a
   separate code path.  Stuart Knight found this bug and supplied the 7.3 fix.

   Revision 2.1  1999/09/14 19:13:09  kagel
   Establishing current  source version id for initial working version of
   newly modularized source.  This is Feature Version 6.1 Source revision
   2.1.

   Revision 1.2  1999/09/14 19:13:09  kagel
   Completed module disconnection.


*/

static const char RCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_synonyms.ec,v 2.14 2018/06/22 16:30:05 art Exp $";
static const char RCSCompile[] = "@(#)$cc: " __FILE__ " " __DATE__ " " __TIME__ " $";

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>

EXEC SQL INCLUDE sqlhdr;
EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;
EXEC SQL INCLUDE datetime;

EXEC SQL INCLUDE myschema.eh;

EXEC SQL INCLUDE prototypes.h;

void print_synonyms( FILE *outfile, FILE *outfileG, int tabauths, int colauths, int grantas )
{
    EXEC SQL BEGIN DECLARE SECTION;
    short btabind;
    EXEC SQL END DECLARE SECTION;
    char tempstr[1024];
    static char prepped = 0;

    if (!prepped) {
	EXEC SQL DECLARE syncurs CURSOR FOR
	    SELECT s.servername, s.dbname, s.owner, s.tabname, t.tabname,
	    s.btabid
	    FROM "informix".syssyntable s, outer "informix".systables t
	    WHERE s.tabid = :systables.tabid
	    AND s.btabid = t.tabid;
	if (sqlca.sqlcode != 0) {
	    fprintf( stderr,
		     "Error declaring syssyntable query! Err: %d, ISAM: %d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    rexit(99);
	}
	prepped = 1;
    }
    EXEC SQL OPEN syncurs;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error opening syssyntable cursor! Err: %d, ISAM: %d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	rexit(99);
    }

    EXEC SQL FETCH syncurs
	INTO :syssyntable.servername,
	     :syssyntable.dbname,
	     :syssyntable.owner,
	     :syssyntable.tabname,
	     :syssyntable.ltabname,
	     :syssyntable.btabid indicator btabind;
    if (replace_owners) {
      strcpy( syssyntable.owner, owner_override );
    }
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Cannot fetch syssyntable record for "
		 "synonym: %s! Err: %d, "
		 "ISAM: %d.\n",
		 systables.tabname,
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	rexit(99);
    }
    EXEC SQL CLOSE syncurs;

    tempstr[0] = (char)0;
    if (syssyntable.dbname[0] != ' ' && syssyntable.dbname[0] != (char)0) {
	strcat( tempstr, syssyntable.dbname );
	if (syssyntable.servername[0] != ' '
	    && syssyntable.servername[0] != (char)0)
	{
	    strcat( tempstr, "@" );
	    strcat( tempstr, syssyntable.servername );
	}
	strcat( tempstr, ":" );
    }
    if (!stripowner
	&& syssyntable.owner[0] != (char)0
	&& syssyntable.owner[0] != ' ')
    {
	strcat( tempstr, "\"" );
	strcat( tempstr, syssyntable.owner );
	strcat( tempstr, "\"." );
    }
    if (syssyntable.tabname[0] != (char)0 && syssyntable.tabname[0] != ' ')
	strcat( tempstr, syssyntable.tabname );
    else
	strcat( tempstr, syssyntable.ltabname );

    /* Private synonyms should keep their owner. -- ASK */
    if (!stripowner || systables.tabtype[0] == 'P') {
	fprintf( outfile,
		 "CREATE %s SYNONYM%s \"%s\".%s FOR %s;\n",
		 (systables.tabtype[0] == 'P' ? "PRIVATE" : "PUBLIC"),
		 create_conditionally ? " IF NOT EXISTS" : "",
		 systables.owner,
		 systables.tabname,
		 tempstr );

	print_auths( systables.tabid,
		     systables.tabname,
		     outfileG,
		     systables.owner,
		     tabauths,
		     colauths,
		     grantas );
    } else {
	fprintf( outfile,
		 "CREATE %s SYNONYM%s %s FOR %s;\n",
		 (systables.tabtype[0] == 'P' ? "PRIVATE" : "PUBLIC"),
		 create_conditionally ? " IF NOT EXISTS" : "",
		 systables.tabname,
		 tempstr );

	print_auths( systables.tabid,
		     systables.tabname,
		     outfileG,
		     systables.owner,
		     tabauths,
		     colauths,
		     grantas );
    }
}

