/* print_triggers.ec - Print out trigger source DDL.
*/
/*
   RCS Header:
   -----------
   $Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_triggers.ec,v 2.18 2018/12/27 14:57:26 art Exp $

   RCS Log:
   --------
   $Log: print_triggers.ec,v $
   Revision 2.18  2018/12/27 14:57:26  art
   typo

   Revision 2.17  2018/12/27 14:49:26  art
   Fixed problem with trigger owners.

   Revision 2.16  2018/02/20 16:47:46  art
   Fixed RANGE/INTERVAL output to supress automatic partitions (not needed).
   Removed extent and lock mode clauses from virtual table definitions.
   Changed to generate INSTEAD of triggers before GRANT statements for 
   VIEWs to eliminate errors at create time from grants to multi-table
   views.

   Revision 2.15  2014/07/29 02:58:04  art
   Eliminated audit triggers and procedures if --no_audit_tabs is passed.

   Revision 2.14  2012/03/12 00:19:49  art
   Added prototype for local trigger state printout function.

   Revision 2.13  2012/03/12 00:16:51  art
   Added support for new option --index-online to add the ONLINE keyword to create index
   statements.  Added support for foreign keys created with the INDEX DISABLED option.

   Revision 2.12  2010/09/21 13:01:55  art
   Removed some debugging trace.

   Revision 2.11  2010/07/28 02:08:25  art
   Missed a trick.  Fixed 2.10.

   Revision 2.10  2010/07/28 01:38:54  art
   Fixed object state printout for triggers.  Was using the name and state of
   the next trigger rather than the current one.

   Revision 2.9  2010/01/06 13:55:18  art
   Added new option --set-owner to override all object owners with a specified owner name.

   Revision 2.8  2009/10/02 06:24:56  art
   Added new option -g to place user privileges into a separate file.
   Improved trigger parser.
   Made indexes and constraint processing recognize and report object state (ie disabled,
   filtering, etc.).
   Improved --filtering option again.

   Revision 2.7  2009/02/11 18:30:34  art
   Fixed problem with LVARCHARS on AIX-64bit.

   Revision 2.6  2009/01/01 17:58:25  art
   Removed RCSWhat ident string which was BB specific.

   Revision 2.5  2004/09/14 17:16:18  kagel
   Fixing ANSI database access.

   Revision 2.4  2003/09/30 21:04:50  kagel
   Added undocumented -o option to support mydbdiff.  Changed some default
   sort orders so the output will be more consistent when tables are altered.
   Increased the Features version to 6.25 to recognize the changes.

   Revision 2.3  1999/12/01 21:11:49  kagel
   no change!

   Revision 2.2  1999/12/01 21:11:49  kagel
   Problem with trigger lines that contained a trailing space.  This space was
   significant but was being stripped because the data column was being fetched
   into a type STRING field.  Changed field definition to type DATA to retain
   the trailing spaces.  Also boundary bug in detecting when bigtxt had to be
   expanded if the new text would make bigtxt EXACTLY full.  Stuart Knight
   found these bugs also and supplied the fix to the latter problem.

   Revision 2.1  1999/09/01 19:27:02  kagel
   Establishing current  source version id for initial working version of
   newly modularized source.  This is Feature Version 6.1 Source revision
   2.1.

   Revision 1.1  1999/09/01 19:27:02  kagel
   Initial revision


*/

static const char RCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_triggers.ec,v 2.18 2018/12/27 14:57:26 art Exp $";
static const char RCSCompile[] = "@(#)$cc: " __FILE__ " " __DATE__ " " __TIME__ " $";

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>

EXEC SQL INCLUDE sqlhdr;
EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;
EXEC SQL INCLUDE datetime;

EXEC SQL INCLUDE myschema.eh;
EXEC SQL INCLUDE prototypes.h;

void print_triggerstate( FILE *outfile, char *state );

extern int NoAuditTabs;

void print_triggers( FILE *outfile, char *in_tabtype, int in_tabid )
{
    EXEC SQL BEGIN DECLARE SECTION;
    string tabname[IDBYTES], trigname[IDBYTES], last_trigname[IDBYTES], state[2], prevstate[2];
    string owner[IDBYTES], last_owner[UNAMELEN];
    systrigbody_t systrigbody;
    string trgmatch[135];
    string query[4097];
    string tabtype[2];
    int tabid;
    EXEC SQL END DECLARE SECTION;
    long ltrigid = -1, len;

    strcpy( tabtype, in_tabtype );
    tabid = in_tabid;
    snprintf( query,
	      4096,
	      "SELECT tabname, trigname, stb.*, st.owner, os.state "
	      "FROM \"informix\".systables s, \"informix\".systriggers st, \"informix\".systrigbody stb, "
	      "     \"informix\".sysobjstate os "
	      "WHERE s.tabid = st.tabid "
	      "  AND s.tabname MATCHES ? "
	      "  AND st.trigid = stb.trigid "
	      "  AND datakey in ( 'D', 'A' ) "
	      "  AND os.tabid = s.tabid AND os.name = st.trigname AND os.objtype = 'T' "
	      "%s "
	      "  AND s.tabtype = ? "
	      "%s "
	      "ORDER BY tabname, trigname, datakey desc, seqno; ",
	      NoAuditTabs ? "  AND trigname NOT MATCHES '*_[idu]tr' " : "",
	      tabid != 0 ? "  AND s.tabid = ? " : "  AND s.tabid != ? " );
    EXEC SQL PREPARE s_trigs FROM :query;
    if (sqlca.sqlcode < 0) {
        fprintf( stderr,
                 "Cannot PREPARE s_trigs.  SQLCode=%d, ISAM=%d.\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1] );
        rexit(87);
    }

    EXEC SQL DECLARE trigs CURSOR FOR s_trigs;
    if (sqlca.sqlcode < 0) {
        fprintf( stderr,
                 "Cannot DECLARE trigs.  SQLCode=%d, ISAM=%d.\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1] );
        rexit(87);
    }

    EXEC SQL OPEN trigs USING :tablename, :tabtype, :tabid;
    if (sqlca.sqlcode < 0) {
        fprintf( stderr,
                 "Cannot OPEN trigs.  SQLCode=%d, ISAM=%d errm=%s.\n",
                 sqlca.sqlcode,
                 sqlca.sqlerrd[1],
		 sqlca.sqlerrm );
        rexit(87);
    }

    len = 0;
    if (!bigtxt || bigtxtsz < 4096) {
        bigtxtsz = 4096;
        bigtxt = (char *)realloc( bigtxt, bigtxtsz );
    }
    bigtxt[0] = (char)0;
    for (;;) {
        EXEC SQL FETCH trigs
	    INTO :tabname, :trigname, :systrigbody, :owner, :state;
        if (replace_owners) {
	    strcpy( owner, owner_override );
	}
        if (sqlca.sqlcode < 0) {
            fprintf( stderr,
                     "Cannot FETCH trigs.  SQLCode=%d, ISAM=%d.\n",
                     sqlca.sqlcode,
                     sqlca.sqlerrd[1] );
            rexit(87);
	} else if (sqlca.sqlcode == SQLNOTFOUND) {
            if (ltrigid != -1) { /* No more triggers.  Output last if any. */
                format_trigger( outfile, bigtxt, prevstate );
		if (tabid != 0)
		    fprintf( outfile, "\n" );
	    	if (prevstate[0] == 'D') {
	    	  if (stripowner)
	    	    fprintf( outfile,
	    		     "SET TRIGGERS %s ",
	    		     last_trigname );
	    	  else
	    	    fprintf( outfile,
	    		     "SET TRIGGERS \"%s\".%s ",
	    		     last_owner,
	    		     last_trigname );
	    	  print_triggerstate( outfile, prevstate );
                }
	    }
            break;
        }
        if (ltrigid != systrigbody.trigid && ltrigid != -1) {
            /* Ahh new trigger.  Output the last one first. */ 
            format_trigger( outfile, bigtxt, prevstate );
	    if (tabid != 0)
	        fprintf( outfile, "\n" );
            if (prevstate[0] == 'D') {
                if (stripowner)
                    fprintf( outfile,
                             "SET TRIGGERS %s ",
                             last_trigname );
                else
                    fprintf( outfile,
                             "SET TRIGGERS \"%s\".%s ",
                             last_owner,
                             last_trigname );
                print_triggerstate( outfile, prevstate );
            }
            len = 0;
            bigtxt[0] = (char)0;
        }
        if ((int)(len + strlen( systrigbody.data )) >= (int)bigtxtsz) {
            bigtxtsz += 1024;
            bigtxt = (char *)realloc( bigtxt, bigtxtsz );
        }
        strcat( bigtxt, systrigbody.data );
        len += strlen( systrigbody.data );
        ltrigid = systrigbody.trigid;
	strcpy( prevstate, state );
	strcpy( last_trigname, trigname );
	strcpy( last_owner, owner );
    }
}


void print_triggerstate( FILE *outfile, char *state )
{
    switch (state[0]) {
      case 'D':
	  fprintf( outfile, " DISABLED;\n\n" );
	  break;
      case 'F':
	  fprintf( outfile, " FILTERING WITHOUT ERRORS;\n\n" );
	  break;
      case 'G':
	  fprintf( outfile, " FILTERING WITH ERRORS;\n\n" );
	  break;
      case 'E':
	  fprintf( outfile, ";\n\n" );
	  break;
    }
}
