/* print_userauths.ec - Print user level permissions.
*/
/*
   RCS Header:
   -----------
   $Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_userauths.ec,v 2.32 2021/12/02 19:01:06 art Exp $

   RCS Log:
   --------
   $Log: print_userauths.ec,v $
   Revision 2.32  2021/12/02 19:01:06  art
   Fix some more.

   Revision 2.31  2021/12/02 18:32:29  art
   Fixed handling table names created with CAPs under DELIMIDENT.

   Revision 2.30  2021/01/12 16:42:40  art
   Added new options --no-setsessionauth and --comment-setsessionauth
   Needed because even the database owner cannot grant SETSESSIONAUTH to
   itself so some scripts will fail if these auths are included. They 
   should be processed manually. The -no suppresses and the --comment
   comments out these privs.

   Revision 2.29  2018/02/26 23:39:57  art
   Changed npused to double to reflect new catalog tables.

   Revision 2.28  2018/02/26 21:57:46  art
   Changed to not use the same global string as myschema.ec's a1 cursor.

   Revision 2.27  2014/08/31 01:23:13  art
   Removed privilege GRANT statements for views.  Not really needed.
   Removed printing of SETSESSIONAUTH privilege for the owner of the database.
   Not needed and it will fail.  If target db has a different owner, likely
   the intent is not to allow the privilege to the owner of the source
   schema anyway.

   Revision 2.26  2014/03/18 15:26:08  art
   Added support for GRANT SETSESSIONAUTH.  FYI: Not supported by dbschema!

   Revision 2.25  2012/01/27 18:35:45  art
   Fixed order by clause when -o is passed in.

   Revision 2.24  2011/03/03 14:35:22  art
   Formatting

   Revision 2.23  2010/12/06 16:28:04  art
    Fixed two warnings and an SQL bug.  Thanks to Oyvind Gjarstad

   Revision 2.22  2010/11/15 00:34:45  art
   Completed support for external directives.

   Revision 2.20  2010/03/17 11:48:59  art
   Added support for GRANT DEFAULT ROLE.

   Revision 2.19  2009/11/23 18:47:05  art
   Added support for EXTERNAL tables (11.50xc6+).
   Improved 64bit compiler support.  Removed warnings.

   Revision 2.18  2009/10/02 06:24:56  art
   Added new option -g to place user privileges into a separate file.
   Improved trigger parser.
   Made indexes and constraint processing recognize and report object state (ie disabled,
   filtering, etc.).
   Improved --filtering option again.

   Revision 2.17  2009/02/11 18:30:34  art
   Fixed problem with LVARCHARS on AIX-64bit.

   Revision 2.16  2009/01/01 17:58:25  art
   Removed RCSWhat ident string which was BB specific.

   Revision 2.15  2004/09/14 17:37:24  kagel
   Typo.

   Revision 2.14  2004/09/14 17:14:48  kagel
   Fixing ANSI database access.

   Revision 2.13  2004/08/04 17:42:59  kagel
   *** empty log message ***

   Revision 2.12  2004/08/04 17:35:39  kagel
   Fixed dropping temp table.

   Revision 2.11  2004/08/04 17:29:45  kagel
   Fixed code to not use CASE for 7.2x.

   Revision 2.10  2004/07/28 20:54:45  kagel
   Added support for UDR language authorizations.

   Revision 2.9  2003/10/29 16:01:10  kagel
   Improved backward compatility by IFDEFing use of SQL CASE clause.  Replaced
   with a temp table join to map codes to sort-order id.

   Revision 2.8  2003/09/30 20:53:20  kagel
   Added undocumented -o option to support mydbdiff.  Changed some default
   sort orders so the output will be more consistent when tables are altered.
   Increased the Features version to 6.25 to recognize the changes.

   Revision 2.7  2000/09/13 13:16:47  kagel
   Added quotes to user and role names to support upper and mixed case
   names.

   Revision 2.6  1999/11/03 20:00:46  kagel
   Trying to fix bad code generator output.

   Revision 2.5  1999/11/03 19:24:01  kagel
   Fixed new identifier sizes.

   Revision 2.4  1999/11/03 16:38:36  kagel
   Just moved some data definitions from myschema.ec to print_userauths.ec.

   Revision 2.3  1999/11/03 15:07:32  kagel
   Changed return type to void.

   Revision 2.2  1999/10/12 17:08:31  kagel
   Removed some testing traces left around.

   Revision 2.1  1999/09/28 21:06:30  kagel
   Establishing current  source version id for initial working version of
   newly modularized source.  This is Feature Version 6.1 Source revision
   2.1.

   Revision 1.1  1999/09/22 21:00:00  kagel
   Initial revision


*/

static const char RCSHeader[] = "@(#)$Header: /home/art/GoogleDrive/Latest\040Art's\040Stuff/myschema.d/RCS/print_userauths.ec,v 2.32 2021/12/02 19:01:06 art Exp $";
static const char RCSCompile[] = "@(#)$cc: " __FILE__ " " __DATE__ " " __TIME__ " $";

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>

EXEC SQL INCLUDE sqlhdr;
EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;
EXEC SQL INCLUDE datetime;

EXEC SQL INCLUDE myschema.eh;

EXEC SQL INCLUDE prototypes.h;

EXEC SQL BEGIN DECLARE SECTION;
string username[UNAMELEN], defrole[UNAMELEN];
int priv;
sysroleauth_t sysroleauth;
typedef struct SYSSURROGATEAUTH_S {
    string trusteduser[33];
    string surrogateuser[33];
} syssurrogateauth_t;
syssurrogateauth_t suauths;
EXEC SQL END DECLARE SECTION;

void print_userauths( FILE *outfile, FILE *outfileG, int SupressSetSessionAuth )
{
    EXEC SQL BEGIN DECLARE SECTION;	
    string ustatement[4096];
    EXEC SQL END DECLARE SECTION;	
    int usecase = 1;
    /* Select for User Privileges. */
    EXEC SQL IFDEF NOINT8;
    usecase = 1;
    EXEC SQL ENDIF;

    if (!usecase || (major_vers == 7 && minor_vers < 30) || major_vers < 7) {
	exec sql
	    create temp table myschema_auth_map( type int, usertype char(1) );
	exec sql insert into myschema_auth_map values ( 1, "D" );
	exec sql insert into myschema_auth_map values ( 2, "R" );
	exec sql insert into myschema_auth_map values ( 3, "C" );
	exec sql insert into myschema_auth_map values ( 0, "G" );
	strcpy( ustatement,
		"SELECT username, type, \" \" "
		"FROM \"informix\".sysusers su, myschema_auth_map ma "
		"WHERE su.usertype = ma.usertype ");
    } else if (major_vers >= 10 ) {
	strcpy( ustatement,
		"SELECT username, CASE usertype WHEN 'D' THEN 1 WHEN 'R' THEN 2 "
		"    WHEN 'C' THEN 3 WHEN 'G' THEN 0 WHEN 'U' THEN 5 END usertype, "
		"    defrole "
		"FROM \"informix\".sysusers " );  /* 'U' users have a role granted but no
						connectivity yet. */
    } else {
	strcpy( ustatement,
		"SELECT username, CASE usertype WHEN 'D' THEN 1 WHEN 'R' THEN 2 "
		"    WHEN 'C' THEN 3 WHEN 'G' THEN 0 WHEN 'U' THEN 5 END usertype, "
		"    \" \" "
		"FROM \"informix\".sysusers " );  /* 'U' users have a role granted but no
						connectivity yet. */
    }
    if (doOrder == NameOrder)
	strcat( ustatement, "ORDER BY 2 ASC, 1 ASC" );
    else
	strcat( ustatement, "ORDER BY 2 ASC " );

    EXEC SQL PREPARE s_user from :ustatement;
    if (sqlca.sqlcode) {
	fprintf( stderr,
		 "Cannot PREPARE select statement for users, error: %d.\n",
		 sqlca.sqlcode );
	fprintf( stderr, "Error at character %d.\nStatement is:\n",
		 sqlca.sqlerrd[4] );
	fprintf( stderr, "%s", ustatement );
	rexit(18);
    }
    EXEC SQL DECLARE user CURSOR FOR s_user;
    if (sqlca.sqlcode ) {
	fprintf( stderr,
		 "Cannot declare user cursor; Error = %d\n",
		 sqlca.sqlcode );
	rexit(17);
    }
    EXEC SQL OPEN user;
    if (sqlca.sqlcode) {
	fprintf( stderr, "Cannot open user cursor; Error = %d, Isam = %d\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	rexit(16);
    }

    EXEC SQL FETCH user INTO :username, :priv, :defrole;
    while (!sqlca.sqlcode) {
	switch (priv) {
	case 1:
	    fprintf( outfileG, "GRANT DBA TO \"%s\";\n", username );
	    break;
	case 2:
	    fprintf( outfileG, "GRANT RESOURCE TO \"%s\";\n", username );
	    break;
	case 3:
	    fprintf( outfileG, "GRANT CONNECT TO \"%s\";\n", username );
	    break;
	case 0:
	    if (ids_server)
		fprintf( outfile,
			 "CREATE ROLE%s \"%s\";\n",
			 create_conditionally ? " IF NOT EXISTS" : "",
			 username );
	    break;
	}
	if (defrole[0] != (char)0)
	    fprintf( outfile,
		     "GRANT DEFAULT ROLE \"%s\" TO \"%s\";\n",
		     defrole,
		     username );

	EXEC SQL FETCH user INTO :username, :priv, :defrole;
    }
    if (sqlca.sqlcode && sqlca.sqlcode != SQLNOTFOUND) {
	fprintf( stderr,
		 "Cannot FETCH cursor user; Error = %d\n",
		 sqlca.sqlcode );
	rexit( 21 );
    }
    fprintf( outfile, "\n" );
    fprintf( outfileG, "\n" );

    if (!usecase || (major_vers == 7 && minor_vers < 30)) {
	exec sql drop table myschema_auth_map;
    }

    if (!conn5 && ids_server) {
	/* Grant role code. */
	EXEC SQL DECLARE c_roles CURSOR FOR
	    SELECT rolename, grantee, is_grantable
	    FROM "informix".sysroleauth 
	    ORDER BY rolename, grantee;
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Error DECLARing CURSOR c_roles: Err=%d, ISAM=%d\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    rexit( 21 );
	}
	EXEC SQL OPEN c_roles;
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Error OPENing CURSOR c_roles: Err=%d, ISAM=%d\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    rexit( 21 );
	}
	EXEC SQL FETCH c_roles INTO :sysroleauth;
	while (!sqlca.sqlcode) {
	    fprintf( outfileG,
		     "GRANT \"%s\" TO \"%s\"%s;\n",
		     sysroleauth.rolename,
		     sysroleauth.grantee,
		     ((sysroleauth.is_grantable[0] == 'y')
		      ? " WITH GRANT OPTION"
		      : "") );
	    EXEC SQL FETCH c_roles INTO :sysroleauth;
	}
	if (sqlca.sqlcode && sqlca.sqlcode != SQLNOTFOUND) {
	    fprintf( stderr,
		     "Cannot FETCH cursor c_roles; Error = %d\n",
		     sqlca.sqlcode );
	    rexit( 21 );
	}
	EXEC SQL CLOSE c_roles;
	fprintf( outfileG, "\n" );
    }
    if (major_vers >= 11 && SupressSetSessionAuth != 1) {
	/* Surrogate user autorization. */
	int auths = 0;

	EXEC SQL DECLARE suauth_c CURSOR FOR
	    SELECT trusteduser, surrogateuser
	    FROM syssurrogateauth
	    WHERE NOT trusteduser = surrogateuser
	    ORDER BY ROWID;
	if (sqlca.sqlcode ) {
	    fprintf( stderr,
		     "Cannot declare suauth_c cursor; Error = %d\n",
		     sqlca.sqlcode );
	    rexit(20);
	}

	EXEC SQL OPEN suauth_c;
	for (;;) {
	    EXEC SQL FETCH suauth_c INTO :suauths;
	    if (sqlca.sqlcode != 0)
		break;
	    fprintf( outfile,
		     "%sGRANT SETSESSIONAUTH ON \"%s\" TO \"%s\";\n",
		     SupressSetSessionAuth == 2 ? "-- " : "",
		     suauths.surrogateuser,
		     suauths.trusteduser );
	    auths++;
	}
	EXEC SQL CLOSE suauth_c;
	if (auths)
	    fprintf( outfile, "\n" );
    }
}

void print_langauths( FILE *outfile ) 
{
    EXEC SQL BEGIN DECLARE SECTION;
    typedef struct SLAUTH_S {
        string grantor[32];
        string grantee[32];
        long langid;
        char langauth[1];
        string langname[30];
    } slauth_t;

    slauth_t auth;
    string slauth_txt[500];
    
    EXEC SQL END DECLARE SECTION;

    strcpy( slauth_txt, 
	    "SELECT grantor, grantee, sa.langid, langauth, langname "
	    "FROM \"informix\".sysroutinelangs sl, \"informix\".syslangauth sa "
	    "WHERE sl.langid = sa.langid "
	    "  AND langauth = 'u' "
	    "  AND grantee != 'DBA' "
	    "ORDER BY 1, 2, 3;" );
    EXEC SQL PREPARE sl_auth_s FROM :slauth_txt;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Error PREPARing CURSOR slauth: Err=%d, ISAM=%d\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	fprintf( stderr, "Stmt: %s\n", slauth_txt );
	rexit( 21 );
    }
    EXEC SQL DECLARE slauth CURSOR FOR sl_auth_s;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Error DECLARing CURSOR slauth: Err=%d, ISAM=%d\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	rexit( 21 );
    }
    EXEC SQL OPEN slauth;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Error OPENing CURSOR slauth: Err=%d, ISAM=%d, <%s>\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1],
		 sqlca.sqlerrm );
	rexit( 21 );
    }
    for(;;) {
	EXEC SQL FETCH slauth INTO :auth;
	if (sqlca.sqlcode != 0)
	    break;
	fprintf( outfile,
		 "GRANT usage ON LANGUAGE %s TO \"%s\";\n",
		 auth.langname,
		 auth.grantee );
    }
    if (sqlca.sqlcode != 0 && sqlca.sqlcode != SQLNOTFOUND) {
	    fprintf( stderr,
		     "Cannot FETCH cursor slauth; Error = %d\n",
		     sqlca.sqlcode );
	    rexit( 21 );
    }
    EXEC SQL FREE slauth;
    fprintf( outfile, "\n" );
}
