/*
  Program Id: sqlstruct

  Program Description: this program will generate a "C" struct for a
  specified SQL statement. 

  Author: Art S. Kagel from schema.ec 
  originally taken from Rick Friedman's dbschema_cc for Sybase

  Date: May of 1992  

  Version: 1.0

  Comments:    

  Revisions:

  Usage:    sqlstruct -n structname [-F sqlfile|-P] [-h host] [-f|-c|-m|-s] [-?] -d database [filename]

  RCS Header
  ----------
  $Header: /home/art/GoogleDrive/Latest/RCS/sqlstruct.ec,v 1.41 2021/12/21 17:12:40 art Exp $

  RCS Log
  -------
  $Log: sqlstruct.ec,v $
  Revision 1.41  2021/12/21 17:12:40  art
  Updated Copyright notice.

  Revision 1.40  2021/04/27 19:17:14  art
  Added -l to generate a comma separated list of column names suitable
  for a projection clause or some coding situations.

  Revision 1.39  2016/06/27 21:24:36  art
  Updated copyright notice.

  Revision 1.38  2015/03/10 15:10:39  art
  Added -S - generate SQL schema and -T - generate SQL temp table schema.

  Revision 1.37  2013/07/09 16:18:24  art
  Fixed problem with old 18byte object names.  Added support for 128byte names.

  Revision 1.36  2013/02/10 22:18:19  art
  Updated Copyright

  Revision 1.35  2012/05/30 01:25:33  art
  Protected TRUE and FALSE from pre-definition (mostly on AIX).

  Revision 1.34  2011/05/23 17:17:11  art
  Added support for 4GL RECORD structures (-G).

  Revision 1.33  2011/05/23 16:04:57  art
  Updated the copyright statement.

  Revision 1.32  2010/08/25 13:18:02  art
  *** empty log message ***

  Revision 1.31  2010/08/25 13:16:54  art
  *** empty log message ***

  Revision 1.30  2010/08/25 13:16:15  art
  Typo.

  Revision 1.29  2010/08/25 13:12:34  art
  Fixed main loop exit and end of input detection.
  Fixed compile time warning on some platforms.

  Revision 1.28  2009/07/23 17:29:24  art
  Changed include of getopt.h to local include.

  Revision 1.27  2009/04/12 21:49:32  art
  Finished cleanup.  Added -V for revision and copyright output.

  Revision 1.26  2009/04/01 16:36:23  art
  Removed all BB-isms and added support for LVARCHAR, BIGINT, and BIGSERIAL

  Revision 1.25  2004/07/30 17:57:10  kagel
  Fixed complex type output.

  Revision 1.24  2003/05/30 17:33:42  kagel
  Eliminated problem if progid contains semi-colon.

  Revision 1.23  2003/05/30 16:45:35  kagel
  Add -D to insert the progid description in the output and
  -S to make the input structure prefix into a suffix.

  Revision 1.22  2003/05/22 20:59:38  kagel
  Fixed -u processing.

  Revision 1.21  2003/05/22 20:44:27  kagel
  Added new options to eliminate struct names and change the prefix for
  input struct and typedef names.

  Revision 1.20  2003/05/22 18:26:37  kagel
  missing '
  semi-colon.

  Revision 1.19  2003/05/22 18:24:21  kagel
  Made wildcard work for -P also.

  Revision 1.18  2003/05/22 18:11:09  kagel
  Added ability to specify a wildcard for progid on the commandline or
  multiple progids or SQL statements at prompts.

  Revision 1.17  2003/05/22 17:09:14  kagel
  Removed message when processing update statements also error return
  code removed.

  Revision 1.16  2002/12/26 20:24:22  kagel
  fixups.

  Revision 1.15  2002/12/26 20:19:26  kagel
  Added remaining data types.

  Revision 1.14  2002/12/26 16:48:55  kagel
  Fixed bug in reallocation of the SQL strings if SQL exceeds allocation.

  Revision 1.13  2002/12/24 20:48:00  kagel
  Changed fetching #conversions to a cursor.

  Revision 1.12  2002/07/16 15:49:49  kagel
  Added support for input structure of progids.

  Revision 1.11  2002/03/05 18:54:38  kagel
  Cleanup help.

  Revision 1.9  2002/03/05 18:49:57  kagel
  Cosmetic.

  Revision 1.8  2002/03/05 18:48:40  kagel
  Fixed syntax of structures.

  Revision 1.7  2002/03/05 18:46:52  kagel
  Added datetime support.

  Revision 1.6  2002/03/05 18:43:02  kagel
  Added read SQL from stdin.

  Revision 1.5  2001/11/28 17:44:18  kagel
  Added new options -m & -s to specify other string handling.

  Revision 1.4  2001/10/18 15:11:21  kagel
  Added -s & -e.

  Revision 1.3  2001/10/18 14:50:12  kagel
  Added support for progmenu appid/progids.

  Revision 1.2  2001/06/15 18:32:57  kagel
  Added some stuff.  Fixed -f, added -c.  Fixed char output.

  Revision 1.1  2001/06/15 17:58:47  kagel
  Initial revision

*/
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#if defined(__sysv__) || defined(i386) || defined(__GNUC__)
#include <string.h>
#else
#include <strings.h>
#endif
#include <errno.h>
#include "getopt.h"

EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqlda;
EXEC SQL INCLUDE sqltypes;
EXEC SQL INCLUDE sqlstype;
EXEC SQL INCLUDE datetime;
EXEC SQL INCLUDE varchar;

static char rcs_header[] = "$Header: /home/art/GoogleDrive/Latest/RCS/sqlstruct.ec,v 1.41 2021/12/21 17:12:40 art Exp $";

extern struct sqlca_s sqlca;

EXEC SQL BEGIN DECLARE SECTION;

typedef struct QDATA_RETN_S {
    short col_seq;
    short sqltype;
    short ret_len;
    dtime_t timestamp;  /* DATETIME */
} qdata_retn_t;

qdata_retn_t *conversions;
long nconverts, cnvrt;

string databasename[129], host[19], statement[2048], connection[41];
string stmtline[76], chartype[20];
char *sqlstr = (char *)NULL;
char strname[1025] = "";
short seq;
int incr = 1;

EXEC SQL END DECLARE SECTION;

#define TBSFACT 16777216L
#if !defined( FALSE )
#define FALSE (0)
#endif
#if !defined( TRUE )
#define TRUE (!FALSE)
#endif

#define SQLINCR 2048

int namestruct = TRUE;

typedef enum { pt_sql, pt_file, pt_stdin } prompt_type;

void usage( char *program );
void rexit( int );
void strip( char *str, int len );

int main(int argc, char *argv[])
{
    int i, j, c;
    int fglflag = 0, errflag=0, dflag=0, tflag=0, hflag=0, eflag=0, sqlflag=0, tmpflag=0;
    FILE *outfile = stdout, *infile = stdin;
    char filename[1025], comma[2];
    static char sqlfilename[1025];
    prompt_type prompt;
    char *nl, *end, line[1025];
    long sqllen = 0, sqlmax = 0, do_listing = 0;
    int newline;
    ifx_sqlda_t *dastruct;
    ifx_sqlvar_t *col;

    strcpy( chartype, "none" );
    host[0] = (char)0;
    prompt = pt_sql;

    while ((c = getopt(argc, argv, "GSTVd:h:ufclmeF:n:s?")) != EOF){
        switch(c) {
	case 'V':
	    fprintf( stderr,
		     "sqlstruct.ec Revison: %s.  Copyright 1990-2022, Art S. Kagel\n",
		     "$Revision: 1.41 $" );
	    return 0;
	    break;
	case 'l':
	    do_listing = 1;
	    break;
	case 'T':
	    tmpflag = !tmpflag;
	case 'S':
	    sqlflag = !sqlflag;
	    break;
	case 'G':
	    fglflag = !fglflag;
	    break;
	case 'u':
	    namestruct = FALSE;
	    break;
	case 'n':
	    strncpy( strname, optarg, 1025 );
	    break;
	case 'c':
	    strcpy( chartype, "char" );
	    incr = 1;
	    break;
	case 'f':
	    strcpy( chartype, "fixchar" );
	    incr = 0;
	    break;
	case 'm':
	    strcpy( chartype, "char" );
	    incr = 0;
	    break;
	case 's':
	    strcpy( chartype, "string" );
	    incr = 1;
	    break;
	case 'd':
	    if (dflag)
		errflag++;
	    else{
		dflag++;
		strncpy(databasename, optarg, sizeof databasename);
		databasename[sizeof databasename - 1] = (char)0;
	    }
	    break;
	case 'h':
	    if (hflag)
		errflag++;
	    else{
		hflag++;
		strncpy(host, optarg, 18);
		host[18] = (char)0;
	    }
	    break;
	case 'e':
	    eflag = !eflag;
	    break;
	case 'F':
	    /* Use file containing SQL text. */
	    if (prompt != pt_sql) {
		fprintf( stderr, "Only one of -P, -F, or -a & -p permitted!\n" );
		errflag++;
	    }
	    prompt = pt_file;
	    strncpy( sqlfilename, optarg, 1025 );
	    if (strcmp( sqlfilename, "-" ) == 0)
		prompt = pt_stdin;
	    break;
	case '?':
	default:
	    errflag++;
	    break;
        }
    }

    if (optind < argc) {
	strcpy( filename, argv[optind] );
	outfile = fopen( filename, "w" );
	if (!outfile) {
	    fprintf( stderr, "Cannot open output: %s.\n", filename );
	    exit(3);
	}
    }

    if (errflag){
	usage( argv[0] );
        exit (1);
    }

    if (dflag == 0){
	fprintf( stderr, "Database is required.\n" );
	usage( argv[0] );
        exit (1);
    }

    if (strcmp(databasename,"")==0)
	{
	    fprintf( stderr, "Database is required.\n" );
	    usage( argv[0] );
	    exit (1);
	}

    if (strlen(strname) <= 0 && do_listing == 0) {
	fprintf( stderr, "Structure name (-n<name>) required!\n" );
	usage( argv[0] );
	exit(1);
    }

    if ((eflag + fglflag + sqlflag + do_listing) > 1) {
	fprintf( stderr, "-e, -S, -T, -l, and -G are mutually exclusive!\n" );
	usage( argv[0] );
	exit(1);
    }

    if (sqlflag && !namestruct) {
	fprintf( stderr, "Cannot generate an unnamed schema!\n" );
	usage( argv[0] );
	exit(1);
    }

    strcpy( connection, databasename );
    if (hflag) {
	strcat( connection, "@" );
	strcat( connection, host );
	EXEC SQL
	    CONNECT TO :connection;
    } else {
	EXEC SQL DATABASE :connection;
    }
    if ( sqlca.sqlcode )
	{
	    fprintf( stderr,
		     "Cannot access host (%s) & database (%s); error=%d\n",
		     host,
		     databasename,
		     sqlca.sqlcode );
	    rexit(1);
	}

    for (;;) {
	if (prompt == pt_file) {
	    infile = fopen( sqlfilename, "r" );
	} else if (prompt == pt_stdin)
	    infile = stdin;

	end = (char *)0;
	newline = 0;
	sqlmax = SQLINCR;
	sqlstr = (char *)malloc( sqlmax );
	if (!sqlstr) {
	    fprintf( stderr,
		     "Error allocating SQL string.  Errno = %d.\n",
		     errno );
	    exit(1);
	}
	sqlstr[0] = (char)0;
	sqllen = 0;
	if (prompt != pt_file)
	    fprintf( stderr, "> " );
	while (fgets( line, 1025, infile )) {
	    newline = ((nl = strchr( line, '\n' )) != (char *)0);
	    if (newline)
		*nl = (char)0;
	    if ((end = strchr( line, ';' )) != (char *)0)
		*end = (char)0;
	    sqllen += strlen(line);
	    if (sqllen+2 >= sqlmax) {
		while (sqllen >= sqlmax)
		    sqlmax += SQLINCR;
		errno = 0;
		sqlstr = (char *)realloc( sqlstr, sqlmax );
		if (!sqlstr || errno) {
		    fprintf( stderr,
			     "Error expanding SQL string.  Errno = %d.\n",
			     errno );
		    exit(1);
		}
	    }
	    strcat( sqlstr, line );
	    if (newline) {
		strcat( sqlstr, " " );
		sqllen++;
	    }

	    if (end)
		break;
	    newline = 0;
	    if (prompt != pt_file)
		fprintf( stderr, "> " );
	}
	strcat( sqlstr, ";\n" );

	if (ferror( infile )) {
	    fprintf( stderr, "Error reading SQL.  Errno=%d.\n", errno );
	    exit(1);
	}
	if (strcmp( chartype, "none" ) == 0)
	    if (eflag)
		strcpy( chartype, "string" );
	    else
		strcpy( chartype, "char" );

	/* Break out on an empty input string. */
	strip( sqlstr, strlen( sqlstr ) );
	if (strlen( sqlstr ) <= 2) {
	    fprintf( stderr, "\n" );
	    break;
	}

	EXEC SQL
	    PREPARE stmt1 FROM :sqlstr;
	if (sqlca.sqlcode) {
	    fprintf( stderr,
		     "Cannot PREPARE statement, error: %d.\n",
		     sqlca.sqlcode );
	    fprintf( stderr, "Error at character %d.\nStatement is:\n",
		     sqlca.sqlerrd[4] );
	    fprintf( stderr, "%s\n", sqlstr );
	    rexit(2);
	}

	EXEC SQL DESCRIBE stmt1 INTO dastruct;
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Cannot DESCRIBE statement, error: %d.\n",
		     sqlca.sqlcode );
	    fprintf( stderr, "%s\n", sqlstr );
	    rexit(2);
	}

	if (sqlca.sqlcode == SQ_UPDATE ||
	    sqlca.sqlcode == SQ_DELETE ||
	    sqlca.sqlcode == SQ_INSERT)
	    {
		/* Update, delete, & insert statements do not return data.  OK. */
		fprintf( stderr,
			 "Statement does not return data! Type=%d.\n",
			 sqlca.sqlcode );
		fprintf( stderr, "%s\n", sqlstr );

		fflush( outfile );
		continue;
	    }

	if (sqlca.sqlcode != (SQ_SELECT - SQ_SELECT)
	    && sqlca.sqlcode != SQ_SELINTO
	    && sqlca.sqlcode != SQ_EXECPROC)
	    {
		fprintf( stderr,
			 "Statement does not return data! Type=%d.\n",
			 sqlca.sqlcode );
		fprintf( stderr, "%s\n", sqlstr );
		continue;
	    } else if (sqlca.sqlcode == SQ_EXECPROC && dastruct->sqld <= 0) {
	    fprintf( stderr, "Stored procedure does not return data!\n" );
	    fprintf( stderr, "%s\n", sqlstr );
	    continue;
	}

	i = 0;
	if (eflag)
	    fprintf( outfile, "\nEXEC SQL BEGIN DECLARE SECTION;\n" );

	if (fglflag)
	    fprintf( outfile, "\nDEFINE %s RECORD\n", strname );
	else if (sqlflag)
	    fprintf( outfile, "\nCREATE %sTABLE %s (\n", tmpflag ? "TEMP ": "", strname );
	else if (do_listing == 1) {
	    fprintf( outfile, "\n" );
	} else {
	    if (namestruct)
		fprintf( outfile, "\ntypedef struct %s_s {\n", strname );
	    else
		fprintf( outfile, "\ntypedef struct {\n" );
	}

	cnvrt = 0;
	for (col = dastruct->sqlvar, i = 0; i < dastruct->sqld; col++, i++) {
	    if ((sqlflag || do_listing == 1) && i < (dastruct->sqld - 1))
		strcpy( comma, "," );
	    else
		comma[0] = (char)0;
	    if (do_listing == 1) {
		fprintf( outfile, "%s%s ", col->sqlname, comma );
	    } else {
		if (cnvrt < nconverts && (i+1) >= conversions[cnvrt].col_seq) {
		    while (cnvrt < nconverts && (i+1) > conversions[cnvrt].col_seq)
			cnvrt++;
		    if (cnvrt < nconverts && (i+1) == conversions[cnvrt].col_seq) {
			col->sqltype = conversions[cnvrt].sqltype;
			col->sqllen  = conversions[cnvrt].ret_len;
			cnvrt++;
		    }
		}
		switch (MASKNONULL(col->sqltype)) {
		case SQLSMINT:
		case CSHORTTYPE:
		    if (fglflag || sqlflag)
			fprintf( outfile, "    %s SMALLINT%s\n", col->sqlname, comma );
		    else
			fprintf( outfile, "    short %s;\n", col->sqlname );
		    break;
		case SQLSMFLOAT:
		case CFLOATTYPE:
		    if (fglflag || sqlflag)
			fprintf( outfile, "    %s SMALLFLOAT%s\n", col->sqlname, comma );
		    else
			fprintf( outfile, "    float %s;\n", col->sqlname );
		    break;
		case SQLFLOAT:
		case CDOUBLETYPE:
		    if (fglflag || sqlflag)
			fprintf( outfile, "    %s FLOAT%s\n", col->sqlname, comma );
		    else
			fprintf( outfile, "    double %s;\n", col->sqlname );
		    break;
		case SQLDECIMAL:
		case SQLMONEY:
		case CMONEYTYPE:
		case CDECIMALTYPE:
		    if (fglflag || sqlflag)
			if (PRECDEC( col->sqllen ) == 0)
			    fprintf( outfile, "    %s DECIMAL(%d)%s\n", col->sqlname, PRECTOT( col->sqllen ), comma );
			else
			    fprintf( outfile, "    %s DECIMAL(%d,%d)%s\n", col->sqlname, PRECTOT( col->sqllen ), PRECDEC( col->sqllen ), comma );
		    else
			if (strcmp(col->sqlname, "(count)" ) == 0)
			    fprintf( outfile, "    int64_t count;\n" );
			else
			    fprintf( outfile, "    dec_t %s;\n", col->sqlname );
		    break;
		case SQLINT:
		case SQLSERIAL:
		case CLONGTYPE:
		case CINTTYPE:
		    if (fglflag || sqlflag)
			fprintf( outfile, "    %s INT%s\n", col->sqlname, comma );
		    else
			fprintf( outfile, "    int %s;\n", col->sqlname );
		    break;
		case SQLDATE:
		    if (fglflag || sqlflag)
			fprintf( outfile, "    %s DATE%s\n", col->sqlname, comma );
		    else
			fprintf( outfile, "    int %s;\n", col->sqlname );
		    break;
		case SQLCHAR:
		case SQLNCHAR:
		case CCHARTYPE:
		case CSTRINGTYPE:
		case CFIXCHARTYPE:
		    if (fglflag || sqlflag)
			fprintf( outfile, "    %s CHAR(%d)%s\n", col->sqlname, col->sqllen, comma );
		    else
			fprintf( outfile,
				 "    %s %s[%d];\n",
				 chartype,
				 col->sqlname,
				 col->sqllen + incr );
		    break;
		case SQLVCHAR:
		case SQLNVCHAR:
		    if (fglflag || sqlflag)
			if (VCMIN(col->sqllen) == 0)
			    fprintf( outfile, "    %s VARCHAR(%d)%s\n", col->sqlname, VCMAX(col->sqllen), comma );
			else
			    fprintf( outfile, "    %s VARCHAR(%d, %d)%s\n", col->sqlname, VCMAX( col->sqllen ), VCMIN( col->sqllen ), comma );
		    else
			fprintf( outfile,
				 "    %s %s[%d];\n",
				 chartype,
				 col->sqlname,
				 VCMAX(col->sqllen) + incr );
		    break;
		case SQLLVARCHAR:
		    if (fglflag || sqlflag)
			fprintf( outfile, "    %s LVARCHAR(%d)%s\n", col->sqlname, col->sqllen, comma );
		    else
			fprintf( outfile,
				 "    %s %s[%d];\n",
				 chartype,
				 col->sqlname,
				 col->sqllen + incr );
		    break;
		case CINVTYPE:
		case SQLINTERVAL:
		    if (fglflag || sqlflag) {
			int fr1, p1, pt1, fr2, p2, pt2;
			char part1[31], part2[31];
			char dtparts[7][10] = { "YEAR",
						"MONTH",
						"DAY",
						"HOUR",
						"MINUTE",
						"SECOND",
						"FRACTION" };

			p1 = (col->sqllen / 16) % 16;
			pt1 = (p1 <= 10) ? (p1 / 2) : 6;
			fr2 = (col->sqllen % 16) - 10;
			p2 = col->sqllen % 16;
			pt2 = (p2 <= 10) ? (p2 / 2) : 6;
			if (pt1 != 6) {
			    fr1 = (col->sqllen / 256) + p1 - p2;
			    sprintf( part1, "%s(%d)", dtparts[pt1], fr1 );
			} else
			    strcpy( part1, dtparts[pt1] );
			if (pt2 == 6) {
			    sprintf( part2, "%s(%d)", dtparts[pt2], fr2 );
			} else
			    strcpy( part2, dtparts[pt2] );
			fprintf( outfile,
				 "    %s INTERVAL %s TO %s%s\n",
				 col->sqlname,
				 part1,
				 part2,
				 comma );
		    } else 
			fprintf( outfile,
				 "    intrvl_t %s;\n",
				 col->sqlname );
		    break;
		case SQLDTIME:
		case CDTIMETYPE:
		    if (fglflag || sqlflag) {
			int pt1, fr2, pt2;
			char part1[31], part2[31];
			char dtparts[7][10] = { "YEAR",
						"MONTH",
						"DAY",
						"HOUR",
						"MINUTE",
						"SECOND",
						"FRACTION" };

			pt1 = (((col->sqllen / 16) % 16) <= 10) ?
			    (((col->sqllen / 16) % 16) / 2) : 6;
			strcpy( part1, dtparts[pt1] );
			pt2 = ((col->sqllen % 16) <= 10) ?
			    ((col->sqllen % 16) / 2) : 6;
			if (pt2 == 6) {
			    fr2 = (col->sqllen % 16) - 10;
			    sprintf( part2, "%s(%d)", dtparts[pt2], fr2 );
			} else
			    strcpy( part2, dtparts[pt2] );
			fprintf( outfile,
				 "    %s DATETIME %s TO %s%s\n",
				 col->sqlname,
				 part1,
				 part2,
				 comma );
		    } else
			fprintf( outfile,
				 "    dtime_t %s;\n",
				 col->sqlname );
		    break;
#if defined (SQLINT8)
		case SQLSERIAL8:
		case SQLINT8:
		case CINT8TYPE:
#if defined (SQLINFXBIGINT)
		case SQLINFXBIGINT:
		case SQLBIGSERIAL:
		case CBIGINTTYPE:
#endif
		    if (fglflag || sqlflag)
			fprintf( outfile, "    %s BIGINT%s\n", col->sqlname, comma );
		    else
			fprintf( outfile,
				 "    %s %s;\n",
				 eflag ? "bigint" : "int64_t",
				 col->sqlname );
		    break;
		case SQLBOOL:
		case CBOOLTYPE:
		    if (fglflag || sqlflag)
			fprintf( outfile, "    %s BOOLEAN%s\n", col->sqlname, comma );
		    else
			fprintf( outfile, "    boolean %s;\n", col->sqlname );
		    break;
#endif
		default:
		    fprintf( outfile,
			     "   <Unknown type #%d, len=%d> %s;\n",
			     col->sqltype,
			     col->sqllen,
			     col->sqlname );
		}
	    }
	}
	if (fglflag)
	    fprintf( outfile, "END RECORD\n\n" );
	else if (sqlflag)
	    fprintf( outfile, ");\n\n" );
	else if (do_listing == 1) {
	    fprintf( outfile, "\n" );
	} else {
	    fprintf( outfile, "} %s_t, *%s_tp;\n\n", strname, strname );
	    if (eflag)
		fprintf( outfile, "EXEC SQL END DECLARE SECTION;\n\n" );
	}
	
	fflush( outfile );
    }
    fclose( outfile );
    rexit(0);
}

void rexit(int stat)
{
    EXEC SQL
	CLOSE DATABASE;
    EXEC SQL
	DISCONNECT ALL;
    exit(stat);
}

void usage( char *program )
{
    fprintf( stderr, "Usage:\n" );
    fprintf( stderr, "%s -? | -V\n", program );
    fprintf( stderr,
	     "%s -n structname [-F sqlfile] [-h host] -d database [-f|-c|-m|-s|-S|-T|-l]\n"
	     "\t\t[-e|-G] [filename]\n",
	     program );
    fprintf( stderr,
	     "\t-n - Name of output structure.  Required (except if -l is passed).\n" );
    fprintf( stderr,
	     "\t-u - Generate unnamed structure.  "
	     "(Default: name structure \n\t     <structname>_s)\n" );
    fprintf( stderr,
	     "\t-F - Read SQL statement from sqlfile "
	     "(-F- implies read from stdin.\n" );
    fprintf( stderr,
	     "\t-e - generate ESQL style structs including DECLARE's.\n" );
    fprintf( stderr,
	     "\t-G - generate 4GL style RECORD structures.\n" );
    fprintf( stderr,
	     "\t-S - generate an SQL schema for a table matching the query results.\n" );
    fprintf( stderr,
	     "\t-T - generate an SQL schema for a temp table matching the query results.\n" );
    fprintf( stderr,
	     "\t-l - generate a comma separated list of column names suitable for a \n"
	     "\t     projection clause.\n" );
    fprintf( stderr,
	     "\t-m - generate 'char' type for strings in headers rather than\n"
	     "\t     'string' type.\n" );
    fprintf( stderr,
	     "\t-f - generate 'fixchar' type for strings in ESQL headers rather than\n"
	     "\t     'string' type.  (Space padded no NULL "
	     "terminator.)\n" );
    fprintf( stderr,
	     "\t-c - generate 'char' type for strings in ESQL headers."
	     "\n\t     (Space padded with NULL terminator.)\n" );
    fprintf( stderr,
	     "\t-s - generate 'string' type for strings in ESQL headers.\n"
	     "\t     (NULL terminated with trailing spaces stripped for CHAR and all\n"
	     "\t     but explicit trailing spaces stripped for VARCHAR columns.)\n" );
    fprintf( stderr,
	     "\t-h - Connect to host.  Host defaults to current host.\n" );
    fprintf( stderr,
	     "\t-d - Connect to database.\n" );
    fprintf( stderr,
	     "\t-V - Print version and copyright and exit.\n" );
    fprintf( stderr,
	     "\t-? - Print this usage.\n\n" );
    
    fprintf( stderr,
	     "Note that only statements that return data are supported.\n\n" );
    fprintf( stderr,
	     "The default datatype for CHAR and VARCHAR type fields is 'string' when -e is"
	     "\nspecified, 'char' otherwise.\n\n" );
}

void strip( char *str, int len )
{
    char *p;
    int alen = len;

    for (p = (str + len - 1); p > str; p--, alen--) {
	if (*p == ' ' || *p == (char)0)
	    *p = (char)0;
	else
	    break;
    }
    p = str;
    while (*p == ' ')
	p++;
    if (p > str)
	strcpy( str, p );
}
