/* ul.ec - Unload a select set to a binary fixed length file or reload such an
   unload file.

   Need to link in libgen.a (-lgen) to get the basename() function.
*/
/*
  RCS Header:
  -----------
  $Header: /home/art/GoogleDrive/Latest/RCS/ul.ec,v 1.63 2022/08/24 10:50:32 art Exp $

  RCS Log:
  --------
  $Log: ul.ec,v $
  Revision 1.63  2022/08/24 10:50:32  art
  Fixed include files for AIX. Thanks to Doug Lawry for the assist.

  Revision 1.62  2022/02/07 15:52:51  art
  Added -I to invode IO in O_DIRECT mode bypassing the system
  cache.

  Revision 1.61  2021/12/21 17:12:40  art
  Updated Copyright notice.

  Revision 1.60  2020/11/19 21:49:57  art
  Made inlude changes for better Solaris support.

  Revision 1.59  2017/12/26 16:15:16  art
  Fixed to compile under latest Linux & GCC 4.8

  Revision 1.58  2016/06/27 21:33:45  art
  Updated copyright notice.

  Revision 1.57  2013/11/20 15:37:03  art
  Fixed HPUX detection YET AGAIN!

  Revision 1.56  2013/04/29 22:24:12  art
  Fixed location of inet.h on HPUX.

  Revision 1.55  2013/02/10 22:18:19  art
  Updated Copyright

  Revision 1.54  2012/03/12 00:40:00  art
  Added support for NCHAR and NVARCHAR type columns.  Thanks to Aleksander Kamenik
  for pointing out the missing support.

  Revision 1.53  2011/12/21 20:47:07  art
  Added new option -R which begins all transactions with the WITHOUT REPLICATION
  clause which prevents the operation from being replicated via Enterprise
  Replication.  Thanks to Øyvind Gjerstad for the suggestion.

  Revision 1.52  2011/05/13 16:16:11  art
  Documented -N.

  Revision 1.51  2011/04/29 16:44:50  art
  Updated copyright statement.

  Revision 1.50  2011/04/29 16:32:34  art
  Fixed printf format warnings.
  Added support for bigint and int8 (int8 not right yet).
  Added -N option to export in native byte order rather than network byte order.
  Note: default file format is compatible with external tables in "informix" format.

  Revision 1.49  2011/04/29 14:45:39  art
  Moved status printouts to stderr.  Implemented "-f -" for reading or writing from/to
  stdin/stdout.

  Revision 1.48  2009/09/09 18:10:54  art
  Fixed file size handling to allow files larger than 2GB.
  Fixed handling of BOOLEAN.
  Fixed a bug in DATETIME handling as string.
  Improved help.
  Fixed variable character default to handling as STRING to eliminate soft trailing spaces.

  Revision 1.47  2009/07/22 14:36:25  art
  Replaced all str*cpy's in the getopt loop with snprintf for added security.

  Revision 1.46  2009/05/22 04:10:43  art
  Tried to fix the strlower macro for non-gcc compilers.

  Revision 1.45  2009/05/13 21:08:10  art
  Made to work for mixed case table name arguments when DELIMIDENT is disabled.

  Revision 1.44  2009/02/23 22:23:29  art
  Fixed BIGINT/BIGSERIAL and INT8/SERIAL8 code to be ifdef'd out in older compilers.

  Revision 1.43  2008/12/10 06:03:28  art
  Added -F to the Usage output.

  Revision 1.42  2008/12/10 05:48:22  art
  Added one to length for variable LVARCHAR data (no -e) to allow space
  for the NULL terminator if the column was full.

  Revision 1.41  2008/12/10 05:23:09  art
  Made code 64bit compile safe.
  Fixed a bug in DECIMAL type handling.
  Added a warning to the Usage about not using LVARCHAR for binary data storage.

  Revision 1.40  2008/12/08 18:53:48  art
  Added support for INT8, SERIAL8, BIGINT, BIGSERIAL, and LVARCHAR columns.
  Also added support for unloading/loading LVARCHAR as FIXCHAR with -F option.
  Thanks to Ravi Krishna for suggesting the enhancement/improvement.

  Revision 1.39  2008/04/16 18:35:27  art
  Removed RCSwhat.

  Revision 1.38  2006/07/11 12:59:48  kagel
  Fixed #defines for HPUX.

  Revision 1.37  2004/08/17 20:21:26  kagel
  Typo

  Revision 1.36  2004/08/17 16:30:01  kagel
  Added support for large files for HP platforms.

  Revision 1.35  2004/01/14 18:50:21  kagel
  Made to compile and run on HP/ia64/HP-UX.

  # Revision 1.34  2003/10/30  17:59:55  17:59:55  kagel (art kagel 802-3153 NY - EAST 55TH STREET R+D NY INTERNAL SYSTEMS)
  # Removed references to 9.xx extention field to hold qualifiers.  Changed to
  # use a parallel array of qualifiers.
  # 
  Revision 1.33  2003/01/27 19:34:43  kagel
  Portability improvement - minor.

  Revision 1.32  2002/12/30 18:50:13  kagel
  Better fix.

  Revision 1.31  2002/12/30 18:46:39  kagel
  Fixed _H_LOCALDEF definition to compile on AIX with later SDK versions that
  already fix the problem that was meant to correct.

  Revision 1.30  2002/12/30 17:18:30  kagel
  Fixed typo (missing braces).

  Revision 1.28  2002/12/27 18:05:57  kagel
  Removing trace.  Need to make datetime changes for intervals as well.

  Revision 1.27  2002/12/27 17:26:45  kagel
  Added code to sab other binary fields in the dtime_t columns.

  Revision 1.23  2002/10/28 19:34:51  kagel
  Fixed some problems with NULL CHAR columns.

  Revision 1.22  2002/10/28 17:30:43  kagel
  Blocked binary datetime processing.  Fixed DATE type.

  Revision 1.21  2002/10/28 16:51:09  kagel
  Fixed mapping maxsize to unsigned long long.  Fixed all conversions.

  Revision 1.20  2002/10/28 15:31:04  kagel
  Fixed portable output format.

  Revision 1.19  2002/07/26 15:32:19  kagel
  Changed file format to always unload/load integers/dates/floats/doubles
  to Network Byte Order for portability of output files.

  Revision 1.18  2001/07/17 17:32:40  kagel
  last changes.

  Revision 1.16  1999/12/27 16:53:06  kagel
  Got name/behavior flags backwards.

  Revision 1.15  1999/12/27 15:23:13  kagel
  Fixed 2 typos and added #include for basename.

  Revision 1.14  1999/12/27 14:54:03  kagel
  Added automatic detection of program name.  If argv[0] == "bload"
  then the -l (load) flag is set, if argv[0] == "bunload" then the
  -u (unload) flag is set.  Otherwise the -l or -u flag is required.
  Giving the -l or -u specifies load or unload behavior regardless of
  the program name.

  Revision 1.13  1999/10/12 16:30:16  kagel
  Fixed #if defined() for AIX from __AIX__ to _AIX.

  Revision 1.12  1999/10/11 14:02:57  kagel
  Fixed #include of fcntl.h from sys/ to default directory.  The former
  location is system dependent.

  Revision 1.11  1999/10/04 17:10:35  kagel
  Fixed a few platform specific compilation problems affecting mainly
  AIX.  Check the #if defined( __AIX__ ) for correctness if this is your
  platform.

  Revision 1.10  1999/09/07 14:30:53  kagel
  Fixed CSHORTTYPE handling.  Broken when BLOB handling added.

  Revision 1.9  1999/08/16 19:20:45  kagel
  Removed ieeefp.h which is not needed and caused compilations errors
  on some systems.

  Revision 1.8  1999/07/14 23:53:50  kagel
  Fixed error dumping and added NULL column handling.

  Revision 1.7  1999/07/14 18:31:04  kagel
  Fixed usage display.

  Revision 1.6  1999/07/14 18:16:52  kagel
  Fixed COMMIT processing, eliminated testing display_count it is always non-
  zero!  Added reporting of number of records logged in error and total
  records input.

  Revision 1.5  1999/07/13 18:21:56  kagel
  Added -e errlogfile flag and feature.  Rows in error now are copied to
  the error log in ASCII delimited., dbload compatible, format.

  Revision 1.4  1999/07/08 19:03:53  kagel
  Added -V Version flag.  Added -D flag to export datetime in ASCII default
  now binary as old ESQL bug seems fixed.   Added BLOB support to unload.
  Load code not complete.

  # Revision 1.3  1998/06/22  20:57:16  kagel
  # Added ability to specify unbuffered output/ buffer size/ O_SYNC of output
  # file.  Minor improvements to the usage report.
  #
  # Revision 1.2  1998/04/01  19:38:17  kagel
  # Fixed bug in -d option handling. Enhanced several error messages.
  # Need to add INTERVAL type handling.
  #
  # Revision 1.1  1998/02/10  19:04:58  kagel
  # Initial revision
  #

*/

static char RCSHeader[] = "$Header: /home/art/GoogleDrive/Latest/RCS/ul.ec,v 1.63 2022/08/24 10:50:32 art Exp $";

/* Enable large file support
 * Mask differences in how the various OS's handle 64bit file offsets & 
 *  files larger than 2GB-1
 */

#if !defined __gcc__
#	define inline 
#endif

#if defined sun
#       define _FILE_OFFSET_BITS 64
#       define _LARGEFILE_SOURCE
#elif defined __DGUX__
#       define _FILE_OFFSET_BITS 64
#       define _LONGLONG_ENABLED 1
#elif defined _AIX
#       define _LARGE_FILES
#elif defined __hpux
#       define _APP32_64BIT_OFF_T
#endif

#if !defined LITTLE_ENDIAN
#	define LITTLE_ENDIAN 1234
#endif

/* Needed as AIX defines a locale control structure loc_t & while some SDK
 * versions include the fix dependent on the var _H_LOCALDEF, some define it
 * unconditionally and some not at all, so we'll be nice here and make it
 * conditional. */
EXEC SQL IFNDEF _H_LOCALEDEF;
EXEC SQL DEFINE _H_LOCALEDEF; 
EXEC SQL ENDIF;
 
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <sys/time.h>
#include <sys/resource.h>
#if defined _AIX
#include <fcntl.h>
#else
#include <asm-generic/fcntl.h>
#endif
#include <libgen.h>
#if defined( __hpux ) && defined( __hppa )
/* On HP only need to include this header to get the hton* and n*toh function
   prototypes. */
#include <arpa.inet.h>
#elif defined( __hpux ) 
#include <arpa/inet.h>
#elif defined( linux ) 
#include <arpa/inet.h>
#include <ctype.h>
#elif defined( __svr4__ )
#include <arpa/inet.h>
#include <ctype.h>
#endif
#include "getopt.h"

EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqlda;
EXEC SQL INCLUDE sqltypes;
EXEC SQL INCLUDE decimal;
EXEC SQL INCLUDE datetime;
EXEC SQL INCLUDE locator.h;

#if defined( NO_LARGE_FILES )
#define MAXFILESIZE (long long)2147483647
#else
#define MAXFILESIZE ((long long)4*1024*1024*1024*1024*1024)
#endif

int has_trans = 0, is_ansi = 0;	/*  True if DB has transactions, is ANSI.  */
int DTascii = 1, display_count = 0, verbose = 0, fixvchar = 1;
unsigned long long maxsize = MAXFILESIZE;
long maxrecs;
int use_ins_curs = 0, compatible = 0, use_sync = 0, no_buf = 0, bufsz = 0, FixLVChar = 0;
int use_direct = 0;
char *blobbuff;
long bbsize = 0, global_length = 0;
off_t bpos = 0;
char errfile[1025] = "ulerr.log", delim[2] = "|";
int EndianNess;
int NoSwab = 0;
int no_replicate = 0;

char Revision[] = "$Revision: 1.63 $";

#if !defined( __GNUC__ )
#define strlower( t, s, m ) {					\
	char *target = (t), *source = (s); int max = (m);	\
	while (max-- && *source) {				\
	    *target = tolower(*source);				\
	    target++; source++;					\
	}							\
	*target = 0;						\
    }
#else
inline void strlower( char *target, char *source, int max )
{
    while (max-- && *source) {
        *target++ = tolower(*source++);
    }
    *target = 0;
}
#endif

void usage( char * );
int load( char *insrtstmt, char *fname, char *prog );
int unload( char *slct, char *fname, char *prog );
int openit( loc_t *loc, int flags, int bsize );
int writeit( loc_t *loc, char *buf, int nbytes );
int readit( loc_t *loc, char *buf, int nbytes );
int closeit( loc_t *loc );
void dumprec( struct sqlda *ldesc );
int strip( char *str, int len );
int charcpy( unsigned char *target, unsigned char *source, int len );
int DetermineEndianNess();


int main( int argc, char **argv )
{
    int n, ch;
    int ret = 0, Delimident_Set = 0;
    EXEC SQL BEGIN DECLARE SECTION;
    static char dbname[129];
    EXEC SQL END DECLARE SECTION;
    static char  *qstr;
    static char  filename[1024];
    short load_flag, unload_flag;

    load_flag = unload_flag = 0;
    EndianNess = DetermineEndianNess();

    if ( argc == 1 ) {
	usage( argv[0] );
	exit( -1 );
    }

    if (strcmp( "bload", basename( argv[0] ) ) == 0) {
	load_flag = 1;
	unload_flag = 0;
    } else if (strcmp( "bunload", basename( argv[0] ) ) == 0) {
	load_flag = 0;
	unload_flag = 1;
    }

    if (getenv( "DELIMIDENT" ) != (char *)NULL) {
	Delimident_Set = 1;
    }

    while ((ch = getopt( argc, argv, "lnuviCd:Fe:f:q:c:s:r:B:DINRSUV" )) != EOF) {
	switch (ch) {
	case 'R':
	    no_replicate = 1;
	    break;
	case 'N':
	    NoSwab = 1;
	    break;
	case 'n':
	    fixvchar = 0;
	    break;
	case 'v':
	    verbose = 1;
	    break;
	case 'F':
	    FixLVChar = 1;
	    break;
	case 'V':
	    fprintf( stderr,
		     "\nul.ec: %s Copyright 1998-2022 Art S. Kagel.\n\n",
		     strtok( Revision, "$" ) );
	    exit( 0 );
	    break;
	case 'B':
	    bufsz = atol( optarg );
	    break;
	case 'D':
	    DTascii = 0;
	    break;
	case 'U':
	    no_buf = 1;
	    break;
	case 'S':
	    use_sync = 1;
	    break;
	case 'I':
	    use_direct = 1;
	    break;
	case 's':
	    /* Max output file size. */
	    maxsize = (unsigned long long)atol( optarg );
	    break;
	case 'r':
	    /* Max output file records. */
	    maxrecs = (int)atol( optarg );
	    break;
	case 'l':
	    /* load ?  */
	    load_flag = 1;
	    unload_flag = 0;
	    break;
	case 'u':
	    /* unload ?  */
	    unload_flag = 1;
	    load_flag = 0;
	    break;
	case 'd':
	    /* database name */
	    snprintf( dbname, sizeof dbname, "%s", optarg );
	    break;
	case 'e': 
	    /* Error log file name */
	    snprintf( errfile, sizeof errfile, "%s", optarg );
	    break;
	case 'f': 
	    /* output file name */
	    snprintf( filename, sizeof filename, "%s", optarg );
	    break;
	case 'q':   
	    /* query string */
	    qstr = malloc( strlen( optarg ) + 2 );
	    snprintf( qstr, (strlen( optarg ) + 2), "%s", optarg );
	    break;
	case 'c':   
	    /* Number of rows between traces and flushing. */
	    display_count = atol( optarg );
	    break;
	case 'i':
	    /* Use an insert cursor? */
	    use_ins_curs = 1;
	    break;
	case 'C':
	    /* Enable compatibility mode. */
	    compatible=1;
	    break;
	default:
	    usage( argv[0] );
	    exit(2);
	}
    }

    if (bufsz < 0) {
	fprintf( stderr, "Invalid I/O buffersize (%d).\n", bufsz );
	exit(2);
    } else if ((bufsz % 1024)) {
	bufsz += 1024 - (bufsz % 1024);
	fprintf( stderr,
		 "Odd buffersize adjusting to next K (%d)\n",
		 bufsz );
    }

    if (qstr == (char *)0) {
	fprintf( stderr, "-q <SELECT_stmt> option is required!\n" );
	usage( argv[0] );
	exit(3);
    }

    if (use_sync && use_direct) {
	fprintf( stderr, "-S & -I cannot be used together!\n" );
	usage( argv[0] );
	exit(4);
    }

    if (qstr[0] == '@') {
	/* Query file name specified!  Read the query in. */
	long qrysz = 2048;
	char line[2049];
	FILE *qfile;

	qfile = fopen( qstr, "r" );
	qstr = realloc( qstr, qrysz );
	qstr[0] = (char)0;
	while (fgets( line, 2049, qfile )) {
	    while ((long)(strlen( qstr ) + strlen( line ) + 1) >= qrysz) {
		qrysz += 2048;
		qstr = realloc( qstr, qrysz );
	    }
	    strcat( qstr, line );
	}
	fclose( qfile );
    }

    if (strlen( qstr ) == 0) {
	fprintf( stderr, "SQL SELECT or INSERT statement required!\n" );
	usage( argv[0] );
	exit(2);
    }

    if (verbose) 
	fprintf( stderr, "%s data with: %s;\n", load_flag ? "Inserting" : "Selecting", qstr );

    if (!unload_flag && !load_flag) {
	fprintf( stderr, "Must supply either -u (unload) or -l (load)!\n" );
	usage( argv[0] );
	exit(2);
    }

    if (maxsize == (unsigned long long)0 && maxrecs == 0L)
	maxsize = (unsigned long long)MAXFILESIZE;

    if (strlen( dbname ) == 0) {
	fprintf( stderr, "Must supply database (-d)!\n" );
	usage( argv[0] );
	exit(2);
    }

    if (strlen( filename ) == 0) {
	fprintf( stderr, "Must supply file (-f%s)!\n", filename );
	usage( argv[0] );
	exit(2);
    }

    if (display_count <= 0)
	display_count = 1000;

    /* Use the status code returned from the DATABASE
     * statement to check if the database has transactions.
     */

    EXEC SQL WHENEVER ERROR CONTINUE;

    EXEC SQL DATABASE :dbname;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Cannot open database.  Code: %d, ISAM: %d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	return 19;
    }
    has_trans = (sqlca.sqlwarn.sqlwarn1 == 'W');
    is_ansi   = (sqlca.sqlwarn.sqlwarn2 == 'W');

    if ((!has_trans || is_ansi) && no_replicate) {
	fprintf( stderr,
		 "Transactions without replication (-R) specified but database %s %s.  Ignoring.\n",
		 dbname,
		 is_ansi ? "is ANSI mode." : "does not support transactions" );
	no_replicate = 0;
    }

    if (unload_flag)
	ret = unload( qstr, filename, argv[0] );
    else if (load_flag)
	ret = load( qstr, filename, argv[0] );
          
    return ret;
}

/* ---------------------------------------------------------------------- */
int unload( char *slct, char *fname, char *prog)
{
    register int pos;
    register char *cp;
    register int len;
    register int i;
    struct sqlda *udesc;
    struct sqlvar_struct *col;
    FILE *unlfile;
    char delim = '|';
    char *buffer = NULL;
    long long  counter = 0, filerecs = 0, ret = 0, align;
    char fname2[1025];
    static int fnum = 0;
    struct rlimit rlp;
    int nblobs = 0, flen, rlen, outbsize = 0, blobn;
    int *bbufs[1024];
    char *outbuff, *iptr, *optr, str[100];
    long long filebytes = (long long)0;
    short *indicators, *qualifiers;
    intrvl_t *ivp;
    EXEC SQL BEGIN DECLARE SECTION;
    char slctstmt[70000];
    loc_t *locators;
    EXEC SQL END DECLARE SECTION;

    snprintf( slctstmt, sizeof slctstmt, "%s", slct );

    EXEC SQL PREPARE usqlobj FROM :slctstmt;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error: %d(%d) on prepare.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	switch (sqlca.sqlcode) {
	case -200:  /* Identifier too long. */
	case -201:  /* Syntax error, offer some help. */
	case -202:  /* Illegal character. */
	case -203:  /* Illegal integer. */
	case -204:  /* Illegal floating point value. */
	case -206:  /* Table not found. */
	case -217:  /* Column not in any table in the query. */
	case -220:  /* Missing FROM clause. */
	case -236:  /* Number of columns in INSERT does not match VALUES */
	case -253:  /* Identifier too long. */
	case -282:  /* Unmatched quote. */
	case -292:  /* Unnamed column in INSERT table does not accept NULL */
	case -294:  /* Column not in GROUP BY */
	case -309:  /* ORDER BY column not selected */
	case -324:  /* Ambiguous column */
	    fprintf( stderr,
		     "SQL: %s;\n",
		     slctstmt );
	    fprintf( stderr,
		     "Problem found at offset: %d, text msg: %s.\n",
		     sqlca.sqlerrd[4],
		     sqlca.sqlerrm );
	    break;
	default:
	    break;
	}
	return 1;
    }

    EXEC SQL DESCRIBE usqlobj INTO udesc;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error: %d(%d) on describe.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	return 2;
    }
    if (strcmp( fname, "-" ) == 0)
	unlfile = stdout;
    else
	unlfile = fopen(fname,"w");
    if (unlfile == NULL) {
	fprintf( stderr,
		 "Cannot open initial output file '%s', errno=%d\n",
		 fname,
		 errno );
	return 3;
    }
    if (no_buf)
	setvbuf( unlfile, (char *)NULL, _IONBF, 0 );
    else if (bufsz)
	setvbuf( unlfile, (char *)NULL, _IOFBF, bufsz );

    if (use_sync) {
	fcntl( fileno( unlfile ), O_SYNC, 0 );
    } else if (use_direct) {
	fcntl( fileno( unlfile ), O_DIRECT, 0 );
    }

    /* Step 1: analyze udesc to determine type and memory requirements
     *         of each item in the select list.  rtypalign() returns a
     *         pointer to the next appropriate boundary (starting at
     *         pos) for the indicated data type.
     */
    pos = 0;

    for (col = udesc->sqlvar, i = 0; i < udesc->sqld; col++, i++) {
	long fld_len = col->sqllen;

	switch(col->sqltype) {
	case SQLSMFLOAT:
	    col->sqltype = CFLOATTYPE;
	    fld_len = col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    break;

	case SQLFLOAT:
	    col->sqltype = CDOUBLETYPE;
	    fld_len = col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    if (compatible && (pos % 8))
		/* In compatibility mode double is aligned on 8byte boundary */
		fld_len = col->sqllen = 8 + 4;
	    break;

	case SQLMONEY:
	case SQLDECIMAL:
	    col->sqltype = CDECIMALTYPE;
	    fld_len = col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    break;

	case SQLSERIAL:
	case SQLINT:
	    col->sqltype = CINTTYPE;
	    fld_len = col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    break;

#if defined( SQLINT8 )
	case SQLSERIAL8:
	case SQLINT8:
	    col->sqltype = SQLINT8;
	    fld_len = col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    break;
#endif
            
#if defined( SQLINFXBIGINT )
	case SQLBIGSERIAL:
	case SQLINFXBIGINT:
	    col->sqltype = CBIGINTTYPE;
	    fld_len = col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    break;
#endif
            
	case SQLSMINT:
	    col->sqltype = CSHORTTYPE;
	    if (compatible) {
		/* In compatibility mode smallint is aligned on 4byte boundary
		   This will allocate enough room in the buffer for this. */
		if (pos % 4)
		    fld_len = col->sqllen = 4;
	    } else {
		fld_len = col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    }
	    break;

	case SQLVCHAR:
	case SQLNVCHAR:
	    if (fixvchar)
		col->sqltype = CFIXCHARTYPE; /* get all bytes */
	    else
		col->sqltype = CSTRINGTYPE; /* ignore virtual trailing spaces */
	    fld_len = col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    break;

	case SQLBOOL:
	    col->sqltype = CFIXCHARTYPE; /* get all bytes */
	    fld_len = col->sqllen =  1;
	    break;

	case SQLCHAR:
	case SQLNCHAR:
	    col->sqltype = CFIXCHARTYPE; /* get all bytes */
	    fld_len = col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    break;

#if defined( SQLLVARCHAR )
	case SQLLVARCHAR:
	    if (FixLVChar == 1) {
		col->sqltype = CFIXCHARTYPE; /* get all bytes */
		fld_len = col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    } else {
		col->sqllen += 1;  /* Allow for the NULL! */
		col->sqltype = CLVCHARTYPE;
		fld_len = col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    }
	    break;
#endif

	case SQLDATE:
	    col->sqltype = CINTTYPE;    
	    fld_len = col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    break;

	case SQLINTERVAL:
	    if (DTascii) {
		col->sqltype = CFIXCHARTYPE;
		fld_len = col->sqllen = rtypmsize( col->sqltype, 25 );
	    } else {
		col->sqltype = CINVTYPE;
		fld_len = rtypmsize( col->sqltype, col->sqllen );
	    }
	    break;

	case SQLDTIME:
	    if (DTascii) {
		col->sqltype = CFIXCHARTYPE;
		fld_len = col->sqllen = rtypmsize( col->sqltype, 26 );
	    } else {
		col->sqltype = CDTIMETYPE;
		fld_len = rtypmsize( col->sqltype, col->sqllen );
	    }
	    break;

	case SQLBYTES:
	case SQLTEXT:
	    col->sqltype = CLOCATORTYPE;
	    fld_len = col->sqllen = sizeof (loc_t);
	    break;

	default:
	    /*  Other
		data types.  Do nothing if we see
		an unsupported type. */
	    fprintf( stderr, "Unsupported data type: %d, column #%d \n",
		     col->sqltype,
		     i+1 );
	    return 4;
	}
	pos = (int) rtypalign( pos, col->sqltype ) + fld_len;
    }

    /* Step 2: Allocate memory to receive a row of the table returned
     *         by the SELECT statement.  The variable pos has an integer
     *         value equal to the number of bytes occupied by the row.
     */
    fprintf( stderr, "Buffer length = %d.\n", pos );
    buffer = (char *)malloc( pos );
    if (buffer == (char *)NULL) {
	fprintf(stderr,"Out of memory\n");
	return 5;
    }
    indicators = (short *)malloc( udesc->sqld * sizeof (short) );
    memset( indicators, 0, udesc->sqld * sizeof (short) );
    qualifiers = (short *)malloc( udesc->sqld * sizeof (short) );
    memset( qualifiers, 0, udesc->sqld * sizeof (short) );

    /* Initialize buffer to Nulls */
    memset( buffer, 0, pos );

    /* Step 3: Set pointers in the allocated memory to receive each
     *         item in the select list.
     */
    pos = 0;
    cp = buffer;
    rlen = 0;
    for (col = udesc->sqlvar, i = 0; i < udesc->sqld; col++, i++) {
	switch(col->sqltype) {
	case CLOCATORTYPE:
	    locators = (loc_t *)realloc( locators, (nblobs+1) * sizeof (loc_t) );
	    align = (int) rtypalign( (int)(cp - buffer), col->sqltype );
	    cp = buffer + align;
	    col->sqldata = (char *)&(locators[nblobs]);
	    cp += col->sqllen;
	    rlen += 4; /* Only need room in output row for BLOB length. */
	    locators[nblobs].loc_loctype = LOCUSER;
	    locators[nblobs].loc_bufsize = 4096;
	    bbufs[nblobs] = (int *)malloc( 4096 );
	    if (nblobs > 1023) {
		fprintf( stderr, ">1023 blob columns in record\n" );
		exit(9);
	    }
	    locators[nblobs].loc_buffer = (char *)bbufs[nblobs];
	    locators[nblobs].loc_write = writeit;
	    locators[nblobs].loc_open = openit;
	    locators[nblobs].loc_close = closeit;
	    nblobs++;
	    break;
	case CDOUBLETYPE:
	    if (compatible) {
		align = (int) rtypalign( (int)(cp - buffer), CDOUBLETYPE );
		col->sqllen = 8;
		if (align % 8)
		    cp = buffer + align + 4;
		else
		    cp = buffer + align;
		col->sqldata = cp;
		cp += col->sqllen;
	    } else {
		align = (int) rtypalign( (int)(cp - buffer), col->sqltype );
		cp = buffer + align;
		col->sqldata = cp;
		cp += col->sqllen;
		rlen += col->sqllen;
	    }
	    break;
	case CSHORTTYPE:
	    if (compatible) {
		align = (int) rtypalign( (int)(cp - buffer), CINTTYPE );
		col->sqllen = 2;
		if (align % 4)
		    cp = buffer + align + 2;
		else
		    cp = buffer + align;
		col->sqldata = cp;
		cp += col->sqllen;
		rlen += col->sqllen;
	    } else {
#if !defined( NOFIX_DOUBLE )
		/* Missing code.  May break old files.  If so:
		   Compile with '-DNOFIX_DOUBLE' */
		align = (int) rtypalign( (int)(cp - buffer), col->sqltype );
		cp = buffer + align;
		col->sqldata = cp;
		cp += col->sqllen;
		rlen += col->sqllen;
#endif
	    }
	    break;
	case CDTIMETYPE:
	    align = (int) rtypalign( (int)(cp - buffer), col->sqltype );
	    cp = buffer + align;
	    col->sqldata = cp;
	    ((dtime_t *)(col->sqldata))->dt_qual =
		TU_ENCODE(TU_LEN(col->sqllen),
			  TU_START(col->sqllen),
			  TU_END(col->sqllen) );
	    /* Save the qualifier for later */
	    qualifiers[i] = htons(((dtime_t *)(col->sqldata))->dt_qual);
	    col->sqllen = ((dtime_t *)(col->sqldata))->dt_qual;
	    cp += (flen = rtypmsize( col->sqltype, col->sqllen ));
	    rlen += flen;
	    break;
	case CINVTYPE:
	    align = (int) rtypalign( (int)(cp - buffer), col->sqltype );
	    cp = buffer + align;
	    col->sqldata = cp;
	    ((intrvl_t *)(col->sqldata))->in_qual =
		TU_IENCODE(TU_LEN(col->sqllen),
			   TU_START(col->sqllen),
			   TU_END(col->sqllen) );
	    /* Save the qualifier for later */
	    qualifiers[i] = htons(((intrvl_t *)(col->sqldata))->in_qual);
	    col->sqllen = ((intrvl_t *)(col->sqldata))->in_qual;
	    cp += (flen = rtypmsize( col->sqltype, col->sqllen ));
	    rlen += flen;
	    break;
	default:
	    align = (int) rtypalign( (int)(cp - buffer), col->sqltype );
	    cp = buffer + align;
	    col->sqldata = cp;
	    cp += col->sqllen;
	    rlen += col->sqllen;
	    break;
	}
	col->sqlitype = CSHORTTYPE;
	col->sqlilen  = sizeof (short);
	col->sqlidata = (char *)indicators;
	col->sqlind = indicators++;
    }
    pos = (long)cp - (long)buffer;

    if (nblobs) {
	fprintf( stderr,
		 "Length of fixed portion of output records = %d\n",
		 rlen );
    } else {
	fprintf( stderr, "Length of output records = %d.\n", pos );
	if (!maxrecs)
	    maxrecs = ((long)maxsize / pos);
	else
	    maxsize = ((unsigned long long)(maxrecs * pos));
    }

    getrlimit( RLIMIT_FSIZE, &rlp );
    if ((unsigned long long)maxsize > (unsigned long long)rlp.rlim_cur) {
	fprintf( stderr, "Maximum number of rows (-r) cannot be written!\n" );
	fprintf( stderr,
		 "Request exceeds user's ulimit (%llu), adjusting maximum.\n",
		 (unsigned long long)rlp.rlim_cur );
	maxrecs = (long)((unsigned long long)rlp.rlim_cur /
			 (unsigned long long)pos);
	maxsize = (unsigned long long)rlp.rlim_cur;
    }

    fprintf( stderr, "Maximum rows per file = %ld.\n", maxrecs );
    EXEC SQL WHENEVER ERROR CONTINUE;

    /* Step 4: Fetch each row of the query and write to the output file.
     */
    EXEC SQL DECLARE usqlcurs CURSOR FOR usqlobj;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Error declaring query cursor!  Code=%d, ISAM=%d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	ret = 17;
    }

    if (has_trans && !is_ansi)
	EXEC SQL BEGIN WORK;

    EXEC SQL OPEN usqlcurs;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Error opening query cursor!  Code=%d, ISAM=%d.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	ret = 18;
    }

    sqlca.sqlcode = 0;
    col = udesc->sqlvar;
    while (sqlca.sqlcode == 0) {

	bpos = 0;
	global_length = 0;
	/* Initialize buffer to Nulls */
	/* memset( buffer, 0, pos ); */

	for (col = udesc->sqlvar, i = 0; i < udesc->sqld; col++, i++) {
	    switch(col->sqltype) {
	    case CDTIMETYPE:
		/* Get the saved qualifier. */
		if (NoSwab)
		    ((dtime_t *)(col->sqldata))->dt_qual = qualifiers[i];
		else
		    ((dtime_t *)(col->sqldata))->dt_qual = htons(qualifiers[i]);
		break;
	    case CINVTYPE:
		/* Get the saved qualifier. */
		if (NoSwab)
		    ((intrvl_t *)(col->sqldata))->in_qual = qualifiers[i];
		else
		    ((intrvl_t *)(col->sqldata))->in_qual = htons(qualifiers[i]);
		break;
	    default:
		break;
	    }
	}
	EXEC SQL FETCH usqlcurs USING DESCRIPTOR udesc;

	if (sqlca.sqlcode == -452) {
	    fprintf( stderr,
		     "Error opening BLOB.  Invalid length suspected.\n" );
	} else if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Error fetching row!  Code=%d, ISAM=%d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    ret = 6;
	    break;
	} else if (sqlca.sqlcode == 100)
	    break;

	for (col = udesc->sqlvar, i = 0; i < udesc->sqld; col++, i++) {
	    if (*col->sqlind == -1) {
		/* Column is NULL.  Put out recognizable NULL value. */
		switch(col->sqltype) {
		case CCHARTYPE:
		case CSTRINGTYPE:
		case CFIXCHARTYPE:
		    if (compatible) 
			memset( col->sqldata, 0xFF, col->sqllen );
		    else
			memset( col->sqldata, 0, col->sqllen );
		    break;
		case CSHORTTYPE:
		    if (NoSwab)
			*(short *)col->sqldata = SHRT_MIN;
		    else
			*(short *)col->sqldata = htons(SHRT_MIN);
		    break;
		case CINTTYPE:
		case CLONGTYPE:
		case CDATETYPE:
		    if (NoSwab)
			*(int *)col->sqldata = INT_MIN;
		    else
			*(int *)col->sqldata = htonl(INT_MIN);
		    break;
#if defined( CBIGINTTYPE )
		case CBIGINTTYPE:
		    if (NoSwab) {
			((int *)col->sqldata)[0] = 0xffffffff;
			((int *)col->sqldata)[1] = 0xffffffff;
		    } else {
			((int *)col->sqldata)[0] = htonl(0xffffffff);
			((int *)col->sqldata)[1] = htonl(0xffffffff);
		    }
		    break;
#endif
#if defined( SQLINT8 )
		case SQLINT8:
		    memset( (char *)col->sqldata, 0xff, 10 );
		    break;
#endif
		case CFLOATTYPE:
		    if (NoSwab)
			*(int *)col->sqldata = 0x7fffffff;
		    else
			*(int *)col->sqldata = htonl(0x7fffffff);
		    break;
		case CDOUBLETYPE:
		    if (NoSwab) {
			((int *)col->sqldata)[0] = 0x7fffffff;
			((int *)col->sqldata)[1] = 0xffffffff;
		    } else {
			((int *)col->sqldata)[0] = htonl(0x7fffffff);
			((int *)col->sqldata)[1] = htonl(0xffffffff);
		    }
		    break;
		case CDECIMALTYPE:
		case CMONEYTYPE:
		    (*(dec_t *)col->sqldata).dec_pos = DECPOSNULL;
		    break;
		case CDTIMETYPE:
		    (*(dtime_t *)col->sqldata).dt_dec.dec_pos = DECPOSNULL;
		    break;
		case CINVTYPE:
		    (*(intrvl_t *)col->sqldata).in_dec.dec_pos = DECPOSNULL;
		    break;
		}
	    } else {
		int templ, tmplp1, tmplp2;
		short temps;
		float tempf;
		double tempd;

		switch(col->sqltype) {
		case CINTTYPE:
		case CLONGTYPE:
		    /* Swab data to network format. */
		    if (!NoSwab) {
			templ = htonl( *(unsigned int *)(col->sqldata) );
			*(int *)(col->sqldata) = templ;
		    }
		    break;
		case CSHORTTYPE:
		    /* Swab data to network format. */
		    if (!NoSwab) {
			temps = htons( *(unsigned short *)(col->sqldata) );
			*(short *)(col->sqldata) = temps;
		    }
		    break;
#if defined( CBIGINTTYPE )
		case CBIGINTTYPE:
		    /* Swab data to network format, swab words if needed. */
		    if (!NoSwab) {
			if (EndianNess == LITTLE_ENDIAN) {
			    tmplp1 = htonl(((unsigned int *)(col->sqldata))[1]);
			    tmplp2 = htonl(((unsigned int *)(col->sqldata))[0]);
			} else {
			    tmplp1 = htonl(((unsigned int *)(col->sqldata))[0]);
			    tmplp2 = htonl(((unsigned int *)(col->sqldata))[1]);
			}
			((int *)(col->sqldata))[0] = tmplp1;
			((int *)(col->sqldata))[1] = tmplp2;
		    }
		    break;
#endif
		case CFLOATTYPE:
		    if (!NoSwab) {
			/* Swab data to network format. */
			tempf = (float)htonl( *(unsigned int *)(col->sqldata) );
			*(float *)(col->sqldata) = tempf;
		    }
		    break;
		case CDECIMALTYPE:
		case CMONEYTYPE:
		    if (!NoSwab) {
			dec_t *dec = (dec_t *)(col->sqldata);

			/* Swab binary fields in the dec_t structure to NBO. */
			dec->dec_exp = htons( dec->dec_exp );
			dec->dec_pos = htons( dec->dec_pos );
			dec->dec_ndgts = htons( dec->dec_ndgts );
			break;
		    }
		case CDOUBLETYPE:
		    if (!NoSwab) {
			/* Swab data to network format, swab words if needed. */
			if (EndianNess == LITTLE_ENDIAN) {
			    tmplp1 = htonl(((unsigned int *)(col->sqldata))[1]);
			    tmplp2 = htonl(((unsigned int *)(col->sqldata))[0]);
			} else {
			    tmplp1 = htonl(((unsigned int *)(col->sqldata))[0]);
			    tmplp2 = htonl(((unsigned int *)(col->sqldata))[1]);
			}
			((int *)(col->sqldata))[0] = tmplp1;
			((int *)(col->sqldata))[1] = tmplp2;
		    }
		    break;
		case CDTIMETYPE:
		    if (!NoSwab) {
			dtime_t *dtp = (dtime_t *)(col->sqldata);

			/* Get the saved netorder qualifier. */
			/* Swab binary fields in the dec_t structure to NBO. */
			dtp->dt_qual = qualifiers[i];
			dtp->dt_dec.dec_exp = htons( dtp->dt_dec.dec_exp );
			dtp->dt_dec.dec_pos = htons( dtp->dt_dec.dec_pos );
			dtp->dt_dec.dec_ndgts = htons( dtp->dt_dec.dec_ndgts );
			break;
		    }
		case CINVTYPE:
		    if (!NoSwab) {
			intrvl_t *itvl = (intrvl_t *)(col->sqldata);

			/* Get the saved netorder qualifier. */
			itvl->in_qual = qualifiers[i];
			itvl->in_dec.dec_exp = htons( itvl->in_dec.dec_exp );
			itvl->in_dec.dec_pos = htons( itvl->in_dec.dec_pos );
			itvl->in_dec.dec_ndgts = htons( itvl->in_dec.dec_ndgts );
			break;
		    }
		}
	    }
	}
	if (nblobs) {
	    if ((unsigned long long)(filebytes + rlen + global_length) > maxsize)
		{
		    fclose( unlfile );
		    sprintf( fname2, "%s.%d", fname, ++fnum );
		    fprintf( stderr,
			     "File #%d is full, opening next file <%s>\n",
			     fnum-1,
			     fname2 );
		    unlfile = fopen( fname2, "w" );
		    if (unlfile == NULL) {
			fprintf( stderr,
				 "Cannot open next output file '%s', errno = %d\n",
				 fname,
				 errno );
			return 7;
		    }
		    if(no_buf)
			setvbuf( unlfile, (char *)NULL, _IONBF, 0 );
		    else if (bufsz)
			setvbuf( unlfile, (char *)NULL, _IOFBF, bufsz );

		    if (use_sync) {
			fcntl( fileno( unlfile ), O_SYNC, 0 );
		    } else if (use_direct) {
			fcntl( fileno( unlfile ), O_DIRECT, 0 );
		    }
		    filerecs = 0;
		    filebytes = 0;
		}
	    rlen = (rlen + global_length);
	    if (outbsize < rlen) {
		outbsize = ((rlen + 1023) / 1024) * 1024;
		outbuff = (char *)realloc( outbuff, outbsize );
	    }
	    filebytes += rlen;
	    blobn = 0;
	    /* Build output buffer. */
	    iptr = udesc->sqlvar->sqldata;
	    optr = outbuff;
	    for (col = udesc->sqlvar, i = 0; i < udesc->sqld; col++, i++) {
		switch(col->sqltype) {
		case CLOCATORTYPE:
		    memcpy( optr, iptr, (size_t)(col->sqldata - iptr) );
		    optr += (int)(col->sqldata - iptr);
		    memcpy( optr, (char *)&locators[blobn].loc_size, 4 );
		    optr += 4;
		    iptr = col->sqldata + col->sqllen;
		    blobn++;
		    break;
		default:
		    break;
		}
	    }
	    col--;
	    memcpy( optr, iptr, (size_t)(col->sqldata + col->sqllen - iptr) );
	    optr += (long)(col->sqldata - iptr);
	    memcpy( optr, blobbuff, global_length );

	    /* Write out the buffer. */
	    fwrite( outbuff, rlen, 1, unlfile );
	} else {
	    if (filerecs >= maxrecs) {
		fclose( unlfile );
		sprintf( fname2, "%s.%d", fname, ++fnum );
		fprintf( stderr,
			 "File #%d is full, opening next file <%s>\n",
			 fnum-1,
			 fname2 );
		unlfile = fopen( fname2, "w" );
		if (unlfile == NULL) {
		    fprintf( stderr,
			     "Cannot open next output file '%s', errno = %d\n",
			     fname,
			     errno );
		    return 7;
		}
		if(no_buf)
		    setvbuf( unlfile, (char *)NULL, _IONBF, 0 );
		else if (bufsz)
		    setvbuf( unlfile, (char *)NULL, _IOFBF, bufsz );

		if (use_sync) {
		    fcntl( fileno( unlfile ), O_SYNC, 0 );
		} else if (use_direct) {
		    fcntl( fileno( unlfile ), O_DIRECT, 0 );
		}
		filerecs = 0;
	    }
	    fwrite( udesc->sqlvar->sqldata, pos, 1, unlfile );
	}
	counter++;
	filerecs++;
	if ((counter % display_count) == 0) {
	    fprintf( stderr, "Unloading: %lld\n", counter );
	}
    } /* end fetch loop */

    fprintf( stderr, "Unloaded: %lld records.\n", counter );

    fclose( unlfile );
    if (sqlca.sqlcode && sqlca.sqlcode != SQLNOTFOUND) {
	fprintf( stderr,
		 "Cursor close failed sqlcode = %d, ISAM: %d\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
    }

    free( buffer );
    if (sqlca.sqlcode && sqlca.sqlcode != SQLNOTFOUND) {
	fprintf( stderr,
		 "Cursor free failed sqlcode = %d, ISAM: %d\n", 
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
    }
        
    EXEC SQL CLOSE usqlcurs;
    if (has_trans)
	EXEC SQL COMMIT WORK;
    return ret;
}

/* -------------------------------------------------------------------- */

int load( char *insrtstmt, char *fname, char *prog )
{
    register int pos;
    register char *cp;
    register int len;
    register int i, j;
    struct sqlda *idesc = (struct sqlda *)NULL;
    struct sqlvar_struct *col, *col2;
    FILE *ldfile;
    char delim = '|';
    char *buffer = NULL;
    char file_buff[1001];
    char tstr[100], str[100];
    long long  counter = 0, read_count = 0, ret = 0, align, errrecs = 0;
    intrvl_t *ivp;
    short *indicators, *qualifiers;
    EXEC SQL BEGIN DECLARE SECTION;
    char *stmt, ins_curs[19];
    EXEC SQL END DECLARE SECTION;

    EXEC SQL WHENEVER ERROR CONTINUE;

    strcpy( ins_curs, "ins_curs" );
    stmt = insrtstmt;
    EXEC SQL PREPARE insrt FROM :stmt;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "Error: %d(%d) on insert prepare.\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
	switch (sqlca.sqlcode) {
	case -200:  /* Identifier too long. */
	case -201:  /* Syntax error, offer some help. */
	case -202:  /* Illegal character. */
	case -203:  /* Illegal integer. */
	case -204:  /* Illegal floating point value. */
	case -206:  /* Table not found. */
	case -217:  /* Column not in any table in the query. */
	case -220:  /* Missing FROM clause. */
	case -236:  /* Number of columns in INSERT does not match VALUES */
	case -253:  /* Identifier too long. */
	case -282:  /* Unmatched quote. */
	case -292:  /* Unnamed column in INSERT table does not accept NULL */
	case -294:  /* Column not in GROUP BY */
	case -309:  /* ORDER BY column not selected */
	case -324:  /* Ambiguous column */
	case -391:  /* Cannot insert a null into column column-name. */
	case -415:  /* Data conversion error.  Generic. */
	case -450:  /* Illegal ESQL locator. */
	case -451:  /* Locator buffer size too small. */
	case -1210: /* Date could not be converted to month/day/year
		       format. */
	case -1260: /* It is not possible to convert between the specified
		       types. */
	    fprintf( stderr,
		     "Problem found at offset: %d, text msg: %s.\n",
		     sqlca.sqlerrd[4],
		     sqlca.sqlerrm );
	    fprintf( stderr,
		     "<%s>\n",
		     insrtstmt );
	    break;
	default:
	    break;
	}
	return 9;
    }

    EXEC SQL DESCRIBE insrt INTO idesc;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr, "Error: %d on insert describe.\n", sqlca.sqlcode );
	return 10;
    } else if (idesc->sqld == 0) {
	fprintf( stderr,
		 "Error: INSERT statment does not contain a VALUES clause!\n"
		 "You MUST include a VALUES clause with a replaceable "
		 "parameter marker (?) for\neach column being inserted, ex: "
		 "INSERT INTO fred VALUES (?,?,?)\n\n" );
	return 21;
    }

    if (use_ins_curs) {
	if (has_trans)
	    EXEC SQL DECLARE :ins_curs CURSOR WITH HOLD FOR insrt;
	else
	    EXEC SQL DECLARE :ins_curs CURSOR FOR insrt;
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Error: %d(%d) on insert cursor declare.\n", 
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    return 11;
	}
        
	EXEC SQL OPEN :ins_curs;
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Error: %d(%d) on insert cursor declare.\n", 
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    return 12;
	}         
    }     

    if (strcmp( fname, "-" ) == 0)
	ldfile = stdin;
    else
	ldfile = fopen( fname, "r" );
    if (ldfile == NULL) {
	fprintf( stderr,
		 "Cannot open input file '%s', errno = %d\n",
		 fname,
		 errno );
	return 13;
    }

    /* Step 1: analyze udesc to determine type and memory requirements
     *         of each item in the select list.  rtypalign() returns a
     *         pointer to the next appropriate boundary (starting at
     *         pos) for the indicated data type.
     */
    pos = 0;

    for (col = idesc->sqlvar, i = 0; i < idesc->sqld; col++, i++) {
	long fld_len = col->sqllen;

	switch(col->sqltype) {
	case SQLSMFLOAT:
	    col->sqltype = CFLOATTYPE;
	    fld_len = col->sqllen = rtypmsize(col->sqltype,col->sqllen);
	    break;

	case SQLFLOAT:
	    col->sqltype = CDOUBLETYPE;
	    fld_len = col->sqllen = rtypmsize(col->sqltype,col->sqllen);
	    if (compatible && (pos % 8))
		/* In compatibility mode double is aligned on 8byte boundary */
		fld_len = col->sqllen = 12;
	    break;

	case SQLMONEY:
	case SQLDECIMAL:
	    col->sqltype = CDECIMALTYPE;
	    fld_len = col->sqllen = rtypmsize(col->sqltype,col->sqllen);
	    break;

	case SQLSERIAL:
	case SQLINT:
	    col->sqltype = CINTTYPE;
	    fld_len = col->sqllen = rtypmsize(col->sqltype,col->sqllen);
	    break;

#if defined( SQLINT8 )
	case SQLSERIAL8:
	case SQLINT8:
	    col->sqltype = SQLINT8;
	    fld_len = col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    break;
#endif

#if defined( SQLINFXBIGINT )
	case SQLBIGSERIAL:
	case SQLINFXBIGINT:
	    col->sqltype = CBIGINTTYPE;
	    fld_len = col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    break;
#endif
            
	case SQLSMINT:
	    col->sqltype = CSHORTTYPE;
	    if (compatible) {
		/* In compatibility mode smallint is aligned on 4byte boundary
		   This will allocate enough room in the buffer for this. */
		if (pos % 4)
		    fld_len = col->sqllen = 4;
	    } else {
		fld_len = col->sqllen = rtypmsize(col->sqltype,col->sqllen);
	    }
	    break;

	case SQLVCHAR:
	case SQLNVCHAR:
	    if (fixvchar)
		col->sqltype = CFIXCHARTYPE; /* get all bytes */
	    else
		col->sqltype = CSTRINGTYPE; /* ignore virtual trailing spaces */
	    fld_len = col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    break;

	case SQLBOOL:
	    col->sqltype = CFIXCHARTYPE; /* get all bytes */
	    fld_len = col->sqllen = 1;
	    break;

	case SQLCHAR:
	case SQLNCHAR:
	    col->sqltype = CFIXCHARTYPE; /* get all bytes */
	    fld_len = col->sqllen = rtypmsize(col->sqltype,col->sqllen);
	    break;

#if defined( SQLLVARCHAR )
	case SQLLVARCHAR:
	    if (FixLVChar == 1) {
		col->sqltype = CFIXCHARTYPE; /* get all bytes */
		fld_len = col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    } else {
		col->sqllen += 1;  /* Allow for the NULL! */
		col->sqltype = CLVCHARTYPE;
		fld_len = col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    }
	    break;
#endif

	case SQLDATE:
	    col->sqltype = CINTTYPE;    
	    fld_len = col->sqllen = rtypmsize(col->sqltype,col->sqllen);
	    break;

	case SQLINTERVAL:
	    if (DTascii) {
		col->sqltype = CFIXCHARTYPE;
		fld_len = col->sqllen = rtypmsize( col->sqltype, 25 );
	    } else {
		col->sqltype = CINVTYPE;
		fld_len = rtypmsize( col->sqltype, col->sqllen );
	    }
	    break;

	case SQLDTIME:
	    if (DTascii) {
		col->sqltype = CFIXCHARTYPE;
		fld_len = col->sqllen = rtypmsize(col->sqltype, 26 );
	    } else {
		col->sqltype = CDTIMETYPE;
		fld_len = rtypmsize(col->sqltype,col->sqllen);
	    }
	    break;

	case SQLBYTES:
	case SQLTEXT:
	    col->sqltype = CLOCATORTYPE;
	    fld_len = col->sqllen = sizeof (loc_t);
	    break;

	default:
	    /*  Other
		data types.  Do nothing if we see
		an unsupported type. */
	    fprintf( stderr,
		     "Unsupported data type: %d, column #%d \n",
		     col->sqltype,
		     i+1 );		             
	    return 13;
	}
	pos = (int) rtypalign( pos, col->sqltype ) + fld_len;
	col->sqlind = NULL;
    }

    /* Step 2: Allocate memory to receive a row of the table returned
     *         by the SELECT statement.  The variable pos has an integer
     *         value equal to the number of bytes occupied by the row.
     */
    fprintf( stderr, "Buffer length = %d.\n", pos );
    buffer = malloc( pos );
    if (buffer == NULL) {
	fprintf( stderr, "Cannot allocate memory\n" );
	return 14;
    }
    indicators = (short *)malloc( idesc->sqld * sizeof (short) );
    memset( indicators, 0, idesc->sqld * sizeof (short) );
    qualifiers = (short *)malloc( idesc->sqld * sizeof (short) );
    memset( qualifiers, 0, idesc->sqld * sizeof (short) );

    /* Initialize buffer to Nulls */
    memset( buffer, 0, pos );

    /* Step 3: Set pointers in the allocated memory to receive each
     *         item in the select list.
     */
    pos = 0;
    cp = buffer;
    for (col = idesc->sqlvar, i = 0; i < idesc->sqld; col++, i++) {
	switch(col->sqltype) {
	case CSHORTTYPE:
	    if (compatible) {
		align = (int) rtypalign( (int)(cp - buffer), CINTTYPE );
		col->sqllen = 2;
		if (align % 4)
		    cp = buffer + align + 2;
		else
		    cp = buffer + align;
		col->sqldata = cp;
		cp += col->sqllen;
	    } else {
		align = (int) rtypalign( (int)(cp - buffer), col->sqltype );
		cp = buffer + align;
		col->sqldata = cp;
		cp += col->sqllen;
	    }
	    break;
	case CDOUBLETYPE:
	    if (compatible) {
		align = (int) rtypalign( (int)(cp - buffer), CDOUBLETYPE );
		col->sqllen = 8;
		if (align % 8)
		    cp = buffer + align + 4;
		else
		    cp = buffer + align;
		col->sqldata = cp;
		cp += col->sqllen;
	    } else {
		align = (int) rtypalign( (int)(cp - buffer), col->sqltype );
		cp = buffer + align;
		col->sqldata = cp;
		cp += col->sqllen;
	    }
	    break;
	case CDTIMETYPE:
	    align = (int) rtypalign( (int)(cp - buffer), col->sqltype );
	    cp = buffer + align;
	    col->sqldata = cp;
	    ((dtime_t *)(col->sqldata))->dt_qual =
		TU_ENCODE(TU_LEN(col->sqllen),
			  TU_START(col->sqllen),
			  TU_END(col->sqllen) );
	    /* Save the qualifier for later */
	    qualifiers[i] = ntohs(((dtime_t *)(col->sqldata))->dt_qual);
	    col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    cp += col->sqllen;
	    break;
	case CINVTYPE:
	    align = (int) rtypalign( (int)(cp - buffer), col->sqltype );
	    cp = buffer + align;
	    col->sqldata = cp;
	    ((intrvl_t *)(col->sqldata))->in_qual =
		TU_IENCODE(TU_LEN(col->sqllen),
			   TU_START(col->sqllen),
			   TU_END(col->sqllen) );
	    /* Save the qualifier for later */
	    qualifiers[i] = ntohs(((intrvl_t *)(col->sqldata))->in_qual);
	    col->sqllen = rtypmsize( col->sqltype, col->sqllen );
	    cp += col->sqllen;
	    break;
	default:
	    align = (int) rtypalign( (int)(cp - buffer), col->sqltype );
	    cp = buffer + align;
	    col->sqldata = cp;
	    cp += col->sqllen;
	    break;
	}
	col->sqlitype = CSHORTTYPE;
	col->sqlilen  = sizeof (short);
	col->sqlidata = (char *)indicators;
	col->sqlind = indicators++;
	
    }
    pos = (long)cp - (long)buffer;
    fprintf( stderr, "Input record length assumed: %d.\n", pos );

    /*  Read data and convert all to ASCII format, and
     *  insert into database.
     */

    if (has_trans && !is_ansi)
	if (no_replicate)
	    EXEC SQL BEGIN WORK WITHOUT REPLICATION;
	else
	    EXEC SQL BEGIN WORK;

    /*
     * read the data from the load file.
     */

    fprintf( stderr, "\n" );


    col = idesc->sqlvar;
    while ((read_count = fread( col->sqldata, pos, 1, ldfile )) != 0) {
	unsigned int templ, templ1, templ2;
	unsigned short temps;

	/* Scan data for NULL value indications and flag sqlvar as NULL. */
	for (col2 = idesc->sqlvar, i = 0; i < idesc->sqld; col2++, i++) {
	    switch(col2->sqltype) {
	    case CCHARTYPE:
	    case CSTRINGTYPE:
	    case CFIXCHARTYPE:
		if (!compatible && *col2->sqldata == (char)0) {
		    *col2->sqlind = -1;
		} else if (compatible) {
		    for (j=0; j<col2->sqllen; j++) {
			if (col2->sqldata[j] != (char)0xFF) {
			    *col2->sqlind = 0;
			    break;
			}
		    }
		    memset( col2->sqldata, (char)0, col2->sqllen );
		}
		break;
	    case CSHORTTYPE:
		/* Swab data to host format. */
		temps = ntohs( *(unsigned short *)col2->sqldata );
		*(short *)col2->sqldata = temps;
		if (*(short *)col2->sqldata == SHRT_MIN)
		    *col2->sqlind = -1;
		else
		    *col2->sqlind = 0;
		break;
	    case CINTTYPE:
	    case CLONGTYPE:
	    case CDATETYPE:
		/* Swab data to host format. */
		templ = ntohl( *(unsigned int *)col2->sqldata );
		*(int *)col2->sqldata = templ;
		if (*(int *)col2->sqldata == INT_MIN)
		    *col2->sqlind = -1;
		else
		    *col2->sqlind = 0;
		break;
#if defined( CBIGINTTYPE )
	    case CBIGINTTYPE:
		/* Swab data to host format. */
		templ1 = ntohl( ((unsigned int *)col2->sqldata)[0] );
		templ2 = ntohl( ((unsigned int *)col2->sqldata)[1] );
		if (EndianNess == LITTLE_ENDIAN) {
		    ((unsigned int *)col2->sqldata)[0] = templ2;
		    ((unsigned int *)col2->sqldata)[1] = templ1;
		} else {
		    ((unsigned int *)col2->sqldata)[0] = templ1;
		    ((unsigned int *)col2->sqldata)[1] = templ2;
		}

		if (((unsigned int *)col2->sqldata)[0] == 0xffffffff &&
		    ((unsigned int *)col2->sqldata)[1] == 0xffffffff)
		    *col2->sqlind = -1;
		else
		    *col2->sqlind = 0;
		break;
#endif
#if defined( SQLINT8 )
	    case SQLINT8:
		/* Not swabbing INT8/SERIAL8 */
		if (((unsigned short *)col2->sqldata)[0] == 0xffff &&
		    ((unsigned short *)col2->sqldata)[1] == 0xffff &&
		    ((unsigned short *)col2->sqldata)[2] == 0xffff &&
		    ((unsigned short *)col2->sqldata)[3] == 0xffff &&
		    ((unsigned short *)col2->sqldata)[4] == 0xffff)
		    *col2->sqlind = -1;
		else
		    *col2->sqlind = 0;
		break;
#endif
	    case CFLOATTYPE:
		/* Swab data to host format. */
		templ = ntohl( *(unsigned int *)col2->sqldata );
		*(int *)col2->sqldata = templ;
		if (*(int *)col2->sqldata == 0x7fffffff)
		    *col2->sqlind = -1;
		else
		    *col2->sqlind = 0;
		break;
	    case CDOUBLETYPE:
		/* Swab data to host format. */
		templ1 = ntohl( ((unsigned int *)col2->sqldata)[0] );
		templ2 = ntohl( ((unsigned int *)col2->sqldata)[1] );
		if (EndianNess == LITTLE_ENDIAN) {
		    ((unsigned int *)col2->sqldata)[0] = templ2;
		    ((unsigned int *)col2->sqldata)[1] = templ1;
		} else {
		    ((unsigned int *)col2->sqldata)[0] = templ1;
		    ((unsigned int *)col2->sqldata)[1] = templ2;
		}

		if (((unsigned int *)col2->sqldata)[0] == 0x7fffffff &&
		    ((unsigned int *)col2->sqldata)[1] == 0xffffffff)
		    *col2->sqlind = -1;
		else
		    *col2->sqlind = 0;
		break;
	    case CDECIMALTYPE:
	    case CMONEYTYPE:
		if ((*(dec_t*)col2->sqldata).dec_pos == DECPOSNULL)
		    *col2->sqlind = -1;
		else {
		    dec_t *dec = (dec_t *)(col2->sqldata);

		    /* Swab binary fields to host order. */
		    dec->dec_exp = ntohs( dec->dec_exp );
		    dec->dec_pos = ntohs( dec->dec_pos );
		    dec->dec_ndgts = ntohs( dec->dec_ndgts );
		    *col2->sqlind = 0;
		}
		break;
	    case CDTIMETYPE:
		if ((*(dtime_t *)col2->sqldata).dt_dec.dec_pos == DECPOSNULL) {
		    *col2->sqlind = -1;
		} else {
		    dtime_t *dtp = (dtime_t *)(col2->sqldata);

		    /* Swab binary fields to host order. */
		    dtp->dt_qual = ntohs( dtp->dt_qual );
		    dtp->dt_dec.dec_exp = ntohs( dtp->dt_dec.dec_exp );
		    dtp->dt_dec.dec_pos = ntohs( dtp->dt_dec.dec_pos );
		    dtp->dt_dec.dec_ndgts = ntohs( dtp->dt_dec.dec_ndgts );
		    *col2->sqlind = 0;
		}
		break;
	    case CINVTYPE:
		if ((*(intrvl_t *)col2->sqldata).in_dec.dec_pos == DECPOSNULL)
		    *col2->sqlind = -1;
		else {
		    intrvl_t *intvl = (intrvl_t *)(col2->sqldata);

		    /* Swab binary fields to host order. */
		    intvl->in_qual = ntohs( intvl->in_qual );
		    intvl->in_dec.dec_exp = ntohs( intvl->in_dec.dec_exp );
		    intvl->in_dec.dec_pos = ntohs( intvl->in_dec.dec_pos );
		    intvl->in_dec.dec_ndgts = ntohs( intvl->in_dec.dec_ndgts );
		    *col2->sqlind = 0;
		}
		break;
	    }
	}

	/* 
	 * insert data using descriptor 
	 */
	if (use_ins_curs) 
	    EXEC SQL PUT :ins_curs using descriptor idesc;
	else
	    EXEC SQL EXECUTE insrt using descriptor idesc ;


	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Error: %d, ISAM: %d on insert after %lld rows.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1],
		     counter );
	    fprintf( stderr,
		     "\tSQLErrM is: %s.\n",
		     sqlca.sqlerrm );
	    if (use_ins_curs) {
		/* Need error recovery for PUT cursor, ala dbcopy.  Must cache
                   input records in an array so we can readd those that follow
                   the one that errored out. */
	    }
	    dumprec( idesc );
	    errrecs++;
	} else
	    counter++;
	if ((counter % display_count) == 0) {
	    fprintf( stderr, "Loading: %lld\r", counter );
	    if (use_ins_curs) {
		EXEC SQL FLUSH :ins_curs;
	    }
	    if (has_trans) {
		EXEC SQL COMMIT WORK;
		if (!is_ansi)
		    if (no_replicate)
			EXEC SQL BEGIN WORK WITHOUT REPLICATION;
		    else
			EXEC SQL BEGIN WORK;
	    }
	}
    } /* end while */

    if (use_ins_curs) {
	EXEC SQL FLUSH :ins_curs;
    }
    if (has_trans)
	EXEC SQL COMMIT WORK;

    fclose( ldfile );

    if (use_ins_curs) {
	EXEC SQL CLOSE :ins_curs;
	if (sqlca.sqlcode && sqlca.sqlcode != SQLNOTFOUND) {
	    fprintf( stderr,
		     "Cursor close failed sqlcode = %d, ISAM: %d\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	}
    }

    free( buffer );

    if (use_ins_curs) {
	EXEC SQL FREE :ins_curs;
	if (sqlca.sqlcode && sqlca.sqlcode != SQLNOTFOUND) {
	    fprintf( stderr,
		     "Cursor free failed sqlcode = %d, ISAM: %d\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	}
    }

    EXEC SQL FREE insrt;
    if (sqlca.sqlcode && sqlca.sqlcode != SQLNOTFOUND) {
	fprintf( stderr,
		 "Statement free failed sqlcode = %d, ISAM: %d\n",
		 sqlca.sqlcode,
		 sqlca.sqlerrd[1] );
    }

    if (counter) {
	fprintf( stderr, "\nLoaded %lld records\n" , counter );
    }
    if (errrecs) {
	fprintf( stderr, "Logged %lld records with errors.\n", errrecs );
    }
    fprintf( stderr, "%lld Input records processed.\n", (counter + errrecs) );
	
    return ret;
}

int openit( loc_t *loc, int flags, int bsize )
{
    if ((flags & LOC_WONLY) && loc->loc_size < -1) {
	fprintf( stderr, "Blob fetched with length: %d.\n", loc->loc_size );
	return -1;
    }
    loc->loc_status = 0;
    loc->loc_xfercount = 0L;
    global_length = 0 ;

    return 0;
}

inline long min( long a, long b )
{
    return (a < b) ? a : b;
}

/* BLOB Fetch function. */
int writeit( loc_t *loc, char *buf, int nbytes )
{
    unsigned long toread;

    if ((bpos + nbytes) > bbsize) {
	bbsize += (2 * (nbytes < 1024 ? 1024 : nbytes));
	blobbuff = (char *)realloc( blobbuff, bbsize );
    }
    if (loc->loc_size != -1)
	toread = min( nbytes, (loc->loc_size - loc->loc_xfercount) );
    else
	toread = nbytes;
    memcpy( &blobbuff[bpos], buf, toread );
    bpos += toread;

    loc->loc_xfercount += toread;
    global_length += toread;

    return toread;
}

/* BLOB Insert function. */
int readit( loc_t *loc, char *buf, int nbytes )
{
    unsigned long towrite;

    towrite = min( nbytes, (loc->loc_size - loc->loc_xfercount) );
    memcpy( buf, &blobbuff[bpos], towrite );
    bpos += towrite;

    loc->loc_xfercount += towrite;
    global_length += towrite;

    return towrite;
}

int closeit( loc_t *loc )
{
    loc->loc_status = 0;
    if (loc->loc_oflags & LOC_WONLY) {
	/* Fetching, cleanup. */
	loc->loc_indicator = 0;
	loc->loc_size = loc->loc_xfercount;
    }

    return 0;
}

int strip( char *str, int len )
{
    char *p;
    int alen = len;

    for (p = (str + len - 1); p > str; p--, alen--) {
	if (*p == ' ' || *p == (char)0)
	    *p = (char)0;
	else
	    break;
    }
    return alen;
}

int charcpy( unsigned char *target, unsigned char *source, int len )
{
    int rlen = 0;

    while (len) {
	if (*source == *delim) {
	    *target++ = '\\';
	    *target = *source;
	    target++;
	} else if (*source < ' ' || *source > '~') {
	    /* Non-printable, convert to hex. */
	    target += sprintf( (char *)target, "\\%2.2x", *source );
	} else {
	    *target = *source;
	    target++;
	}
	source++;
	len--;
	rlen++;
    }
    /*    *target++ = *delim; */
    *target = (char)0;
    rlen++;

    return rlen;
}

void dumprec( struct sqlda *ldesc )
{
    static unsigned char *iobuf = (unsigned char *)NULL;
    static int iobsz = 10240, nxt = 1, recurs = 0;
    static char string[1025], tstring[1025];
    loc_t *loc;
    struct sqlvar_struct *col;
    int i, slen = 0, flen = 0, alen;
    static FILE *errlog = (FILE *)NULL;

    if (iobuf == (unsigned char *)NULL) {
	iobuf = (unsigned char *)malloc( iobsz );
	if (!iobuf) {
	    fprintf( stderr,
		     "Cannot allocate memory for blob.  Errno=%d\n",
		     errno );
	    exit( errno );
	}
    }
    if (errlog == (FILE *)NULL) {
        errlog = fopen( errfile, "w" );
	if (errlog == (FILE *)NULL) {
	    fprintf( stderr, "Cannot open error log file!  Errno=%d", errno );
	    exit( errno );
	}
    }
		 
    if (fseek( errlog, 0, SEEK_END ) == -1) {
	fprintf( stderr, "Error seeking end of error log: %d\n", errno );
	fclose( errlog );
	sprintf( string, "%s.%d", errfile, nxt++ );
	fprintf( stderr, "Opening next log: %s\n", string );
	errlog = fopen( string, "w" );
	if (errlog == (FILE *)NULL) {
	    fprintf( stderr,
		     "Cannot open next log file (%d), exiting.\n",
		     errno );
	    fprintf( stderr, "Record in error was:\n%s\n", iobuf );
	    exit( 99 );
	}
    }

    for (col = ldesc->sqlvar, i = 0; i < ldesc->sqld; col++, i++)
	{
	    errno = 0;
	    if (*col->sqlind == -1) {
		flen = 0;
		string[0] = (char)0;
	    } else {
		switch(col->sqltype)
		    {
		    case CFLOATTYPE:
			flen = sprintf( string,
					"%f",
					(double)*(float *)col->sqldata );
			break;

		    case CDOUBLETYPE:
			flen = sprintf( string,
					"%f",
					*(double *)col->sqldata );
			break;

		    case CDECIMALTYPE:
			dectoasc( (dec_t *)col->sqldata, tstring, 1025, -1 );
			strip( tstring, 1024 );
			flen = sprintf( string, "%s", tstring );
			break;

		    case CLONGTYPE:
		    case CINTTYPE:
			flen = sprintf( string, "%d", *(int *)col->sqldata );
			break;
            
		    case CSHORTTYPE:
			flen = sprintf( string, "%hd", *(short *)col->sqldata );
			break;

#if defined( SQLINT8 )
		    case SQLINT8:
			ifx_int8toasc( (ifx_int8_t *)col->sqldata, tstring, 1024 );
			strip( tstring, 1024 );
			flen = sprintf( string, "%s", tstring );
			break;
#endif
#if defined( CBIGINTTYPE )
		    case CBIGINTTYPE:
			flen = sprintf( string, "%lld", *(long long *)col->sqldata );
			break;
#endif

		    case CFIXCHARTYPE:
		    case CCHARTYPE:
		    case CSTRINGTYPE:
			alen = strip( (char *)col->sqldata, col->sqllen );
			flen = charcpy( (unsigned char *)string,
					(unsigned char *)col->sqldata,
					alen );
			break;

		    case CDATETYPE:
			rdatestr( *(int *)col->sqldata, tstring );
			flen = sprintf( string, "%s", tstring );
			break;

		    case CDTIMETYPE:
			dttoasc( (dtime_t *)col->sqldata, tstring );
			flen = sprintf( string, "%s", tstring );
			break;

		    case CINVTYPE:
			intoasc( (intrvl_t *)col->sqldata, tstring );
			flen = sprintf( string, "%s", tstring );
			break;

		    case CLOCATORTYPE:
			loc = (loc_t *)col->sqldata;
			if (iobsz < (3 * loc->loc_size)) {
			    iobsz = 3 * loc->loc_size;
			    iobuf = realloc( iobuf, iobsz );
			}
			flen = charcpy( iobuf,
					(unsigned char *)loc->loc_buffer,
					loc->loc_size );
			if (!fwrite( iobuf, flen, 1, errlog )) {
			    /* Recursively call me, once, so whole rec is output. */
			    if (!recurs) {
				fprintf( stderr,
					 "Error writing to error log: %d at %d\n",
					 errno,
					 __LINE__ );
				fclose( errlog );
				sprintf( string, "%s.%d", errfile, nxt++ );
				fprintf( stderr, "Opening next log: %s\n", errfile );
				errlog = fopen( string, "w" );
				if (errlog == (FILE *)NULL) {
				    fprintf( stderr,
					     "Cannot open next log file (%d), exiting.\n",
					     errno );
				    fprintf( stderr,
					     "Record in error was:\n%s\n",
					     iobuf );
				    exit( 99 );
				}
				recurs = 1;
				dumprec( ldesc );
				recurs = 0;
			    } else {
				fprintf( stderr,
					 "Failed to write to new logfile also!\n" );
				exit(1);
			    }
			    return;
			}
			if (fputs( delim, errlog ) == EOF) {
			    /* Recursively call me, once, so whole rec is output. */
			    if (!recurs) {
				fprintf( stderr,
					 "Error writing to error log: %d at %d\n",
					 errno,
					 __LINE__ );
				fclose( errlog );
				sprintf( string, "%s.%d", errfile, nxt++ );
				fprintf( stderr, "Opening next log: %s\n", errfile );
				errlog = fopen( string, "w" );
				if (errlog == (FILE *)NULL) {
				    fprintf( stderr,
					     "Cannot open next log file (%d), exiting.\n",
					     errno );
				    fprintf( stderr,
					     "Record in error was:\n%s\n",
					     iobuf );
				    exit( 99 );
				}
				recurs = 1;
				dumprec( ldesc );
				recurs = 0;
			    } else {
				fprintf( stderr,
					 "Failed to write to new logfile also!\n" );
				exit(1);
			    }
			    return;
			}
			slen += flen + 1;
			continue;

		    default:
			fprintf( stderr, "dumprec: Invalid data type: %d \n",
				 col->sqltype );		             
			return;
		    }
	    }
	    if (fprintf( errlog, "%s%s", string, delim ) <= 0) {
		/* Recursively call me, once, so whole rec is output. */
		if (!recurs) {
		    fprintf( stderr,
			     "Error writing to error log: %d at %d\n",
			     errno,
			     __LINE__ );
		    fclose( errlog );
		    sprintf( string, "%s.%d", errfile, nxt++ );
		    fprintf( stderr, "Opening next log: %s\n", errfile );
		    errlog = fopen( string, "w" );
		    if (errlog == (FILE *)NULL) {
			fprintf( stderr,
				 "Cannot open next log file (%d), exiting.\n",
				 errno );
			fprintf( stderr, "Record in error was:\n%s\n", iobuf );
			exit( 99 );
		    }
		    recurs = 1;
		    dumprec( ldesc );
		    recurs = 0;
		} else {
		    fprintf( stderr,
			     "Failed to write to new logfile also!\n" );
		    exit(1);
		}
	    }
	    fflush( errlog );
	    slen += flen + 1;
	}
    slen += fprintf( errlog, "\n" );
}


/* ----------------------------------------------------------------------- */


void usage( char *prog )
{

    fprintf( stderr, "\n");
    fprintf( stderr,
	     "%s - unload / load binary images of Informix tables.\n\n",
	     prog );

    fprintf( stderr, "Usage: \n" );	
    fprintf( stderr, "\t%s -V\n-OR-\n", prog );
    fprintf( stderr, 
	     "\t%s -u -d <db_name> -q <SELECT_stmt> -f <file_name> [-c [count]]"
	     " \n\t\t[-s maxsize] [-r maxrows] [-F] [-C] [-B bufsz] [-U] [-D]"
	     "[-S] [-N]\n-OR-\n",
	     prog );
    fprintf( stderr, 
	     "\t%s -l -d <db_name> -q <INSERT_stmt> -f <file_name> [-c [count]] [-R]\n"
	     "\t\t[-i] [-C] [-B bufsz] [-U] [-S] [-F] [-n] [-v] [-N]\n\t\t[-e <error_log_file>]\n\n",
	     prog );

    fprintf( stderr, "\t\t-V - Print version information and exit.\n" );
    fprintf(stderr,
	    "\t\t-R -Do not replicate deletions.  Begins transactions with the WITHOUT \n"
            "\t\t    REPLICATION clause included.\n" );
    fprintf( stderr,
	     "\t\t-B - set I/O buffer size to bufsz "
	     "(default: site dependent)\n" ); 
    fprintf( stderr, "\t\t-U - use unbuffered I/O (default: buffered).\n" );
    fprintf( stderr,
	     "\t\t-S - use synchronous I/O (default: asynchronous).\n" );
    fprintf( stderr,
	     "\t\t-I - use DIRECT I/O to bypass system cache. (default: cached).\n" );
    fprintf( stderr, "\t\t-l - load data from a file. If executed as 'bload' -l is assumed.\n" );
    fprintf( stderr, "\t\t-u - unload data to a file. If executed as 'bunload' -u is assumed.\n" );
    fprintf( stderr, "\t\t-d - database name.\n" );
    fprintf( stderr, "\t\t-q - SQL statement in quotes\n" );
    fprintf( stderr, "\t\t-f - load/unload file name\n" );
    fprintf( stderr, "\t\t-e - records in error log file name.  (Default: ulerr.log)"
	     "\n\t\t    Rows are output in dbload compatible ASCII delimited format.\n" );
    fprintf( stderr, "\t\t-c - display counter (default no counter)\n" );
    fprintf( stderr,
	     "\t\t     if present and no argument displays every 1000 rows.\n" );
    fprintf( stderr,
	     "\t\t-i - use insert cursor, commit after count rows \n"
	     "\t\t    (from -c option).\n" );
    fprintf( stderr, "\t\t-s - maximum size (bytes) for unload files.\n" );
    fprintf( stderr, "\t\t-r - maximum size (rows) for unload files.\n" );
    fprintf( stderr,
	     "\t\t-N - NoSwab mode.  Do not swab bytes to/from network byte order.\n"
	     "\t\t    Files produced with -N can be used as EXTERNAL TABLE files\n"
	     "\t\t    with FORMAT \"INFORMIX\" and %s can read files produced by\n"
	     "\t\t    inserts into such an EXTERNAL TABLE.\n",
	     prog );
    fprintf( stderr,
	     "\t\t-C - Compatibility mode.  Restores old ul60 alignment bug\n"
	     "\t\t    which generated/loaded larger output records for\n"
	     "\t\t    certain column types.\n" );
    fprintf( stderr,
	     "\t\t-F - Fetch, save, and reload LVARCHAR columns as FIXCHAR.  Note\n"
	     "\t\t     that this will cause tables loaded from the saved file to\n"
	     "\t\t     be padded with spaces to the full length of the source\n"
	     "\t\t     table's column.\n" );
    fprintf( stderr,
             "\t\t-n - Fetch, save, and reload VARCHAR columns as STRING.  Note\n"
	     "\t\t     that the default is to use FIXCHAR which will padd out \n"
	     "\t\t     VARCHAR columns with hard spaces to their full length.\n" );
    fprintf( stderr,
	     "\t\t-D - Unload DATETIME & INTERVAL columns in binary.  \n"
	     "\t\t    (Default: unload as text).\n" );
    fprintf( stderr, 
	     "\t\t-v - Verbose mode.  Add trace.\n" );
    fprintf( stderr,
	     "\tNote: It is safe to use %s to export and import binary data \n"
	     "\t      packed into CHAR and VARCHAR columns.  However, while %s \n"
	     "\t      internally uses a fixed length format for these and LVARCHAR data,\n"
	     "\t      LVARCHAR columns cannot be safely loaded with binary data. The \n"
	     "\t      data cannot be reliably extracted in any known format.\n\n",
	     prog,
	     prog);

}  

int DetermineEndianNess()
{
    int testint = 0x04030201, ii, result = 0, mult=1000;
    char *p = (char *)&testint;

    for ( ii=0; ii < 4; ii++) {
	result += mult * p[ii];
	mult /= 10;
    }    

    return result;
}
