/*  
 *   #define NOTIMINGS 1
 * Enable the #define above to eliminate references to UNIX time functions and
 * code
 *
 */
/*
  Program Id: dostats

  Program Description: this program will generate sql to update statistics
  on a table or database;

  Author: Art S. Kagel 

  Creation Date: May 16, 1996

  Version: $Id: dostats.ec,v 1.183 2021/12/21 17:12:40 art Exp $

  Usage:    dostats [-h hostname] -d databasename [-t tablename] [-f filename]
  See usage function for current usage.

  RCS Header
  ----------
  $Header: /home/art/GoogleDrive/Latest/RCS/dostats.ec,v 1.183 2021/12/21 17:12:40 art Exp $

  RCS Log
  -------
  $Log: dostats.ec,v $
  Revision 1.183  2021/12/21 17:12:40  art
  Updated Copyright notice.

  Revision 1.182  2016/06/27 21:20:07  art
  Updated copyright notice.

  Revision 1.181  2013/02/10 22:18:19  art
  Updated Copyright

  Revision 1.180  2011/08/03 17:48:35  art
  Fixed bug when running against an ANSI mode database.  Periods missing from
  some statements between the owner and object names.

  Revision 1.179  2011/05/23 16:04:57  art
  Updated the copyright statement.

  Revision 1.178  2011/05/11 22:13:25  art
  *** empty log message ***

  Revision 1.177  2011/05/11 20:24:40  art
  *** empty log message ***

  Revision 1.176  2011/05/11 19:54:39  art
  Fixed procedure so it will not error out recompiling itself.

  Revision 1.175  2011/05/09 19:25:59  art
  Fixed problem with creating the stored procedure.

  Revision 1.174  2009/10/20 10:59:48  art
  Fixed missing table alias.  Thanks to Richard Harnden for pointing it out.

  Revision 1.173  2009/07/29 05:15:25  art
  Changed a variable name that causes multiple defines on Solaris with -static included.

  Revision 1.172  2009/07/29 04:17:14  art
  Tweaked usage.

  Revision 1.171  2009/07/29 04:13:41  art
  Fixed usage.

  Revision 1.170  2009/07/29 04:08:57  art
  Added new feature --aus-thresholds to set -A, -B, & -Q from the Auto
  Update Statistics threshold settings.

  Revision 1.169  2009/07/29 03:12:59  art
  Improved --one-time.  Routines now delete the dostats_stmts rows for that procedure
  if clear_it is 't'.  Now checks for the ph_group record before trying to insert it
  and reports any errors.

  Revision 1.168  2009/07/27 18:11:07  art
  Due to very large databases causing the creation of a procedure
  that's too large to create, have modified dostats to write the
  commands into a new table added to sysadmin, dostats_stmts,
  The procedure that's created now simply fetches the statements
  from there and executes them immediate.

  Revision 1.167  2009/07/23 16:51:12  art
  Made inclusion of my getopt.h unconditional.  Needed on all platforms except Linux now.

  Revision 1.166  2009/07/19 02:03:18  art
  Improved usage a bit.  Fixed problems with some double quoted strings that
  confused Windows versions of ESQL/C to single quotes.

  Revision 1.165  2009/07/16 15:03:33  art
  Missed one.

  Revision 1.164  2009/07/16 14:57:09  art
  Fixed problem when --procedure names a new procedure.

  Revision 1.163  2009/07/14 18:34:16  art
  Fixed printf format problems in error messages.

  Revision 1.162  2009/07/09 20:20:55  art
  Changes for windows compatibility.

  Revision 1.161  2009/06/25 01:51:29  art
  Fixed typo.

  Revision 1.160  2009/06/24 02:45:10  art
  Added feature to mark a scheduled task as a one-time run.  The procedure
  will disable its schedule record after it completes its run.
  Increased the feature version to 6.00 for the scheduling feature suite!

  Revision 1.159  2009/06/24 01:53:27  art
  Fixed initializers for new scehduling options.

  Revision 1.158  2009/06/24 01:48:34  art
  Fixed missing qualifier for IDS version prior to 9.xx

  Revision 1.157  2009/06/24 01:46:32  art
  Fixed missing qualification for IDS versions prior to 9.xx

  Revision 1.156  2009/06/11 23:30:54  art
  Added --procedure option to allow one to name the procedure that --schedule uses,
  replace an existing scheduled procedure, or just create a procedure to run
  without scheduling.

  Revision 1.155  2009/06/11 16:03:12  art
  Added options (--schedule etc.) to schedule actual run using the sysadmin
  database and a stored procedure.

  Revision 1.154  2009/05/13 20:56:17  art
  Made to work for mixed case table and procedure name arguments when DELIMIDENT is disabled.
  This is needed for some scripting applications where object names are kept in a mixed case
  repository but IDS keeps all object names in lower case without DELIMIDENT.

  Revision 1.153  2009/04/03 15:36:29  art
  FIxed many ANSI mode problems again.  Thanks to Kernoal Stephens for his
  persistence in providing feedback.

  Revision 1.152  2009/04/02 15:25:48  art
  Fixed problem with handling tables with very large numbers of rows.
  However this makes dostats no longer compile on systems with SDK
  versions earlier than 3.50 as they do not understand BIGINT.

  Revision 1.151  2009/02/27 15:30:05  art
  Typo

  Revision 1.150  2009/02/27 15:13:45  art
  Updated copyright.
  Fixed support for versions prior to 7.24 (I think) that don't support 
  DBINFO('version',..

  Revision 1.149  2009/02/11 18:12:53  art
  Fixed reference to sysroutinelangs.

  Revision 1.148  2008/12/10 07:05:35  art
  Updated copyright message.

  Revision 1.147  2008/12/10 07:03:14  art
  Fixed problem with attempting to process C & Java UDRs.
  Fixed problem with missing semi-colons for COMMIT WORK statements in ANSI mode
  DBs. Fixed problem with alignment of table owners for ANSI mode DBs.

  Revision 1.146  2008/06/17 02:30:30  art
  Fixed problem with not writing COMMIT WORK to the output file for ANSI
  databases when not processing commands immediately.

  Revision 1.145  2008/05/04 19:36:49  art
  Fixed bug in -n.

  Revision 1.144  2008/04/25 14:12:08  art
  Added new option -n errlimit to ignore errors during processing.

  Revision 1.143  2008/04/16 18:12:46  art
  Fixed SET LOCK MODE statements - again.

  Revision 1.142  2008/04/06 04:21:33  art
  Checking in latest.  1.137-1.141 missing.

  Revision 1.142  2007/10/22 15:12:45  kagel
  typos

  Revision 1.141  2007/10/22 15:08:37  kagel
  Can't use host var in SET LOCK MODE.

  Revision 1.140  2007/10/22 15:01:24  kagel
  Added -w option to specify lock mode wait behavior.

  Revision 1.139  2007/09/07 16:20:21  kagel
  Thanks to John Carlson for pointing out the problem with ' systdistold' 
  handling in servers upgraded to 11.10 from earlier releases.

  Revision 1.138  2007/07/31 21:38:50  kagel
  Added code submitted by Douglas Palmer to eliminate some IDS 11 procs
  from update stats.  Still need to validate the need for this though.

  Revision 1.137  2007/05/29 17:33:25  kagel
  Fixed typo in DBUPSPACE warning.  Thanks to Alinda Watson for pointing this
  out.

  Revision 1.136  2007/03/20 14:41:24  kagel
  More code cleanup.

  Revision 1.135  2007/03/19 18:07:44  kagel
  Removed extraneous space from some MEDIUM commands output.

  Revision 1.132  2007/03/19 17:30:01  kagel
  Fixed optimized server detection.  Thanks to Richard Harnden for
  identifying this bug and suggesting a fix.

  Revision 1.131  2007/03/14 19:43:24  kagel
  Added support for the IDS 11 SAMPLING SIZE option.

  Revision 1.130  2007/02/22 18:13:06  kagel
  Removed unneccessary restriction requiring -t'none' with includes.
  Code does not require this restriction to work.

  Revision 1.129  2007/01/31 15:29:48  kagel
  Fixed version portability problem (some esql/c versions can't handle
  more than one declaration on a line with an initializer for esql
  specific types.
  Also a bit of code cleanup - fixing missing prototypes/header files,
  removing most unreferenced vars, etc.

  Revision 1.128  2006/10/16 16:44:17  kagel
  Fix problem with version detection missing if -S specified.
  Add -v2 for extra verbosity (currently prints version detected).

  Revision 1.127  2006/08/17 16:21:15  kagel
  Fixed so that an error message is printed if no databases are selected

  Revision 1.125  2006/06/28 16:14:30  kagel
  Attempt to fix SEGV processing some OUT & INOUT parameters to routines.
  Added hidden diagnostic arg '-z' to output raw and cooked parameter lists.

  Revision 1.124  2006/06/19 22:11:18  kagel
  Fixed Gary Quiring and Colin Dawson's problem!  Procs work!

  Revision 1.123  2006/06/19 15:56:58  kagel
  Replaced all strcpy's with snprintf's and expanded procstmt.

  Revision 1.122  2006/06/13 20:14:40  kagel
  Replaced all calls to sprintf with snprintf to prevent memory overruns.
  Expanded several temp strings to hold longer usernames, etc.

  Revision 1.121  2006/03/27 20:41:45  kagel
  Fixed bug in new procedure code affecting versions prior to 9.xx.

  Revision 1.120  2006/03/27 19:42:54  kagel
  Made all error messages consistent.  All contain the string 'FAILED'
  so that scripts can parse the output more easily.

  Revision 1.119  2006/03/22 14:59:29  kagel
  Needed to strip out parameter types for 'out' and 'inout' parameters.

  Revision 1.118  2006/03/22 04:04:55  kagel
  Fixed V9+ & SPECIFIC stored procedure updates.

  Revision 1.117  2006/02/16 14:35:00  kagel
  Uncommented code removed for testing.

  Revision 1.116  2006/01/18 20:36:03  kagel
  Expanded tempstr to prevent overflow, corrected unclosed comment that probably 
  commented out some code.  Compilers were silently passing the errant comment.
  Thanks to Roy Mercer again for pointing that one out.

  Revision 1.115  2006/01/18 18:26:15  kagel
  Fixed problem handling tables with very large numbers of columns or very long 
  column names when optimized server handling is enabled.  Thanks to Roy Mercer 
  for reporting the error.

  Revision 1.114  2005/12/09 17:40:06  kagel
  Removed trace.  Last changes (1.113) was trying to improve the 'working on' 
  output with new optimizations enabled.  Old code was confused.  Better but not 
  perfect yet.

  Revision 1.113  2005/10/24 16:48:40  kagel
  Fixes???

  Revision 1.112  2005/09/15 16:26:47  kagel
  FIxed problem with garbage ownername for first table under some odd conditions.

  Revision 1.111  2005/09/07 15:31:24  kagel
  Fixed -G - some test code was left in.

  Revision 1.110  2005/09/06 20:16:14  kagel
  Expanded -i! & -x! to allow either a WHERE clause or a full sub-query.
  Either way, multiple -i! or -x! options are allowed as is mixing the !
  versions with the other -x & -i flavors.  Multiple -i! option filter results
  are OR'd and multiple -x! options filter results are AND'd.  Now the user
  has complete flexibility to use a file of names or a table of names.  Users
  can encode run codes in the name list table and reference those codes in the
  -i/-x sub-query to select tables for multiple parallel runs.
  A stored procedure that returns just a list of tablenames can also be used
  in the select list of the sub-query for more complex filters.
  Upped the Feature version to 5.10

  Revision 1.109  2005/09/02 17:37:44  kagel
  Fixed problem with processing keys multiple times if multiple indexes have
  the same key columns but with different ordering (ASC/DESC).  Improved
  include and exclude sub-query process entry.  Removed some restrictions on
  mixing includes and excludes.

  Revision 1.108  2005/08/31 21:46:31  kagel
  Added -G option to always get actual rowcounts for browse (-b) mode
  operations using COUNT(*).  This is needed for databases with thousands of
  tables where the select from the unindexed sysptnhdr is taking too long.

  Also added option to specify a WHERE clause to a subquery on systables for
  the IN clause instead of listing tables singly or by a file.

  Revision 1.107  2005/08/04 18:19:40  kagel
  Added new option -N to supress the SPECIFIC keyword in 9.xx.  Apparently
  there is a 'bug' that causes massive logical log usage if SPECIFIC is
  included.  This is only needed if you have overloaded function names
  so one can suppress if otherwise.

  Revision 1.106  2005/07/12 19:02:33  kagel
  Changed MEDIUM to not include the DISTRIBUTIONS ONLY clause so MED columns 
  also have LOW stats included.  Fixed problem with runtimes getting confused 
  for some operation paths when -e/-E is enabled.

  Revision 1.105  2005/07/07 16:25:17  kagel
  Fixed bug causing loss of MEDIUM & HIGH stats for the last table processed
  when server optimization support is enabled.
  This whole thing has to be rewritten just because there is multiplication
  of identical code everywhere - AARRGGHH.  But let's get this out the door.

  Revision 1.104  2005/06/30 19:35:29  kagel
  Partial rewrite to enable support for new optimized server
  code in 7.31xD2 and 9.30xC3.  New options -o to override
  the default to support the optimizations if present and
  to do things the old way if not present.  Using -o will
  permit one to generate a script on a new server to run on
  an old one or vice-versa.

  Revision 1.103  2005/06/17 17:33:42  kagel
  Fixed -a problem some more ;-)

  Revision 1.102  2005/06/17 16:17:11  kagel
  FIxed problem with running -a with -S and a shared memory connection.

  Revision 1.101  2004/09/10 17:11:58  kagel
  Missed one WITH HOLD.
  Forgot to uncomment CLOSE.

  Revision 1.100  2004/09/10 17:01:57  kagel
  Oops, left some code in test configuration.  This is good.

  Revision 1.99  2004/09/10 16:55:38  kagel
  Fixed problems with ANSI databases: table owners and all cursors made
  WITH HOLD cursors.

  Revision 1.98  2004/06/01 21:02:10  kagel
  Added support for environment variables to specify username and password
  for remote server support.

  Revision 1.97  2003/10/13 17:14:15  kagel
  Fixed compile problems on non-386 linux releases.

  Revision 1.96  2003/05/13 17:43:41  kagel
  Corrected SET PDQPRIORITY output for tables when table only.

  Revision 1.95  2003/05/13 17:08:53  kagel
  Updated copyright info.

  Revision 1.94  2003/05/13 16:53:14  kagel
  Missing comma.  Also increased Features Version to 4.12 to recognise the
  new -Q option.

  Revision 1.93  2003/05/13 16:50:42  kagel
  Cosmetic.  Also 1.92 added SET LOCK MODE TO WAIT 10 to eliminate problems
  (-252 -113) processing transient tables.

  Revision 1.92  2003/05/13 16:46:01  kagel
  Added sanity check on table PDQ setting.  Added verbose output for -Q.

  Revision 1.91  2003/05/13 16:06:13  kagel
  Fixed ESQL/C compiler warnings about getrowcount().

  Revision 1.90  2003/05/13 15:55:17  kagel
  Fixed bug in output of PDQPRIORITY, semi-colon was missing.
  Added new option: -Q<tablePDQ> to force a value of PDQPRIORITY to be output
  for running update statistics for tables.  This implies that the SET
  PDQPRIORITY preceeding running update statistics for procedures is now
  Always output.  If no -Q the environment variable PDQPRIORITY if set is
  used.

  Revision 1.89  2003/03/27 22:14:29  kagel
  Couple of related fixes came out of testing with Jay.

  Revision 1.88  2003/03/27 21:32:40  kagel
  Updated getrowcount() for partitions which have no sysptnhdr record, still
  not clear what kind of tables that would be but several users have reported
  the problem, now instead of erroring out getrowcount() just do a SELECT
  COUNT(*) from the table.  Thanks to Jay Konigsberg for his persistence
  in trying to help me track down the source of this problem, for pointing out
  the obvious solution, and for testing the results.

  Revision 1.87  2003/02/13 16:39:51  kagel
  Changed to include getopt.h on Linux and if HAVE_GETOPT_H is #defined.

  Revision 1.86  2003/02/10 17:36:44  kagel
  Added #if defined to allow compilation using c4gl versions that do not
  recognize ANSI style parameters as host variables.

  Revision 1.85  2003/02/03 22:11:15  kagel
  Added a prompt to the interactive entry of the table/database include/exclude
  list

  Revision 1.84  2003/02/03 19:07:09  kagel
  Increased features version to 4.11 to recognize new -g option.

  Revision 1.83  2003/02/03 18:41:11  kagel
  Added -g (enable fragment columns to HIGH) to usage.

  Revision 1.82  2003/02/03 18:35:06  kagel
  Completed fragmentation expression support.
  ###
  # Release Notes for Revisions 1.2->1.81 SNIPPED for brevity. 6/30/2005
  ###

  # Revision 1.1  1996/10/23  21:23:23  kagel
  # Initial revision
  #
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <errno.h>
#if !defined(NOTIMINGS)
#include <sys/time.h>
#endif
#if defined(i386) || defined(linux)
#include <string.h>
#else
#include <strings.h>
#endif
#include <signal.h>

#if defined(_WINDOWS_)
#include <windows.h>
#define SIGQUIT SIGBREAK
#endif

#include "getopt.h"  /* Not <getopt.h> need the delivered header. */

#define TBSFACT 16777216L
#if !defined( FALSE )
#define FALSE (0)
#endif
#if !defined( TRUE )
#define TRUE (!FALSE)
#endif
#define Version "6.00"
#define Revision "$Revision: 1.183 $"

EXEC SQL INCLUDE sqlhdr;
EXEC SQL INCLUDE sqlca;
EXEC SQL INCLUDE sqltypes;
EXEC SQL INCLUDE datetime;
EXEC SQL INCLUDE locator;

#if !defined(bigint)
#define bigint int
#endif

static char rcs_header[] = "$Header: /home/art/GoogleDrive/Latest/RCS/dostats.ec,v 1.183 2021/12/21 17:12:40 art Exp $";

static struct sqlca_s local_sqlca;

EXEC SQL BEGIN DECLARE SECTION;
struct {
    int tabid, partnum;
    bigint nrows;
    int days;
    short colno[16], colind, daysind;
    string tabname[129];
    string tabowner[33];
    string keystring[2049];
} tabinfo, lastidx;  

typedef struct rowcounts_s {
    int partnum;
    bigint nrows;
} rowcounts_t;

typedef struct S_BLOBBUFF {
    char *buff, *base;
    int bbsize, global_length, full_length;
} blobbuff_t;

rowcounts_t *rowcounts = (rowcounts_t *)NULL;
string databasename[129], host[1000], connhost[1000], username[1000], passwd[1000];
string *statement = (char *)NULL;
string statement2[2048], waitstr_dostats[200];
string tablename[129], columnnames[32768][129], procname[129], *dbstmt;
string specificname[129];
string procowner[33];
lvarchar *arglist;
char *arglist_ptr;
short arg_ind, specific_ind;
loc_t exprtext;
int evalpos;
blobbuff_t tblobbuff, *blobbuff;
int stmt_len = 0, dbstmt_len = 0;
char tempstr[65535], pdqstmt_1[100], pdqstmt_2[100];
char procstmt[65535];
short colno;
int NoDistOnly = 0;  /* If TRUE do not include 'DISTRIBUTIONS ONLY' clause. */
int NoLevel = 0;     /* If TRUE do not include statistics level keywords. */
int ForceDistOnly = 0; /* If TRUE use HIGH ... DISTRIBUTIONS ONLY and LOW even
			  for single column indexes. */
int doBrowse = 0;		 /* Find tables with row counts changed more */
double BrowseThreshHold = 15.00; /* than BrowseThreshHold % since last run */
int haveBrowseThreshHold = 0;
int doPhysicalCount = 0;
int doDistAging = 0;	/* Find tables with distributions older than DistAgingLimit days. */
int DistAgingLimit = 30;
int haveDistAgingLimit = 0;
int tablePDQ = 0;
int procedurePDQ = 0;
int QuotedIdents = 0;
double nrecs;
int numparts;
int major_vers, minor_vers;
string level_vers[256], server_type[256];
int ServerOptimized = FALSE;
int ServerOptimizationOverridden = FALSE;
int UseServerOptimization = TRUE;
int waitlen = 10;
char schedprocname[129], schedprocfilename[1025];
char schedule_stmt[65535];
int tbl_chk, OK;
EXEC SQL END DECLARE SECTION;

int conn = 0, Release = 7, processed = 0;
FILE *outfile = NULL;
int dotable = 1, doindex = 1, docolumn = 1, tabmedium = 1, dropdists = 0,
    is_se = 0, doprocs = 0, proconly = 0, ExpandReport = 0, have_smi = 1,
    once = 0, verbose = 0, singledb = 0, dosystem = 0, doexplain = 0, doPDQ = 0,
    doFragExpr = 0, useShared = 0, DoSpecific = 1, have_waitlen = 0;
int dumpargs = 0;

#if !defined(NOTIMINGS)
struct timeval totstart, totend, totdiff;
struct timeval dbsstart, dbsend, dbsdiff;
struct timeval tabstart, tabend, tabdiff;
struct timeval stmtstart, stmtend, stmtdiff;
int secs, msecs;
#endif

/* FUNCTION PROTOTYPES */
#if !defined(NOTIMINGS)
void timediff( struct timeval, struct timeval, int *, int * );
#endif
void usage( char *program );
void exit( int stat );
void dostats_shutdown( int sig );
#define STR_STEP 1024
void readlist( char **, int *, char **, int *, char * );
void appendStr( char **, int *, char * );
double minres( int nrecs );
void InitKeyStr( char *keystr );
short ParseKeyStr( );
long long getrowcount( char *tabowner, char *tabname, int tabid, int partnum );
int openit( loc_t *loc, int flags, int bsize );
int writeit( loc_t *loc, char *buf, int nbytes );
int closeit( loc_t *loc );
void StripTypeList( char *list );
int cfind( char *item, char collist[][129], int ncols );
int is_Server_Optimized( int major, int minor, char *level );
int Store_Statement( char *statement );
int Write_Statement( char *statement );

int (*Out_Func)( char * );

#if !defined( __GNUC__ )
#if defined( max )
#undef max
#endif

#if defined( strlower )
#undef strlower
#endif

#define max(a,b) ((a) > (b) ? (a) : (b) )
#define strlower( t, s, m ) {					\
	char *target = (t), *source = (s); int max = (m);	\
	while (max-- && *source) {				\
	    *target++ = tolower(*source++);			\
	}							\
	*target = 0;						\
    }
#else
#ifndef max
inline int max( int a, int b)
{
    return (a > b ? a : b);
}
#endif

inline void strlower( char *target, char *source, int max )
{
    while (max-- && *source) {
        *target++ = tolower(*source++);
    }
    *target = 0;
}
#endif


void dostats_shutdown( int sig )
{
    sqlbreak();
    if (conn == 1 && Release != 5) {
	EXEC SQL SET CONNECTION 'WORK';
	sqlbreak();
    } else if (Release != 5) {
	EXEC SQL SET CONNECTION 'DBASES';
	sqlbreak();
    }
    EXEC SQL DISCONNECT ALL;
    if (outfile != (FILE *)NULL) {
	fflush( outfile );
	fclose( outfile );
    }
    exit( 1 );
}

int main(int argc, char *argv[])
{
    int i, c, errflag=0, dflag=0, tflag=0, hflag=0, pflag=0, mflag=0,
	ANSIMode=0;
    int alldb = 0;
    char lasttbl[129], filename[1024], *h, *pdqenv;
    static short columns[32768];
    int ncols, column, column2, leadsame, jj;
    int tableset = 0, indexset = 0, columnset = 0, samplesize = 0;
    double resolution = 0.0, mresolution = 0.0, confidence = 0.0;
    int resset = 0,  mresset = 0, confset = 0, sampleset = 0, err = 0;
    int do_scheduling = 0, do_proconly = 0, do_procedure = 0;
    char pwd[1025];
    FILE *schedfile = (FILE *)NULL;
    char sch_starttime[50], sch_lasttime[50], sch_frequency[50];
    char sch_sunday[4], sch_monday[4], sch_tuesday[4], sch_wednesday[4];
    char sch_thursday[4], sch_friday[4], sch_saturday[4];
    char confstring[30], resstring[30], mresstring[30], samplestring[100];
    int  sch_have_freq = 0, sch_have_start = 0, sch_have_end = 0, sch_onetime = 0;
    char sch_freq_type;
    static char revision[1024];
    char ownerstr[500], lastowner[500];
    char *includes = (char *)NULL, *excludes = (char *)NULL;
    char *subquery = (char *)NULL, *exquery = (char *)NULL;
    char *dbincludes = (char *)NULL, *dbexcludes = (char *)NULL;
    int incl_len = 0, excl_len = 0, dbexcl_len = 0, qry_len = 0, xqry_len = 0;
    int sawlistincl = 0, sawlistexcl = 0, sawsubincl = 0, sawsubexcl = 0;
    int noexit = 0, use_aus_thresholds = 0;
    char *at, *pstemp_env;
    struct option longopts[] = {
	{ "schedule", no_argument, (int *)NULL, 0x100 },
	{ "first-time", required_argument, (int *)NULL, 0x101 },
	{ "last-time", required_argument, (int *)NULL, 0x102 },
	{ "frequency", required_argument, (int *)NULL, 0x103 },
	{ "procedure", required_argument, (int *)NULL, 0x104 },
	{ "one-time", no_argument, (int *)NULL, 0x105 },
	{ "to-file", required_argument, (int *)NULL, 'f' },
	{ "aus-thresholds", no_argument, (int *)NULL, 0x106 },
        { (char *)NULL, no_argument, (int *)NULL, -1 }
    };
    int longint;

    signal( SIGINT, dostats_shutdown );
    signal( SIGQUIT, dostats_shutdown );

    if (getenv( "DELIMIDENT" ) != (char *)NULL) {
	QuotedIdents = 1;
    }

    snprintf( tablename, sizeof tablename, "%s", "*" );
    host[0] = (char)0;
    memset( ownerstr, 0, sizeof ownerstr );
    strcpy( sch_starttime, "'00:00:01'" );
    strcpy( sch_lasttime, "'00:00:05'" );
    sch_frequency[0] = (char)0;
    strcpy( sch_sunday, "'f'" );
    strcpy( sch_monday, "'f'" );
    strcpy( sch_tuesday, "'f'" );
    strcpy( sch_wednesday, "'f'" );
    strcpy( sch_thursday, "'f'" );
    strcpy( sch_friday, "'f'" );
    strcpy( sch_saturday, "'f'" );
    while ((c = getopt_long( argc,
			     argv,
			     "56Dc:d:f:h:i:n:r:t:x:A:B:P:Q:R:CEFGILNSTVXabeglmopsv::w:zZ:",
			     longopts,
			     &longint )) != EOF )
    {
	switch(c) {
	  case 0x100:
	      /* Enable writing commands to the IDS 11 scheduler. */
	      do_scheduling = 1;
	      break;
	  case 0x104:
	      do_proconly = 1;
	      snprintf( schedprocname, sizeof schedprocname, "%s", optarg );
	      break;
	  case 0x101:
	      /* First time to execute scheduled commands. */
	      snprintf( sch_starttime, sizeof sch_starttime, "'%s'", optarg );
	      sch_have_start = 1;
	      break;
	  case 0x102:
	      /* Last time to execute scheduled commands. */
	      snprintf( sch_lasttime, sizeof sch_lasttime, "'%s'", optarg );
	      sch_have_end = 1;
	      break;
	  case 0x105:
	      /* Set procedure scheduled to run once only then disable
		 itself. */
	      sch_onetime = 1;
	      break;
	  case 0x106:
	      /* Set aging, browse, and PDQPRIORITY using the Auto
		 Update Statistics configuration. */
	      use_aus_thresholds = 1;
	      break;
	  case 0x103:
	      /* Frequency at which to execute scheduled commands. */
	      sch_have_freq = 1;
	      sch_freq_type = optarg[0];
	      switch (sch_freq_type) {
		case 'd':
		case 'D':
		    snprintf( sch_frequency, sizeof sch_frequency, "'1 00:00:00'" );
		    strcpy( sch_sunday, "'t'" );
		    strcpy( sch_monday, "'t'" );
		    strcpy( sch_tuesday, "'t'" );
		    strcpy( sch_wednesday, "'t'" );
		    strcpy( sch_thursday, "'t'" );
		    strcpy( sch_friday, "'t'" );
		    strcpy( sch_saturday, "'t'" );
		    break;
		case 'w':
		case 'W':
		    snprintf( sch_frequency, sizeof sch_frequency, "'7 00:00:00'" );
		    switch (optarg[1]) {
		      case '0':
			  strcpy( sch_sunday, "'t'" );
			  break;
		      case '1':
			  strcpy( sch_monday, "'t'" );
			  break;
		      case '2':
			  strcpy( sch_tuesday, "'t'" );
			  break;
		      case '3':
			  strcpy( sch_wednesday, "'t'" );
			  break;
		      case '4':
			  strcpy( sch_thursday, "'t'" );
			  break;
		      case '5':
			  strcpy( sch_friday, "'t'" );
			  break;
		      case '6':
			  strcpy( sch_saturday, "'t'" );
			  break;
		    }
		    break;
		case 'm':
		case 'M':
		    snprintf( sch_frequency, sizeof sch_frequency, "'30 00:00:00'" );
		    switch (optarg[1]) {
		      case '0':
			  strcpy( sch_sunday, "'t'" );
			  break;
		      case '1':
			  strcpy( sch_monday, "'t'" );
			  break;
		      case '2':
			  strcpy( sch_tuesday, "'t'" );
			  break;
		      case '3':
			  strcpy( sch_wednesday, "'t'" );
			  break;
		      case '4':
			  strcpy( sch_thursday, "'t'" );
			  break;
		      case '5':
			  strcpy( sch_friday, "'t'" );
			  break;
		      case '6':
			  strcpy( sch_saturday, "'t'" );
			  break;
		    }
		    break;
	      }
	      break;
	  case 'z':
	      dumpargs = 1;
	      break;
	  case 'G':
	      doPhysicalCount = 1;
	      break;
	  case 'N':
	      DoSpecific = 0;
	      break;
	  case 'o':
	      ServerOptimizationOverridden = TRUE;
	      break;
	  case 'm':
	      mflag = 1;
	      break;
	  case 'n':
	      noexit = atoi( optarg );
	      break;
	  case 'p':
	      pflag++;
	      break;
	  case 's':
	      is_se = 1;
	      singledb = 1;
	      have_smi = 0;
	      break;
	  case 'S':
	      singledb = 1;
	      useShared = 1;
	      break;
	  case 'v':
	      if (optarg == (char *)NULL)
		  verbose = 1;
	      else
		  verbose = atoi( optarg );
	      break;
	  case 'b':
	      doBrowse = 1;
	      break;
	  case 'g':
	      doFragExpr = 1;
	      break;
	  case 'e':
	      ExpandReport = 1;
	      break;
	  case 'a':
	      doDistAging = 1;
	      break;
	  case 'E':
	      ExpandReport = 2;
	      break;
	  case 'F':
	      ForceDistOnly = 1;
	      break;
	  case 'X':
	      doexplain = 1;
	      break;
	  case 'B':
	      BrowseThreshHold = atof( optarg );
	      haveBrowseThreshHold = 1;
	      break;
	  case 'A':
	      DistAgingLimit = atoi( optarg );
	      haveDistAgingLimit = 1;
	      break;
	  case 'P':
	      procedurePDQ = atol( optarg );
	      break;
	  case 'Q':
	      tablePDQ = atol( optarg );
	      doPDQ = 1;
	      break;
	  case '5':
	      /* For 5.xx databases.  R5.xx can only do UPDATE STATISTICS LOW.
		 Also the R5.xx syntax does not support the 'LOW' keyword.
	      */
	      Release = 5;
	      NoLevel = TRUE;
	      NoDistOnly = TRUE;
	      tabmedium = 0;
	      tableset = 1;
	      singledb = 1;
	      have_smi = 0;
	      break;
	  case 'D':
	  case '6':
	      Release = 6;
	      NoDistOnly = TRUE;
	      break;
	  case 'd':
	      if (dflag)
		  errflag++;
	      else{
		  dflag++;
		  if (QuotedIdents) {
		      strncpy(databasename, optarg, 128);
		      databasename[128] = (char)0;
		  } else {
		      strlower(databasename, optarg, 128);
		  }
	      }
	      break;
	  case 't':
	      if (tflag)
		  errflag++;
	      else{
		  tflag++;
		  if (QuotedIdents) {
		      strncpy(tablename, optarg, 128);
		      tablename[128] = (char)0;
		  } else {
		      strlower(tablename, optarg, 128);
		  }
	      }
	      break;
	  case 'f':
	      if (strcmp( "-", optarg ) == 0) {
		  snprintf( filename, sizeof filename, "%s", "stdout" );
		  outfile = stdout;
	      } else {
		  snprintf( filename, sizeof filename, "%s", optarg );
		  outfile = fopen( filename, "w" );
	      }
	      fprintf( stderr, "Writing commands to: %s\n", filename );
	      if (!outfile) {
		  fprintf( stderr,
			   "Cannot open output: %s, errno=%d.\n",
			   filename,
			   errno );
		  exit(3);
	      }
	      break;
	  case 'h':
	      if (hflag)
		  errflag++;
	      else{
		  hflag++;
		  strncat(host, optarg, 18);
		  host[19] = (char)0;
	      }
	      break;
	  case 'i':
	      /* Form table include list. */
	      if (optarg[0] == '@') {
		  readlist( &includes,
			    &incl_len,
			    (char **)NULL,
			    (int *) NULL,
			    optarg );
		  sawlistincl = 1;
	      } else if (optarg[0] == '!') {
		  appendStr( &subquery, &qry_len, " OR a.tabname IN (" );
		  if (strncasecmp( &(optarg[1]), "select", 6 ) != 0)
		      appendStr( &subquery, &qry_len, "select tabname from systables where " );
		  appendStr( &subquery, &qry_len, &(optarg[1]) );
		  appendStr( &subquery, &qry_len, ") " );
		  sawsubincl = 1;
	      } else {
		  if (includes)
		      appendStr( &includes, &incl_len, "','" );
		  else
		      appendStr( &includes, &incl_len, "('" );
		  appendStr( &includes, &incl_len, optarg );
		  sawlistincl = 1;
	      }
	      break;
	  case 'x':
	      /* Form table exclude list. */
	      if (optarg[0] == '@') {
		  readlist( &excludes,
			    &excl_len,
			    &dbexcludes,
			    &dbexcl_len,
			    optarg );
		  sawlistexcl = 1;
	      } else if (optarg[0] == '!') {
		  appendStr( &exquery, &xqry_len, " AND a.tabname NOT IN (" );
		  if (strncasecmp( &(optarg[1]), "select", 6 ) != 0)
		      appendStr( &exquery, &xqry_len, "select tabname from systables where " );
		  appendStr( &exquery, &xqry_len, &(optarg[1]) );
		  appendStr( &exquery, &xqry_len, ") " );
		  sawsubexcl = 1;
	      } else if (strchr( optarg, ':' )) {
		  /* Database exclusion. */
		  if (dbexcludes)
		      appendStr( &dbexcludes, &dbexcl_len, "','" );
		  else
		      appendStr( &dbexcludes, &dbexcl_len, "AND name NOT IN ('" );
		  appendStr( &dbexcludes, &dbexcl_len, strtok( optarg, ":" ) );
	      } else {
		  /* Table exclusion. */
		  if (excludes)
		      appendStr( &excludes, &excl_len, "','" );
		  else
		      appendStr( &excludes, &excl_len, "('" );
		  appendStr( &excludes, &excl_len, optarg );
		  sawlistexcl = 1;
	      }
	      break;
	  case 'T':
	      tableset = 1;
	      tabmedium = 1;
	      break;
	  case 'I':
	      indexset = 1;
	      break;
	  case 'r':
	      mresolution = atof( optarg );
	      mresset = 1;
	      break;
	  case 'R':
	      resolution = atof( optarg );
	      resset = 1;
	      break;
	  case 'c':
	      confidence = atof( optarg );
	      confset = 1;
	      break;
	  case 'Z':
	      samplesize = atoi( optarg );
	      sampleset = 1;
	      break;
	  case 'C':
	      columnset = 1;
	      break;
	  case 'L':
	      tabmedium = 0;
	      tableset = 1;
	      indexset = 0;
	      dropdists = 1;
	      break;
	  case 'l':
	      tabmedium = 0;
	      tableset = 1;
	      indexset = 0;
	      dropdists = 0;
	      break;
	  case 'w':
	      waitlen = atoi( optarg );
	      have_waitlen = 1;
	      break;
	  case 'V':
	      snprintf( revision, sizeof revision, "%s", Revision );
	      printf( "\ndostats: Features Version %s, Source %s!\n",
		      Version,
		      strtok( revision, "$" ) );
	      fprintf( stderr,
		       "\tCopyright 1996-2022 by Art S. Kagel.\n" );
	      fprintf( stderr,
		       "\tLicense for private usage (including use by "
		       "commercial enterprise\n"
		       "\tfor its own needs) is granted.  Commercial and "
		       "exploitation rights\n"
		       "\tare reserved.\n\n" );
	      exit(0);
	  case '?':
	  default:
	      errflag++;
	      break;
	}
    }

    if (errflag || dflag == 0){ usage( argv[0] ); exit( 1 ); }

    if (optind < argc) {
	fprintf( stderr, "Extra arguments ignored!\n" );
    }

    if (pflag && !tflag && !includes && !excludes && !subquery) {
	proconly = 1;
	doprocs = 1;
    } else if (pflag || (dflag && !tflag && !includes && !excludes && !subquery && !exquery))
	doprocs = 1;

    if (tableset || indexset || columnset)
	dotable = doindex = docolumn = 0;

    if (tableset) {
	dotable = 1;
    }

    if (indexset) {
	doindex = 1;
    }

    if (columnset) {
	docolumn = 1;
    }

    if (mflag) {
	dosystem = 1;
    }

    if (confset && !tabmedium) {
	fprintf( stderr,
		 "Confidence option included but MEDIUM level stats have not "
		 "been requested!  Ignoring.\n" );
	confset = 0;
    }

    if (sampleset && !tabmedium) {
	fprintf( stderr,
		 "Sampling size option included but MEDIUM level stats have not "
		 "been requested!  Ignoring.\n" );
	sampleset = 0;
    }

    if (sampleset && samplesize <= 0) {
	fprintf( stderr,
		 "Invalid minimum sample size (%d) is out of range (0 < MinSamp < nrows)!\n",
		 samplesize );
	exit(1);
    } else if (sampleset)
	snprintf( samplestring, sizeof samplestring, " SAMPLING SIZE %d ", samplesize );

    if (resset && !doindex) {
	fprintf( stderr,
		 "HIGH Resolution option included but HIGH stats"
		 " have not been requested!  Ignoring.\n" );
	resset = 0;
    }

    if (mresset && !tabmedium) {
	fprintf( stderr,
		 "MEDIUM Resolution option included but MEDIUM stats"
		 " have not been requested!  Ignoring.\n" );
	mresset = 0;
    }

    if (confset && (confidence > 0.9900000000 || confidence <= 0.80)) {
	fprintf( stderr,
		 "Confidence level specified is out of range(0.80 -> 0.99)!\n" );
	exit(1);
    } else if (confset) {
	snprintf( confstring, sizeof confstring, " %15.10f", confidence );
	if (!mresset) {
	    mresset = 1;
	    mresolution = 2.5;
	}
    }

    if (resset && (resolution > 10.0 || resolution < 0.005)) {
	fprintf( stderr,
		 "HIGH Resolution specified is out of "
		 "range(max(0.005,(1/nrows)) -> 10.0%%)!\n" );
	exit(1);
    } else if (resset)
	snprintf( resstring, sizeof resstring, " RESOLUTION %15.10f", resolution );

    if (mresset && (mresolution > 10.0 || mresolution < 0.005)) {
	fprintf( stderr,
		 "MEDIUM Resolution specified is out of "
		 "range((max(0.005,(1/nrows)) -> 10.0%%)!\n" );
	exit(1);
    } else if (mresset)
	snprintf( mresstring, sizeof mresstring, " RESOLUTION %15.10f", mresolution );

    if (tflag && strcmp( tablename, "none" ) == 0 && !includes && !subquery ) {
	fprintf( stderr,
		 "Includes required if tablename is \"none\"!\n" );
	usage( argv[0] );
	exit( 1 );
    } else if (tflag && strcmp( tablename, "none" ) == 0 && excludes && exquery && !includes && !subquery) {
	fprintf( stderr,
		 "Excludes prohibited if tablename is \"none\"!\n" );
	usage( argv[0] );
	exit( 1 );
    } else if (strcmp( tablename, "*" ) == 0 && includes && subquery && !excludes && !exquery) {
	fprintf( stderr,
		 "Tablename template other than '*' required with includes!\n" );
	usage( argv[0] );
	exit( 1 );
    }

    if (procedurePDQ < 0 || procedurePDQ > 100) {
	fprintf( stderr,
		 "Store procedure PDQ priority out-of-range(0-100): %d!\n",
		 procedurePDQ );
	exit( 1 );
    }

    if (useShared && doBrowse) {
	fprintf( stderr,
		 "-b option not possible with a shared memory connection!\n"
		 "Remove -S and change to a networked or pipe connection.\n" );
	exit( 1 );
    }

    if (Release == 5 && doBrowse) {
	fprintf( stderr,
		 "-b option not valid for OL5 servers!\n" );
	exit( 1 );
    }

    if (is_se && doBrowse) {
	fprintf( stderr,
		 "-b option not valid for SE servers!\n" );
	exit( 1 );
    }

    if (haveBrowseThreshHold && !doBrowse) {
	fprintf( stderr,
		 "Browse threshhold (-B) entered without browse request(-b)!\n" );
	exit( 1 );
    }

    if (BrowseThreshHold < 0.0001 || BrowseThreshHold > 10000.00 ) {
	fprintf( stderr,
		 "Browse threshhold out-of-range(0.0001-10000.0): %f!\n",
		 BrowseThreshHold );
	exit( 1 );
    }

    if (Release == 5 && doDistAging) {
	fprintf( stderr,
		 "-a option not valid for OL5 servers!\n" );
	exit( 1 );
    }

    if (is_se && doDistAging) {
	fprintf( stderr,
		 "-a option not valid for SE servers!\n" );
	exit( 1 );
    }

    if (is_se && doFragExpr) {
	doFragExpr = 0;
	fprintf( stderr,
		 "-g option not valid for SE servers!  Disabling.\n" );
    }

    if (Release == 5 && doFragExpr) {
	doFragExpr = 0;
	fprintf( stderr,
		 "-g option not valid for OL5 servers!  Disabling.\n" );
    }

    if (haveDistAgingLimit && !doDistAging) {
	fprintf( stderr,
		 "Distribution Aging Limit (-A) entered without Aging request(-a)!\n" );
	exit( 1 );
    }

    if (DistAgingLimit < 0) {
	fprintf( stderr,
		 "Distribution Aging Limit out-of-range(>= 0): %d!\n",
		 DistAgingLimit );
	exit( 1 );
    }

    if (!doPDQ) {
	pdqenv = getenv( "PDQPRIORITY" );
	if (pdqenv) {
	    tablePDQ = atol( pdqenv );
	} else
	    tablePDQ = 0;
    }

    if (tablePDQ < 0 || tablePDQ > 100) {
	fprintf( stderr,
		 "Table PDQPRIORITY out-of-range(0-100): %d!",
		 tablePDQ );
	exit( 1 );
    }

    do_procedure = do_scheduling + do_proconly;

    if (!(do_scheduling || do_proconly) && sch_onetime) {
	fprintf( stderr, "--one-time passed without --schedule!\n" );
	exit( 1 );
    }

    if (!do_scheduling && sch_have_start) {
	fprintf( stderr, "--first-time passed without --schedule!\n" );
	exit( 1 );
    }

    if (!do_scheduling && sch_have_end) {
	fprintf( stderr, "--last-time passed without --schedule!\n" );
	exit( 1 );
    }

    if (!do_scheduling && sch_have_freq) {
	fprintf( stderr, "--frequency passed without --schedule!\n" );
	exit( 1 );
    }

    if (do_scheduling && outfile != (FILE *)NULL) {
	fprintf( stderr, "--schedule is not compatible with -f/--to-file!\n" );
	exit( 1 );
    }

    if (do_scheduling && singledb) {
	fprintf( stderr, "--schedule requires a non-shared-memory connection type!\n" );
	exit( 1 );
    }

    snprintf( pdqstmt_1, sizeof pdqstmt_1, "SET PDQPRIORITY %d;", procedurePDQ );
    snprintf( pdqstmt_2, sizeof pdqstmt_2, "SET PDQPRIORITY %d;", tablePDQ );
     
    if (strcmp( databasename, "" ) == 0
	|| ( tflag == 1 && strcmp( tablename, "" ) == 0 ))
    {
	usage( argv[0] );
	exit( 1 );
    }
    
    if (includes && sawlistincl)
	appendStr( &includes, &incl_len, "') " );
    else if (includes)
	appendStr( &includes, &incl_len, ") " );
    
    if (excludes && sawlistexcl)
	appendStr( &excludes, &excl_len, "') " );
    else if (excludes)
	appendStr( &excludes, &excl_len, ") " );

    if (dbexcludes)
	appendStr( &dbexcludes, &dbexcl_len, "') " );

    if (do_scheduling && !sch_have_freq) {
	/* Scheduling enabled but no frequency set.  Default to Weekly
	   on Sunday. */
	snprintf( sch_frequency, sizeof sch_frequency, "'7 00:00:00'" );
	strcpy( sch_sunday, "'t'" );
	sch_have_freq = 1;
    }

    if (do_scheduling && sch_have_freq && !sch_have_start) {
	sprintf( sch_starttime, "'00:00:00'" );
    }

    if (do_scheduling && sch_have_freq  && !sch_have_end) {
	sprintf( sch_lasttime, "'00:00:05'" );
    }

    if (verbose > 0) {
	fprintf( stderr, "Verbose mode level %d               ENABLED....\n", verbose );
	if (QuotedIdents)
	    fprintf( stderr,
		     "Delimited identifiers               ENABLED....\n" );
	fprintf( stderr,
		 "PDQPRIORITY for table processing set to: %d.\n",
		 tablePDQ );
	if (doprocs) {
	    fprintf( stderr,
		     "Stored Procedure processing         ENABLED....\n" );
	    fprintf( stderr,
		     "   Default PDQ priority for procedures set to: %d.\n",
		     procedurePDQ );
	} else
	    fprintf( stderr,
		     "Stored Procedure processing         DISABLED...\n" );
	if (is_se)
	    fprintf( stderr,
		     "Standard Engine handling            ENABLED....\n" );
	else
	    fprintf( stderr,
		     "Standard Engine handling            DISABLED...\n" );
	if (singledb)
	    fprintf( stderr,
		     "Shared Memory Connection handling   ENABLED....\n" );
	else
	    fprintf( stderr,
		     "Shared Memory Connection handling   DISABLED...\n" );
	if (ExpandReport)
	    fprintf( stderr,
		     "Expanded Reporting                  ENABLED....\n" );
	else
	    fprintf( stderr,
		     "Expanded Reporting                  DISABLED...\n" );
	if (ExpandReport == 2)
	    fprintf( stderr,
		     "Expanded Time Reporting             ENABLED....\n" );
	else
	    fprintf( stderr,
		     "Expanded Time Reporting             DISABLED...\n" );
	if (ForceDistOnly)
	    fprintf( stderr,
		     "Force Distributions Only            ENABLED....\n" );
	else
	    fprintf( stderr,
		     "Force Distributions Only            DISABLED...\n" );
	if (Release == 5)
	    fprintf( stderr,
		     "Release 5 handling                  ENABLED....\n" );
	else
	    fprintf( stderr,
		     "Release 5 handling                  DISABLED...\n" );
	if (NoLevel != TRUE)
	    fprintf( stderr,
		     "Statistics Levels                   ENABLED....\n" );
	else
	    fprintf( stderr,
		     "Statistics Levels                   DISABLED...\n" );
	if (NoDistOnly != TRUE)
	    fprintf( stderr,
		     "Distributions Only clauses          ENABLED....\n" );
	else
	    fprintf( stderr,
		     "Distributions Only clauses          DISABLED...\n");
	if (dosystem)
	    fprintf( stderr,
		     "System catalog processing           ENABLED....\n");
	else
	    fprintf( stderr,
		     "System catalog processing           DISABLED...\n");
	if (tflag)
	    fprintf( stderr,
		     "Table Name Template                 SUPPLIED (%s)...\n",
		     tablename );
	else
	    fprintf( stderr,
		     "Table Name Template                 NOT SUPPLIED...\n");
	if (proconly)
	    fprintf( stderr, "Only Stored Procedures will be processed.\n" );
	if (hflag)
	    fprintf( stderr,
		     "Host Name                           SUPPLIED (%s)...\n",
		     host );
	else
	    fprintf( stderr,
		     "Host Name                           NOT SUPPLIED...\n");
	if (do_scheduling) {
	    fprintf( stderr,
		     "Command Execution                   SCHEDULED...\n" );
	    if (do_proconly)
		fprintf( stderr,
			 "    Command procedure: %s\n",
			 schedprocname );
	} else if (!outfile)
	    fprintf( stderr,
		     "Command Execution                   IMMEDIATE...\n" );
	else
	    fprintf( stderr,
		     "Command Execution                   DEFERRED...\n" );
	
	if (subquery != (char *)NULL)
	    fprintf( stderr,
		     "Table Inclusion SubQuery Processed:\n\t%s\n",
		     subquery );
	if (includes != (char *)NULL)
	    fprintf( stderr,
		     "Table Inclusion List Processed:\n\ttabname IN %s\n",
		     includes );
	if (exquery != (char *)NULL)
	    fprintf( stderr,
		     "Table Exclusion SubQuery Processed:\n\t%s\n",
		     exquery );
	if (excludes != (char *)NULL)
	    fprintf( stderr,
		     "Table Exclusion List Processed:\n\ttabname NOT IN %s\n",
		     excludes );
	if (dbexcludes != (char *)NULL)
	    fprintf( stderr,
		     "Database Exclusion List Processed:\n\t%s\n",
		     (dbexcludes + 4) );
	if (tableset || indexset || columnset) {
	    if (tabmedium)
		fprintf( stderr,
			 "Table Level Medium Statistics       ENABLED...\n" );
	    else
		fprintf( stderr,
			 "Table Level Low Statistics          ENABLED...\n" );
	    if (dropdists)
		fprintf( stderr, "Distributions will be dropped.\n" );
	    if (!indexset)
		fprintf( stderr,
			 "Index Lead High Statistics          DISABLED...\n" );
	    if (!columnset)
		fprintf( stderr,
			 "Index Key LOW Statistics            DISABLED...\n" );
	}
	if (mresset)
	    fprintf( stderr,
		     "Medium Statistics Resolution set to: %f\n",
		     mresolution );
	if (resset)
	    fprintf( stderr,
		     "High Statistics Resolution set to: %f\n",
		     resolution );
	if (confset)
	    fprintf( stderr,
		     "Medium Statistics Confidence set to: %f\n",
		     confidence );
	if (doexplain)
	    fprintf( stderr,
		     "SET EXPLAIN                         ENABLED...\n" );
	if (doBrowse)
	    fprintf( stderr,
		     "Table Browsing                      ENABLED...\n" );
	if (haveBrowseThreshHold)
	    fprintf( stderr,
		     "Table browse threshhold set to: %f%%\n",
		     BrowseThreshHold );
	if (doDistAging)
	    fprintf( stderr,
		     "Distribution Aging                  ENABLED...\n" );
	if (haveDistAgingLimit)
	    fprintf( stderr,
		     "Distribution Aging limit set to: %d days\n",
		     DistAgingLimit );
	if (doFragExpr)
	    fprintf( stderr,
		     "Fragment expression parsing enabled\n" );
	if (ServerOptimizationOverridden)
	    fprintf( stderr,
		     "Optimized server handling override  ENABLED....\n" );
	else
	    fprintf( stderr,
		     "Optimized server handling override  DISABLED...\n" );
	if (waitlen > 0)
	    fprintf( stderr, "Lock mode set to wait %d seconds.\n", waitlen );
	else if (waitlen == 0)
	    fprintf( stderr, "Lock mode set to not wait.\n" );
	else  /* (waitlen < 0) */
	    fprintf( stderr, "Lock mode set to wait forever.\n" );
	    
    }
    
    /* Access requested database */
    
    if ((at=strchr( databasename, '@' )) != (char *)0) {
	/* Database name contains embedded server name.  Extract it. */
	if (host[0] == (char)0) {
	    *at++ = (char)0;
	    strncpy( host, at, 129 );
	} else {
	    fprintf( stderr, "Host specified in -h and -d!  Choose one.\n" );
	    usage( argv[0] );
	    exit( 1 );
	}
    }

    if (host[0] == (char)0) {
	/* No host specified.  Connect to default. */
	h = getenv( "INFORMIXSERVER" );
	if (h)
	    snprintf( host, sizeof host, "%s", h );
	else {
	    fprintf( stderr,
		     "Cannot open database connection.  "
		     "INFORMIXSERVER not set.\n"
		     "Set INFORMIXSERVER environment variable or include "
		     "'-h hostname' option.\n" );
	    usage( argv[0] );
	    exit( 1 );
	}
    }
    
    if (username[0] == (char)0) {
	/* No username specified.  Connect to default. */
	h = getenv( "INFORMIXUSER" );
	if (h)
	    snprintf( username, sizeof username, "%s", h );
    }
    
    
    if (passwd[0] == (char)0) {
	/* No username specified.  Connect to default. */
	h = getenv( "INFORMIXPASS" );
	if (h)
	    snprintf( passwd, sizeof passwd, "%s", h );
    }

    snprintf( connhost, sizeof connhost, "@%s", host );
    
    if (have_smi && (!singledb || doBrowse || doDistAging)) {
	if (username[0] != (char)0) {
	    if (passwd[0] != (char)0)
		EXEC SQL CONNECT TO :connhost AS 'DBASES'
		    USER :username USING :passwd
		    WITH CONCURRENT TRANSACTION;
	    else {
		fprintf( stderr,
			 "Cannot specify INFORMIXUSER without INFORMIXPASS!\n" );
		usage( argv[0] );
		exit( 1 );
	    }
	} else
	    EXEC SQL CONNECT TO :connhost AS 'DBASES'
		    WITH CONCURRENT TRANSACTION;
	if (sqlca.sqlcode) {
	    fprintf( stderr,
		     "FAILED: Cannot connect to %s (@%s); error=%d, isam=%d\n\n",
		     "SMI DATABASE",
		     host,
		     (int)sqlca.sqlcode,
		     (int)sqlca.sqlerrd[1] );
	    if (sqlca.sqlcode == -27000) {
		if (doBrowse) {
		    fprintf( stderr,
			     "-b option not possible with a shared memory connection!\n"
			     "Remove -S and change to a networked or pipe connection.\n" );
		} else {
		    fprintf( stderr,
			     "You cannot use a shared memory connection in "
			     "$INFORMIXSERVER or the -h option \nwithout the -S "
			     "option.  Specify a pipe or network connection or -S "
			     "and rerun.\n\n" );
		}
	    }
	    exit(1);
	}
	EXEC SQL DATABASE "sysmaster";
	if (sqlca.sqlcode < 0 && !is_se) {
	    /* No sysmaster database and not SE.  Must be 5.xx engine.  Just 
	       try database.  OK, let's be nice and set 5.0 parameters. */
	    Release = 5;
	    have_smi = 0;
	    NoLevel = 1;
	    NoDistOnly = TRUE;
	    tabmedium = 0;
	    tableset = 1;
	    singledb = 1;
	    alldb = 0;
	    doBrowse = 0;
	    doDistAging = 0;
	    doFragExpr = 0;
	    ServerOptimized = FALSE;
	} else if (sqlca.sqlcode == 0) {
	    if (!singledb) {
		/* Found sysmaster.  Continue.  Oh, cancel any -5 option. */
		if (strlen( databasename ) == 3 &&
		    (strcmp( databasename, "ALL" ) == 0 ||
		     strcmp( databasename, "all" ) == 0))
		    snprintf( databasename, sizeof databasename, "%s", "*" );
		appendStr( &dbstmt, &dbstmt_len,
			   "SELECT name FROM \"informix\".sysdatabases "
			   "WHERE name != 'sysmaster' "
			   "AND name matches ? " );
		if (dbexcludes)
		    appendStr( &dbstmt, &dbstmt_len, dbexcludes );
		appendStr( &dbstmt, &dbstmt_len, " ORDER BY name");
		EXEC SQL PREPARE getdb
		    FROM :dbstmt;
		if (sqlca.sqlcode) {
		    fprintf( stderr,
			     "FAILED: Cannot PREPARE select statement for databases, error: %d.\n",
			     (int)sqlca.sqlcode );
		    fprintf( stderr, "Error at character %d.\nStatement is:\n",
			     (int)sqlca.sqlerrd[4] );
		    fprintf( stderr, "SELECT name FROM sysdatabases" );
		    exit(2);
		}
		EXEC SQL DECLARE dbc CURSOR WITH HOLD FOR getdb;
		if (sqlca.sqlcode) {
		    fprintf( stderr,
			     "FAILED: Cannot declare dbc cursor; Error = %d\n",
			     (int)sqlca.sqlcode );
		    exit(1);
		}
		EXEC SQL OPEN dbc using :databasename;
		if (sqlca.sqlcode) {
		    fprintf( stderr,
			     "FAILED: Cannot open dbc cursor; Error = %d\n",
			     (int)sqlca.sqlcode );
		    exit(1);
		}
	    }
	    if ((doBrowse || doDistAging) && !doPhysicalCount) {
		int part;

		EXEC SQL SELECT count(*)
		    INTO :numparts
		    FROM "informix".sysptnhdr;
		rowcounts =
		    (rowcounts_t *)malloc( numparts * sizeof (rowcounts_t) );

		EXEC SQL DECLARE rowcount_c CURSOR WITH HOLD FOR
		    SELECT partnum, nrows
		    FROM "informix".sysptnhdr
		    ORDER BY partnum;
		if (sqlca.sqlcode) {
		    fprintf( stderr,
			     "FAILED: Cannot declare rowcount_c cursor. "
			     "Error = %d, isam=%d\n",
			     (int)sqlca.sqlcode,
			     (int)sqlca.sqlerrd[1] );
		    exit(1);
		}
		EXEC SQL OPEN rowcount_c;
		if (sqlca.sqlcode) {
		    fprintf( stderr,
			     "FAILED: Cannot open rowcount_c cursor. "
			     "Error = %d, isam=%d\n",
			     (int)sqlca.sqlcode,
			     (int)sqlca.sqlerrd[1] );
		    exit(1);
		}
		for (part = 0; part < numparts; part++) {
		    EXEC SQL FETCH rowcount_c INTO :rowcounts[part];
		    if (sqlca.sqlcode)
			break;
		}
		if (sqlca.sqlcode < 0) {
		    fprintf( stderr,
			     "FAILED: Error fetching rowcounts.  Error=%d, isam=%d\n\n",
			     (int)sqlca.sqlcode,
			     (int)sqlca.sqlerrd[1] );
		    exit( 1 );
		}
		EXEC SQL CLOSE rowcount_c;
	    } else
		numparts = 0;
	    if (singledb) {
		EXEC SQL DISCONNECT 'DBASES';
		if (sqlca.sqlcode) {
		    fprintf( stderr,
			     "FAILED: Cannot safely disconnect from sysmaster! "
			     "Error=%d, isam=%d\n\n",
			     (int)sqlca.sqlcode,
			     (int)sqlca.sqlerrd[1] );
		    exit( 1 );
		}
	    }
	} else {
	    /* ELSE must be SE.  Just have to try the database. */
	    have_smi = 0;
	    singledb = 1;
	    alldb = 0;
	}
    }

    if (username[0] != (char)0) {
	if (passwd[0] != (char)0)
	    EXEC SQL CONNECT TO :connhost AS 'WORK'
		USER :username USING :passwd
		WITH CONCURRENT TRANSACTION;
	else {
	    fprintf( stderr,
		     "Cannot specify INFORMIXUSER without INFORMIXPASS!\n" );
	    usage( argv[0] );
	    exit( 1 );
	}
    } else
	EXEC SQL CONNECT TO :connhost AS 'WORK'
		WITH CONCURRENT TRANSACTION;
    if (sqlca.sqlcode) {
	fprintf( stderr,
		 "FAILED: Cannot connect to %s (@%s); error=%d, isam=%d\n\n",
		 "WORK DATABASE",
		 host,
		 (int)sqlca.sqlcode,
		 (int)sqlca.sqlerrd[1] );
	if (sqlca.sqlcode == -27000)
	    fprintf( stderr,
		     "FAILED: You cannot use a shared memory connection in "
		     "$INFORMIXSERVER or the -h option \nwithout the -S "
		     "option.  Specify a pipe or network connection or -S "
		     "and rerun.\n\n" );
	exit(1);
    }
    if (waitlen == 0)
	strcpy( waitstr_dostats, "SET LOCK MODE TO NOT WAIT" );
    else if (waitlen < 0)
	strcpy( waitstr_dostats, "SET LOCK MODE TO WAIT" );
    else /* (waitlen > 0) */
	sprintf( waitstr_dostats, "SET LOCK MODE TO WAIT %d", waitlen );
    EXEC SQL EXECUTE IMMEDIATE :waitstr_dostats;
    if (sqlca.sqlcode < 0) {
	fprintf( stderr,
		 "Warning: error setting lock mode to%s wait; "
		 "error=%d, isam=%d.\n",
		 ((waitlen == 0) ? " not" : ""),
		 (int)sqlca.sqlcode,
		 (int)sqlca.sqlerrd[1] );
    }
    if (have_smi) {
	/* Version detection. */
	EXEC SQL DATABASE 'sysmaster';
	if (sqlca.sqlcode == -387) {
	    fprintf( stderr,
		     "FAILED: No permissions to database(sysmaster@%s).  "
		     "Rerun as: informix.\n",
		     host );
	    exit(1);
	} else if (sqlca.sqlcode) {
	    fprintf( stderr,
		     "FAILED: Cannot access database (sysmaster); error=%d, isam=%d\n",
		     (int)sqlca.sqlcode,
		     (int)sqlca.sqlerrd[1] );
	    exit(1);
	}
	EXEC SQL SELECT DBINFO( 'version', 'major' )
	    INTO :major_vers
	    FROM "informix".systables where tabid = 1;
	if (sqlca.sqlcode == -728) {
	    /* Older versions do not support DBINFO( 'version', ... */
	    Release = major_vers = 7;
	    minor_vers = 23;
	    level_vers[0] = (char)0;
	    snprintf( server_type, sizeof server_type, "Informix Dynamic Server" );
	} else if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "FAILED: Error fetching major version id.  SQL=%d, ISAM=%d.\n",
		     (int)sqlca.sqlcode,
		     (int)sqlca.sqlerrd[1] );
	    exit( 1 );
	} else {
	    Release = major_vers;
	    EXEC SQL SELECT DBINFO( 'version', 'minor' )
		INTO :minor_vers
		FROM "informix".systables where tabid = 1;
	    if (sqlca.sqlcode < 0) {
		fprintf( stderr,
			 "FAILED: Error fetching minor version id.  SQL=%d, ISAM=%d.\n",
			 (int)sqlca.sqlcode,
			 (int)sqlca.sqlerrd[1] );
		exit( 1 );
	    }
	    EXEC SQL SELECT DBINFO( 'version', 'level' )
		INTO :level_vers
		FROM "informix".systables where tabid = 1;
	    if (sqlca.sqlcode < 0) {
		fprintf( stderr,
			 "FAILED: Error fetching patch level string.  SQL=%d, ISAM=%d.\n",
			 (int)sqlca.sqlcode,
			 (int)sqlca.sqlerrd[1] );
		exit( 1 );
	    }
	    EXEC SQL SELECT DBINFO( 'version', 'server-type' )
		INTO :server_type
		FROM "informix".systables where tabid = 1;
	    if (sqlca.sqlcode < 0) {
		fprintf( stderr,
			 "FAILED: Error fetching server type string.  SQL=%d, ISAM=%d.\n",
			 (int)sqlca.sqlcode,
			 (int)sqlca.sqlerrd[1] );
		exit( 1 );
	    }
	}

	if (do_scheduling && major_vers < 11) {
	    /* No scheduler available! */
	    fprintf( stderr,
		     "Task scheduling not supported by server version: %d.%d.%s\n",
		     major_vers,
		     minor_vers,
		     level_vers );
	    exit( 1 );
	}

	if (use_aus_thresholds && major_vers < 11) {
	    /* No scheduler available! */
	    fprintf( stderr,
		     "Auto Update Statistics thresholds not supported by server version: %d.%d.%s\n",
		     major_vers,
		     minor_vers,
		     level_vers );
	    exit( 1 );
	} else if (use_aus_thresholds) {
	    if (!doPDQ)
		EXEC SQL SELECT values
		    INTO :tablePDQ
		    FROM sysadmin:"informix".ph_threshold
		    WHERE name = 'AUS_PDQ';
	    if (!haveDistAgingLimit)
		EXEC SQL SELECT values
		    INTO :DistAgingLimit
		    FROM sysadmin:"informix".ph_threshold
		    WHERE name = 'AUS_AGE';
	    if (!haveBrowseThreshHold)
		EXEC SQL SELECT values
		    INTO :BrowseThreshHold
		    FROM sysadmin:"informix".ph_threshold
		    WHERE name = 'AUS_CHANGE';
	}

	ServerOptimized = is_Server_Optimized( major_vers, minor_vers, level_vers );

	UseServerOptimization = (ServerOptimized != ServerOptimizationOverridden);
	if (verbose >= 1)
	    fprintf( stderr,
		     "   Optimized server %s (%s %d.%d.%s). \n\tOptimizations %s.\n",
		     ServerOptimized ? "detected" : "not detected",
		     server_type,
		     major_vers,
		     minor_vers,
		     level_vers,
		     UseServerOptimization ? "followed" : "avoided" );
	else if (verbose > 0)
	    fprintf( stderr,
		     "   Optimized server %s.  Optimizations %s.\n",
		     ServerOptimized ? "detected" : "not detected",
		     UseServerOptimization ? "followed" : "avoided" );

	if (UseServerOptimization) {
	    if (tablePDQ == 0) {
		tablePDQ = 10;
		if (verbose > 0)
		    fprintf( stderr,
			     "   PDQPRIORITY for table processing reset to: %d%% to support optimized server.\n",
			     tablePDQ );
	    }
	    pstemp_env = getenv( "PSORT_NPROCS" );
	    if (pstemp_env == (char *)NULL || strlen( pstemp_env ) == 0) {
		fprintf( stderr,
			 "   Warning!  Optimized servers produce statistics fastest with PSORT_NPROCS set.\n" );
	    }
	    pstemp_env = getenv( "PSORT_DBTEMP" );
	    if (pstemp_env == (char *)NULL || strlen( pstemp_env ) == 0) {
		fprintf( stderr,
			 "   Warning!  Optimized servers produce statistics fastest with PSORT_DBTEMP set.\n" );
	    }
	    pstemp_env = getenv( "DBUPSPACE" );
	    if (pstemp_env == (char *)NULL || strlen( pstemp_env ) == 0) {
		fprintf( stderr,
			 "   Warning!  Optimized servers produce statistics fastest with DBUPSPACE set.\n" );
	    }
	}
	if (sampleset && major_vers < 11) {
	    fprintf( stderr, "   Minimum sample size specified but unsupported.  Ignoring.\n" );
	    sampleset = 0;
	}
    }

    if (do_procedure) {

	/* Connection to sysadmin for scheduling if needed. */
	if (username[0] != (char)0) {
	    if (passwd[0] != (char)0)
		EXEC SQL CONNECT TO :connhost AS 'SCHEDULER'
		    USER :username USING :passwd
		    WITH CONCURRENT TRANSACTION;
	    else {
		fprintf( stderr,
			 "Cannot specify INFORMIXUSER without INFORMIXPASS!\n" );
		usage( argv[0] );
		exit( 1 );
	    }
	} else
	    EXEC SQL CONNECT TO :connhost AS 'SCHEDULER'
		    WITH CONCURRENT TRANSACTION;
	if (sqlca.sqlcode) {
	    fprintf( stderr,
		     "FAILED: Cannot connect to %s (@%s); error=%d, isam=%d\n\n",
		     "sysadmin",
		     host,
		     (int)sqlca.sqlcode,
		     (int)sqlca.sqlerrd[1] );
	    if (sqlca.sqlcode == -27000) {
		fprintf( stderr,
			 "Task scheduling is not possible with a shared memory connection!\n"
			 "Change to a networked or pipe connection type.\n" );
	    }
	    exit(1);
	}
	EXEC SQL DATABASE "sysadmin";
	if (sqlca.sqlcode == -387) {
	    fprintf( stderr,
		     "Cannot open sysadmin database SQL: %d, ISAM: %d.\n"
		     "  You must be user informix or a DBSA user to use task scheduling!\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit( 1 );
	} else if (sqlca.sqlcode == -329 && sqlca.sqlerrd[1] == -111) {
	    fprintf( stderr,
		     "Sysadmin database does not exist!\nSysadmin must exist to use --procedure or --schedule\n" );
	    exit( 1 );
	} else if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Cannot open sysadmin database.  SQL: %d  ISAM: %d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit( 1 );
	}

	EXEC SQL SELECT count(*) INTO :tbl_chk FROM systables WHERE tabname = 'dostats_stmts';
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Cannot verify dostats_stmts table.  SQL: %d  ISAM: %d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit( 1 );
	}
	if (tbl_chk == 0) {
	    /* Table does not already exist.  Create it. */
	    EXEC SQL CREATE TABLE dostats_stmts (
		     function char(128),
		     sequence int,
		     continued boolean,
		     statement lvarchar(768) );
	    if (sqlca.sqlcode < 0) {
		fprintf( stderr,
			 "Cannot create dostats_stmts table.  SQL: %d  ISAM: %d.\n",
			 sqlca.sqlcode,
			 sqlca.sqlerrd[1] );
		exit( 1 );
	    }
	    EXEC SQL CREATE INDEX dostats_stmts_pk on dostats_stmts (function, sequence, continued desc);
	    if (sqlca.sqlcode < 0) {
		fprintf( stderr,
			 "Cannot index dostats_stmts table.  SQL: %d  ISAM: %d.\n",
			 sqlca.sqlcode,
			 sqlca.sqlerrd[1] );
		exit( 1 );
	    }
	}
	Out_Func = Store_Statement;
    } else {
	Out_Func = Write_Statement;
    }

    if (!singledb) {
	EXEC SQL SET CONNECTION 'DBASES';
	if (sqlca.sqlcode) {
	    fprintf( stderr,
		     "FAILED %d: Cannot set connection to (%s); error=%d, isam=%d\n",
		     __LINE__,
		     "DBASES",
		     (int)sqlca.sqlcode,
		     (int)sqlca.sqlerrd[1] );
	    exit(1);
	}
    }

    alldb = have_smi;
#if !defined(NOTIMINGS)
    if (ExpandReport)
	gettimeofday( &totstart, NULL );
#endif
    /* Loop on all <selected> databases. */
    for (;;) {
#if !defined(NOTIMINGS)
	if (ExpandReport)
	    gettimeofday( &dbsstart, NULL );
#endif
	/* Switch to connection for database work. */
	if (!have_smi || singledb) {
	    /* This is a 5.xx engine no sysmaster db simulate fetch. */
	    if (once) {
		sqlca.sqlcode = 100;
		break;
	    } else {
		sqlca.sqlcode = 0;
		once = 1;
	    }
	} else {
	    EXEC SQL SET CONNECTION 'DBASES';
	    conn = 1;
	    if (sqlca.sqlcode) {
		fprintf( stderr,
			 "FAILED %d: Cannot set connection to (%s); "
			 "error=%d, isam=%d\n",
			 __LINE__,
			 "DBASES",
			 (int)sqlca.sqlcode,
			 (int)sqlca.sqlerrd[1] );
		exit(1);
	    }
	    EXEC SQL FETCH dbc INTO :databasename;
	    if (sqlca.sqlcode == SQLNOTFOUND) {
		if (!processed)
		    fprintf( stderr,
			     "FAILED\nNo databases selected!.\n" );
		break;
	    } else if (sqlca.sqlcode == -329) {
		fprintf( stderr,
			 "FAILED: No permissions to database.  Rerun as: informix.  "
			 "Continuing.\n" );
		continue;
	    } else if (sqlca.sqlcode < 0) {
		fprintf( stderr,
			 "FAILED: Cannot fetch dbc cursor, SQLerr=%d, ISAM=%d.\n",
			 (int)sqlca.sqlcode,
			 (int)sqlca.sqlerrd[1] );
		exit(1);
	    }
	    processed++;
	}
	if (verbose >= 0)
	    fprintf( stderr, "\nWorking on database: %s.\n", databasename );

	if (do_procedure) {
	    if (schedfile != (FILE *)NULL) {
		/* Finish up: insert scheduler record. */

		if (do_proconly) {
		    if (verbose >= 2)
			fprintf( stderr, "Dropping procedure(1): %s.\n", schedprocname );
		    snprintf( schedule_stmt,
			      sizeof schedule_stmt,
			      "DROP PROCEDURE %s;",
			      schedprocname );
		    EXEC SQL EXECUTE IMMEDIATE :schedule_stmt;
		    if (sqlca.sqlcode != 0 && sqlca.sqlcode != -674) {
			fprintf( stderr,
				 "Error dropping old copy of procedure!  SQL: %d  ISAM: %d\n",
				 sqlca.sqlcode,
				 sqlca.sqlerrd[1] );
			exit( 1 );
		    }
		}

		if (do_scheduling)
		    fprintf( stderr,
			     "Scheduling stats update for database: %s.\n",
			     databasename );
		else
		    fprintf( stderr,
			     "Creating stats procedure %s in database: %s.\n",
			     schedprocname,
			     databasename );

		EXEC SQL CREATE ROUTINE FROM :schedprocfilename;
		if (sqlca.sqlcode != 0) {
		    fprintf( stderr,
			     "Error creating scheduling procedure: %s!  SQL: %d  ISAM: %d\n",
			     schedprocname,
			     sqlca.sqlcode,
			     sqlca.sqlerrd[1] );
		    exit( 1 );
		}
		unlink( schedprocfilename );
		EXEC SQL SET CONNECTION 'SCHEDULER';
		if (sqlca.sqlcode != 0) {
		    fprintf( stderr,
			     "Error changing connections to SCHEDULER!  SQL: %d  ISAM: %d\n",
			     sqlca.sqlcode,
			     sqlca.sqlerrd[1] );
		    exit( 1 );
		}
		if (do_scheduling) {
		    EXEC SQL SELECT count(*) INTO :OK FROM ph_group WHERE group_name = 'DOSTATS';
		    if (sqlca.sqlcode != 0) {
			fprintf( stderr,
				 "Error verifying DOSTATS task group record!  SQL: %d  ISAM: %d\n",
				 sqlca.sqlcode,
				 sqlca.sqlerrd[1] );
			exit( 1 );
		    }
		    if (OK == 0) {
			EXEC SQL INSERT INTO ph_group (group_id, group_name, group_description)
			    VALUES ( 0, "DOSTATS", "Update statistics generated by dostats(c)." );
			if (sqlca.sqlcode < 0 && sqlca.sqlcode != -268) {
			    fprintf( stderr,
				     "Error inserting DOSTATS task group record!  SQL: %d  ISAM: %d\n",
				     sqlca.sqlcode,
				     sqlca.sqlerrd[1] );
			    exit( 1 );
			}
		    }
		    
		    /* Add scheduling record here. --ASK-- */
		    snprintf( schedule_stmt,
			      sizeof schedule_stmt,
			      "INSERT INTO ph_task( tk_id, "
			      "       tk_name, tk_type, tk_group, tk_description, tk_execute, "
			      "       tk_start_time, tk_stop_time, tk_frequency, tk_Sunday, tk_Monday, "
			      "       tk_Tuesday, tk_Wednesday, tk_Thursday, tk_Friday, tk_Saturday ) "
			      "VALUES ( 0, '%s_task', 'TASK', 'DOSTATS', 'dostats run for %s', "
			      "   'EXECUTE PROCEDURE %s:%s($DATA_TASK_ID, $DATA_SEQ_ID, %s);', %s, %s, %s, %s, %s, %s, %s, %s, %s, %s );",
			      schedprocname,
			      schedprocname,
			      databasename,
			      schedprocname,
			      sch_onetime ? "\"t\"" : "\"f\"",
			      sch_starttime[0] == (char)(0) ? "NULL" : sch_starttime,
			      sch_lasttime[0] == (char)(0) ? "NULL" : sch_lasttime,
			      sch_frequency[0] == (char)(0) ? "NULL" : sch_frequency,
			      sch_sunday,
			      sch_monday,
			      sch_tuesday,
			      sch_wednesday,
			      sch_thursday,
			      sch_friday,
			      sch_saturday );
		    EXEC SQL EXECUTE IMMEDIATE :schedule_stmt;
		    if (sqlca.sqlcode != 0) {
			fprintf( stderr,
				 "Error inserting task scheduler record!  SQL: %d  ISAM: %d\n%s\n%*.*s\n",
				 sqlca.sqlcode,
				 sqlca.sqlerrd[1],
				 schedule_stmt,
				 sqlca.sqlerrd[4],
				 sqlca.sqlerrd[4],
				 "^" );
			exit( 1 );
		    }
		}
	    }

	    if (!do_proconly) {
		snprintf( schedprocname,
			  sizeof schedprocname,
			  "updstats_%s_%s_%d",
			  databasename,
			  tflag ? tablename : "all",
			  getpid() );
	    }

	    getcwd( pwd, sizeof pwd );

	    snprintf( schedprocfilename,
		      sizeof schedprocfilename,
		      "%s/%s.sql",
		      pwd,
		      schedprocname );
	    schedfile = fopen( schedprocfilename, "wt" );
	    outfile = schedfile;

	    fprintf( schedfile,
		     "CREATE PROCEDURE %s( taskid int, seqid int, clearit boolean );\n",
		     schedprocname );
	    fprintf( schedfile,
		     "    DEFINE cont, trailing boolean;\n" );
	    fprintf( schedfile,
		     "    DEFINE part char(800);\n" );
	    fprintf( schedfile,
		     "    DEFINE l_statement char(32767);\n" );
	    fprintf( schedfile,
		     "    DEFINE seq int;\n\n" );
	    fprintf( schedfile,
		     "    ON EXCEPTION IN (-705) \n    END EXCEPTION WITH RESUME;\n\n" );
	    if (verbose > 1) {
		fprintf( schedfile,
			 "    SET DEBUG FILE TO '%s.txt'; \n", schedprocname );
		fprintf( schedfile,
			 "    TRACE ON; \n" );
	    }
	    if (waitlen < 0)
		fprintf( schedfile,
			 "    SET LOCK MODE TO WAIT; \n" );
	    else if (waitlen == 0)
		fprintf( schedfile,
			 "    SET LOCK MODE TO NOT WAIT; \n" );
	    else 
		fprintf( schedfile,
			 "    SET LOCK MODE TO WAIT %d; \n", waitlen );
	    fprintf( schedfile,
		     "    LET l_statement = ''; \n" );
	    fprintf( schedfile,
		     "    LET trailing = 'f'; \n" );
	    fprintf( schedfile,
		     "    FOREACH SELECT sequence, continued, statement \n" );
	    fprintf( schedfile,
		     "            INTO seq, cont, part \n" );
	    fprintf( schedfile,
		     "            FROM sysadmin:dostats_stmts \n" );
	    fprintf( schedfile,
		     "            WHERE function = '%s' \n", schedprocname );
	    fprintf( schedfile,
		     "            ORDER BY 1, 2  desc \n" );
	    if (verbose > 1) {
		fprintf( schedfile, "    TRACE part;\n" );
		fprintf( schedfile, "    TRACE l_statement;\n" );
	    }
	    fprintf( schedfile,
		     "        IF (cont) THEN \n" );
	    fprintf( schedfile,
		     "            IF (trailing) THEN \n" );
	    fprintf( schedfile,
		     "                LET part = ' ' || part; \n" );
	    fprintf( schedfile,
		     "            END IF \n" );
	    fprintf( schedfile,
		     "            LET l_statement = rtrim( l_statement ) || rtrim( part ); \n" );
	    fprintf( schedfile,
		     "            IF (substr( part, 768 ) = ' ') THEN \n" );
	    fprintf( schedfile,
		     "               LET trailing = 't'; \n" );
	    fprintf( schedfile,
		     "            ELSE \n" );
	    fprintf( schedfile,
		     "               LET trailing = 'f'; \n" );
	    fprintf( schedfile,
		     "            END IF \n" );
	    fprintf( schedfile,
		     "        ELSE \n" );
	    fprintf( schedfile,
		     "            IF (trailing) THEN \n" );
	    fprintf( schedfile,
		     "                LET part = ' ' || part; \n" );
	    fprintf( schedfile,
		     "            END IF \n" );
	    fprintf( schedfile,
		     "            LET l_statement = rtrim( l_statement ) || rtrim( part ); \n" );
	    fprintf( schedfile,
		     "            EXECUTE IMMEDIATE l_statement; \n" );
	    fprintf( schedfile,
		     "            IF SQLCODE = -705 THEN \n" );
	    fprintf( schedfile,
		     "                CONTINUE FOREACH; \n" );
	    fprintf( schedfile,
		     "            END IF \n" );
	    fprintf( schedfile,
		     "            LET trailing = 'f'; \n" );
	    fprintf( schedfile,
		     "            LET l_statement = ''; \n" );
	    fprintf( schedfile,
		     "        END IF \n" );
	    fprintf( schedfile,
		     "    END FOREACH; \n" );
	    fprintf( schedfile,
		     "    IF (clearit = 't') THEN \n"
		     "        UPDATE sysadmin:ph_task SET tk_enable = 'f' WHERE tk_id = taskid; \n"
		     "        DELETE FROM sysadmin:dostats_stmts WHERE function = '%s'; \n "
		     "    END IF\n",
		     schedprocname );
	    if (verbose > 1)
		fprintf( schedfile, "    TRACE OFF; \n" );
	    fprintf( schedfile, "end procedure \n" );
	    fprintf( schedfile,
		     "DOCUMENT \n" );
	    fprintf( schedfile,
		     "    'Procedure created by dostats!'; \n" );
	    fflush( schedfile );
	    fclose( schedfile );
	}

	if (have_smi && !singledb) {
	    /* Switch to connection for table work. */
	    EXEC SQL SET CONNECTION 'WORK';
	    conn = 2;
	    if (sqlca.sqlcode) {
		fprintf( stderr,
			 "FAILED: %d: Cannot set connection to (%s); "
			 "error=%d, isam=%d\n",
			 __LINE__,
			 "WORK",
			 (int)sqlca.sqlcode,
			 (int)sqlca.sqlerrd[1] );
		exit(1);
	    }
	}

	EXEC SQL DATABASE :databasename;
	if (sqlca.sqlcode == -387) {
	    fprintf( stderr,
		     "FAILED: No permissions to database(%s@%s).  "
		     "Rerun as: informix.  "
		     "Continuing.\n",
		     databasename,
		     host );
	    continue;
	} else if (sqlca.sqlcode) {
	    fprintf( stderr,
		     "FAILED: Cannot access database (%s); error=%d, isam=%d\n",
		     databasename,
		     (int)sqlca.sqlcode,
		     (int)sqlca.sqlerrd[1] );
	    exit(1);
	}
	if (sqlca.sqlwarn.sqlwarn1 == 'W' && sqlca.sqlwarn.sqlwarn2 == 'W')
	    ANSIMode = 1;
	else
	    ANSIMode = 0;

	if (!do_procedure && outfile != (FILE *)NULL) {
	    fprintf( outfile, "DATABASE %s;\n", databasename );
	}
	if (!proconly) {
	    if (doexplain) {
		if (outfile == (FILE *)NULL)
		    EXEC SQL SET EXPLAIN ON;
		else
		    Out_Func( "SET EXPLAIN ON;" );
	    }
	    if (outfile == (FILE *)NULL)
		EXEC SQL EXECUTE IMMEDIATE :pdqstmt_2;
	    else
		Out_Func( pdqstmt_2 );
	    if (!statement) { /* First database, build statement. */
		/* Select for tables and index leading columns. */
		if (is_se)
		    appendStr( &statement,
			       &stmt_len,
			       "SELECT a.tabid, 0, 0, a.tabname, a.owner, "
			       "abs(b.part1), abs(b.part2), "
			       "abs(b.part3), abs(b.part4), "
			       "abs(b.part5), abs(b.part6), "
			       "abs(b.part7), abs(b.part8), "
			       "0, 0, 0, 0, 0, 0, 0, 0, " );
		else if (Release >= 9)
		    appendStr( &statement,
			       &stmt_len,
			       "SELECT a.tabid, a.partnum, a.nrows, a.tabname, a.owner, "
			       "indexkeyarray_out(b.indexkeys) "
			       "       AS allparts, " );
		else
		    appendStr( &statement,
			       &stmt_len,
			       "SELECT a.tabid, a.partnum, a.nrows, a.tabname, a.owner, "
			       "abs(b.part1), abs(b.part2), "
			       "abs(b.part3), abs(b.part4), "
			       "abs(b.part5), abs(b.part6), "
			       "abs(b.part7), abs(b.part8), "
			       "abs(b.part9), abs(b.part10), "
			       "abs(b.part11), abs(b.part12), "
			       "abs(b.part13), abs(b.part14), "
			       "abs(b.part15), abs(b.part16), " );
		if (doDistAging)
		    appendStr( &statement,
			       &stmt_len,
			       "(  SELECT (today - MIN(d.constructed)) "
			       "   FROM \"informix\".sysdistrib d "
			       "   WHERE d.tabid = a.tabid ) " );
		else
		    appendStr( &statement,
			       &stmt_len,
			       "0 " );
		if (Release >= 9) {
		    appendStr( &statement,
			       &stmt_len,
			       "FROM \"informix\".systables a, outer (\"informix\".sysindices b, outer "
			       "\"informix\".sysobjstate s) " );
		} else {
		    appendStr( &statement,
			       &stmt_len,
			       "FROM \"informix\".systables a, outer (\"informix\".sysindexes b, outer "
			       "\"informix\".sysobjstate s) " );
		}
		appendStr( &statement,
			   &stmt_len,
			   "WHERE a.tabid = b.tabid AND tabtype = 'T' " );
		if (!dosystem)
		    appendStr( &statement, &stmt_len, "AND a.tabid >= 100 " );
		appendStr( &statement, &stmt_len,
			   "AND a.tabid = s.tabid "
			   "AND s.objtype = 'I' "
			   "AND b.idxname = s.name "
			   "AND s.state != 'D' " );
		if (tflag && strcmp( tablename, "none" ) != 0) {
		    if (includes || subquery)
			appendStr( &statement,
				   &stmt_len,
				   "AND (a.tabname MATCHES '" );
		    else
			appendStr( &statement,
				   &stmt_len,
				   "AND a.tabname MATCHES '" );

		    appendStr( &statement, &stmt_len, tablename );
		    appendStr( &statement, &stmt_len, "' " );
		} else if (includes || subquery) {
		    appendStr( &statement, &stmt_len, "AND (a.tabname IS NULL " );
		}
		if (subquery) {
		    appendStr( &statement, &stmt_len, subquery );
		    appendStr( &statement, &stmt_len, ") " );
		}
		if (includes ) {
		    appendStr( &statement, &stmt_len, "OR a.tabname IN " );
		    appendStr( &statement, &stmt_len, includes );
		    appendStr( &statement, &stmt_len, ") " );
		}
		if (exquery) {
		    appendStr( &statement, &stmt_len, exquery );
		}
		if (excludes) {
		    appendStr( &statement,
			       &stmt_len,
			       "AND a.tabname NOT IN " );
		    appendStr( &statement, &stmt_len, excludes );
		}
		if (Release >= 9) {
		    /*		    appendStr( &statement,
				    &stmt_len,
				    "AND indexkeyarray_out(b.indexkeys) NOT LIKE '<%' " ); 
				    -- Enable to skip functional keys without columns -- */
		    appendStr( &statement,
			       &stmt_len,
			       "ORDER BY 1, 3, 4, 5; " );
		} else
		    appendStr( &statement,
			       &stmt_len,
			       "ORDER BY 1, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, "
			       "13, 14, 15, 16, 17, 18; " );
#if defined( TESTING )
		if (verbose > 0)
		    fprintf( stderr,
			     "Selecting tables with:\n%s\n\n",
			     statement );
#endif
	    }
	    EXEC SQL
		PREPARE stmt1 FROM :statement;
	    if (sqlca.sqlcode != 0) {
		fprintf( stderr,
			 "FAILED: Cannot PREPARE select statement for table data, "
			 "error: %d.\n",
			 (int)sqlca.sqlcode );
		fprintf( stderr, "Error at character %d.\nStatement is:\n",
			 (int)sqlca.sqlerrd[4] );
		fprintf( stderr, "%s\n\n",statement );
		exit(2);
	    }
	
	    EXEC SQL
		DECLARE a1 CURSOR WITH HOLD FOR stmt1;
	    if (sqlca.sqlcode != 0) {
		fprintf( stderr,
			 "FAILED: Cannot declare a1 cursor; Error = %d\n",
			 (int)sqlca.sqlcode );
		exit(1);
	    }
	
	    EXEC SQL
		OPEN a1;
	    if (sqlca.sqlcode == SQLNOTFOUND) {
		fprintf( stderr, "FAILED: No table matching \"%s\".\n", tablename );
		exit(3);
	    } else if (sqlca.sqlcode != 0) {
		fprintf( stderr,
			 "FAILED: Cannot open a1 cursor; Error = %d\n",
			 (int)sqlca.sqlcode );
		exit(1);
	    }

	    lasttbl[0] = (char)0;
	    i = 0;
	    /* Loop on tables/indexes. */
	    memset( &lastidx, 0, sizeof lastidx );
#if !defined(NOTIMINGS)
	    if (ExpandReport)
		gettimeofday( &tabstart, NULL );
#endif
		
	    for (;;) {
		if (Release >= 9) {
		    EXEC SQL FETCH a1
			INTO :tabinfo.tabid,
			:tabinfo.partnum,
			:tabinfo.nrows,
			:tabinfo.tabname,
			:tabinfo.tabowner,
			:tabinfo.keystring:tabinfo.colind,
			:tabinfo.days : tabinfo.daysind;
		} else {
		    EXEC SQL FETCH a1
			INTO :tabinfo.tabid,
			:tabinfo.partnum,
			:tabinfo.nrows,
			:tabinfo.tabname,
			:tabinfo.tabowner,
			:tabinfo.colno[0]:tabinfo.colind,
			:tabinfo.colno[1],
			:tabinfo.colno[2],
			:tabinfo.colno[3],
			:tabinfo.colno[4],
			:tabinfo.colno[5],
			:tabinfo.colno[6],
			:tabinfo.colno[7],
			:tabinfo.colno[8],
			:tabinfo.colno[9],
			:tabinfo.colno[10],
			:tabinfo.colno[11],
			:tabinfo.colno[12],
			:tabinfo.colno[13],
			:tabinfo.colno[14],
			:tabinfo.colno[15],
			:tabinfo.days : tabinfo.daysind;
		}
		if (sqlca.sqlcode ) {
		    if (sqlca.sqlcode == SQLNOTFOUND)
			break;
		    else {
			fprintf( stderr,
				 "FAILED: Cannot FETCH table data (a1).  SQLCODE = %d, ISAM = %d.\nStatement is:\n",
				 (int)sqlca.sqlcode,
				 (int)sqlca.sqlerrd[1] );
			fputs( statement, stderr );
			fputs( "\n\n", stderr );
			exit(1);
		    }
		} else {
		    int skip = 0;

		    if (ANSIMode == 1)
		        snprintf( ownerstr, sizeof ownerstr, "\"%s\".", tabinfo.tabowner );
		    else
		        memset( ownerstr, 0, sizeof ownerstr );

		    if (Release >= 9) {
			/* For IDS 9.xx+ keys */
			if (tabinfo.colind > 0)
			    fprintf( stderr,
				     "WARNING: Key information too long (%d)!  "
				     "Some keys may not be processed.\n",
				     tabinfo.colind );
			/* Need to decode the keystring into colno's. */
			memset( tabinfo.colno, 0, sizeof tabinfo.colno );
			InitKeyStr( tabinfo.keystring );
			for (i=0; i < 16; i++) {
			    tabinfo.colno[i] = ParseKeyStr();
			    if (tabinfo.colno[i] == 0)
				break;
			}
		    }

		    /* Ignore multiple indexes on a table with identical keys
                       except for the direction of each (ASC/DESC). */
		    if (tabinfo.tabid == lastidx.tabid &&
			memcmp( &(tabinfo.colno), &(lastidx.colno), sizeof tabinfo.colno ) == 0)
			continue;

		    if (doBrowse || doDistAging) {
			long long rowcnt;
			
			rowcnt = getrowcount( tabinfo.tabowner, tabinfo.tabname, tabinfo.tabid, tabinfo.partnum );
			if (doBrowse) {
			    /* Verify rowcount out of range, else continue. */
			    double diff;
			    if (tabinfo.nrows > 0 || rowcnt != tabinfo.nrows) {
				if (tabinfo.nrows > 0) {
				    diff = (((double)abs( tabinfo.nrows - rowcnt )
					     / (double)tabinfo.nrows)
					    * (double)100.00);
				    if (diff < BrowseThreshHold)
					skip = 1; /* Skip this table */
				} else /* No rows before.  Force stats to update! */
				    skip = 0;
			    } else /* No rows, no change. */
				skip = 1;
			}
			if (doDistAging && skip) {
			    /* Use Distribution Aging to include tables. */ 
			    if ((tabinfo.daysind < 0 && rowcnt > 0)
				|| (tabinfo.daysind == 0 && tabinfo.days > DistAgingLimit))
				skip = 0;
			} else if (doDistAging && !doBrowse) {
			    /* Use Distribution Aging to eliminate tables. */
			    if ((tabinfo.daysind == 0 && tabinfo.days <= DistAgingLimit)
				|| (tabinfo.daysind < 0 && rowcnt == 0))
				skip = 1;
			}
			if (skip) /* Skip this table it does not pass Browse and Aging tests. */
			    continue;
		    }
		    if (strcmp( lasttbl, tabinfo.tabname ) != 0) {

		        if (lasttbl[0] == (char)0)
			    if (verbose >= 0)
				fprintf( stderr,
					 "    Working on %s%s%s%s:\n",
					 ownerstr,
					 QuotedIdents ? "\"" : "",
					 tabinfo.tabname,
					 QuotedIdents ? "\"" : "" );

			/* Unless this is the first table, if optimizations followed process last table. */
			if (lasttbl[0] != (char)0 && UseServerOptimization) {
			    int col, highlen, medlen;
			    int medfirst = 1, highfirst = 1;
			    char highcols[32768], medcols[32768];

			    medlen = 32767 - (56 + strlen( lasttbl ));
			    highlen = medlen + 2;
			    highcols[0] = (char)0;
			    medcols[0] = (char)0;
			    for (col=0; col < ncols; col++) {
				if ((strlen( highcols ) + strlen( columnnames[col] ) + 1 ) > highlen) {
				    /* Name list is getting too long.  Issue command now and start over. */
				    snprintf( tempstr, sizeof tempstr,
					      "UPDATE STATISTICS HIGH FOR TABLE %s%s%s%s (%s)%s%s;",
					      lastowner,
					      QuotedIdents ? "\"" : "",
					      lasttbl,
					      QuotedIdents ? "\"" : "",
					      highcols,
					      resset ? resstring : "",
					      (NoDistOnly) ?
					      "" : " DISTRIBUTIONS ONLY");
				    if (doindex) {
					if (outfile == (FILE *)NULL) {
					    if (verbose >= 0)
						fprintf( stderr, "\t%s(HIGH)...", highcols );
#if !defined(NOTIMINGS)
					    if (ExpandReport)
					        gettimeofday( &stmtstart, NULL );
#endif
					    EXEC SQL EXECUTE IMMEDIATE :tempstr;
					    if (sqlca.sqlcode == -313) {
						fprintf( stderr,
							 "FAILED\nNo permissions to update statistics in"
							 " database: %s@%s.\nRerun as informix.  "
							 "Continuing with stored procedures anyway.\n",
							 databasename,
							 host );
						break;
					    } else if (sqlca.sqlcode == -740) {
						nrecs = getrowcount( tabinfo.tabowner, tabinfo.tabname, tabinfo.tabid, tabinfo.partnum );
						fprintf( stderr,
							 "FAILED\nMinimum resolution for this table is: %f\n",
							 minres( nrecs ) );
						exit(3);
					    } else if (sqlca.sqlcode) {
						fprintf( stderr,
							 "FAILED\nCannot Update statistics for table: "
							 "%s%s%s%s, column %s; "
							 "\n\tSQLCODE = %d, ISAM = %d, POS = %d\n",
							 ownerstr,
							 QuotedIdents ? "\"" : "",
							 tabinfo.tabname,
							 QuotedIdents ? "\"" : "",
							 columnnames[column],
							 (int)sqlca.sqlcode,
							 (int)sqlca.sqlerrd[1],
							 (int)sqlca.sqlerrd[4] );
						if (sqlca.sqlcode == -201)
						    fprintf( stderr, "%s\n", tempstr );
						if (!noexit) {
						    exit(1);
						} else {
						    noexit--;
						}
					    } else
#if !defined(NOTIMINGS)
						if (ExpandReport) {
						    gettimeofday( &stmtend, NULL );
						    timediff( stmtstart,
							      stmtend,
							      &secs,
							      &msecs );
						    if (ExpandReport == 2) {
							fprintf( stderr,
								 "SUCCESSFUL\t(%d hrs %d mins %d.%03.3d Secs)\n",
								 (secs / 3600),
								 ((secs % 3600) / 60),
								 (secs % 60),
								 msecs );
						    } else {
							fprintf( stderr,
								 "SUCCESSFUL\t(%d.%03.3d Secs)\n",
								 secs,
								 msecs );
						    }
						} else 
#endif
						    if (verbose >= 0)
							fprintf( stderr, "SUCCESSFUL\n" );
					} else {
					    if (verbose >= 0)
						fprintf( stderr, "\t%s(HIGH)", highcols );
					    Out_Func( tempstr );
					    if (verbose >= 0)
						fprintf( stderr, "\n" );
					}
				    }
				    if (ANSIMode)
					if (!outfile) {
					    if (verbose >= 2)
						fprintf( stderr, "CW st: %d.\n", __LINE__ );
					    EXEC SQL COMMIT WORK;
					} else {
					    Out_Func( "COMMIT WORK;" );
					    if (verbose >= 2)
						fprintf( stderr, "CW st: %d.\n", __LINE__ );
					}
				    highfirst = 1;
				    highcols[0] = (char)0;
				}
				if ((strlen( medcols ) + strlen( columnnames[col] ) + 1) > medlen) {
				    /* Name list is getting too long.  Issue command now and start over. */
				    snprintf( tempstr, sizeof tempstr,
					      "UPDATE STATISTICS MEDIUM FOR TABLE %s%s%s%s (%s)%s%s%s;",
					      lastowner,
					      QuotedIdents ? "\"" : "",
					      lasttbl,
					      QuotedIdents ? "\"" : "",
					      medcols,
					      sampleset ? samplestring : "",
					      mresset ? mresstring : "",
					      (mresset && confset) ? confstring : "" );
				    if (doindex) {
					if (outfile == (FILE *)NULL) {
#if !defined(NOTIMINGS)
					    if (ExpandReport)
					        gettimeofday( &stmtstart, NULL );
#endif
					    if (verbose >= 0)
						fprintf( stderr, "\t%s(MEDIUM)...", medcols );
					    EXEC SQL EXECUTE IMMEDIATE :tempstr;
					    if (sqlca.sqlcode == -313) {
						fprintf( stderr,
							 "FAILED\nNo permissions to update statistics in"
							 " database: %s@%s.\nRerun as informix.  "
							 "Continuing with stored procedures anyway.\n",
							 databasename,
							 host );
						break;
					    } else if (sqlca.sqlcode == -740) {
						nrecs = getrowcount( tabinfo.tabowner, tabinfo.tabname, tabinfo.tabid, tabinfo.partnum );
						fprintf( stderr,
							 "FAILED\nMinimum resolution for this table is: %f\n",
							 minres( nrecs ) );
						exit(3);
					    } else if (sqlca.sqlcode) {
						fprintf( stderr,
							 "FAILED\nCannot Update statistics for table: "
							 "%s%s%s%s, column %s; "
							 "\n\tSQLCODE = %d, ISAM = %d, POS = %d\n",
							 ownerstr,
							 QuotedIdents ? "\"" : "",
							 tabinfo.tabname,
							 QuotedIdents ? "\"" : "",
							 columnnames[column],
							 (int)sqlca.sqlcode,
							 (int)sqlca.sqlerrd[1],
							 (int)sqlca.sqlerrd[4] );
						if (sqlca.sqlcode == -201)
						    fprintf( stderr, "%s\n", tempstr );
                                                if (!noexit) {
                                                    exit(1);
                                                } else {
                                                    noexit--;
                                                }
					    } else
#if !defined(NOTIMINGS)
						if (ExpandReport) {
						    gettimeofday( &stmtend, NULL );
						    timediff( stmtstart,
							      stmtend,
							      &secs,
							      &msecs );
						    if (ExpandReport == 2) {
							fprintf( stderr,
								 "SUCCESSFUL\t(%d hrs %d mins %d.%03.3d Secs)\n",
								 (secs / 3600),
								 ((secs % 3600) / 60),
								 (secs % 60),
								 msecs );
						    } else {
							fprintf( stderr,
								 "SUCCESSFUL\t(%d.%03.3d Secs)\n",
								 secs,
								 msecs );
						    }
						} else 
#endif
						    if (verbose >= 0)
							fprintf( stderr, "SUCCESSFUL\n" );
					} else {
					    if (verbose >= 0)
						fprintf( stderr, "\t%s(MEDIUM)", medcols );
					    Out_Func( tempstr );
					    if (verbose >= 0)
						fprintf( stderr, "\n" );
					}
					if (ANSIMode)
					    if (!outfile) {
						if (verbose >= 2)
						    fprintf( stderr, "CW st: %d.\n", __LINE__ );
						EXEC SQL COMMIT WORK;
					    } else {
						Out_Func( "COMMIT WORK;" );
						if (verbose >= 2)
						    fprintf( stderr, "CW st: %d.\n", __LINE__ );
					    }
				    }
				    medfirst = 1;
				    medcols[0] = (char)0;
				}
				if (columns[col] & 1) {
				    /* Include column in HIGH. */
				    if (highfirst) {
					highfirst = 0;
				    } else {
					strcat( highcols, ", " );
				    }
				    strcat( highcols, columnnames[col] );
				} else {
				    /* Include column in MEDIUM. */
				    if (medfirst) {
					medfirst = 0;
				    } else {
					strcat( medcols, ", " );
				    }
				    strcat( medcols, columnnames[col] );
				}
			    }

			    if (strlen( highcols )) {
				snprintf( tempstr, sizeof tempstr,
					  "UPDATE STATISTICS HIGH FOR TABLE %s%s%s%s (%s)%s%s;",
					  lastowner,
					  QuotedIdents ? "\"" : "",
					  lasttbl,
					  QuotedIdents ? "\"" : "",
					  highcols,
					  resset ? resstring : "",
					  (NoDistOnly) ?
					  "" : " DISTRIBUTIONS ONLY");
				if (doindex) {
				    if (outfile == (FILE *)NULL) {
					if (verbose >= 0)
					    fprintf( stderr, "\t%s(HIGH)...", highcols );
#if !defined(NOTIMINGS)
					if (ExpandReport)
					    gettimeofday( &stmtstart, NULL );
#endif
					EXEC SQL EXECUTE IMMEDIATE :tempstr;
					if (sqlca.sqlcode == -313) {
					    fprintf( stderr,
						     "FAILED\nNo permissions to update statistics in"
						     " database: %s@%s.\nRerun as informix.  "
						     "Continuing with stored procedures anyway.\n",
						     databasename,
						     host );
					    break;
					} else if (sqlca.sqlcode == -740) {
					    nrecs = getrowcount( tabinfo.tabowner, tabinfo.tabname, tabinfo.tabid, tabinfo.partnum );
					    fprintf( stderr,
						     "FAILED\nMinimum resolution for this table is: %f\n",
						     minres( nrecs ) );
					    exit(3);
					} else if (sqlca.sqlcode) {
					    fprintf( stderr,
						     "FAILED\nCannot Update statistics for table: "
						     "%s%s%s%s, column %s; "
						     "\n\tSQLCODE = %d, ISAM = %d, POS = %d\n",
						     ownerstr,
						     QuotedIdents ? "\"" : "",
						     tabinfo.tabname,
						     QuotedIdents ? "\"" : "",
						     columnnames[column],
						     (int)sqlca.sqlcode,
						     (int)sqlca.sqlerrd[1],
						     (int)sqlca.sqlerrd[4] );
					    if (sqlca.sqlcode == -201)
						fprintf( stderr, "%s\n", tempstr );
					    if (!noexit) {
						exit(1);
					    } else {
						noexit--;
					    }
					} else
#if !defined(NOTIMINGS)
					    if (ExpandReport) {
						gettimeofday( &stmtend, NULL );
						timediff( stmtstart,
							  stmtend,
							  &secs,
							  &msecs );
						if (ExpandReport == 2) {
						    fprintf( stderr,
							     "SUCCESSFUL\t(%d hrs %d mins %d.%03.3d Secs)\n",
							     (secs / 3600),
							     ((secs % 3600) / 60),
							     (secs % 60),
							     msecs );
						} else {
						    fprintf( stderr,
							     "SUCCESSFUL\t(%d.%03.3d Secs)\n",
							     secs,
							     msecs );
						}
					    } else 
#endif
						if (verbose >= 0)
						    fprintf( stderr, "SUCCESSFUL\n" );
				    } else {
					if (verbose >= 0)
					    fprintf( stderr, "\t%s(HIGH)", highcols );
					Out_Func( tempstr );
					if (verbose >= 0)
					    fprintf( stderr, "\n" );
				    }
				    if (ANSIMode)
					if (!outfile) {
					    if (verbose >= 2)
						fprintf( stderr, "CW st: %d.\n", __LINE__ );
					    EXEC SQL COMMIT WORK;
					} else {
					    Out_Func( "COMMIT WORK;" );
					    if (verbose >= 2)
						fprintf( stderr, "CW st: %d.\n", __LINE__ );
					}
				}
				highcols[0] = (char)0;
			    }
			    if (strlen( medcols )) {
				snprintf( tempstr, sizeof tempstr,
					  "UPDATE STATISTICS MEDIUM FOR TABLE %s%s%s%s (%s)%s%s%s;",
					  lastowner,
					  QuotedIdents ? "\"" : "",
					  lasttbl,
					  QuotedIdents ? "\"" : "",
					  medcols,
					  sampleset ? samplestring : "",
					  mresset ? mresstring : "",
					  (mresset && confset) ? confstring : "" );
				if (doindex) {
				    if (outfile == (FILE *)NULL) {
					if (verbose >= 0)
					    fprintf( stderr, "\t%s(MEDIUM)...", medcols );
#if !defined(NOTIMINGS)
					if (ExpandReport)
					    gettimeofday( &stmtstart, NULL );
#endif
					EXEC SQL EXECUTE IMMEDIATE :tempstr;
					if (sqlca.sqlcode == -313) {
					    fprintf( stderr,
						     "FAILED\nNo permissions to update statistics in"
						     " database: %s@%s.\nRerun as informix.  "
						     "Continuing with stored procedures anyway.\n",
						     databasename,
						     host );
					    break;
					} else if (sqlca.sqlcode == -740) {
					    nrecs = getrowcount( tabinfo.tabowner, tabinfo.tabname, tabinfo.tabid, tabinfo.partnum );
					    fprintf( stderr,
						     "FAILED\nMinimum resolution for this table is: %f\n",
						     minres( nrecs ) );
					    exit(3);
					} else if (sqlca.sqlcode) {
					    fprintf( stderr,
						     "FAILED\nCannot Update statistics for table: "
						     "%s%s%s%s, column %s; "
						     "\n\tSQLCODE = %d, ISAM = %d, POS = %d\n",
						     ownerstr,
						     QuotedIdents ? "\"" : "",
						     tabinfo.tabname,
						     QuotedIdents ? "\"" : "",
						     columnnames[column],
						     (int)sqlca.sqlcode,
						     (int)sqlca.sqlerrd[1],
						     (int)sqlca.sqlerrd[4] );
					    if (sqlca.sqlcode == -201)
						fprintf( stderr, "%s\n", tempstr );
					    if (!noexit) {
						exit(1);
					    } else {
						noexit--;
					    }
					} else
#if !defined(NOTIMINGS)
					    if (ExpandReport) {
						gettimeofday( &stmtend, NULL );
						timediff( stmtstart,
							  stmtend,
							  &secs,
							  &msecs );
						if (ExpandReport == 2) {
						    fprintf( stderr,
							     "SUCCESSFUL\t(%d hrs %d mins %d.%03.3d Secs)\n",
							     (secs / 3600),
							     ((secs % 3600) / 60),
							     (secs % 60),
							     msecs );
						} else {
						    fprintf( stderr,
							     "SUCCESSFUL\t(%d.%03.3d Secs)\n",
							     secs,
							     msecs );
						}
					    } else 
#endif
						if (verbose >= 0)
						    fprintf( stderr, "SUCCESSFUL\n" );
				    } else {
					if (verbose >= 0)
					    fprintf( stderr, "\t%s(MEDIUM)", medcols );
					Out_Func( tempstr );
					if (verbose >= 0)
					    fprintf( stderr, "\n" );
				    }
				    if (ANSIMode)
					if (!outfile) {
					    if (verbose >= 2)
						fprintf( stderr, "CW st: %d.\n", __LINE__ );
					    EXEC SQL COMMIT WORK;
					} else {
					    Out_Func( "COMMIT WORK;" );
					    if (verbose >= 2)
						fprintf( stderr, "CW st: %d.\n", __LINE__ );
					}
				}
				medcols[0] = (char)0;
			    }
			}
#if !defined(NOTIMINGS)
			if (outfile == (FILE *)NULL && ExpandReport && lasttbl[0] != (char)0) {
			    gettimeofday( &tabend, NULL );
			    timediff( tabstart,
				      tabend,
				      &secs,
				      &msecs );
			    if (ExpandReport == 2) {
				fprintf( stderr,
					 "    Table %s%s%s%s completed \t(%d hrs %d mins %d.%03.3d Secs)\n",
					 lastowner,
					 QuotedIdents ? "\"" : "",
					 lasttbl,
					 QuotedIdents ? "\"" : "",
					 (secs / 3600),
					 ((secs % 3600) / 60),
					 (secs % 60),
					 msecs );
			    } else 
				fprintf( stderr,
					 "    Table %s%s%s%s completed \t(%d.%03.3d Secs)\n",
					 lastowner,
					 QuotedIdents ? "\"" : "",
					 lasttbl,
					 QuotedIdents ? "\"" : "",
					 secs,
					 msecs );
			    gettimeofday( &tabstart, NULL );
			}
#endif
			/* New table.  Update stats to medium for the whole table. */
			memset( &lastidx, 0, sizeof lastidx );
			if (dotable) {
#if !defined(NOTIMINGS)
			    if (ExpandReport)
				gettimeofday( &stmtstart, NULL );
#endif
			    
			    if (tabmedium) {
				/* If no indexes, or Vers 5 or 6, do basic stats also. */
				if (!UseServerOptimization) {
				    snprintf( tempstr, sizeof tempstr,
					      "UPDATE STATISTICS MEDIUM FOR TABLE %s%s%s%s%s%s%s%s;",
					      ownerstr,
					      QuotedIdents ? "\"" : "",
					      tabinfo.tabname,
					      QuotedIdents ? "\"" : "",
					      sampleset ? samplestring : "",
					      mresset ? mresstring : "",
					      (mresset && confset) ? confstring : "",
					      (NoDistOnly || tabinfo.colind == -1)
					      ? "" : " DISTRIBUTIONS ONLY");
				}
			    } else {
				snprintf( tempstr, sizeof tempstr,
					  "UPDATE STATISTICS %s FOR TABLE %s%s%s%s%s;",
					  NoLevel ? "" : "LOW",
					  ownerstr,
					  QuotedIdents ? "\"" : "",
					  tabinfo.tabname,
					  QuotedIdents ? "\"" : "",
					  (NoDistOnly || !dropdists)
					  ? "" : " DROP DISTRIBUTIONS");
			    }
			    if (!tabmedium || !UseServerOptimization) {
				if (outfile == (FILE *)NULL) {
				    if (verbose >= 0)
					fprintf( stderr,
						 "%s%s%s(%s)...",
						 QuotedIdents ? "\"" : "",
						 tabinfo.tabname,
						 QuotedIdents ? "\"" : "",
						 tabmedium ? "MEDIUM" : "LOW" );
#if !defined(NOTIMINGS)
				    if (ExpandReport)
					gettimeofday( &stmtstart, NULL );
#endif
				    EXEC SQL EXECUTE IMMEDIATE :tempstr;
				    if (sqlca.sqlcode == -313) {
					fprintf( stderr,
						 "FAILED\nNo permissions to update statistics in"
						 " database: %s@%s.\nRerun as informix.  "
						 "Continuing with stored procedures anyway.\n",
						 databasename,
						 host );
					break;
				    } else if (sqlca.sqlcode == -740) {
					nrecs = getrowcount( tabinfo.tabowner, tabinfo.tabname, tabinfo.tabid, tabinfo.partnum );
					fprintf( stderr,
						 "FAILED\nMinimum resolution for this table is: %f\n",
						 minres( nrecs ) );
					exit(3);
				    } else if (sqlca.sqlcode) {
					fprintf( stderr,
						 "FAILED\nCannot Update statistics for table: %s%s%s%s; "
						 "\n\tSQLCODE = %d, ISAM = %d, POS = %d\n",
						 ownerstr,
						 QuotedIdents ? "\"" : "",
						 tabinfo.tabname,
						 QuotedIdents ? "\"" : "",
						 (int)sqlca.sqlcode,
						 (int)sqlca.sqlerrd[1],
						 (int)sqlca.sqlerrd[4] );
					if (sqlca.sqlcode == -201)
					    fprintf( stderr, "%s\n", tempstr );
					if (!noexit) {
					    exit(1);
					} else {
					    noexit--;
					}
				    } else {
#if !defined(NOTIMINGS)
					if (ExpandReport) {
					    gettimeofday( &stmtend, NULL );
					    timediff( stmtstart,
						      stmtend,
						      &secs,
						      &msecs );
					    if (ExpandReport == 2) {
						fprintf( stderr,
							 "SUCCESSFUL\t:"
							 "(%d hrs %d mins %d.%03.3d Secs)\n",
							 (secs / 3600),
							 ((secs % 3600) / 60),
							 (secs % 60),
							 msecs );
					    } else 
						fprintf( stderr,
							 "SUCCESSFUL\t(%d.%03.3d Secs)\n",
							 secs,
							 msecs);
					} else 
#endif
					    if (verbose >= 0)
						fprintf( stderr, "SUCCESSFUL\n" );
				    }
				} else {
				    if (verbose >= 0)
					fprintf( stderr,
						 "%s%s%s(%s)",
						 QuotedIdents ? "\"" : "",
						 tabinfo.tabname,
						 QuotedIdents ? "\"" : "",
						 tabmedium ? "MEDIUM" : "LOW" );
				    Out_Func( tempstr );
				    if (verbose >= 0)
					fprintf( stderr, "\n" );
				}
				if (ANSIMode)
				    if (!outfile) {
					if (verbose >= 2)
					    fprintf( stderr, "CW st: %d.\n", __LINE__ );
					EXEC SQL COMMIT WORK;
				    } else {
					Out_Func( "COMMIT WORK;" );
					if (verbose >= 2)
					    fprintf( stderr, "CW st: %d.\n", __LINE__ );
				    }
			    }
			}
			snprintf( lasttbl, sizeof lasttbl, "%s", tabinfo.tabname );
			if (ANSIMode == 1)
			    snprintf( lastowner, sizeof lastowner, "\"%s\".", tabinfo.tabowner );
			else
			    memset( lastowner, 0, sizeof lastowner );

			ncols = 0;
			snprintf( statement2, sizeof statement2,
				  "SELECT colno, colname "
				  "    FROM \"informix\".syscolumns "
				  "    WHERE tabid = %d"
				  "    ORDER BY colno",
				  tabinfo.tabid );
			EXEC SQL PREPARE stmt2 FROM :statement2;
			if (sqlca.sqlcode != 0) {
			    fprintf( stderr,
				     "FAILED\nCannot PREPARE stmt2.  SQL=%d, ISAM=%d, "
				     "POS=%d.\n",
				     (int)sqlca.sqlcode,
				     (int)sqlca.sqlerrd[1],
				     (int)sqlca.sqlerrd[4] );
			    exit(1);
			}
			EXEC SQL DECLARE curs2 CURSOR WITH HOLD FOR stmt2;
			if (sqlca.sqlcode != 0) {
			    fprintf( stderr,
				     "FAILED\nCannot DECLARE curs2.  "
				     "SQL=%d, ISAM=%d, POS=%d.\n",
				     (int)sqlca.sqlcode,
				     (int)sqlca.sqlerrd[1],
				     (int)sqlca.sqlerrd[4] );
			    exit(1);
			}
			EXEC SQL OPEN curs2;
			if (sqlca.sqlcode != 0) {
			    fprintf( stderr,
				     "FAILED\nCannot OPEN curs2.  SQL=%d, ISAM=%d.\n",
				     (int)sqlca.sqlcode,
				     (int)sqlca.sqlerrd[1] );
			    exit(1);
			}
			/* Initialize flags. */
			memset( columns, 0, sizeof columns );  
			for (ncols=0; ncols < 32768; ncols++) {
			    EXEC SQL FETCH curs2
				INTO :colno, :columnnames[ncols];
			    if (sqlca.sqlcode < 0) {
				fprintf( stderr,
					 "FAILED\nCannot FETCH curs2.  "
					 "SQL=%d, ISAM=%d.\n",
					 (int)sqlca.sqlcode,
					 (int)sqlca.sqlerrd[1] );
				exit(1);
			    } else if (sqlca.sqlcode == SQLNOTFOUND)
				break;
			}
		    } else {
			if (verbose >= 0)
			    fprintf( stderr,
				     "    Working on %s%s%s%s:\n",
				     ownerstr,
				     QuotedIdents ? "\"" : "",
				     lasttbl,
				     QuotedIdents ? "\"" : "" );
		    }
		    if (ANSIMode == 1)
		        snprintf( ownerstr, sizeof ownerstr, "\"%s\".", tabinfo.tabowner );
		    else
		        memset( ownerstr, 0, sizeof ownerstr );
		    if (tabinfo.colind >= 0) {
			if (strcmp( lasttbl, tabinfo.tabname ) != 0) {
			    if (verbose >= 0)
				fprintf( stderr,
					 "    Working on %s%s%s%s:\n",
					 ownerstr,
					 QuotedIdents ? "\"" : "",
					 tabinfo.tabname,
					 QuotedIdents ? "\"" : "" );
			}
			/* This table has index(es).  Update statistics high for the
			   first column in each index and low for the whole key. */
			column = ((tabinfo.colno[0] > 0) ?
				  tabinfo.colno[0] :
				  (-1 * tabinfo.colno[0])) - 1;
			/* Make certain we did not do this columns already. */
			if ((columns[column] % 2) == 0) { 
			    columns[column] |= 1;  /* Set flag 'HIGH DONE' for this column. */
			    if (UseServerOptimization || !(NoDistOnly || (tabinfo.colno[1]==0 && !ForceDistOnly)))
				columns[column] |= 4; /* Set 'DISTRIBUTIONS ONLY DONE' flag for this column. */
#if !defined(NOTIMINGS)
			    if (ExpandReport)
				gettimeofday( &stmtstart, NULL );
#endif

			    if (!UseServerOptimization) {
				snprintf( tempstr, sizeof tempstr,
					  "UPDATE STATISTICS HIGH FOR TABLE %s%s%s%s (%s)%s%s;",
					  ownerstr,
					  QuotedIdents ? "\"" : "",
					  tabinfo.tabname,
					  QuotedIdents ? "\"" : "",
					  columnnames[column],
					  resset ? resstring : "",
					  (NoDistOnly || (tabinfo.colno[1]==0 && !ForceDistOnly)) ?
					  "" : " DISTRIBUTIONS ONLY");
				if (doindex) {
				    if (outfile == (FILE *)NULL) {
					if (verbose >= 0)
					    fprintf( stderr, "\t%s(HIGH)...", columnnames[column] );
#if !defined(NOTIMINGS)
					if (ExpandReport)
					    gettimeofday( &stmtstart, NULL );
#endif
					EXEC SQL EXECUTE IMMEDIATE :tempstr;
					if (sqlca.sqlcode == -313) {
					    fprintf( stderr,
						     "FAILED\nNo permissions to update statistics in"
						     " database: %s@%s.\nRerun as informix.  "
						     "Continuing with stored procedures anyway.\n",
						     databasename,
						     host );
					    break;
					} else if (sqlca.sqlcode == -740) {
					    nrecs = getrowcount( tabinfo.tabowner, tabinfo.tabname, tabinfo.tabid, tabinfo.partnum );
					    fprintf( stderr,
						     "FAILED\nMinimum resolution for this table is: %f\n",
						     minres( nrecs ) );
					    exit(3);
					} else if (sqlca.sqlcode) {
					    fprintf( stderr,
						     "FAILED\nCannot Update statistics for table: "
						     "%s%s%s%s, column %s; "
						     "\n\tSQLCODE = %d, ISAM = %d, POS = %d\n",
						     ownerstr,
						     QuotedIdents ? "\"" : "",
						     tabinfo.tabname,
						     QuotedIdents ? "\"" : "",
						     columnnames[column],
						     (int)sqlca.sqlcode,
						     (int)sqlca.sqlerrd[1],
						     (int)sqlca.sqlerrd[4] );
					    if (sqlca.sqlcode == -201)
						fprintf( stderr, "%s\n", tempstr );
					    if (!noexit) {
						exit(1);
					    } else {
						noexit--;
					    }
					} else
#if !defined(NOTIMINGS)
					    if (ExpandReport) {
						gettimeofday( &stmtend, NULL );
						timediff( stmtstart,
							  stmtend,
							  &secs,
							  &msecs );
						if (ExpandReport == 2) {
						    fprintf( stderr,
							     "SUCCESSFUL\t(%d hrs %d mins %d.%03.3d Secs)\n",
							     (secs / 3600),
							     ((secs % 3600) / 60),
							     (secs % 60),
							     msecs );
						} else {
						    fprintf( stderr,
							     "SUCCESSFUL\t(%d.%03.3d Secs)\n",
							     secs,
							     msecs );
						}
					    } else 
#endif
						if (verbose >= 0)
						    fprintf( stderr, "SUCCESSFUL\n" );
				    } else {
					if (verbose >= 0)
					    fprintf( stderr, "\t%s(HIGH)", columnnames[column] );
					Out_Func( tempstr );
					if (verbose >= 0)
					    fprintf( stderr, "\n" );
				    }
				    if (ANSIMode)
					if (!outfile) {
					    if (verbose >= 2)
						fprintf( stderr, "CW st: %d.\n", __LINE__ );
					    EXEC SQL COMMIT WORK;
					} else {
					    Out_Func( "COMMIT WORK;" );
					    if (verbose >= 2)
						fprintf( stderr, "CW st: %d.\n", __LINE__ );
					}
				}
			    }
			}
			/* For multi-column indexes starting with common column
			   lists do update statistics high for first different
			   column. */
			if (lastidx.tabname[0] != (char)0) {
			    /* Don't need to do this for the first index. */
			    leadsame = 0;
			    for (jj=0; lastidx.colno[jj] && tabinfo.colno[jj]; jj++) {
				/* Compare first columns of index to last index. */
				if (lastidx.colno[jj] == tabinfo.colno[jj])
				    leadsame = 1;
				else
				    break;
			    }
			    /* If there is work to do make sure we did not do this
			       column already! */
			    column = ((tabinfo.colno[jj] > 0) ?
				      tabinfo.colno[jj] :
				      (-1 * tabinfo.colno[jj])) - 1;
			    if (doindex &&
				!NoLevel &&
				leadsame &&
				column >= 0 &&
				(columns[column] % 2) == 0)
			    {
				columns[column] |= 1;  /* Set 'HIGH DONE' flag for this column. */
				if (UseServerOptimization || !(NoDistOnly || (tabinfo.colno[1]==0 && !ForceDistOnly)))
				    columns[column] |= 4; /* Set 'DISTRIBUTIONS ONLY DONE' flag for this column. */
#if !defined(NOTIMINGS)
				if (ExpandReport)
				    gettimeofday( &stmtstart, NULL );
#endif
			    
				if (!UseServerOptimization) {
				    snprintf( tempstr, sizeof tempstr,
					      "UPDATE STATISTICS HIGH FOR TABLE %s%s%s%s ( %s )%s%s;",
					      ownerstr,
					      QuotedIdents ? "\"" : "",
					      tabinfo.tabname,
					      QuotedIdents ? "\"" : "",
					      columnnames[column],
					      resset ? resstring : "",
					      (NoDistOnly ||
					       (tabinfo.colno[1]==0
						&& !ForceDistOnly))
					      ? "" : " DISTRIBUTIONS ONLY" );
				    if (outfile == (FILE *)NULL) {
					if (verbose >= 0)
					    fprintf( stderr, "\t%s(HIGH)...", columnnames[column] );
#if !defined(NOTIMINGS)
					if (ExpandReport)
					    gettimeofday( &stmtstart, NULL );
#endif
					EXEC SQL EXECUTE IMMEDIATE :tempstr;
					if (sqlca.sqlcode == -313) {
					    fprintf( stderr,
						     "FAILED\nNo permissions to update statistics in"
						     " database: %s@%s.\nRerun as informix.  "
						     "Continuing with stored procedures anyway.\n",
						     databasename,
						     host );
					    break;
					} else if (sqlca.sqlcode == -740) {
					    nrecs = getrowcount( tabinfo.tabowner, tabinfo.tabname, tabinfo.tabid, tabinfo.partnum );
					    fprintf( stderr,
						     "FAILED\nMinimum resolution for this table is: %f\n",
						     minres( nrecs ) );
					    exit( 3 );
					} else if (sqlca.sqlcode) {
					    fprintf( stderr,
						     "FAILED\nCannot Update statistics for table: "
						     "%s%s%s%s, column %s; "
						     "\n\tSQLCODE = %d, ISAM = %d, POS = %d\n",
						     ownerstr,
						     QuotedIdents ? "\"" : "",
						     tabinfo.tabname,
						     QuotedIdents ? "\"" : "",
						     columnnames[column],
						     (int)sqlca.sqlcode,
						     (int)sqlca.sqlerrd[1],
						     (int)sqlca.sqlerrd[4] );
					    if (sqlca.sqlcode == -201)
						fprintf( stderr, "%s\n", tempstr );
					    if (!noexit) {
						exit(1);
					    } else {
						noexit--;
					    }
					} else
#if !defined(NOTIMINGS)
					    if (ExpandReport) {
						gettimeofday( &stmtend, NULL );
						timediff( stmtstart,
							  stmtend,
							  &secs,
							  &msecs );
						if (ExpandReport == 2) {
						    fprintf( stderr,
							     "SUCCESSFUL\t(%d hrs %d mins %d.%03.3d Secs)\n",
							     (secs / 3600),
							     ((secs % 3600) / 60),
							     (secs % 60),
							     msecs );
						} else 
						    fprintf( stderr,
							     "SUCCESSFUL\t(%d.%03.3d Secs)\n",
							     secs,
							     msecs);
					    } else 
#endif
						if (verbose >= 0)
						    fprintf( stderr, "SUCCESSFUL\n" );
				    } else {
					Out_Func( tempstr );
				    }
				    if (ANSIMode)
					if (!outfile) {
					    if (verbose >= 2)
						fprintf( stderr, "CW st: %d.\n", __LINE__ );
					    EXEC SQL COMMIT WORK;
					} else {
					    Out_Func( "COMMIT WORK;" );
					    if (verbose >= 2)
						fprintf( stderr, "CW st: %d.\n", __LINE__ );
					}
				}
			    }
			    /* Look backwards at lastidx also because that column
			       may not have been done yet. */
			    if (lastidx.colno[jj] != 0) {
				column2 = ((lastidx.colno[jj] > 0) ?
					   lastidx.colno[jj] :
					   (-1 * lastidx.colno[jj])) - 1;
				if (doindex && !NoLevel && leadsame && (columns[column2] % 2) == 0)
				{
				    if (ANSIMode == 1)
					snprintf( lastowner, sizeof lastowner, "\"%s\".", lastidx.tabowner );
				    else
					memset( lastowner, 0, sizeof lastowner );
				    columns[column2] |= 1;  /* Set flag 'HIGH DONE' for this column. */
				    if (UseServerOptimization || !(NoDistOnly || (tabinfo.colno[1]==0 && !ForceDistOnly)))
					columns[column2] |= 4; /* Set 'DISTRIBUTIONS ONLY DONE' flag for this column. */
#if !defined(NOTIMINGS)
				    if (ExpandReport)
					gettimeofday( &stmtstart, NULL );
#endif
			    
				    if (!UseServerOptimization) {
					snprintf( tempstr, sizeof tempstr,
						  "UPDATE STATISTICS HIGH FOR TABLE %s%s%s%s ( %s)%s%s;",
						  lastowner,
						  QuotedIdents ? "\"" : "",
						  lastidx.tabname,
						  QuotedIdents ? "\"" : "",
						  columnnames[column2],
						  resset ? resstring : "",
						  (NoDistOnly ||
						   (tabinfo.colno[1]==0 &&
						    !ForceDistOnly))
						  ? "" : " DISTRIBUTIONS ONLY" );
					if (outfile == (FILE *)NULL) {
					    if (verbose >= 0)
						fprintf( stderr, "\t%s(HIGH)...", columnnames[column2] );
#if !defined(NOTIMINGS)
					    if (ExpandReport)
					        gettimeofday( &stmtstart, NULL );
#endif
					    EXEC SQL EXECUTE IMMEDIATE :tempstr;
					    if (sqlca.sqlcode == -313) {
						fprintf( stderr,
							 "FAILED\nNo permissions to update statistics in"
							 " database: %s@%s.\nRerun as informix.  "
							 "Continuing with stored procedures anyway.\n",
							 databasename,
							 host );
						break;
					    } else if (sqlca.sqlcode == -740) {
						nrecs = getrowcount( lastidx.tabowner, lastidx.tabname, lastidx.tabid, lastidx.partnum );
						fprintf( stderr,
							 "FAILED\nMinimum resolution for this table is: %f\n",
							 minres( nrecs ) );
						exit(3);
					    } else if (sqlca.sqlcode) {
						fprintf( stderr,
							 "FAILED\nCannot Update statistics for table: "
							 "%s%s%s%s, column %s; "
							 "\n\tSQLCODE = %d, ISAM = %d, POS = %d\n",
							 ownerstr,
							 QuotedIdents ? "\"" : "",
							 lastidx.tabname,
							 QuotedIdents ? "\"" : "",
							 columnnames[column2],
							 (int)sqlca.sqlcode,
							 (int)sqlca.sqlerrd[1],
							 (int)sqlca.sqlerrd[4] );
						if (sqlca.sqlcode == -201)
						    fprintf( stderr, "%s\n", tempstr );
                                                if (!noexit) {
                                                    exit(1);
                                                } else {
                                                    noexit--;
                                                }
					    } else
#if !defined(NOTIMINGS)
						if (ExpandReport) {
						    gettimeofday( &stmtend, NULL );
						    timediff( stmtstart,
							      stmtend,
							      &secs,
							      &msecs );
						    if (ExpandReport == 2) {
							fprintf( stderr,
								 "SUCCESSFUL\t(%d hrs %d mins %d.%03.3d Secs)\n",
								 (secs / 3600),
								 ((secs % 3600) / 60),
								 (secs % 60),
								 msecs );
						    } else 
							fprintf( stderr,
								 "SUCCESSFUL\t(%d.%03.3d Secs)\n",
								 secs,
								 msecs);
						} else 
#endif
						    if (verbose >= 0)
							fprintf( stderr, "SUCCESSFUL\n" );
					} else {
					    if (verbose >= 0)
						fprintf( stderr, "\t%s(HIGH)", columnnames[column2] );
					    Out_Func( tempstr );
					    if (verbose >= 0)
						fprintf( stderr, "\n" );
					}
					if (ANSIMode)
					    if (!outfile) {
						if (verbose >= 2)
						    fprintf( stderr, "CW st: %d.\n", __LINE__ );
						EXEC SQL COMMIT WORK;
					    } else {
						Out_Func( "COMMIT WORK;" );
						if (verbose >= 2)
						    fprintf( stderr, "CW st: %d.\n", __LINE__ );
					    }
				    }
				}
			    }
			}
			if (docolumn && ((tabinfo.colno[1] != 0) ||
					 ((tabinfo.colno[1] == 0) && (columns[column]/4)%2) == 1))
			{
#if !defined(NOTIMINGS)
			    if (ExpandReport)
				gettimeofday( &stmtstart, NULL );
#endif
			    
			    snprintf( tempstr, sizeof tempstr,
				      "UPDATE STATISTICS %s FOR TABLE %s%s%s%s (",
				      NoLevel ? "" : "LOW",
				      ownerstr,
				      QuotedIdents ? "\"" : "",
				      tabinfo.tabname,
				      QuotedIdents ? "\"" : ""
				      );
			    if (verbose >= 0)
				fprintf( stderr, "\t" ); 
			    for (c = 0; c < 16; c++) {
				if (tabinfo.colno[c] != 0) {
				    if (c != 0) {
					strcat( tempstr, ", " );
					if (verbose >= 0)
					    fprintf( stderr, ", " );
				    }
				    column2 = ((tabinfo.colno[c] > 0) ?
					       tabinfo.colno[c] :
					       (-1 * tabinfo.colno[c])) - 1;
				    strcat( tempstr, columnnames[column2] );
				    if (verbose >= 0)
					fprintf( stderr, "%s", columnnames[column2] );
				}
			    }
			    strcat( tempstr, ");" );
			    if (outfile == (FILE *)NULL) {
				if (verbose >= 0)
				    fprintf( stderr, "(LOW)..." );
#if !defined(NOTIMINGS)
				if (ExpandReport)
				    gettimeofday( &stmtstart, NULL );
#endif
				EXEC SQL EXECUTE IMMEDIATE :tempstr;
				if (sqlca.sqlcode == -313) {
				    fprintf( stderr,
					     "FAILED\nNo permissions to update statistics in"
					     " database: %s@%s.\nRerun as informix.  "
					     "Continuing with stored procedures anyway.\n",
					     databasename,
					     host );
				    break;
				} else if (sqlca.sqlcode == -740) {
				    nrecs = getrowcount( tabinfo.tabowner, tabinfo.tabname, tabinfo.tabid, tabinfo.partnum );
				    fprintf( stderr,
					     "FAILED\nMinimum resolution for this table is: %f\n",
					     minres( nrecs ) );
				    exit(3);
				} else if (sqlca.sqlcode) {
				    fprintf( stderr,
					     "FAILED\nCannot Update statistics "
					     "for table: "
					     "%s%s%s%s, column %s; "
					     "\n\tSQLCODE = %d, ISAM = %d, "
					     "POS = %d\n",
					     ownerstr,
					     QuotedIdents ? "\"" : "",
					     tabinfo.tabname,
					     QuotedIdents ? "\"" : "",
					     columnnames[column2] ,
					     (int)sqlca.sqlcode,
					     (int)sqlca.sqlerrd[1],
					     (int)sqlca.sqlerrd[4] );
				    if (sqlca.sqlcode == -201)
					fprintf( stderr, "%s\n", tempstr );
				    if (!noexit) {
					exit(1);
				    } else {
					noexit--;
				    }
				} else
#if !defined(NOTIMINGS)
				    if (ExpandReport) {
					gettimeofday( &stmtend, NULL );
					timediff( stmtstart,
						  stmtend,
						  &secs,
						  &msecs );
					if (ExpandReport == 2) {
					    fprintf( stderr,
						     "SUCCESSFUL"
						     "\t(%d hrs %d "
						     "mins %d.%03.3d Secs)\n",
						     (secs / 3600),
						     ((secs % 3600) / 60),
						     (secs % 60),
						     msecs );
					} else 
					    fprintf( stderr,
						     "SUCCESSFUL"
						     "\t(%d.%03.3d Secs)\n",
						     secs,
						     msecs);
				    } else 
#endif
					if (verbose >= 0)
					    fprintf( stderr, "SUCCESSFUL\n" );
			    } else {
				if (verbose >= 0)
				    fprintf( stderr, "(LOW)" );
				Out_Func( tempstr );
				if (verbose >= 0)
				    fprintf( stderr, "\n" );
			    }
			}
		    }

		    /* Do we need to look at fragmentation expressions? */
		    if (doFragExpr && !tabinfo.partnum) {
			char *tok;

			/* This is a fragmented table.  Process fragment
			   expression columns if any. */
			exprtext.loc_loctype = LOCUSER;
			exprtext.loc_buffer = (char *)malloc( 4096 );
			if (!exprtext.loc_buffer) {
			    fprintf( stderr, 
				     "FAILED\nCould not allocate memory for fragment expressions. "
				     "Errno=%d.\n", 
				     errno );
			    exit(1);
			}
			exprtext.loc_bufsize = 4096;
			exprtext.loc_write = writeit;
			exprtext.loc_open = openit;
			exprtext.loc_close = closeit;

			tblobbuff.bbsize = 4096;
			tblobbuff.base = (char *)malloc( tblobbuff.bbsize );
			if (!tblobbuff.base) {
			    fprintf( stderr, 
				     "FAILED\nCould not allocate memory for tblob buffer. "
				     "Errno=%d.\n", 
				     errno );
			    exit(1);
			}
			tblobbuff.buff = tblobbuff.base;
			tblobbuff.full_length = tblobbuff.global_length = 0;
			blobbuff = &tblobbuff;

			/* Perhaps prepare this above. --ASK */
			EXEC SQL DECLARE c_fragexpr CURSOR WITH HOLD FOR
			    SELECT evalpos, exprtext
			    FROM "informix".sysfragments
			    WHERE tabid = :tabinfo.tabid
			    AND fragtype = 'T'
			    AND strategy = 'E'
			    ORDER BY evalpos;
			if (sqlca.sqlcode != 0) {
			    fprintf( stderr,
				     "FAILED DECLARING cursor for fragexpr.  Code=%d, ISAM=%d.\n",
				     (int)sqlca.sqlcode,
				     (int)sqlca.sqlerrd[1] );
			    exit(1);
			}
			EXEC SQL OPEN c_fragexpr;
			if (sqlca.sqlcode != 0) {
			    fprintf( stderr,
				     "FAILED OPENING cursor for fragexpr.  Code=%d, ISAM=%d.\n",
				     (int)sqlca.sqlcode,
				     (int)sqlca.sqlerrd[1] );
			    exit(1);
			}
			for(;;) {
			    EXEC SQL FETCH c_fragexpr
				INTO :evalpos, :exprtext;
			    if (sqlca.sqlcode != 0)
				break;
			    tblobbuff.buff[tblobbuff.global_length] = (char)0;
			    strcat( tblobbuff.buff, " " );
			    tblobbuff.buff += strlen( tblobbuff.buff );
			    tblobbuff.full_length += tblobbuff.global_length;
			}
			if (sqlca.sqlcode < 0) {
			    fprintf( stderr,
				     "FAILED FETCHING fragexpr. Code=%d, ISAM=%d.\n",
				     (int)sqlca.sqlcode,
				     (int)sqlca.sqlerrd[1] );
			    exit(1);
			}
			EXEC SQL CLOSE c_fragexpr;
			if (tblobbuff.full_length) {
			    /* OK, got the expression, now parse for column names! */
			    tok = strtok( tblobbuff.base, "( )<>=!\t|" );
			    while (tok) {
				column = cfind( tok, columnnames, ncols );
				if (column >= 0) {
				    /* OK this token at least looks like a column name. */
				    /* Make certain we did not do this column already. */
				    if ((columns[column] % 2) == 0) { 
					columns[column] |= 1;  /* Set flag 'HIGH DONE' for this column. */
					if (UseServerOptimization || !(NoDistOnly || (tabinfo.colno[1]==0 && !ForceDistOnly)))
					    columns[column] |= 4; /* Set 'DISTRIBUTIONS ONLY DONE' flag for this column. */
#if !defined(NOTIMINGS)
					if (ExpandReport)
					    gettimeofday( &stmtstart, NULL );
#endif
					if (!UseServerOptimization) {
					    snprintf( tempstr, sizeof tempstr,
						      "UPDATE STATISTICS HIGH FOR TABLE %s%s%s%s (%s)%s%s;",
						      ownerstr,
						      QuotedIdents ? "\"" : "",
						      tabinfo.tabname,
						      QuotedIdents ? "\"" : "",
						      columnnames[column],
						      resset ? resstring : "",
						      (NoDistOnly || (tabinfo.colno[1]==0 && !ForceDistOnly)) ?
						      "" : " DISTRIBUTIONS ONLY");
					    if (outfile == (FILE *)NULL) {
						if (verbose >= 0)
						    fprintf( stderr, "\t%s(HIGH)...", columnnames[column] );
#if !defined(NOTIMINGS)
						if (ExpandReport)
						    gettimeofday( &stmtstart, NULL );
#endif
						EXEC SQL EXECUTE IMMEDIATE :tempstr;
						if (sqlca.sqlcode == -313) {
						    fprintf( stderr,
							     "FAILED\nNo permissions to update statistics in"
							     " database: %s@%s.\nRerun as informix.  "
							     "Continuing with stored procedures anyway.\n",
							     databasename,
							     host );
						    break;
						} else if (sqlca.sqlcode == -740) {
						    nrecs = getrowcount( tabinfo.tabowner, tabinfo.tabname, tabinfo.tabid, tabinfo.partnum );
						    fprintf( stderr,
							     "FAILED\nMinimum resolution for this table is: %f\n",
							     minres( nrecs ) );
						    exit(3);
						} else if (sqlca.sqlcode) {
						    fprintf( stderr,
							     "FAILED\nCannot Update statistics for table: "
							     "%s%s%s%s, column %s; "
							     "\n\tSQLCODE = %d, ISAM = %d, POS = %d\n",
							     ownerstr,
							     QuotedIdents ? "\"" : "",
							     tabinfo.tabname,
							     QuotedIdents ? "\"" : "",
							     columnnames[column],
							     (int)sqlca.sqlcode,
							     (int)sqlca.sqlerrd[1],
							     (int)sqlca.sqlerrd[4] );
						    if (sqlca.sqlcode == -201)
							fprintf( stderr, "%s\n", tempstr );
						    if (!noexit) {
							exit(1);
						    } else {
							noexit--;
						    }
						} else
#if !defined(NOTIMINGS)
						    if (ExpandReport) {
							gettimeofday( &stmtend, NULL );
							timediff( stmtstart,
								  stmtend,
								  &secs,
								  &msecs );
							if (ExpandReport == 2) {
							    fprintf( stderr,
								     "SUCCESSFUL\t(%d hrs %d mins %d.%03.3d Secs)\n",
								     (secs / 3600),
								     ((secs % 3600) / 60),
								     (secs % 60),
								     msecs );
							} else {
							    fprintf( stderr,
								     "SUCCESSFUL\t(%d.%03.3d Secs)\n",
								     secs,
								     msecs );
							}
						    } else 
#endif
							if (verbose >= 0)
							    fprintf( stderr, "SUCCESSFUL\n" );
					    } else {
						if (verbose >= 0)
						    fprintf( stderr, "\t%s(HIGH)", columnnames[column] );
						Out_Func( tempstr );
						if (verbose >= 0)
						    fprintf( stderr, "\n" );
					    }
					    if (ANSIMode)
						if (!outfile) {
						    if (verbose >= 2)
							fprintf( stderr, "CW st: %d.\n", __LINE__ );
						    EXEC SQL COMMIT WORK;
						} else {
						    Out_Func( "COMMIT WORK;" );
						    if (verbose >= 2)
							fprintf( stderr, "CW st: %d.\n", __LINE__ );
						}
					}
				    }
				}
				tok = strtok( (char *)NULL, "( )<>=!\t|" );
			    }
			}
		    }
		}
		/* In ANSI Mode there are no singleton transactions and the
		   entire dostats session would be a single HUGE transaction
		   possibly tying up resources and locking sysdistrib
		   records.  Therefore we must commit each statement
		   manually. */
		if (ANSIMode)
		    if (!outfile) {
			if (verbose >= 2)
			    fprintf( stderr, "CW st: %d.\n", __LINE__ );
			EXEC SQL COMMIT WORK;
		    } else {
			Out_Func( "COMMIT WORK;" );
			if (verbose >= 2)
			    fprintf( stderr, "CW st: %d.\n", __LINE__ );
		    }
		memcpy( &lastidx, &tabinfo, sizeof tabinfo );
	    }
	    EXEC SQL CLOSE a1;
	    if (lasttbl[0] != (char)0 && UseServerOptimization) {
		int col, highlen, medlen;
		int medfirst = 1, highfirst = 1;
		char highcols[32768], medcols[32768];

		medlen = 32767 - (56 + strlen( lasttbl ));
		highlen = medlen + 2;
		highcols[0] = (char)0;
		medcols[0] = (char)0;
		for (col=0; col < ncols; col++) {
		    if ((strlen( highcols ) + strlen( columnnames[col] ) + 1) > highlen) {
			/* Name list is getting too long.  Issue command now and start over. */
			snprintf( tempstr, sizeof tempstr,
				  "UPDATE STATISTICS HIGH FOR TABLE %s%s%s%s (%s)%s%s;",
				  lastowner,
				  QuotedIdents ? "\"" : "",
				  lasttbl,
				  QuotedIdents ? "\"" : "",
				  highcols,
				  resset ? resstring : "",
				  (NoDistOnly) ?
				  "" : " DISTRIBUTIONS ONLY");
			if (doindex) {
			    if (outfile == (FILE *)NULL) {
				if (verbose >= 0)
				    fprintf( stderr, "\t%s(HIGH)...", highcols );
#if !defined(NOTIMINGS)
				if (ExpandReport)
				    gettimeofday( &stmtstart, NULL );
#endif
				EXEC SQL EXECUTE IMMEDIATE :tempstr;
				if (sqlca.sqlcode == -313) {
				    fprintf( stderr,
					     "FAILED\nNo permissions to update statistics in"
					     " database: %s@%s.\nRerun as informix.  "
					     "Continuing with stored procedures anyway.\n",
					     databasename,
					     host );
				    break;
				} else if (sqlca.sqlcode == -740) {
				    nrecs = getrowcount( tabinfo.tabowner, tabinfo.tabname, tabinfo.tabid, tabinfo.partnum );
				    fprintf( stderr,
					     "FAILED\nMinimum resolution for this table is: %f\n",
					     minres( nrecs ) );
				    exit(3);
				} else if (sqlca.sqlcode) {
				    fprintf( stderr,
					     "FAILED\nCannot Update statistics for table: "
					     "%s%s%s%s, column %s; "
					     "\n\tSQLCODE = %d, ISAM = %d, POS = %d\n",
					     ownerstr,
					     QuotedIdents ? "\"" : "",
					     tabinfo.tabname,
					     QuotedIdents ? "\"" : "",
					     columnnames[column],
					     (int)sqlca.sqlcode,
					     (int)sqlca.sqlerrd[1],
					     (int)sqlca.sqlerrd[4] );
				    if (sqlca.sqlcode == -201)
					fprintf( stderr, "%s\n", tempstr );
				    if (!noexit) {
					exit(1);
				    } else {
					noexit--;
				    }
				} else
#if !defined(NOTIMINGS)
				    if (ExpandReport) {
					gettimeofday( &stmtend, NULL );
					timediff( stmtstart,
						  stmtend,
						  &secs,
						  &msecs );
					if (ExpandReport == 2) {
					    fprintf( stderr,
						     "SUCCESSFUL\t(%d hrs %d mins %d.%03.3d Secs)\n",
						     (secs / 3600),
						     ((secs % 3600) / 60),
						     (secs % 60),
						     msecs );
					} else {
					    fprintf( stderr,
						     "SUCCESSFUL\t(%d.%03.3d Secs)\n",
						     secs,
						     msecs );
					}
				    } else 
#endif
					if (verbose >= 0)
					    fprintf( stderr, "SUCCESSFUL\n" );
			    } else {
				if (verbose >= 0)
				    fprintf( stderr,
					     "\t%s:\n",
					     columnnames[column]  );
				Out_Func( tempstr );
			    }
			    if (ANSIMode)
				if (!outfile) {
				    if (verbose >= 2)
					fprintf( stderr, "CW st: %d.\n", __LINE__ );
				    EXEC SQL COMMIT WORK;
				} else {
				    Out_Func( "COMMIT WORK;" );
				    if (verbose >= 2)
					fprintf( stderr, "CW st: %d.\n", __LINE__ );
				}
			}
			highfirst = 1;
			highcols[0] = (char)0;
		    }
		    if ((strlen( medcols ) + strlen( columnnames[col] ) + 1) > medlen) {
			/* Name list is getting too long.  Issue command now and start over. */
			snprintf( tempstr, sizeof tempstr,
				  "UPDATE STATISTICS MEDIUM FOR TABLE %s%s%s%s (%s)%s%s%s;",
				  lastowner,
				  QuotedIdents ? "\"" : "",
				  lasttbl,
				  QuotedIdents ? "\"" : "",
				  medcols,
				  sampleset ? samplestring : "",
				  mresset ? mresstring : "",
				  (mresset && confset) ? confstring : "" );
			if (doindex) {
			    if (outfile == (FILE *)NULL) {
				if (verbose >= 0)
				    fprintf( stderr, "\t%s(MEDIUM)...", medcols );
#if !defined(NOTIMINGS)
				if (ExpandReport)
				    gettimeofday( &stmtstart, NULL );
#endif
				EXEC SQL EXECUTE IMMEDIATE :tempstr;
				if (sqlca.sqlcode == -313) {
				    fprintf( stderr,
					     "FAILED\nNo permissions to update statistics in"
					     " database: %s@%s.\nRerun as informix.  "
					     "Continuing with stored procedures anyway.\n",
					     databasename,
					     host );
				    break;
				} else if (sqlca.sqlcode == -740) {
				    nrecs = getrowcount( tabinfo.tabowner, tabinfo.tabname, tabinfo.tabid, tabinfo.partnum );
				    fprintf( stderr,
					     "FAILED\nMinimum resolution for this table is: %f\n",
					     minres( nrecs ) );
				    exit(3);
				} else if (sqlca.sqlcode) {
				    fprintf( stderr,
					     "FAILED\nCannot Update statistics for table: "
					     "%s%s%s%s, column %s; "
					     "\n\tSQLCODE = %d, ISAM = %d, POS = %d\n",
					     ownerstr,
					     QuotedIdents ? "\"" : "",
					     tabinfo.tabname,
					     QuotedIdents ? "\"" : "",
					     columnnames[column],
					     (int)sqlca.sqlcode,
					     (int)sqlca.sqlerrd[1],
					     (int)sqlca.sqlerrd[4] );
				    if (sqlca.sqlcode == -201)
					fprintf( stderr, "%s\n", tempstr );
				    if (!noexit) {
					exit(1);
				    } else {
					noexit--;
				    }
				} else
#if !defined(NOTIMINGS)
				    if (ExpandReport) {
					gettimeofday( &stmtend, NULL );
					timediff( stmtstart,
						  stmtend,
						  &secs,
						  &msecs );
					if (ExpandReport == 2) {
					    fprintf( stderr,
						     "SUCCESSFUL\t(%d hrs %d mins %d.%03.3d Secs)\n",
						     (secs / 3600),
						     ((secs % 3600) / 60),
						     (secs % 60),
						     msecs );
					} else {
					    fprintf( stderr,
						     "SUCCESSFUL\t(%d.%03.3d Secs)\n",
						     secs,
						     msecs );
					}
				    } else 
#endif
					if (verbose >= 0)
					    fprintf( stderr, "SUCCESSFUL\n" );
			    } else {
				if (verbose >= 0)
				    fprintf( stderr,
					     "\t%s:\n",
					     columnnames[column]  );
				Out_Func( tempstr );
			    }
			    if (ANSIMode)
				if (!outfile) {
				    if (verbose >= 2)
					fprintf( stderr, "CW st: %d.\n", __LINE__ );
				    EXEC SQL COMMIT WORK;
				} else {
				    Out_Func( "COMMIT WORK;" );
				    if (verbose >= 2)
					fprintf( stderr, "CW st: %d.\n", __LINE__ );
				}
			}
			medfirst = 1;
			medcols[0] = (char)0;
		    }
		    if (columns[col] & 1) {
			/* Include column in HIGH. */
			if (highfirst) {
			    highfirst = 0;
			} else {
			    strcat( highcols, ", " );
			}
			strcat( highcols, columnnames[col] );
		    } else {
			/* Include column in MEDIUM. */
			if (medfirst) {
			    medfirst = 0;
			} else {
			    strcat( medcols, ", " );
			}
			strcat( medcols, columnnames[col] );
		    }
		}

		if (strlen( highcols )) {
		    snprintf( tempstr, sizeof tempstr,
			      "UPDATE STATISTICS HIGH FOR TABLE %s%s%s%s (%s)%s%s;",
			      lastowner,
			      QuotedIdents ? "\"" : "",
			      lasttbl,
			      QuotedIdents ? "\"" : "",
			      highcols,
			      resset ? resstring : "",
			      (NoDistOnly) ?
			      "" : " DISTRIBUTIONS ONLY");
		    if (doindex) {
			if (outfile == (FILE *)NULL) {
			    if (verbose >= 0)
				fprintf( stderr, "\t%s(HIGH)...", highcols );
#if !defined(NOTIMINGS)
			    if (ExpandReport)
			        gettimeofday( &stmtstart, NULL );
#endif
			    EXEC SQL EXECUTE IMMEDIATE :tempstr;
			    if (sqlca.sqlcode == -313) {
				fprintf( stderr,
					 "FAILED\nNo permissions to update statistics in"
					 " database: %s@%s.\nRerun as informix.  "
					 "Continuing with stored procedures anyway.\n",
					 databasename,
					 host );
				break;
			    } else if (sqlca.sqlcode == -740) {
				nrecs = getrowcount( tabinfo.tabowner, tabinfo.tabname, tabinfo.tabid, tabinfo.partnum );
				fprintf( stderr,
					 "FAILED\nMinimum resolution for this table is: %f\n",
					 minres( nrecs ) );
				exit(3);
			    } else if (sqlca.sqlcode) {
				fprintf( stderr,
					 "FAILED\nCannot Update statistics for table: "
					 "%s%s%s%s, column %s; "
					 "\n\tSQLCODE = %d, ISAM = %d, POS = %d\n",
					 ownerstr,
					 QuotedIdents ? "\"" : "",
					 tabinfo.tabname,
					 QuotedIdents ? "\"" : "",
					 columnnames[column],
					 (int)sqlca.sqlcode,
					 (int)sqlca.sqlerrd[1],
					 (int)sqlca.sqlerrd[4] );
				if (sqlca.sqlcode == -201)
				    fprintf( stderr, "%s\n", tempstr );
				if (!noexit) {
				    exit(1);
				} else {
				    noexit--;
				}
			    } else
#if !defined(NOTIMINGS)
				if (ExpandReport) {
				    gettimeofday( &stmtend, NULL );
				    timediff( stmtstart,
					      stmtend,
					      &secs,
					      &msecs );
				    if (ExpandReport == 2) {
					fprintf( stderr,
						 "SUCCESSFUL\t(%d hrs %d mins %d.%03.3d Secs)\n",
						 (secs / 3600),
						 ((secs % 3600) / 60),
						 (secs % 60),
						 msecs );
				    } else {
					fprintf( stderr,
						 "SUCCESSFUL\t(%d.%03.3d Secs)\n",
						 secs,
						 msecs );
				    }
				} else 
#endif
				    if (verbose >= 0)
					fprintf( stderr, "SUCCESSFUL\n" );
			} else {
			    if (verbose >= 0)
				fprintf( stderr, "\t%s(HIGH)...", highcols );
			    Out_Func( tempstr );
			    if (verbose >= 0)
				fprintf( stderr, "\n" );
			}
			if (ANSIMode)
			    if (!outfile) {
				if (verbose >= 2)
				    fprintf( stderr, "CW st: %d.\n", __LINE__ );
				EXEC SQL COMMIT WORK;
			    } else {
				Out_Func( "COMMIT WORK;" );
				if (verbose >= 2)
				    fprintf( stderr, "CW st: %d.\n", __LINE__ );
			    }
		    }
		    highcols[0] = (char)0;
		}
		if (strlen( medcols )) {
		    snprintf( tempstr, sizeof tempstr,
			      "UPDATE STATISTICS MEDIUM FOR TABLE %s%s%s%s (%s)%s%s%s;",
			      lastowner,
			      QuotedIdents ? "\"" : "",
			      lasttbl,
			      QuotedIdents ? "\"" : "",
			      medcols,
			      sampleset ? samplestring : "",
			      mresset ? mresstring : "",
			      (mresset && confset) ? confstring : "" );
		    if (doindex) {
			if (outfile == (FILE *)NULL) {
			    if (verbose >= 0)
				fprintf( stderr, "\t%s(MEDIUM)...", medcols );
#if !defined(NOTIMINGS)
			    if (ExpandReport)
			        gettimeofday( &stmtstart, NULL );
#endif
			    EXEC SQL EXECUTE IMMEDIATE :tempstr;
			    if (sqlca.sqlcode == -313) {
				fprintf( stderr,
					 "FAILED\nNo permissions to update statistics in"
					 " database: %s@%s.\nRerun as informix.  "
					 "Continuing with stored procedures anyway.\n",
					 databasename,
					 host );
				break;
			    } else if (sqlca.sqlcode == -740) {
				nrecs = getrowcount( tabinfo.tabowner, tabinfo.tabname, tabinfo.tabid, tabinfo.partnum );
				fprintf( stderr,
					 "FAILED\nMinimum resolution for this table is: %f\n",
					 minres( nrecs ) );
				exit(3);
			    } else if (sqlca.sqlcode) {
				fprintf( stderr,
					 "FAILED\nCannot Update statistics for table: "
					 "%s%s%s%s, column %s; "
					 "\n\tSQLCODE = %d, ISAM = %d, POS = %d\n",
					 ownerstr,
					 QuotedIdents ? "\"" : "",
					 tabinfo.tabname,
					 QuotedIdents ? "\"" : "",
					 columnnames[column],
					 (int)sqlca.sqlcode,
					 (int)sqlca.sqlerrd[1],
					 (int)sqlca.sqlerrd[4] );
				if (sqlca.sqlcode == -201)
				    fprintf( stderr, "%s\n", tempstr );
				if (!noexit) {
				    exit(1);
				} else {
				    noexit--;
				}
			    } else
#if !defined(NOTIMINGS)
				if (ExpandReport) {
				    gettimeofday( &stmtend, NULL );
				    timediff( stmtstart,
					      stmtend,
					      &secs,
					      &msecs );
				    if (ExpandReport == 2) {
					fprintf( stderr,
						 "SUCCESSFUL\t(%d hrs %d mins %d.%03.3d Secs)\n",
						 (secs / 3600),
						 ((secs % 3600) / 60),
						 (secs % 60),
						 msecs );
				    } else {
					fprintf( stderr,
						 "SUCCESSFUL\t(%d.%03.3d Secs)\n",
						 secs,
						 msecs );
				    }
				} else 
#endif
				    if (verbose >= 0)
					fprintf( stderr, "SUCCESSFUL\n" );
			} else {
			    if (verbose >= 0)
				fprintf( stderr, "\t%s(MEDIUM)", medcols );
			    Out_Func( tempstr );
			    if (verbose >= 0)
				fprintf( stderr, "\n" );
			}
			if (ANSIMode)
			    if (!outfile) {
				if (verbose >= 2)
				    fprintf( stderr, "CW st: %d.\n", __LINE__ );
				EXEC SQL COMMIT WORK;
			    } else {
				Out_Func( "COMMIT WORK;" );
				if (verbose >= 2)
				    fprintf( stderr, "CW st: %d.\n", __LINE__ );
			    }
		    }
		    medcols[0] = (char)0;
		}
	    }

#if !defined(NOTIMINGS)
	    if (outfile == (FILE *)NULL && ExpandReport) {
		gettimeofday( &tabend, NULL );
		timediff( tabstart,
			  tabend,
			  &secs,
			  &msecs );
		if (ExpandReport == 2) {
		    fprintf( stderr,
			     "    Table %s%s%s%s completed "
			     "\t(%d hrs %d mins %d.%03.3d Secs)\n",
			     lastowner,
			     QuotedIdents ? "\"" : "",
			     lasttbl,
			     QuotedIdents ? "\"" : "",
			     (secs / 3600),
			     ((secs % 3600) / 60),
			     (secs % 60),
			     msecs );
		} else {
		    fprintf( stderr,
			     "    Table %s%s%s%s completed \t(%d.%03.3d Secs)\n",
			     lastowner,
			     QuotedIdents ? "\"" : "",
			     lasttbl,
			     QuotedIdents ? "\"" : "",
			     secs,
			     msecs );
		}
	    }
#endif
	    if (doexplain) {
		if (outfile == (FILE *)NULL)
		    EXEC SQL SET EXPLAIN OFF;
		else
		    Out_Func( "SET EXPLAIN OFF;" );
	    }
	}

	if (doprocs) {
	    if (outfile == (FILE *)NULL)
		EXEC SQL EXECUTE IMMEDIATE :pdqstmt_1;
	    else
		Out_Func( pdqstmt_1 );
	    
	    if (major_vers < 9) {
		snprintf( procstmt,
			  sizeof procstmt,
			  "SELECT UNIQUE sp.owner, procname, tabname, tabname "
			  "FROM \"informix\".sysprocedures sp, outer systables "
			  "WHERE procname != ' oldsystdist' and tabid = (-1) * procid " );
	    } else if (major_vers < 11) { /*
					    Expected problem with external routines finally reared its ugly
					    head and had to be fixed.  Limiting processing to ONLY SPL routines.
					  */

		snprintf( procstmt,
			  sizeof procstmt,
			  "SELECT UNIQUE sp.owner, sp.procname, sp.specificname, sp.paramtypes::lvarchar "
			  "FROM \"informix\".sysprocedures sp, \"informix\".sysroutinelangs sr "
			  "WHERE sp.langid = sr.langid and sr.langname = 'spl' and procname != ' oldsystdist'" );
	    } else { /*
		       This block courtesy of Douglas Palmer.  Need to validate why the AND procname...
		     *//*  
			   Expected problem with external routines finally reared its ugly
			   head and had to be fixed.  Limiting processing to ONLY SPL routines.
		       */
		snprintf( procstmt,
			  sizeof procstmt,
			  "SELECT UNIQUE sp.owner, sp.procname, sp.specificname, sp.paramtypes::lvarchar "
			  "FROM \"informix\".sysprocedures sp, \"informix\".sysroutinelangs sr "
			  "WHERE sp.langid = sr.langid and sr.langname = 'spl' and procname != ' oldsystdist' "
			  "  AND procname NOT MATCHES '*seclabel_*' "
		          "  AND procname != ' systdistold' " );
	    }
	    EXEC SQL PREPARE proc_s from :procstmt;
	    if (sqlca.sqlcode < 0) {
		fprintf( stderr,
			 "FAILED PREPARING STATEMENT procs_s.  SQL=%d, ISAM=%d (%s-%s).\n",
			 (int)sqlca.sqlcode,
			 (int)sqlca.sqlerrd[1],
			 sqlca.sqlerrm,
			 sqlca.sqlerrp );
		fprintf( stderr,
			 "%s\n",
			 procstmt );
		exit(9);
	    }
	    EXEC SQL DECLARE procs CURSOR WITH HOLD FOR proc_s;
	    if (sqlca.sqlcode < 0) {
		fprintf( stderr,
			 "FAILED DECLARING CURSOR procs.  SQL=%d, ISAM=%d.\n",
			 (int)sqlca.sqlcode,
			 (int)sqlca.sqlerrd[1] );
		exit(9);
	    }
	    EXEC SQL OPEN procs;
	    if (sqlca.sqlcode < 0) {
		fprintf( stderr,
			 "FAILED OPENING CURSOR procs.  SQL=%d, ISAM=%d.\n",
			 (int)sqlca.sqlcode,
			 (int)sqlca.sqlerrd[1] );
		exit(9);
	    }
#if !defined(NOTIMINGS)
	    if (ExpandReport)
		gettimeofday( &tabstart, NULL );
#endif

	    if (verbose >= 0)
		fprintf( stderr, "\nProcedures:\n" );
	    for (;;) {
#if !defined(NOTIMINGS)
		if (ExpandReport)
		    gettimeofday( &stmtstart, NULL );
#endif

		if (major_vers < 9) {
		    arglist_ptr = "";
		} else {
		    err = ifx_var_flag( &arglist, 1 );
		    if (err) {
			fprintf( stderr, "FAILED (%d) trying to configure memory for procedure types.\n", err );
			exit(9);
		    }
		}
		EXEC SQL FETCH procs
		    INTO :procowner,
		    :procname,
		    :specificname INDICATOR specific_ind,
		    :arglist INDICATOR arg_ind;
		if (sqlca.sqlcode < 0) {
		    fprintf( stderr,
			     "FAILED FETCHING CURSOR procs.  SQL=%d, ISAM=%d.\n",
			     (int)sqlca.sqlcode,
			     (int)sqlca.sqlerrd[1] );
		    exit(9);
		} else if (sqlca.sqlcode == SQLNOTFOUND)
		    break;

		if (ANSIMode == 1)
		    snprintf( ownerstr, sizeof ownerstr, "\"%s\".", procowner );
		else
		    memset( ownerstr, 0, sizeof ownerstr );


		if (arg_ind == -1 || major_vers < 9)
		    arglist_ptr = "";
		else
		    arglist_ptr = (char *)ifx_var_getdata( &arglist );
		if (arglist_ptr == 0)
		    arglist_ptr = "";
		if (Release >= 9 && arg_ind != -1) {
		    if (dumpargs) {
			fprintf( stderr,
				 "Diag(RAW):    %s(%s).\n",
				 procname,
				 arglist_ptr );
			fflush( stderr );
		    }
		    if (strlen( arglist_ptr ) > 0)
			StripTypeList( arglist_ptr );
		    if (dumpargs) {
			fprintf( stderr,
				 "Diag(COOKED): %s(%s).\n\n",
				 procname,
				 arglist_ptr );
			fflush( stderr );
		    }
		    snprintf( tempstr, sizeof tempstr,
			      "UPDATE STATISTICS FOR PROCEDURE %s%s%s%s(%s);",
			      ownerstr,
			      QuotedIdents ? "\"" : "",
			      procname,
			      QuotedIdents ? "\"" : "",
			      arglist_ptr );
		} else if (Release >= 9 && DoSpecific && specific_ind != -1) {
		    snprintf( tempstr, sizeof tempstr,
			      "UPDATE STATISTICS FOR SPECIFIC PROCEDURE %s%s%s%s;",
			      ownerstr,
			      QuotedIdents ? "\"" : "",
			      specificname,
			      QuotedIdents ? "\"" : "" );
		} else
		    snprintf( tempstr, sizeof tempstr,
			      "UPDATE STATISTICS FOR PROCEDURE %s%s%s%s;",
			      ownerstr,
			      QuotedIdents ? "\"" : "",
			      procname,
			      QuotedIdents ? "\"" : "" );
		if (outfile == (FILE *)NULL) {
		    if (verbose >= 0)
			fprintf( stderr,
				 "\t%s%s%s...",
				 QuotedIdents ? "\"" : "",
				 procname,
				 QuotedIdents ? "\"" : "" );
		    EXEC SQL EXECUTE IMMEDIATE :tempstr;
		    if (sqlca.sqlcode < 0) {
			fprintf( stderr,
				 "FAILED\nCannot Update statistics for procedure: %s%s%s. "
				 "SQL=%d, ISAM=%d.\n",
				 QuotedIdents ? "\"" : "",
				 procname,
				 QuotedIdents ? "\"" : "",
				 (int)sqlca.sqlcode,
				 (int)sqlca.sqlerrd[1] );
			if (!noexit) {
			    exit(1);
			} else {
			    noexit--;
			}
		    } else
#if !defined(NOTIMINGS)
			if (ExpandReport) {
			    gettimeofday( &stmtend, NULL );
			    timediff( stmtstart,
				      stmtend,
				      &secs,
				      &msecs );
			    if (ExpandReport == 2) {
				if ((int)strlen( procname ) < 3)
				    fprintf( stderr,
					     "SUCCESSFUL\t\t\t(%d hrs %d mins %d.%03.3d Secs)\n",
					     (secs / 3600),
					     ((secs % 3600) / 60),
					     (secs % 60),
					     msecs );
				else if ((int)strlen( procname ) < 11)
				    fprintf( stderr,
					     "SUCCESSFUL\t\t(%d hrs %d mins %d.%03.3d Secs)\n",
					     (secs / 3600),
					     ((secs % 3600) / 60),
					     (secs % 60),
					     msecs );
				else
				    fprintf( stderr,
					     "SUCCESSFUL\t(%d hrs %d mins %d.%03.3d Secs)\n",
					     (secs / 3600),
					     ((secs % 3600) / 60),
					     (secs % 60),
					     msecs );
			    } else {
				if ((int)strlen( procname ) < 3)
				    fprintf( stderr,
					     "SUCCESSFUL\t\t\t(%d.%03.3d Secs)\n",
					     secs,
					     msecs);
				else if ((int)strlen( procname ) < 11)
				    fprintf( stderr,
					     "SUCCESSFUL\t\t(%d.%03.3d Secs)\n",
					     secs,
					     msecs);
				else
				    fprintf( stderr,
					     "SUCCESSFUL\t(%d.%03.3d Secs)\n",
					     secs,
					     msecs);
			    }
			} else 
#endif
			    if (verbose >= 0)
				fprintf( stderr, "SUCCESSFUL\n" );
		} else {
		    if (verbose >= 0)
			fprintf( stderr,
				 "\t%s%s%s:\n",
				 QuotedIdents ? "\"" : "",
				 procname,
				 QuotedIdents ? "\"" : "" );
		    Out_Func( tempstr );
		}
		if (major_vers >= 9)
		    ifx_var_dealloc( &arglist );
		/* In ANSI Mode there are no singleton transactions and the
		   entire dostats session would be a single HUGE transaction
		   possibly tying up resources and locking sysprocedures
		   records.  Therefore we must commit each statement
		   manually. */
		if (ANSIMode)
		    if (!outfile) {
			if (verbose >= 2)
			    fprintf( stderr, "CW st: %d.\n", __LINE__ );
			EXEC SQL COMMIT WORK;
		    } else {
			Out_Func( "COMMIT WORK;" );
			if (verbose >= 2)
			    fprintf( stderr, "CW st: %d.\n", __LINE__ );
		    }
	    }
#if !defined(NOTIMINGS)
	    if (ExpandReport) {
		gettimeofday( &tabend, NULL );
		timediff( tabstart,
			  tabend,
			  &secs,
			  &msecs );
		if (ExpandReport == 2) {
		    fprintf( stderr,
			     "    Procedures complete "
			     "\t((%d hrs %d mins %d.%03.3d Secs)\n",
			     (secs / 3600),
			     ((secs % 3600) / 60),
			     (secs % 60),
			     msecs );
		} else
		    fprintf( stderr,
			     "    Procedures complete \t(%d.%03.3d Secs)\n",
			     secs,
			     msecs );
	    }
#endif
	}
#if !defined(NOTIMINGS)
	if (outfile == (FILE *)NULL && ExpandReport) {
	    gettimeofday( &dbsend, NULL );
	    timediff( dbsstart,
		      dbsend,
		      &secs,
		      &msecs );
	    if (ExpandReport == 2) {
		fprintf( stderr,
			 "Database complete \t((%d hrs %d mins %d.%03.3d Secs)\n",
			 (secs / 3600),
			 ((secs % 3600) / 60),
			 (secs % 60),
			 msecs );
	    } else
		fprintf( stderr,
			 "Database %s complete \t(%d.%03.3d Secs)\n",
			 databasename,
			 secs,
			 msecs );
	}
#endif
	if (!alldb)
	    break;
	if (outfile != (FILE *)NULL && !do_procedure)
	    fprintf( outfile, "\n" );
    }
    if (alldb) {
	if (verbose >= 2)
	    fprintf( stderr, "CW st: %d.\n", __LINE__ );
	EXEC SQL COMMIT WORK;
    }
#if !defined(NOTIMINGS)
    if (outfile == (FILE *)NULL && ExpandReport) {
	gettimeofday( &totend, NULL );
	timediff( totstart,
		  totend,
		  &secs,
		  &msecs );
	if (ExpandReport == 2) {
	    fprintf( stderr,
		     "Processing complete \t((%d hrs %d mins %d.%03.3d Secs)\n",
		     (secs / 3600),
		     ((secs % 3600) / 60),
		     (secs % 60),
		     msecs );
	} else
	    fprintf( stderr,
		     "Processing complete \t(%d.%03.3d Secs)\n",
		     secs,
		     msecs );
    }
#endif
    if (do_procedure && schedfile != (FILE *)NULL) {
	/* Finish up: insert scheduler record. */

	EXEC SQL SET CONNECTION 'WORK';
	if (sqlca.sqlcode) {
	    fprintf( stderr,
		     "FAILED %d: Cannot set connection to (%s); error=%d, isam=%d\n",
		     __LINE__,
		     "WORK",
		     (int)sqlca.sqlcode,
		     (int)sqlca.sqlerrd[1] );
	    exit(1);
	}

	if (do_proconly) {
	    if (verbose >= 2)
		fprintf( stderr, "Dropping procedure(2): %s.\n", schedprocname );
	    snprintf( schedule_stmt,
		      sizeof schedule_stmt,
		      "DROP PROCEDURE %s;",
		      schedprocname );
	    EXEC SQL EXECUTE IMMEDIATE :schedule_stmt;
	    if (sqlca.sqlcode != 0 && sqlca.sqlcode != -674) {
		fprintf( stderr,
			 "Error dropping old copy of procedure!  SQL: %d  ISAM: %d\n",
			 sqlca.sqlcode,
			 sqlca.sqlerrd[1] );
		exit( 1 );
	    }
	}
	
	if (do_scheduling)
	    fprintf( stderr,
		     "Scheduling stats update for database: %s.\n",
		     databasename );
	else
	    fprintf( stderr,
		     "Creating stats procedure %s in database: %s.\n",
		     schedprocname,
		     databasename );

	EXEC SQL CREATE ROUTINE FROM :schedprocfilename;
	if (sqlca.sqlcode != 0) {
	    fprintf( stderr,
		     "Error creating scheduling procedure: %s!  SQL: %d  ISAM: %d\n",
		     schedprocname,
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit( 1 );
	}
	unlink( schedprocfilename );
	EXEC SQL SET CONNECTION 'SCHEDULER';
	if (sqlca.sqlcode != 0) {
	    fprintf( stderr,
		     "Error changing connections to SCHEDULER!  SQL: %d  ISAM: %d\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	    exit( 1 );
	}

	if (do_scheduling) {
	    EXEC SQL SELECT count(*) INTO :OK FROM ph_group WHERE group_name = 'DOSTATS';
	    if (sqlca.sqlcode != 0) {
		fprintf( stderr,
			 "Error verifying DOSTATS task group record!  SQL: %d  ISAM: %d\n",
			 sqlca.sqlcode,
			 sqlca.sqlerrd[1] );
		exit( 1 );
	    }
	    if (OK == 0) {
		EXEC SQL INSERT INTO ph_group (group_id, group_name, group_description)
		    VALUES ( 0, "DOSTATS", "Update statistics generated by dostats(c)." );
		if (sqlca.sqlcode < 0 && sqlca.sqlcode != -268) {
		    fprintf( stderr,
			     "Error inserting DOSTATS task group record!  SQL: %d  ISAM: %d\n",
			     sqlca.sqlcode,
			     sqlca.sqlerrd[1] );
		    exit( 1 );
		}
	    }
	    /* Add scheduler record here. --ASK-- */
	    snprintf( schedule_stmt,
		      sizeof schedule_stmt,
		      "INSERT INTO ph_task( tk_id, "
		      "       tk_name, tk_type, tk_group, tk_description, tk_execute, "
		      "       tk_start_time, tk_stop_time, tk_frequency, tk_Sunday, tk_Monday, "
		      "       tk_Tuesday, tk_Wednesday, tk_Thursday, tk_Friday, tk_Saturday ) "
		      "VALUES ( 0, '%s_task', 'TASK', 'DOSTATS', 'dostats run for %s', "
		      "   'EXECUTE PROCEDURE %s:%s($DATA_TASK_ID, $DATA_SEQ_ID, %s);', %s, %s, %s, %s, %s, %s, %s, %s, %s, %s );",
		      schedprocname,
		      schedprocname,
		      databasename,
		      schedprocname,
		      sch_onetime ? "\"t\"" : "\"f\"",
		      sch_starttime[0] == (char)(0) ? "NULL" : sch_starttime,
		      sch_lasttime[0] == (char)(0) ? "NULL" : sch_lasttime,
		      sch_frequency[0] == (char)(0) ? "NULL" : sch_frequency,
		      sch_sunday,
		      sch_monday,
		      sch_tuesday,
		      sch_wednesday,
		      sch_thursday,
		      sch_friday,
		      sch_saturday );
	    EXEC SQL EXECUTE IMMEDIATE :schedule_stmt;
	    if (sqlca.sqlcode != 0) {
		fprintf( stderr,
			 "Error inserting task scheduler record!  SQL: %d  ISAM: %d\n%s\n%*.*s\n",
			 sqlca.sqlcode,
			 sqlca.sqlerrd[1],
			 schedule_stmt,
			 sqlca.sqlerrd[4],
			 sqlca.sqlerrd[4],
			 "^" );
	    }
	}
    }
    return 0;
}


void readlist( char **list, int *len, char **dblist, int *dblen, char *file )
{
    FILE *fil;
    char string[1025];

    if ((int)strlen( file ) == 1)
	fil = stdin;
    else {
	fil = fopen( &file[1], "rt" );
	if (fil == (FILE *)NULL) {
	    fprintf( stderr, "Cannot open list file: %s.\n", &file[1] );
	    exit( 19 );
	}
    }

    if (fil == stdin) {
	fputs( "table> ", stderr );
	fflush( stderr );
    }
    while (fgets( string, 1025, fil )) {
	/* remove trailing newline */
	string[(int)strlen( string ) - 1] = (char)0;

	/* new feature: skip blank lines and comments */
	/* TODO: skip blank lines with blanks and allow leading space? */
	if (! *string || *string == '#')
	    continue;

	if (strchr( string, ':' )) {
	    /* database exclude. */
	    if (dblist == (char **) NULL) {
		fprintf( stderr,
			 "Databases not supported in include lists!\n" );
		exit(20);
	    }

	    if (*dblist)
		appendStr( dblist, dblen, "','" );
	    else
		appendStr( dblist, dblen, " AND name NOT IN ('" );

	    appendStr( dblist, dblen, strtok( string, ":" ) );
	} else {
	    /* Table include/exclude. */
	    if (*list)
		appendStr( list, len, "','" );
	    else
		appendStr( list, len, "('" );

	    appendStr( list, len, string );
	}
	if (fil == stdin) {
	    fputs( "table> ", stderr );
	    fflush( stderr );
	}
    }

    if ((int)strlen( file ) > 1)
	fclose( fil );
}

void appendStr( char **str, int *str_size, char *newstr )
{
    int sum_len;

    if (*str == NULL) {
	*str_size = STR_STEP;
	*str = (char *) malloc( STR_STEP );
	**str = 0;
    }

    sum_len = strlen(*str) + strlen(newstr) + 1;
    if (sum_len > *str_size) {
	/* round up sum_len to multiple of STR_STEP */
	/* optimiser won't eliminate this - it would change the meaning */
	*str_size = (sum_len + STR_STEP - 1) / STR_STEP * STR_STEP;

	*str = (char *)realloc( *str, *str_size );

	if (! *str) {
	    fprintf( stderr, "FAILED\nError allocating memory errno = %d.\n", errno );
	    exit( 255 );
	}
    }
    strcat(*str, newstr);
}

#if !defined(NOTIMINGS)
void timediff( struct timeval start,
	       struct timeval end,
	       int *secs,
	       int *msec )
{
    *secs = end.tv_sec - start.tv_sec;
    *msec = end.tv_usec - start.tv_usec;
    if (*msec < 0) {
	*secs--;
	*msec += 1000000;
    }
    *msec += 500;
    *msec /= 1000;
}
#endif

double minres( int nrecs )
{
    double res;

    if (nrecs > 1) {
	res = (1.0 / (double)nrecs);
	if (res < 0.005)
	    res = 0.005;
	else if (res > 10.0)
	    res = 10.0;
    } else
	res = 10.0;

    return res;
}

int cmpcnts( const void *one, const void *two )
{
    if (((rowcounts_t *)one)->partnum < ((rowcounts_t *)two)->partnum)
	return -1;
    else if (((rowcounts_t *)one)->partnum > ((rowcounts_t *)two)->partnum)
	return 1;
    else
	return 0;
}


/*  #define C4GL_720_COMPILER 1 */
EXEC SQL ifdef C4GL_720_COMPILER;
int just_count_it( tabowner, tabname )
     $PARAMETER char *tabowner;
     $PARAMETER char *tabname;
EXEC SQL else;
int just_count_it( 
     EXEC SQL BEGIN DECLARE SECTION;
     PARAMETER char *tabowner,
     PARAMETER char *tabname
     EXEC SQL END DECLARE SECTION;
     )
EXEC SQL endif;
{
    EXEC SQL BEGIN DECLARE SECTION;
    static int nrows;
    static char altsql[2048];
    EXEC SQL END DECLARE SECTION;
    static char lastowner[33] = "", lasttable[129] = "";

    if (strcmp( lastowner, tabowner ) == 0 && strcmp( lasttable, tabname ) == 0) {
	return nrows;
    } else {
        snprintf( lastowner, sizeof lastowner, "%s", tabowner );
	snprintf( lasttable, sizeof lasttable, "%s", tabname );
    }

    snprintf( altsql,
	      sizeof altsql,
	      "SELECT count(*) FROM \"%s\".%s",
	      tabowner,
	      tabname );
    EXEC SQL PREPARE altcnt1 FROM :altsql;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "FAILED preparing alt table rowcount! Err=%d, isam=%d\n\n",
		 (int)sqlca.sqlcode,
		 (int)sqlca.sqlerrd[1] );
	exit( 1 );
    }

    EXEC SQL EXECUTE altcnt1 INTO :nrows;
    if (sqlca.sqlcode != 0) {
	fprintf( stderr,
		 "FAILED executing alt table rowcount! Err=%d, isam=%d\n\n",
		 (int)sqlca.sqlcode,
		 (int)sqlca.sqlerrd[1] );
	exit( 1 );
    }

    return nrows;
}


EXEC SQL ifdef C4GL_720_COMPILER;
long long getrowcount( tabowner, tabname, tabid, partnum )
     $PARAMETER char *tabowner;
     $PARAMETER char *tabname;
     $PARAMETER int tabid;
     $PARAMETER int partnum;
EXEC SQL else;
long long getrowcount(
     EXEC SQL BEGIN DECLARE SECTION;
     PARAMETER char *tabowner,
     PARAMETER char *tabname,
     PARAMETER int tabid,
     PARAMETER int partnum
     EXEC SQL END DECLARE SECTION;
			   )
EXEC SQL endif;
{
    EXEC SQL BEGIN DECLARE SECTION;
    bigint nrows = 0LL;
    int partn;
    EXEC SQL END DECLARE SECTION;
    rowcounts_t *rc, lcl;
    static int once = 1;

    if (!numparts) {
	/* No counts retrieved from sysptnhdr.  Just count the table. */
        return just_count_it( tabowner, tabname );
    } else if (partnum) {
	lcl.partnum = partnum;
	rc = (rowcounts_t *)bsearch( &lcl,
				     rowcounts,
				     numparts,
				     sizeof (rowcounts_t),
				     cmpcnts );
	if (rc)
	    nrows = rc->nrows;
	else {
	    fprintf( stderr,
		     "Error looking up partnum to match rowcount.  "
		     "Getting counts the hard way.\n" );
	    nrows = just_count_it( tabowner, tabname );
	}
    } else {
	/* Fragmented table.  Get fragment partnums and look them up. */
        if (once) {
	    EXEC SQL DECLARE partn_c CURSOR WITH HOLD FOR
		SELECT partn
		INTO :partn
		FROM "informix".sysfragments
		WHERE tabid = :tabid
		AND fragtype = 'T';
	    if (sqlca.sqlcode != 0) {
		fprintf( stderr,
			 "FAILED declaring cursor for fragment rowcounts! Err=%d, isam=%d\n\n",
			 (int)sqlca.sqlcode,
			 (int)sqlca.sqlerrd[1] );
		exit( 1 );
	    }
	    once = 0;
	}
	EXEC SQL OPEN partn_c;
	if (sqlca.sqlcode != 0) {
	    fprintf( stderr,
		     "FAILED opening cursor for fragment rowcounts! Err=%d, isam=%d\n\n",
		     (int)sqlca.sqlcode,
		     (int)sqlca.sqlerrd[1] );
	    exit( 1 );
	}
	nrows = 0;
	for (;;) {
	    EXEC SQL FETCH partn_c
		INTO :partn;
	    if (sqlca.sqlcode < 0) {
		fprintf( stderr,
			 "FAILED fetching fragment rowcounts! Err=%d, isam=%d\n\n",
			 (int)sqlca.sqlcode,
			 (int)sqlca.sqlerrd[1] );
		exit( 1 );
	    } else if (sqlca.sqlcode == 100)
		break;
	    lcl.partnum = partn;
	    rc = (rowcounts_t *)bsearch( &lcl,
					 rowcounts,
					 numparts,
					 sizeof (rowcounts_t),
					 cmpcnts );
	    if (rc)
		nrows += rc->nrows;
	    else {
		fprintf( stderr,
			 "Error looking up partnum to match rowcount.  "
			 "Getting counts the hard way.\n" );
		return just_count_it( tabowner, tabname );
	    }
	}
	EXEC SQL CLOSE partn_c;
    }

    return nrows;
}

void usage( char *program )
{
    char *pager;
    FILE *errout;
    int  redirected = 0;
 
    errout = fdopen( 2, "w" );
    if (isatty(2)) {
        /* fclose( errout ); -- UNIX more needs stderr open, so... */
        if (!(pager = getenv( "PAGER" )) || (int)strlen(pager)==0)
            pager = "more";
        errout = popen( pager, "w" );
        redirected = 1;
    }

    fprintf(errout,
	    "Usage:\t%s [-h host] -d database [-t table] [-p] [-f filename] [-g] [-N]\n"
	    "\t\t[-i table_1 [-i table_2 [...]]] [-x [table_a|dbase_a:] [-x...]] \n"
	    "\t\t[-i @[file1] [-i @[file2]]] [-x @[file1] [-x @[file2]]] [-G] \n"
	    "\t\t[-i '!where clause'|subquery] [-x '!where clause'|subquery] \n"
	    "\t\t[-D|-6|-5] [-S] [-v[-1|0|1]] [-e|-E] [-s] [-V] [-F] [-T] [-I]\n"
	    "\t\t [-C] [-L] [-o] [-P procPDQ] [-Q tablePDQ] [-w Nsecs] [-X]\n"
	    "\t\t[-R HighRes] [-r MedRes] [-c MedConf] [-Z MinSamp] [-m] \n"
	    "\t\t[-b[-B BrowseThreshHold]] [-a [-A AgingLimit]] [-n errlimit] \n"
	    "\t\t[--procedure --schedule [--frequency [d|w[0-6]|m[0-6]]]\n"
	    "\t\t    [--first-time <datetime>] [--last-time <datetime>]\n"
	    "\t\t    [--one-time]] [--aus-thresholds] \n",
	    program );
    fprintf( errout,
	     "\t-h Host defaults to $INFORMIXSERVER.  (The program uses multiple \n"
	     "\t   connections to the host.  This requires a non-shared memory \n"
	     "\t   connection.) \n" );
    fprintf( errout,
	     "\t-d Database may contain valid \"matches\" clause wildcards.\n" );
    fprintf( errout,
	     "\t   The string 'ALL' is equivalent to '*' and will cause %s to \n"
	     "\t   execute against all databases except sysmaster.\n",
	     program ); 
    fprintf( errout,
	     "\t-t Table may contain valid \"matches\" clause wildcards.\n"
	     "\t   Table defaults to '*', all user tables (No system catalog tables.)\n"
	     "\t   The special name \"none\" can be used with the -i option to specify\n"
	     "\t   a limited list of tables with dissimilar names.\n" );
    fprintf( errout,
	     "\t-G Get physical row count using COUNT(*).  On servers with a very large\n"
	     "\t   number of tables this option can improve runtime as sysptnhdr has no\n"
	     "\t   indices and must be scanned.\n" );
    fprintf( errout,
	     "\t-b Browse mode.  Select only tables which have had their row count \n"
	     "\t   change by a percentage specified by the -B, browse threshhold, \n"
	     "\t   option since statistics were last updated.  Overrides inclusions. \n"
	     "\t   This option requires a non-shared memory connection and so is not \n"
	     "\t   compatible with -S. \n" );
    fprintf( errout,
	     "\t-B Specify browse threshhold.  Specify percent change used by the \n"
	     "\t   browse mode option (-b) to filter tables on which to operate. \n"
	     "\t   Range: 0.0001<->10000.0  (Default: 15.0%%)\n" );
    fprintf( errout,
	     "\t-a Distribution Aging.  Select only tables whose distributions are \n"
	     "\t   more than AgingLimit days old.  If specified with -b adds tables \n"
	     "\t   rejected by browsing that satisfy the AgingLimit.  Otherwise Aging \n"
	     "\t   eliminates tables otherwise included if they do no exceed AgingLimit.\n" );
    fprintf( errout,
	     "\t-A Specify distribution AgingLimit.  Range: >= 0. (Default: 30 days) \n" );
    fprintf( errout,
	     "\t-p Update statistics for stored procedures.  If -t is not also\n"
	     "\t   specified only stored procedures are updated.  (Default: true\n"
	     "\t   if no -t option provided, false if -t, -x, or -i are provided.)\n" );
    fprintf( errout,
	     "\t-Q Set PDQPRIORITY for updating statistics on tables.  \n"
	     "\t   (Default: Environment value of PDQPRIORITY or zero.)\n");
    fprintf( errout,
	     "\t-w Specify wait setting for lock mode.  >0 wait N seconds, <0 wait \n"
	     "\t   forever, ==0 set to not wait.\n" );
    fprintf( errout,
	     "\t-P Set PDQPRIORITY for compiling stored procedures.  A PDQPRIORITY \n"
	     "\t   value for stored procedure compilation is now always produced.\n"
	     "\t   (Default: zero.)\n" );
    fprintf( errout,
	     "\t-f filename writes statistics commands to the named file and does not \n"
	     "\t   execute (default: Execute commands immediately).  If filename is '-' \n"
	     "\t   output is written to stdout (errout should be redirected). \n"
	     "\t   (Long option: --to-file).\n" );
    fprintf( errout,
	     "\t-X Enable SET EXPLAIN ON during the statistics run.  You can review\n"
	     "\t   the sqexplain.out file do see if other options and environment \n"
	     "\t   variables may improve the performance of dostats.  See Informix \n"
	     "\t   manuals about using PDQPRIORITY, PSORT_NPROCS, PSORT_DBTEMP, \n"
	     "\t   DBSPACETEMP, and DBUPSPACE with UPDATE STATISTICS.\n" ); 
    fprintf( errout,
	     "\t-D Needed for 6.00 databases.  Removes new 7.xx syntax.\n" );
    fprintf( errout,
	     "\t-5 Needed for 5.xx databases.  Removes new ODS syntax.  Only permits \n"
	     "\t   a single database to be processed.\n" );
    fprintf( errout,
	     "\t   If omitted and SYSMASTER database not found this option is assumed.\n" );
    fprintf( errout,
	     "\t-6 Synonym for -D for consistency with the -5 option.\n" );
    fprintf( errout,
	     "\t-v Verbose mode.  There are 4 modes: \n"
             "\t   -1 - No progress output.  Only print error messages. \n"
             "\t    0 - Normal progress output.  The -v without argument is equivalent\n"
	     "\t        to -v0. \n"
             "\t    1 - Expands all explicit and inferred options in addition to normal \n"
             "\t        progress output. \n"
	     "\t    2 - Adds even more verbosity. \n" );
    fprintf( errout,
	     "\t-S Shared memory connection handling.  Only permits a single database\n"
	     "\t   to be processed.\n" );
    fprintf( errout,
	     "\t-i Specify tables to forceably include in the tables processed.  One\n"
	     "\t   may enter a single table, the @ sign followed by the name of a file\n"
	     "\t   containing a list of tables one per line, or an exclamation point (!)\n"
	     "\t   followed by either a WHERE filter to be applied to systables to \n"
	     "\t   select tables for inclusion, or a complete sub-query beginning with \n"
	     "\t   the keyword 'select' which will return a list of valid table names \n"
	     "\t   to include.  To aid script use if the argument to -i is simply '@' \n"
	     "\t   stdin is read.  Multiple -i options are concatenated to form a net \n"
	     "\t   list.  Note that all -i options can be used together.  (NB - Database\n"
	     "\t   includes are not implemented.)\n" );
    fprintf( errout,
	     "\t-x Specify tables or databases to forceably exclude from those\n"
	     "\t   processed.  One may enter a single table or database, the @ sign\n"
	     "\t   followed by the name of a file containing a list of tables and/or\n"
	     "\t   databases, one per line, or an exclamation point (!) followed by a\n"
	     "\t   WHERE filter to be applied to systables to select tables to be\n"
	     "\t   excluded, or a complete sub-query beginning with the keyword 'select'\n"
	     "\t   which will return a list of valid table names to exclude.  To aid\n"
	     "\t   script use if the argument to -x is simply '@' stdin is read.  \n"
	     "\t   Multiple -x options are concatenated to form a net list.  Note that \n"
	     "\t   all -x options can be used together.  Database names are specified by\n"
	     "\t   appending a colon (:) to the name (ex: \"sysutils:\" to exclude the\n"
	     "\t   sysutils database).\n" );
    fprintf( errout,
	     "\t-s Recognize Standard Engine limitation of 8 index columns.  Only\n"
	     "\t   permits a single database to be processed.\n" );
    fprintf( errout,
	     "\t-e Expand reporting to output elapsed time of each stage in seconds \n"
	     "\t   and fractions. \n" );
    fprintf( errout,
	     "\t-E Expand reporting to output elapsed time of each stage in hours,\n"
	     "\t   minutes, seconds and fractions. \n" );
#if defined(NOTIMINGS)
    fprintf( errout,
	     "\t   THE TIMING REPORT OPTIONS (-e, -E) have been disabled in this \n"
	     "\t   executable.\n" );
#endif
    fprintf( errout,
	     "\t-V Print out dostats version and source code revision and exit.\n" );
    fprintf( errout,
	     "\t-F Disable single column index optimization.  Use HIGH...DISTRIBUTIONS \n"
	     "\t   ONLY and LOW which is faster on some systems/databases.\n" );
    fprintf( errout,
	     "\t-N No specific keyword for updating stored procedures/functions. \n" );
    fprintf( errout,
	     "\t-o Override detected server optimization level.  By default if the \n"
	     "\t   target engine contains the optimized UPDATE STATISTICS algorithms \n"
	     "\t   introduced in 7.31xD2 & 9.30xC3 all HIGH level distributions are \n"
	     "\t   requested with a minimum number of statements.  For earlier releases \n"
	     "\t   of IDS HIGH level distributions are requested for a single column at \n"
	     "\t   a time.  This option reverses this behavior so that scripts for use \n"
	     "\t   in an advanced engine can be created from a schema loaded on an  \n"
	     "\t   earlier release and vice-versa. \n" );
    fprintf( errout, 
             "\t-n Do not exit on error until errlimit errors are encountered.  Only \n"
             "\t   This only affects errors running update statistics commands. \n" );
    fprintf( errout,
	     "\t--procedure Create the named procedure with the UPDATE STATISTICS \n"
	     "\t\tcommands to run.  Can be used to replace a previously scheduled \n"
	     "\t\tcommand procedure, to create a new named procedure for \n"
	     "\t\tscheduling or just to create a procedure to be manually executed\n"
	     "\t\tlater. If --procedure is not specified, and --schedule is, a \n"
	     "\t\tdefault name is generated for the procedure that is scheduled. \n" );
    fprintf( errout,
	     "\t--schedule Create a stored procedure with the commands and schedule the\n"
	     "\t\tprocedure to run in the future.  Defaults to running once at\n"
	     "\t\tthe following midnight.  Adjust run time and frequency with \n"
	     "\t\tthe following options:\n"
	     "\t\t--frequency How often to run the update statistics \n"
	     "\t\t    procedure created: \n"
	     "\t\t      d - run daily.\n"
	     "\t\t      w - run weekly on day: 0-Sunday, 1-Monday, etc...\n"
	     "\t\t      m - run monthly on the first: 0-Sunday, 1-Monday,\n"
	     "\t\t\tetc.. of month\n"
	     "\t\t--first-time Time of day to run (or to first run) the \n"
	     "\t\t    procedure. \n"
	     "\t\t--last-time  Last time of day to run (if frequency is\n"
	     "\t\t    less than 1 day).\n"
	     "\t\t--one-time  Run the procedure as scheduled then disable the run.\n\n"
	     "\t  You can see your dostats scheduled tasks by querying the task group\n"
	     "\t  (ph_task.tk_group = 'DOSTATS'\n" );
    fprintf( errout,
	     "\t--aus-thresholds Retrieve the Auto Update Statistics settings for \n"
	     "\t   Aging, Browsing (Change), and PDQPRIORITY.  -A, -B, & -Q override \n"
	     "\t   the corresponding individual AUS settings. Aging and Browsing are \n"
	     "\t   only enabled if -a and/or -b are also specified. \n" );
    fprintf( errout,
	     "\t-R HighRes sets resolution for HIGH Level distributions.\n" );
    fprintf( errout,
	     "\t   Calculate distributions by dividing keys into (100 / HighRes)\n" );
    fprintf( errout,
	     "\t   bins each with approximately the same number of keys.\n" );
    fprintf( errout,
	     "\t   Valid range: max(0.005,1/nrows) <= HighRes <= 10.00  \n"
	     "\t   Default: 0.5 (200 bins)\n" );
    fprintf( errout,
	     "\t-r MedRes sets resolution for MEDIUM Level distributions.\n" );
    fprintf( errout,
	     "\t   Calculate distributions by dividing keys into (100 / MedRes)\n" );
    fprintf( errout,
	     "\t   bins each with approximately the same number of keys.\n" );
    fprintf( errout,
	     "\t   Valid range: max(0.005,1/nrows) <= MedRes <= 10.00  \n"
	     "\t   Default: 2.5 (40 bins)\n" );
    fprintf( errout,
	     "\t-c MedConf set confidence level for MEDIUM Level distributions.\n" );
    fprintf( errout,
	     "\t   Valid range: 0.80 <= MedConf <= 0.99, Default: 0.95\n" );
    fprintf( errout,
	     "\t-m Enable processing of system catalog metadata tables.\n"
	     "\t   (Default: disabled)\n" );
    fprintf( errout,
	     "\t-Z Set minimum sample size for MEDIUM resolution.  (0 < MinSamp < nrows)\n" );
    fprintf( errout,
	     " \n"
	     "\t   By default all of -T, -I, & -C flags are set.  Manually specifying \n"
	     "\t   any of -T, -L, -I, or -C disables any of the others not also \n"
	     "\t   explicitly specified.  Enabling -T -I emulates dostats version 1. \n"
	     "\t   Options -T & -L are mutually exclusive if both are given the last \n"
	     "\t   prevails. \n" );
    fprintf( errout,
	     "\t    -T  Enable table level statistics update to MEDIUM. \n" );
    fprintf( errout,
	     "\t    -L  Set table level statistics update to LOW and drop distributions. \n" );
    fprintf( errout,
	     "\t    -l  Set table level statistics update to LOW keeping existing \n"
	     "\t\tdistributions. \n" );
    fprintf( errout,
	     "\t    -I  Enable statistics update of first column of each index to HIGH. \n" );
    fprintf( errout,
	     "\t    -C  Enable statistics update indexed columns to LOW. \n" );
    fprintf( errout,
	     "\t    -g  Enable statistics update fragmentation expression columns to \n"
	     "\t        HIGH (ie treat like first index key columns). \n" );
    fprintf( errout,
	     "\n\tSetting PDQPRIORITY in the %s environment causes an effective \n"
	     "\targument of '-P 0' to be added the commandline to protect procedures. \n",
	     program);
    fprintf( errout,
	     "\n\tThe environment variables INFORMIXUSER and INFORMIXPASS can to used\n"
	     "\tto specify a username and password (respectively) other than the \n"
	     "\tcurrent user's id and password.\n\n" );

    if (redirected) {
        fflush( errout );
        pclose( errout );
    }
}


#if !defined(MAXWORD)
#define MAXWORD 65536
#endif

/* Remove 'out' & 'inout' parameter type entries from parameter type lists. */
void StripTypeList( char *list )
{
    char *out, word[MAXWORD], *wptr;
    int cnt = 0;

    out = list;
    word[0] = (char)0;
    wptr = word;
    while (list != (char)0) {
	switch (*list) {
	  case 0:
	  case ' ':
	  case ',':
	      *wptr = (char)0;
	      if (strcmp( word, "out" ) == 0 || strcmp( word, "inout" ) == 0) {
		  /* Toss this word. */
		  wptr = word;
		  list++; /* Skip the trailing space also. */
	      } else {
		  /* insert this word in the output. */
		  strncpy( out, word, strlen( word ) );	/* Append the word to the output. */
		  wptr = word;				/* Set up for the next word. */
		  out += strlen( word );			/* Adjust the output pointer. */
		  *out = *list;				/* Copy the separator char also. */
		  out++;					/* Adjust pointer again. */
		  if (*list != (char)0)			/* Done? */
		      list++;				/* No - skip separator. */
		  else
		      return;				/* Yes - go back. */
	      }
	      break;
	  default:					/* Build each word here. */
	      *wptr = *list;
	      list++;
	      wptr++;
	      cnt++;
	      if (cnt >= MAXWORD) {
		  fprintf( stderr,
			   "Word size (%d) exceeded.  Word is: <%s> \n",
			   MAXWORD,
			   word );
		  exit( 1 );
	      }
	      break;
	}
    }
}


static short col;
static char *colstr, *opclstr, *tok;
static char *tmp;

void InitKeyStr( char *keystr )
{
    if (keystr[0] == '<') {
	/* Functional key.  2nd token is a list of columns. */
      
	tok = strtok( keystr, "(" ); /* Strip procid */
	if(tok) {
	    tmp = tok + strlen(tok) + 1;
	}
	colstr = strtok( (char *)NULL, "), [" ); /* Get collist */
	tmp = colstr + strlen( colstr ) + 1;

	if (tmp[0] == '[' || tmp[1] == '[') {
	    opclstr = strtok( (char *)NULL, "], " ); /* Strip opclass */
	    tmp = opclstr + strlen( opclstr ) + 1;
	}
	col = (short)atol( colstr );
    } else {
	/* Columns key.  First token we encounter is the colno. */
	tok = strtok( keystr, "[ " );
	if(tok) {
	    tmp = tok + strlen(tok) + 1;
	}
	if (tok) {
	    col = (short)atol( tok );

	    /* Next token is the op class id */
	    tok = strtok( (char *)NULL, "], [" ); /* Strip opclass */
	    if(tok) {
		tmp = tok + strlen(tok) + 1;
	    }
	} else {
	    col = 0;
	}
    }
}

short ParseKeyStr( )
{
    short ret;
    
    ret = abs( col );

    if (ret) {
	tok = strtok( (char *)NULL, "], [()" );

	if(tok) {
	    tmp = tok + strlen(tok) + 1;
	}

	if (tok && tok[0] == '<') {
	    /* Start of another functional key. */
	    colstr = strtok( (char *)NULL, "), [" ); /* Get collist */
	    tmp = colstr + strlen( colstr ) + 1;
	    if (tmp[0] == '[' || tmp[1] == '[') {
		opclstr = strtok( (char *)NULL, "], " ); /* Strip opclass */
		tmp = opclstr + strlen( opclstr ) + 1;
	    }
	    col = (short)atol( colstr );
	} else if (tok) {
	    if(tok) {
		tmp = tok + strlen(tok) + 1;
	    }
	    col = (short)atol( tok );

	    if(tok) {
		tmp = tok + strlen(tok) + 1;
	    }
	    /* Next token is the op class id */
	    tok = strtok( (char *)NULL, "], [()" ); /* Strip opclass */
	    if(tok) {
		tmp = tok + strlen(tok) + 1;
	    }
	} else {
	    col = 0;
	}
    }
    return ret;
}

static off_t pos = 0;
static int inblob = 0;
 
int openit( loc_t *loc, int flags, int bsize )
{
    if (inblob == 1) {
        fprintf( stderr, "Open called for new blob without close.\n"
		 "length = %d.!\n",
		 (int)loc->loc_xfercount );
    }
    loc->loc_status = 0;
    loc->loc_xfercount = 0L;
    inblob = 0;
    blobbuff->global_length = 0 ;
    pos = 0;
    /* Make sure blob buffer is large enough to hold BLOB. */
    if (bsize > blobbuff->bbsize) {
        blobbuff->bbsize = bsize + (1024 - (bsize % 1024));
	blobbuff->buff = (char *)realloc( blobbuff->buff, blobbuff->bbsize );
    }
    memset( blobbuff->buff, 0, blobbuff->bbsize );
 
    return 0;
}
 
int writeit( loc_t *loc, char *buf, int nbytes )
{
    if ((pos + nbytes) > blobbuff->bbsize) {
        blobbuff->bbsize += (2 * (nbytes < 1024 ? 1024 : nbytes));
        blobbuff->buff = (char *)realloc( blobbuff->buff, blobbuff->bbsize );
    }
    if (!inblob)
        inblob = 1;
    memcpy( &(blobbuff->buff[pos]), buf, nbytes );
    pos += nbytes;
 
    loc->loc_xfercount += nbytes;
    blobbuff->global_length += nbytes;
 
    return 0;
}
 
int closeit( loc_t *loc )
{
    unsigned int count;
 
    inblob = 0;
    loc->loc_status = 0;
    if (loc->loc_oflags & LOC_WONLY) {
        /* Fetching, cleanup. */
        loc->loc_indicator = 0;
        loc->loc_size = loc->loc_xfercount;
    }
    count = blobbuff->global_length ;
    if( blobbuff->global_length != loc->loc_xfercount)
        fprintf( stderr,
		 "GLOBAL_LENGTH !=  loc->loc_xfercount <%d, %d> \n",
		 blobbuff->global_length,
		 (int)loc->loc_xfercount );
    pos = 0;
 
    return 0;
}

/*
 * Find a column name in the list of names.  Returns index if found,
 * else -1.  Simple linear search, the list is short.
 */
int cfind( char *item, char collist[][129], int ncols )
{
    int col;

    for (col = 0; col < ncols; col++)
	if (strcmp( item, collist[col] ) == 0)
	    break;
    if (col >= ncols)
	return -1;
    else
	return col;
}

/* Determine whether server version uses optimized update statistics logic. */
int is_Server_Optimized( int major, int minor, char *level )
{
    int level_num;
    char *level_let;

    if (major > 9)
	return 1;
    else if (major == 8 || major < 7) {
	return 0;
    } else {
	level_let = level;
	level_num = atoi( (level + 1) );

	if (major == 9) {
	    if (minor > 30)
		return 1;
	    else if (minor == 30 && (*level_let > 'C' || (*level_let == 'C' && level_num >= 3)))
		return 1;
	    else
		return 0;
	} else if (major == 7) {
	    if (minor > 31)
		return 1;
	    else if (minor == 31 && (*level_let > 'D' || (*level_let == 'D' && level_num >= 2)))
		return 1;
	    else
		return 0;
	} else
	    return 0;
    }
}

EXEC SQL ifdef C4GL_720_COMPILER;
int Store_Statement( statement )
     $PARAMETER char *statement;
EXEC SQL else;
int Store_Statement(
    EXEC SQL BEGIN DECLARE SECTION;
    PARAMETER char *statement
    EXEC SQL END DECLARE SECTION;
    )
EXEC SQL endif;
{
    static char last_proc[129] = "";
    int left;

    EXEC SQL BEGIN DECLARE SECTION;
    static int sequence = 0;
    string part[769];
    char cont_char[2];
    EXEC SQL END DECLARE SECTION;

    if (strcmp( last_proc, schedprocname ) != 0) {
	/* First call for this procedure.  Initialize. */
	sequence = 1;
	strcpy( last_proc, schedprocname );
	EXEC SQL DELETE FROM sysadmin:dostats_stmts WHERE function = :schedprocname;
    }

    if (verbose > 1)
	fprintf( stderr, "In Store_Statement sequence = %d with:\n\t%s\n", sequence, statement );

    cont_char[1] = (char)0;
    left = strlen( statement );
    while (left > 0) {
	snprintf( part, sizeof part, "%s", statement );

	left -= 768;
	statement += 768;

	if (left > 0) {
	    cont_char[0] = 't';
	} else {
	    cont_char[0] = 'f';
	}

	EXEC SQL INSERT INTO sysadmin:dostats_stmts
	    (
	     function,
	     sequence,
	     continued,
	     statement
	     ) VALUES
	    (
	     :schedprocname,
	     :sequence,
	     :cont_char,
	     :part );
	if (sqlca.sqlcode < 0) {
	    fprintf( stderr,
		     "Error inserting statements to dostats_stmts.  SQL: %d, ISAM: %d.\n",
		     sqlca.sqlcode,
		     sqlca.sqlerrd[1] );
	}
	sequence++;
    }
    if (verbose > 1)
	fprintf( stderr, "Bottom of Store_Statement.\n" );
}

int Write_Statement( char *statement )
{
    fprintf( outfile, "%s\n", statement );
    fflush( outfile );
}
